/* agenerator.h - attribute Generator */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef TAGENERATOR_H

#include "definitions.h"
#include "debug.h"
#include "generator.h"
#include "flagtable.h"

namespace faudes {

/**
 * Type definition for flagtable in aGenerator
 */
typedef TFlagTable<aType> aFlagTable;


/**
 * Modell of the common five-tupel consisting of alphabet, stateset,
 * transition relation, initial states, marked states with events marked
 * as controllable / uncontrollable
 *
 * File format of cGenerator is compatible to Generator but differs in the 
 * controllable events section. The controllable events are labeled 
 * "Controllable" and should have their place within a 
 * "EventProperties" section below the "Generator" entry.
 *
 * InsEvent and InsEvents call the Generator class function and do not set
 * the control property of the event. This can be checked with the 
 * ExistsControllableFlag cGenerator method.
 */
template<class TrnSet>
class TaGenerator : public TGenerator<TrnSet> {
	public:

		/**
		 * Creates an emtpy aGenerator object with static eventtable, statetable
		 * and flagtable.
		 */
		TaGenerator(void);

		/** 
		 * aGenerator from a Generator. 
		 *
		 * @param rOtherGen
		 */
		TaGenerator(const TGenerator<TrnSet>& rOtherGen);

		/** 
		 * aGenerator from an aGenerator. Copy constructor 
		 *
		 * @param rOtherGen
		 */
		TaGenerator(const TaGenerator<TrnSet>& rOtherGen);
		
		/**
		 * Creates an aGenerator from file with static eventtable, statetable
		 * and flagtable
		 *
		 * @param pFileName
		 *    Filename
		 *
		 * @exception Exception
		 *   If opening/reading fails an Exception object is thrown (id 2)
		 *   If an uncontrollable event is already flagged controllable (id 100)
		 */
		 TaGenerator(const char* pFileName); 

		/**
		 * Write cgenerator to console
		 */
		virtual void Write(void) const;

		/**
		 * Write cgenerator to file
		 *
		 * @param pFileName
		 *   Filename to write
		 * @param openmode
		 *   ios::openmode (optional)
		 *   Default is truncate file.
		 *
		 * @exception Exception
		 *   If opening/writing fails an Exception object is thrown (id 2)
		 */
		virtual void Write(const char* pFileName, 
				std::ios::openmode openmode = std::ios::out|std::ios::trunc) const;

		/**
		 * Write cgenerator to TokenWriter
		 *
		 * @param rTw
		 *   Reference to TokenWriter
		 *
		 * @exception ios::failure
		 *   Thrown on i/o error.
		 */
		virtual void Write(TokenWriter& rTw) const;

		/**
		 * Write event attributes to TokenWriter
		 *
		 * @param rTw
		 *   Reference to TokenWriter
		 *
		 * @exception ios::failure
		 *   Thrown on i/o error.
		 */
		void WriteEventAttributes(TokenWriter& rTw) const;
		
		/**
		 * Write state attributes to TokenWriter
		 *
		 * @param rTw
		 *   Reference to TokenWriter
		 *
		 * @exception ios::failure
		 *   Thrown on i/o error.
		 */
		void WriteStateAttributes(TokenWriter& rTw) const;

		/**
		 * (Debug-) write cgenerator to console.
		 * This is just an alias function for Write()
		 */
	     virtual void DWrite(void) const;

		/**
		 * (Debug-) write event attributes to console
		 * 
		 * @param rTw
		 *   Reference to TokenWriter
		 *
		 * @exception ios::failure
		 *   Thrown on i/o error.
		 */
		void DWriteEventAttributes(TokenWriter& rTw) const;

		/**
		 * (Debug-) write state attributes to console
		 * 
		 * @param rTw
		 *   Reference to TokenWriter
		 *
		 * @exception ios::failure
		 *   Thrown on i/o error.
		 */
		void DWriteStateAttributes(TokenWriter& rTw) const;
		
		/**
		 * Create a string similar to the output of Write()
		 *
		 * @return
		 *   std::string
		 */
		std::string ToString() const;

		/**
		 * Create a string similar to the output of Write()
		 * for the event attributes
		 *
		 * @return
		 *   std::string
		 */
		 std::string EventAttributesToString(void) const;

		/**
		 * Create a string similar to the output of Write()
		 * for the state attributes
		 *
		 * @return
		 *   std::string
		 */
		 std::string StateAttributesToString(void) const;

		/**
		 * iterator to Begin() of mEventFlagTable
		 * 
		 * @returns
		 *   iterator to Begin of mEventFlagtable
		 */
		aFlagTable::iterator EventFlagTableBegin(void);

		/**
		 * const_iterator to Begin() of mEventFlagTable
		 * 
		 * @returns
		 *   const_iterator to Begin of mEventFlagtable
		 */
		aFlagTable::const_iterator EventFlagTableBegin(void) const;
		
				/**
		 * iterator to End() of mEventFlagTable
		 * 
		 * @returns
		 *   iterator to End of mEventFlagtable
		 */
		aFlagTable::iterator EventFlagTableEnd(void);

		/**
		 * const_iterator to End() of mEventFlagTable
		 * 
		 * @returns
		 *   const_iterator to End of mEventFlagtable
		 */
		aFlagTable::const_iterator EventFlagTableEnd(void) const;
		
		/**
		 * iterator to Begin() of mStateFlagTable
		 * 
		 * @returns
		 *   iterator to Begin of mStateFlagtable
		 */
		aFlagTable::iterator StateFlagTableBegin(void);

		/**
		 * const_iterator to Begin() of mStateFlagTable
		 * 
		 * @returns
		 *   const_iterator to Begin of mStateFlagtable
		 */
		aFlagTable::const_iterator StateFlagTableBegin(void) const;
		
				/**
		 * iterator to End() of mStateFlagTable
		 * 
		 * @returns
		 *   iterator to End of mStateFlagtable
		 */
		aFlagTable::iterator StateFlagTableEnd(void);

		/**
		 * const_iterator to End() of mStateFlagTable
		 * 
		 * @returns
		 *   const_iterator to End of mStateFlagtable
		 */
		aFlagTable::const_iterator StateFlagTableEnd(void) const;
		
		
		/**
		 * Clear aGenerator 
		 *
		 */
		virtual void Clear(void);
  		 
		/**
		 * Read aGenerator from file
		 *
		 * @param pFileName
		 *   Filename containing generator definition
		 * 
		 * @exception Exception
		 *   If opening/reading fails an Exception object is thrown (id 1)
		 *   Exception opening/reading controllable events in file (id 6)
		 *   If an uncontrollable event is already flagged controllable (id 100)
		 */
		virtual void Read(const char* pFileName);

		/**
		 * Read aGenerator from TokenReader
		 *
		 * @param rTr
		 *   Reference to TokenReader
		 *
		 * @exception Exception
		 *   Thrown on i/o error. 
		 * @exception Exception
		 *   Thrown on file format errors (id 50, 51)
		 */
		virtual void Read(TokenReader& rTr);
		
		/**
		 * Read event attributes from file
		 *
		 * @param pFileName
		 *   Filename containing event attributes
		 *   "<EventAttributes>...</EventAttributes>"
	     *
		 * @exception Exception
		 *   If opening/reading fails an Exception object is thrown (id 1)
		 *   Exception opening/reading event attributes in file (id 6)
		 */
		void ReadEventAttributes(const char* pFileName);

		/**
		 * Read event attributes from TokenReader
		 *
		 * @param rTr
		 *   Reference to TokenReader
	     *
		 * @exception ios::failure
		 *   Thrown on i/o error.
		 * @exception Exception
		 *   Thrown on file format errors (51)

		 */
		void ReadEventAttributes(TokenReader& rTr);

		/**
		 * Read state attributes from file
		 *
		 * @param pFileName
		 *   Filename containing event attributes
		 *   "<StateAttributes>...</StateAttributes>"
		 * @param rMinIndexMap
		 *   Map with the minimal state indeces
		 * @param rTmpStateNameMap
	     *   map from state names to state indeces
		 *
		 * @exception Exception
		 *   If opening/reading fails an Exception object is thrown (id 1)
		 *   Exception opening/reading event attributes in file (id 6)
		 */
		void ReadStateAttributes(const char* pFileName, 
          const std::map<Idx,Idx>& rMinIndexMap,
          const std::map<std::string,Idx>& rTmpStateNameMap);

		/**
		 * Read state attributes from TokenReader
		 *
		 * @param rTr
		 *   Reference to TokenReader
		 * @param rMinIndexMap
		 *   Map with the minimal state indeces
		 * @param rTmpStateNameMap
	     *   map from state names to state indeces
		 *
		 * @exception ios::failure
		 *   Thrown on i/o error.
		 * @exception Exception
		 *   Thrown on file format errors (51)

		 */
		void ReadStateAttributes(TokenReader& rTr, 
          const std::map<Idx,Idx>& rMinIndexMap,
          const std::map<std::string,Idx>& rTmpStateNameMap);
  		
		/**
		 * Get copy of EventFlagTable 
		 *
		 * @return
		 *   Copy of this aGenerator's (Event)FlagTable
		 */
		aFlagTable EventFlagTable(void) const;
		
		 /**
		 * Get reference to EventFlagTable 
		 *
		 * @return
		 *   Reference to this aGenerator's (Event)FlagTable
		 */
		aFlagTable& EventFlagTabler(void);

		/**
		 * Get copy of StateFlagTable 
		 *
		 * @return
		 *   Copy of this aGenerator's (State)FlagTable
		 */
		aFlagTable StateFlagTable(void) const;
		
	  /**
		 * Get reference to StateFlagTable 
		 *
		 * @return
		 *   Reference to this aGenerator's (State)FlagTable
		 */ 
		aFlagTable& StateFlagTabler(void);

		/**
		 * Set the EventFlagTable of an existing generator
		 *
		 * @param EventFlagTable
		 *   new EventFlagTable
		 *
		 */
         	void SetEventFlagTable(aFlagTable EventFlagTable);
	
		/**
		 * Add an existing event with an attribute to generator. 
		 *
		 * @param index
		 *   Event index
		 *
		 * @param attribute
		 *   Event attribute
		 */
		void InsAttributedEvent(Idx index,aType attribute);

		/**
		 * Add new named event with attribute to generator. 
		 * An entry in the global eventtable will be made if event is new.
		 *
		 * @param rName
		 *   Name of the event to add
		 * 
		 * @param attribute
         *   Event attribute
		 *
		 * @return 
		 *   New global unique index
		 */
		Idx InsAttributedEvent(const std::string& rName,aType attribute);
	
		/**
		 * Set the attribute of an event(by index)
		 *
		 * @param index
		 *   Event index
		 *
		 * @param attribute
		 *   Event attribute
		 */
		void SetEventAttribute(Idx index,aType attribute);

		/** 
		 * Set the attribute of an event(by name)
		 *
		 * @param rName
		 *   Event name	
		 *
		 * @param attribute
		 *   Event attribute
		 */
		void SetEventAttribute(const std::string& rName,aType attribute);
		
		/**
		 * Clear the attribute of an event(by index)
		 *
		 * @param index
		 *   Event index
		 *
		 * @param attribute
		 *   Event attribute
		 */
		void ClearEventAttribute(Idx index,aType attribute);

		/** 
		 * Clear the attribute of an event(by name)
		 *
		 * @param rName
		 *   Event name	
		 *
		 * @param attribute
		 *   Event attribute
		 */
		void ClearEventAttribute(const std::string& rName,aType attribute);
		 
		/**
		 * Check the attributes of an event (by index)
		 *
		 * @param index
		 *   Event index
		 * @param attributes
		 *   Event attributes to check
		 * @param defaultAttribute
		 *   default Event attribute
		 *
		 * @return
		 *   checks if Event attribute is true
		 */
		 aType GetEventAttribute(Idx index,aType attributes, aType defaultAttribute) const;

		/**
		 * Check the attributes of an event (by name)
		 *
		 * @param rName
		 *   Event name
		 * @param attributes
		 *   Event attributes to check
		 * @param defaultAttribute
		 *   default Event attribute
		 *
		 * @return
		 *   checks if the Event attribute is true
		 */
		 aType GetEventAttribute(const std::string& rName, aType attributes, aType defaultAttribute);

		/**
		 * Check whether event exists in eventflagtable or not
		 *
		 * @param index
		 *   Event index
		 *
		 * @return
		 *   True if exists
		 */
		bool ExistsEventAttribute(Idx index);

		/**
		 * Check whether event exists in eventflagtable or not
		 *
		 * @param rName
		 *   Event name
		 *
		 * @return
		 *   True if exists
		 */
		bool ExistsEventAttribute(const std::string& rName);

		/**
		 * Set the StateFlagTable of an existing generator
		 *
		 * @param StateFlagTable
		 *   new StateFlagTable
		 *
		 */
    		void SetStateFlagTable(aFlagTable StateFlagTable);

		/**
		 * Add an existing state with an attribute to generator. 
		 *
		 * @param index
		 *   Event rName
		 *
		 * @param attribute
		 *   Event attribute
		 */
		void InsAttributedState(Idx index, aType attribute);
		
		/**
		 * Add new named state with attribute to generator. 
		 * An entry in the global statetable will be made.
		 *
		 * @param rName
		 *   Name of the event to add
		 * 
		 * @param attribute
		 *   Event attribute
		 *
		 * @return 
		 *   New global unique index
		 */
		Idx InsAttributedState(const std::string& rName,aType attribute);
	
		/**
		 * Set the attribute of a state (by index)
		 *
		 * @param index
		 *   State index
		 *
		 * @param attribute
		 *   State attribute
		 */
		void SetStateAttribute(Idx index,aType attribute);

	  	/**
		 * Clear the attribute of a state (by index)
		 *
		 * @param index
		 *   State index
		 *
		 * @param attribute
		 *   State attribute
		 */
		void ClearStateAttribute(Idx index,aType attribute);
		 
		/**
		 * Get the attributes of a state (by index)
		 *
		 * @param index
		 *   Event index
		 * @param attributes
     		 *   Event attributes to check
     		 * @param defaultAttribute
     		 *   default Event attribute
		 *
		 * @return
		 *   True / False
		 */
		aType GetStateAttribute(Idx index, aType attribute, aType defaultAttribute) const;

		/**
		 * Check whether state exists in stateflagtable or not
		 *
		 * @param index
		 *   State index
		 *
		 * @return
		 *   True if exists
		 */
		bool ExistsStateAttribute(Idx index);

		/** 
		 * Assignment operator. The Copy method is used
		 *
		 * @param rOtherGen
		 *   Other aGenerator
		 */
		virtual TaGenerator<TrnSet>& operator = (const TaGenerator<TrnSet>& rOtherGen); 

		/** 
		 * Assignment operator for Generator
		 * 
		 * Assignment operator. The Copy method ist used
		 *
		 * @param rOtherGen
		 *   Other aGenerator
		 */
		virtual TaGenerator<TrnSet>& operator = (const TGenerator<TrnSet>& rOtherGen); 
		
		/**
		 * Create a copy with new state indices. 
		 *
		 * @param rGen
		 *   Empty target aGenerator
		 */
		virtual void Copy(TaGenerator<TrnSet>& rGen) const; 
		
		/**
		 * Create Empty aGenerator with proper tables set on stack
         	 * Calls NewaGenerator()
		 *
		 * @return
		 *   New cGenerator
		 */
		TaGenerator<TrnSet> NewGen(void) const;

		/**
		 * Create Empty aGenerator with proper tables set on heap
         	 * Calls NewaGeneratorp()
		 *
		 * @return
		 *   New aGenerator*
		 */
		TaGenerator<TrnSet>* NewGenp(void) const;

		/**
		 * Create Empty aGenerator with proper tables set on stack
		 *
		 * @return
		 *   New cGenerator
		 */
		TaGenerator<TrnSet> NewaGenerator(void) const;

		/**
		 * Create Empty aGenerator with proper tables set on heap
		 *
		 * @return
		 *   New aGenerator*
		 */
		TaGenerator<TrnSet>* NewaGeneratorp(void) const;
		
	private:

	protected:


		/** This aGenerator's event flag table */
		aFlagTable mEventFlagTable;
		
		/** This aGenerator's state flag table */
		aFlagTable mStateFlagTable;

};

		typedef TaGenerator<TransSet> aGenerator;


// TaGenerator<TrnSet>::aGenerator(void)
template<class TrnSet>
TaGenerator<TrnSet>::TaGenerator(void) : TGenerator<TrnSet>() {
	FD_DG("aGenerator(" << this << ")::aGenerator()");
}

// TaGenerator<TrnSet>::aGenerator(rOtherGen)
template<class TrnSet>
TaGenerator<TrnSet>::TaGenerator(const TGenerator<TrnSet>& rOtherGen) 
		: TGenerator<TrnSet>(rOtherGen) {
	FD_DG("aGenerator(" << this << ")::aGenerator(" << &rOtherGen << ")");
}


// TaGenerator<TrnSet>::aGenerator(rOtherGen)
template<class TrnSet>
TaGenerator<TrnSet>::TaGenerator(const TaGenerator<TrnSet>& rOtherGen) 
		: TGenerator<TrnSet>(rOtherGen) {
	FD_DG("aGenerator(" << this << ")::aGenerator(" << &rOtherGen << ")");
	mEventFlagTable = rOtherGen.mEventFlagTable;
	mStateFlagTable = rOtherGen.mStateFlagTable;
}


	
// TaGenerator<TrnSet>::aGenerator(pFileName)
template<class TrnSet>
TaGenerator<TrnSet>::TaGenerator(const char* pFileName) : TGenerator<TrnSet>()  {
	FD_DG("aGenerator(" << this << ")::aGenerator(\"" << pFileName << "\")");
	Read(pFileName);
} 

// virtual
// Write()
template<class TrnSet>
void TaGenerator<TrnSet>::Write(void) const{
	TokenWriter tw;
	Write(tw);
}

// virtual
// Write(file, openmode=default)
template<class TrnSet>
void TaGenerator<TrnSet>::Write(const char* pFileName, std::ios::openmode openmode) const {
	try {
		TokenWriter tw(pFileName, openmode);
		Write(tw);
	}
	catch (faudes::Exception&) {
		std::stringstream errstr;
		errstr << "Exception opening/writing file \"" << pFileName << "\"";
		throw Exception("TaGenerator<TrnSet>::Write", errstr.str(), 2);
	}
}

// virtual
// Write(tw)
template<class TrnSet>
void TaGenerator<TrnSet>::Write(TokenWriter& rTw) const{
	std::map<Idx,Idx> minstateindexmap = TGenerator<TrnSet>::MinStateIndexMap();
	// write generator
	rTw.WriteBegin("aGenerator");
	rTw.WriteBegin("Generator");
	rTw << TGenerator<TrnSet>::mMyName;
	rTw << "\n";
	TGenerator<TrnSet>::mAlphabet.Write(rTw);
	TGenerator<TrnSet>::mStates.Write(rTw, minstateindexmap, true);
	TGenerator<TrnSet>::WriteTransRel(rTw, minstateindexmap);
	TGenerator<TrnSet>::mInitStates.Write(rTw, minstateindexmap, false);
	TGenerator<TrnSet>::mMarkedStates.Write(rTw, minstateindexmap, false);
	rTw.WriteEnd("Generator");
	rTw.Write("\n");
	rTw.Write("\n");
	WriteEventAttributes(rTw);
	WriteStateAttributes(rTw);
	rTw.WriteEnd("aGenerator");
}

// WriteControllableEvents(rTw)
template<class TrnSet>
void TaGenerator<TrnSet>::WriteEventAttributes(TokenWriter& rTw) const {
    FD_DG("aGenerator(" << this << ")::WriteEventAttributes(rTw)");   
	//std::map<Idx,aType> rFlagMap = mEventFlagTable.FlagMapr();
	std::map<Idx,aType>::const_iterator it;
	rTw.WriteBegin("EventAttributes");
	//for(it = rFlagMap.begin(); it != rFlagMap.end(); ++it){
	for(it = EventFlagTableBegin(); it != EventFlagTableEnd(); ++it){
	    if(EventSet::StaticEventSymbolTablep()->EventName(it->first) != ""){
            rTw << EventSet::StaticEventSymbolTablep()->EventName(it->first).c_str();
            rTw << (Idx)it->second;
        }
        else{
            FD_ERR("WriteEventAttributes(" << this << "): index " << it->first 
					<< " does not have a name");
        }
	}
	rTw.WriteEnd("EventAttributes");
}

// WriteObservableEvents(rTw)
template<class TrnSet>
void TaGenerator<TrnSet>::WriteStateAttributes(TokenWriter& rTw) const{
    FD_DG("aGenerator(" << this << ")::WriteStateAttributes(rTw)");
	//std::map<Idx,aType> rFlagMap = mStateFlagTable.FlagMapr();
	std::map<Idx,aType>::const_iterator fit;
	std::map<Idx,Idx> minstateindexmap = TGenerator<TrnSet>::MinStateIndexMap();

	std::map<Idx,Idx>::const_iterator mapit;
    int oldcolumns = rTw.Columns();
    rTw.Columns(6);
    rTw.WriteBegin("StateAttributes");
    bool smalltransrel = false;
    if (TGenerator<TrnSet>::mTransRel.Size() < FD_SMALLTRANSREL) {
    smalltransrel = true;
    }
    
    //for (fit = rFlagMap.begin(); fit != rFlagMap.end(); ++fit) {   
    for(fit = StateFlagTableBegin(); fit != StateFlagTableEnd(); ++fit){
    mapit = minstateindexmap.find(fit->first);
    if (mapit == minstateindexmap.end()) {
      std::cerr << "programming error: index " << fit->first
          << " could not be found in minindexmap" << std::endl;
      abort();
    }
    // write state
    if (smalltransrel) {
      std::string x1 = TGenerator<TrnSet>::StateName(fit->first);
      if (x1 != "") {
        rTw << x1;
      }
      else {
        rTw << mapit->second;
      }
    }
    else {
      rTw << mapit->second;
    }
    
    // write attribute
    rTw << (Idx)(fit->second);
    
    
      }
      rTw.WriteEnd("StateAttributes");
      rTw.Columns(oldcolumns);	
}

// virtual
// DWrite()
template<class TrnSet>
void TaGenerator<TrnSet>::DWrite(void) const {
	TokenWriter tw;
	tw.WriteBegin("aGenerator"); 
	tw.WriteBegin("Generator");
	tw << TGenerator<TrnSet>::mMyName;
	tw << "\n";
	TGenerator<TrnSet>::mAlphabet.Write(tw);
	TGenerator<TrnSet>::mStates.DWrite(tw);
	TGenerator<TrnSet>::DWriteTransRel(tw);
	TGenerator<TrnSet>::mInitStates.DWrite(tw);
	TGenerator<TrnSet>::mMarkedStates.DWrite(tw);
	tw.WriteEnd("Generator");
	tw << "\n";
	DWriteEventAttributes(tw);
	DWriteStateAttributes(tw);
	tw.WriteEnd("aGenerator");
}

// DWriteEventAttributes()
template<class TrnSet>
void TaGenerator<TrnSet>::DWriteEventAttributes(TokenWriter& rTw) const{
    FD_DG("aGenerator(" << this << ")::DWriteEventAttributes(rTw)");
	//std::map<Idx,aType> rFlagMap = mEventFlagTable.FlagMap();
	std::map<Idx,aType>::const_iterator it;
	rTw.WriteBegin("EventAttributes");
	//for(it = rFlagMap.begin(); it != rFlagMap.end(); ++it){
	for(it = EventFlagTableBegin(); it != EventFlagTableEnd(); ++it){
	    if(EventSet::StaticEventSymbolTablep()->EventName(it->first) != ""){
            rTw << EventSet::StaticEventSymbolTablep()->EventName(it->first).c_str();
            rTw << (Idx)it->second;
        }
        else{
            FD_ERR("WriteEventAttributes(" << this << "): index " << it->first 
					<< " does not have a name");
        }
	}
	rTw.WriteEnd("EventAttributes");
}
// DWriteStateAttributes()
template<class TrnSet>
void TaGenerator<TrnSet>::DWriteStateAttributes(TokenWriter& rTw) const{
    FD_DG("aGenerator(" << this << ")::DWriteStateAttributes(rTw)");
	//std::map<Idx,aType> rFlagMap = mStateFlagTable.FlagMap();
	std::map<Idx,aType>::const_iterator fit;
	std::map<Idx,Idx> minstateindexmap = TGenerator<TrnSet>::MinStateIndexMap();

	std::map<Idx,Idx>::const_iterator mapit;
    int oldcolumns = rTw.Columns();
    rTw.Columns(6);
    rTw.WriteBegin("StateAttributes");
    
    //for (fit = rFlagMap.begin(); fit != rFlagMap.end(); ++fit) {   
    for(fit = StateFlagTableBegin(); fit != StateFlagTableEnd(); ++fit){
    rTw << fit->first;
    // write attribute
    rTw << (Idx)(fit->second);
    
    
      }
      rTw.WriteEnd("StateAttributes");
      rTw.Columns(oldcolumns);	
}

// ToString()
template<class TrnSet>
std::string TaGenerator<TrnSet>::ToString() const {
	std::stringstream sstr;
	sstr << "<aGenerator>" << std::endl
		<< /*TGenerator::*/ToString() << std::endl
		<< EventAttributesToString()
		<< StateAttributesToString()
		<< "</aGenerator>" << std::endl;
	return sstr.str();
}

// EventAttributesToString()
template<class TrnSet>
std::string TaGenerator<TrnSet>::EventAttributesToString(void) const {
	std::string ret;
	std::ostringstream ostr;
	//std::map<Idx,aType> rFlagMap = mEventFlagTable.FlagMap();
	std::map<Idx,aType>::const_iterator it;
	ostr << "<EventAttributes> \n ";
	for(it = EventFlagTableBegin(); it != EventFlagTableEnd(); ++it){
	    if(EventSet::StaticEventSymbolTablep()->EventName(it->first) != ""){
            ostr << EventSet::StaticEventSymbolTablep()->EventName(it->first);
            ostr << " " << (Idx)it->second << " " ;
        }
        else{
            FD_ERR("WriteEventAttributesToString(" << this << "): index " << it->first 
					<< " does not have a name");
        }
	}
    ostr << "<EventAttributes> \n";

	ret = ostr.str();
	if (ret.size() > 0) ret.erase(ret.size()-1);
	return ret;
}

// StateAttributesToString()
template<class TrnSet>
std::string TaGenerator<TrnSet>::StateAttributesToString(void) const {
    std::string ret;
	std::ostringstream ostr;
	//std::map<Idx,aType> rFlagMap = mEventFlagTable.FlagMap();
	std::map<Idx,aType>::const_iterator fit;
	ostr << "<StateAttributes> \n ";
    for(fit = StateFlagTableBegin(); fit != StateFlagTableEnd(); ++fit){
        ostr << fit->first << " " ;
        // write attribute
        ostr << (Idx)(fit->second) << " ";
    }
    ostr << "<StateAttributes> \n";
    ret=ostr.str();
    return ret;
}

// iterator EventFlagTableBegin()
template<class TrnSet>
aFlagTable::iterator TaGenerator<TrnSet>::EventFlagTableBegin(void) {
  return mEventFlagTable.Begin();
}

// const_iterator EventFlagTableBegin()
template<class TrnSet>
aFlagTable::const_iterator TaGenerator<TrnSet>::EventFlagTableBegin(void) const{
  return mEventFlagTable.Begin();
}

// iterator EventFlagTableEnd()
template<class TrnSet>
aFlagTable::iterator TaGenerator<TrnSet>::EventFlagTableEnd(void) {
  return mEventFlagTable.End();
}

// const_iterator EventFlagTableEnd()
template<class TrnSet>
aFlagTable::const_iterator TaGenerator<TrnSet>::EventFlagTableEnd(void) const{
  return mEventFlagTable.End();
}

// iterator StateFlagTableBegin()
template<class TrnSet>
aFlagTable::iterator TaGenerator<TrnSet>::StateFlagTableBegin(void) {
  return mStateFlagTable.Begin();
}

// const_iterator StateFlagTableBegin()
template<class TrnSet>
aFlagTable::const_iterator TaGenerator<TrnSet>::StateFlagTableBegin(void) const{
  return mStateFlagTable.Begin();
}

// iterator StateFlagTableEnd()
template<class TrnSet>
aFlagTable::iterator TaGenerator<TrnSet>::StateFlagTableEnd(void) {
  return mStateFlagTable.End();
}

// const_iterator StateFlagTableEnd()
template<class TrnSet>
aFlagTable::const_iterator TaGenerator<TrnSet>::StateFlagTableEnd(void) const{
  return mStateFlagTable.End();
}

// virtual
// Clear()
template<class TrnSet>
void TaGenerator<TrnSet>::Clear(void){
    FD_DG("aGenerator(" << this << ")::Clear()");
    
    TGenerator<TrnSet>::Clear();
    mEventFlagTable.Clear();
    mStateFlagTable.Clear();
}

// virtual
// Read(pFileName)
template<class TrnSet>
void TaGenerator<TrnSet>::Read(const char* pFileName) {
  try {
    TokenReader tr(pFileName);
    Read(tr);
  }
  catch (faudes::Exception& ex) {  
    std::stringstream errstr;
    errstr << "Exception opening/reading file \"" << pFileName << "\"";
    throw Exception("TaGenerator<TrnSet>::Read", ex.What(),1); //errstr.str(), 1);
  }
}

// virtual 
// Read(rTr)
template<class TrnSet>
void TaGenerator<TrnSet>::Read(TokenReader& rTr){
    FD_DG("aGenerator(" << this << ")::Read(rTr)");
    // HELPERS:
    std::map<Idx,Idx> minstateindexmap;
    std::map<std::string,Idx> tmpstringmap;
    // ALGORITHM:
    rTr.ReadBegin("aGenerator");
    rTr.ReadBegin("Generator");
    // get the Generator "name" token
    TGenerator<TrnSet>::ReadGeneratorName(rTr);
    // read mAlphabet
    TGenerator<TrnSet>::ReadAlphabet(rTr);
    // read stateset
    TGenerator<TrnSet>::ReadStates(rTr, minstateindexmap, tmpstringmap);
    // read transrel
    TGenerator<TrnSet>::ReadTransRel(rTr, minstateindexmap, tmpstringmap);
    // read istates
    TGenerator<TrnSet>::ReadInitStates(rTr, minstateindexmap, tmpstringmap);
    // read mstates
    TGenerator<TrnSet>::ReadMarkedStates(rTr, minstateindexmap, tmpstringmap);
    // read end (synthax check only)
    rTr.ReadEnd("Generator");
    
   	ReadEventAttributes(rTr);
   	ReadStateAttributes(rTr,minstateindexmap,tmpstringmap);

   	rTr.ReadEnd("aGenerator");
}


// ReadEventAttributes(pFileName)
template<class TrnSet>
void TaGenerator<TrnSet>::ReadEventAttributes(const char* pFileName) {
	try {
        TokenReader tr(pFileName);
		ReadEventAttributes(tr);
	}
	catch (faudes::Exception&) {
		std::stringstream errstr;
		errstr << "Exception opening/reading file: no \"EventAttributes\" Section \""
        << pFileName << "\"";
		throw Exception("TaGenerator<TrnSet>::ReadEventAttributes", errstr.str(), 150);
	}
}

// ReadEventAttributes(rTr)
template<class TrnSet>
void TaGenerator<TrnSet>::ReadEventAttributes(TokenReader& rTr) {
	FD_DG("aGenerator(" << this << ")::ReadEventAttributes(" << rTr.FileName() << ")");
	Token token;
	int readState=1;
	int evIndex=0;
	aType evAttribute;
	
    try{
	    rTr.ReadBegin("EventAttributes");
    }
    catch(Exception&){
        std::stringstream errstr;
		errstr << "Exception opening/reading file: no \"EventAttributes\" Section \""
        << rTr.FileName() << "\"";
		throw Exception("TaGenerator<TrnSet>::ReadEventAttributes", errstr.str(), 150);
    }
    if(!rTr.Peek(token) ){
        std::stringstream errstr;
        errstr << "event name expected";
        throw Exception("TaGenerator<TrnSet>::ReadEventAttributes", errstr.str(), 151);
    }
    while ((rTr.Peek(token)) && (token.Type() != Token::End)){
        switch (readState) {
          // 0: just read a token first
//          case 0:
//            break;

            // 1: the ev token
          case 1: 
            rTr.Get(token);
            if (token.Type() == Token::Index) {
              std::stringstream errstr;
              errstr << "No event indeces allowed"	<< rTr.FileLine();
              throw Exception("TaGenerator<TrnSet>::ReadEventAttributes", errstr.str(), 151);
            }
            else if (token.Type() == Token::Name) {
              evIndex = TGenerator<TrnSet>::EventIndex(token.StringValue());
              if (evIndex == 0) {
                std::stringstream errstr;
                errstr << "token " << token.StringValue() << " not known in mAlphabet" 
                  << rTr.FileLine();
                throw Exception("TaGenerator<TrnSet>::ReadEventAttributes", errstr.str(), 151);
              }
            }
            break;
            // 2: the attribute token
          case 2:
            rTr.Get(token);
            if (token.Type() == Token::Index) {
              evAttribute = (int)token.IndexValue();
            }
            else {
                std::stringstream errstr;
                errstr << "attribute for event " << TGenerator<TrnSet>::EventName(evIndex) << " does not have"
                << " the correct format";
                throw Exception("TaGenerator<TrnSet>::ReadEventAttributes", errstr.str(), 151);
            }
            SetEventAttribute(evIndex,evAttribute);
            // reset readstate after x2 token
            readState = 0;
            break;
            // default: should not happen
          default:
            std::cerr << "should not happen" << std::endl;
            assert(false);
            break;
        }
        readState++;	    
    }
    // check if next expected token would have been an event token
    if (readState != 1) {
        std::stringstream errstr;
        errstr << "Wrong number of tokens in EventAttributes section " << rTr.FileLine();
        throw Exception("TaGenerator<TrnSet>::ReadEventAttributes", errstr.str(), 150);
    }

    rTr.ReadEnd("EventAttributes");
}

// ReadStateAttributes(pFileName)
template<class TrnSet>
void TaGenerator<TrnSet>::ReadStateAttributes(const char* pFileName, 
    const std::map<Idx,Idx>& rMinIndexMap,
    const std::map<std::string,Idx>& rTmpStateNameMap) {
    FD_DG("aGenerator(" << this << ")::ReadStateAttributes(" << pFileName << ")");
	try {
		TokenReader tr(pFileName);
		ReadStateAttributes(tr,rMinIndexMap,rTmpStateNameMap);
	}
	catch (faudes::Exception&) {
		std::stringstream errstr;
		errstr << "Exception opening/reading file: no \"StateAttributes\" Section \""
        << pFileName << "\"";
		throw Exception("TaGenerator<TrnSet>::ReadStateAttributes", errstr.str(), 150);
	}
}

// ReadStateAttributes(rTr)
template<class TrnSet>
void TaGenerator<TrnSet>::ReadStateAttributes(TokenReader& rTr, 
    const std::map<Idx,Idx>& rMinIndexMap,
    const std::map<std::string,Idx>& rTmpStateNameMap) {
    FD_DG("aGenerator(" << this << ")::ReadStateAttributes(" << rTr.FileName() << ")");
   	Token token;
	int readState=1;
	Idx stateIndex=0;
	aType stateAttribute;
    try{
	    rTr.ReadBegin("StateAttributes");
    }
    catch(Exception&){
        std::stringstream errstr;
		errstr << "Exception opening/reading file: no \"StateAttributes\" Section \""
        << rTr.FileName() << "\"";
		throw Exception("TaGenerator<TrnSet>::ReadStateAttributes", errstr.str(), 150);
    }
    if(!rTr.Peek(token) ){
        std::stringstream errstr;
        errstr << "state name expected";
        throw Exception("TaGenerator<TrnSet>::ReadStateAttributes", errstr.str(), 151);
    }
    while ((rTr.Peek(token)) && (token.Type() != Token::End)){
        switch (readState) {
          // 0: just read a token first
//          case 0:
//            break;

            // 1: the ev token
          case 1: 
            rTr.Get(token);
            if (token.Type() == Token::Index) {
                std::map<Idx,Idx>::const_iterator it = rMinIndexMap.find(token.IndexValue());
                if (it != rMinIndexMap.end()) {
                    stateIndex = it->second;
                }
                else {
                    std::stringstream errstr;
                    errstr << "token " << token.IndexValue() << " not found in rMinIndexMap "
                       << rTr.FileLine();
                    throw Exception("TaGenerator<TrnSet>::ReadStateAttributes", errstr.str(), 152);
                }
            }
            else if (token.Type() == Token::Name) {
                std::map<std::string,Idx>::const_iterator smit = rTmpStateNameMap.find(token.StringValue());
                if (smit != rTmpStateNameMap.end()) {
                    stateIndex = smit->second;
                }
                else {
                    std::stringstream errstr;
                    errstr << "token " << token.StringValue() << " not known in stateset " 
                      << rTr.FileLine();
                    throw Exception("TaGenerator<TrnSet>::ReadStateAttributes", errstr.str(), 152);
                }
            }
            break;
            // 2: the attribute token
          case 2:
            rTr.Get(token);
            if (token.Type() == Token::Index) {
              stateAttribute = (int)token.IndexValue();
            }
            else {
                std::stringstream errstr;
                errstr << "attribute for state " << stateIndex << " does not have"
                << " the correct format";
                throw Exception("TaGenerator<TrnSet>::ReadStateAttributes", errstr.str(), 150);
            }
            SetStateAttribute(stateIndex,stateAttribute);
            // reset readstate after x2 token
            readState = 0;
            break;
            // default: should not happen
          default:
            std::cerr << "should not happen" << std::endl;
            assert(false);
            break;
        }
        readState++;	    
    }
    // check if next expected token would have been a state token
    if (readState != 1) {
        std::stringstream errstr;
        errstr << "Wrong number of tokens in StateAttributes section " << rTr.FileLine();
        throw Exception("TaGenerator<TrnSet>::ReadStateAttributes", errstr.str(), 150);
    }
    rTr.ReadEnd("StateAttributes");
}

// EventFlagTable()
template<class TrnSet>
aFlagTable TaGenerator<TrnSet>::EventFlagTable(void) const {
	return mEventFlagTable;
}

// EventFlagTabler()
template<class TrnSet>
aFlagTable& TaGenerator<TrnSet>::EventFlagTabler(void) {
	return mEventFlagTable;
}

// StateFlagTable()
template<class TrnSet>
aFlagTable TaGenerator<TrnSet>::StateFlagTable(void) const {
	return mStateFlagTable;
}

// StateFlagTabler()
template<class TrnSet>
aFlagTable& TaGenerator<TrnSet>::StateFlagTabler(void){
	return mStateFlagTable;
}


// SetEventFlagTable(EventFlagTable);
template<class TrnSet>
void TaGenerator<TrnSet>::SetEventFlagTable(aFlagTable EventFlagTable){
FD_DG("Generator(" << this << ")::SetEventFlagTable(EventFlagTable)");
    mEventFlagTable = EventFlagTable;
}

// InsAttributedEvent(index, attribute)
template<class TrnSet>
void TaGenerator<TrnSet>::InsAttributedEvent(Idx index, aType attribute) {
	FD_DG("Generator(" << this << ")::InsAttributedEvent(" << index << "," << attribute << ")");
    TGenerator<TrnSet>::mAlphabet.InsEvent(index);
    mEventFlagTable.ClrFlags(index,0xFFFFFFFF);
	SetEventAttribute(index,attribute);
}

// InsAttributedEvent(rName, attribute)
template<class TrnSet>
Idx TaGenerator<TrnSet>::InsAttributedEvent(const std::string& rName, aType attribute) {
	FD_DG("Generator(" << this << ")::InsAttributedEvent(\"" << rName << "," << attribute << "\")");
	Idx index = TGenerator<TrnSet>::mAlphabet.InsEvent(rName);
	SetEventAttribute(index,attribute);
	return index;
}

// SetEventAttribute(index, attribute)
template<class TrnSet>
void TaGenerator<TrnSet>::SetEventAttribute(Idx index, aType attribute) {
    if(TGenerator<TrnSet>::mAlphabet.ExistsEvent(index) ){
	    mEventFlagTable.SetFlags(index, attribute);
    }
}

// SetEventAttribute(rName, attribute)
template<class TrnSet>
void TaGenerator<TrnSet>::SetEventAttribute(const std::string& rName, aType attribute) {
	SetEventAttribute(TGenerator<TrnSet>::EventIndex(rName),attribute);
}

// ClearEventAttribute(index, attribute)
template<class TrnSet>
void TaGenerator<TrnSet>::ClearEventAttribute(Idx index, aType attribute) {
	mEventFlagTable.ClrFlags(index, attribute);
}

// ClearEventAttribute(rName, attribute)
template<class TrnSet>
void TaGenerator<TrnSet>::ClearEventAttribute(const std::string& rName, aType attribute) {
	ClearEventAttribute(TGenerator<TrnSet>::EventIndex(rName),attribute);
}

// EventAttribute(index)
template<class TrnSet>
aType TaGenerator<TrnSet>::GetEventAttribute(Idx index,aType attributes, aType defaultAttribute) const {
    return mEventFlagTable.GetFlags(index,attributes,defaultAttribute);
}

// EventAttribute(rName)
template<class TrnSet>
aType TaGenerator<TrnSet>::GetEventAttribute(const std::string& rName, aType attribute, aType defaultAttribute) {
	return GetEventAttribute(TGenerator<TrnSet>::EventIndex(rName), attribute, defaultAttribute);
}

// ExistsEventAttribute(index)
template<class TrnSet>
bool TaGenerator<TrnSet>::ExistsEventAttribute(Idx index) {    
    return mEventFlagTable.ExistsIndex(index);
}

// ExistsEventAttribute(name)
template<class TrnSet>
bool TaGenerator<TrnSet>::ExistsEventAttribute(const std::string& rName) {
	return mEventFlagTable.ExistsIndex(TGenerator<TrnSet>::EventIndex(rName));
}

// SetStateFlagTable(StateFlagTable);
template<class TrnSet>
void TaGenerator<TrnSet>::SetStateFlagTable(aFlagTable StateFlagTable){
FD_DG("Generator(" << this << ")::SetStateFlagTable(StateFlagTable)");
    mStateFlagTable = StateFlagTable;
}

// InsAttributedState(index, attribute)
template<class TrnSet>
void TaGenerator<TrnSet>::InsAttributedState(Idx index, aType attribute) {
	FD_DG("Generator(" << this << ")::InsAttributedState(\"" << index << "," << attribute << "\")");
	TGenerator<TrnSet>::mStates.Insert(index);
	mStateFlagTable.ClrFlags(index,0xFFFFFFFF);
	SetStateAttribute(index,attribute);
}

// InsAttributedState(rName, attribute)
template<class TrnSet>
Idx TaGenerator<TrnSet>::InsAttributedState(const std::string& rName, aType attribute) {
	FD_DG("Generator(" << this << ")::InsAttributedState(\"" << rName << "," << attribute << "\")");
	Idx index = TGenerator<TrnSet>::mStates.InsState(rName);
	SetStateAttribute(index,attribute);
	return index;
}

// SetStateAttribute(index, attribute)
template<class TrnSet>
void TaGenerator<TrnSet>::SetStateAttribute(Idx index, aType attribute) {
    if( TGenerator<TrnSet>::mStates.ExistsState(index) ){
	    mStateFlagTable.SetFlags(index, attribute);
    }
}

// SetStateAttribute(index, attribute)
template<class TrnSet>
void TaGenerator<TrnSet>::ClearStateAttribute(Idx index, aType attribute) {
	mStateFlagTable.ClrFlags(index, attribute);
}

// StateAttribute(index)
template<class TrnSet>
aType TaGenerator<TrnSet>::GetStateAttribute(Idx index, aType attributes, aType defaultAttribute) const {
    return mStateFlagTable.GetFlags(index,attributes,defaultAttribute);
}

// ExistsStateAttribute(index)
template<class TrnSet>
bool TaGenerator<TrnSet>::ExistsStateAttribute(Idx index) {
	return mStateFlagTable.ExistsIndex(index);
}

//virtual
// operator = Klaus Schmidt 10.04.2006
template<class TrnSet>
TaGenerator<TrnSet>& TaGenerator<TrnSet>::operator = (const TaGenerator<TrnSet>& rOtherGen) {
	FD_DG("aGenerator(" << this << ")::operator = " << &rOtherGen);
	// create a copy
	rOtherGen.Copy(*this);

	return *this;
} 

//virtual
// operator = // Generator Klaus Schmidt 10.04.2006
template<class TrnSet>
TaGenerator<TrnSet>& TaGenerator<TrnSet>::operator = (const TGenerator<TrnSet>& rOtherGen) {
	FD_DG("aGenerator(" << this << ")::operator = " << &rOtherGen);
	// create a copy
	TaGenerator<TrnSet> tmpgen(rOtherGen );
	tmpgen.Copy(*this);

	return *this;
} 
		
  
// virtual
//// copy Klaus Schmidt 10.04.2006
template<class TrnSet>
void TaGenerator<TrnSet>::Copy(TaGenerator<TrnSet>& rGen) const {
	FD_DG("aGenerator(" << this << ")::Copy(rGen&)");
	// HELPERS:
	std::map<Idx,Idx> statereindexmap;

	// call base class method
	TGenerator<TrnSet>::Copy(rGen,statereindexmap);
	// copy the eventflagtable
	rGen.mEventFlagTable = mEventFlagTable;
	// copy the stateflagtable with the new state indeces
	mStateFlagTable.Copy(rGen.mStateFlagTable,statereindexmap);
} 

// NewGen()
template<class TrnSet>
TaGenerator<TrnSet> TaGenerator<TrnSet>::NewGen(void) const {
	return NewaGenerator();
}

// NewGenp()
template<class TrnSet>
TaGenerator<TrnSet>* TaGenerator<TrnSet>::NewGenp(void) const {
	return NewaGeneratorp();
}

// NewcGenerator()
template<class TrnSet>
TaGenerator<TrnSet> TaGenerator<TrnSet>::NewaGenerator(void) const {
	return aGenerator();
}

// NewcGeneratorp()
template<class TrnSet>
TaGenerator<TrnSet>* TaGenerator<TrnSet>::NewaGeneratorp(void) const {
	return new aGenerator();
}

} // namespace faudes

#define TAGENERATOR_H
#endif

