/* aparallel.cpp -- parallel composition */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "aparallel.h"

/* this Macro for cast from derived class of Generator to Generator */ 
#define CAST_GEN *static_cast<Generator*>
#define CONSTCAST_GEN *static_cast<const Generator*>

namespace faudes {

// Parallel(rGen1, rGen2, res)
void Parallel(const aGenerator& rGen1, const aGenerator& rGen2, 
    aGenerator& rResGen){
    FD_DF("aParallel(" << &rGen1 << "," << &rGen2 << ")");

    rResGen.EventFlagTabler().Clear();
    rResGen.StateFlagTabler().Clear();
    Parallel(CONSTCAST_GEN(&rGen1),CONSTCAST_GEN(&rGen2),CAST_GEN(&rResGen));
}

//InvProject(rGen1, rEvents)
void InvProject(aGenerator& rGen1, const EventSet& rEvents){
    FD_DF("InvProject(" << &rGen1 << "," << &rEvents << ")");
    aFlagTable eventTable = rGen1.EventFlagTable();
    aFlagTable stateTable = rGen1.StateFlagTable();
    try{
        InvProject(CAST_GEN(&rGen1),rEvents);
        rGen1.SetStateFlagTable(stateTable);
        rGen1.SetEventFlagTable(eventTable);
    }
    catch(Exception& ex){
        std::stringstream errstr;
        errstr << ex.What();
        throw Exception("InvProject(aGenerator,EventSet)", errstr.str(), 506);
    }  
}

//InvProject(rGen1, rGen2)
void InvProject(aGenerator& rGen1, const aGenerator& rGen2){
    FD_DF("InvProject(" << &rGen1 << "," << &rGen2 << ")");

    try{
        InvProject(rGen1,rGen2.Alphabet() );
        EventSet::const_iterator iterator;
        EventSet alphabetDifference = rGen2.Alphabet() - rGen1.Alphabet();
        for(iterator = alphabetDifference.Begin(); iterator != alphabetDifference.End();iterator++){
            rGen1.SetEventAttribute(*iterator,rGen2.GetEventAttribute(*iterator,0x11111111,0x00000000) );
        }
        for(iterator = rGen1.Alphabet().Begin(); iterator != rGen1.Alphabet().End(); iterator++){
            rGen1.SetEventAttribute(*iterator,rGen1.GetEventAttribute(*iterator,0x11111111,0x00000000) );
        }
    }
    catch(Exception& ex){
        std::stringstream errstr;
        errstr << ex.What();
        throw Exception("InvProject(aGenerator,aGenerator)", errstr.str(), 506);
    }  
}

// StateMin(rGen, rResGen)
void StateMin(aGenerator& rGen, aGenerator& rResGen) {
    FD_DF("StateMin(" << &rGen << "," << &rResGen << ")");
	std::vector<StateSet> subsets;
	std::vector<Idx> newindices;
	rResGen.EventFlagTabler().Clear();
	rResGen.StateFlagTabler().Clear();

	StateMin(rGen, rResGen, subsets, newindices);
	rResGen.SetEventFlagTable(rGen.EventFlagTable() );

}

// StateMin(rGen, rResGen, rSubsets, rNewIndices)
void StateMin(aGenerator& rGen, aGenerator& rResGen, 
		std::vector<StateSet>& rSubsets, std::vector<Idx>& rNewIndices){
	FD_DF("StateMin(" << &rGen << "," << &rResGen << "," << &rSubsets << "," << &rNewIndices << ")");
	rResGen.EventFlagTabler().Clear();
	rResGen.StateFlagTabler().Clear();
	StateMin(CAST_GEN(&rGen), CAST_GEN(&rResGen), rSubsets, rNewIndices);
	rResGen.SetEventFlagTable(rGen.EventFlagTable() );   			
}
		
// Project(rGen&, rProjectAlphabet&)
void Project(aGenerator& rGen, const EventSet& rProjectAlphabet) {
    FD_DF("Project(" << &rGen << "," << &rProjectAlphabet << ")");
    
    aFlagTable eventTable = rGen.EventFlagTable();
    rGen.StateFlagTabler().Clear();
    Project(CAST_GEN(&rGen),rProjectAlphabet);
    rGen.SetEventFlagTable(eventTable);
    EventSet alphabetDiff = rGen.Alphabet() - rProjectAlphabet;
    EventSet::iterator eit;
    for(eit = alphabetDiff.Begin(); eit != alphabetDiff.End(); ++eit){
        rGen.EventFlagTabler().DelFlag(*eit);
    }
}

// Deterministic(rGen&, rResGen&)
void Deterministic(const aGenerator& rGen, aGenerator& rResGen) {
    FD_DF("Deterministic(" << &rGen << "," << &rResGen << ")");
	// temporary vectors
	std::vector<StateSet> power_states;
	std::vector<Idx> det_states;
	rResGen.StateFlagTabler().Clear();
	Deterministic(CONSTCAST_GEN(&rGen), power_states, det_states, 
       CAST_GEN(&rResGen) );
	rResGen.SetEventFlagTable(rGen.EventFlagTable() );
}

// Deterministic(rGen&, rEntryStatesMap&, rResGen&)
void Deterministic(const aGenerator& rGen, std::map<Idx,StateSet>& rEntryStatesMap,
		aGenerator& rResGen) {
    FD_DF("Deterministic(" << &rGen << "," << &rEntryStatesMap << "," << &rResGen << ")");
	// prepare result:
	rEntryStatesMap.clear();
	// helpers:
	std::vector<StateSet> power_states;
	std::vector<Idx> det_states;
	// call Deterministic function
	Deterministic(CONSTCAST_GEN(&rGen), power_states, det_states, CAST_GEN(&rResGen) );
  // build entry states map
  std::vector<StateSet>::size_type i;
  for (i = 0; i < power_states.size(); ++i) {
    rEntryStatesMap.insert(std::pair<Idx,StateSet>(
          det_states[i], power_states[i]));
  }
  rResGen.SetEventFlagTable(rGen.EventFlagTable() );
}
    
// DetProject(rGen, rProjectAlphabet, rResGen&)
void DetProject(const aGenerator& rGen, const EventSet& rProjectAlphabet, aGenerator& rResGen) {
	FD_DF("DetProject(" << &rGen << "," << &rProjectAlphabet << "," << &rResGen << ")");
    // algorithm:
	// temporary copy of rGen
	aGenerator tmp = rGen.NewGen();
	// temporarily assign rGen to rResGen
	rResGen = rGen;
	// project rResGen with respect to rProjectAlphabet
	Project(rResGen, rProjectAlphabet);
	// put deterministic result into tmp
	Deterministic(rResGen, tmp);
	// minimize states and rewrite result to rResGen
	StateMin(tmp, rResGen);
}

// DetProject(rGen, rProjectAlphabet, rEntryStatesMap&, rResGen&)
void DetProject(const aGenerator& rGen, const EventSet& rProjectAlphabet,
		std::map<Idx,StateSet>& rEntryStatesMap, aGenerator& rResGen) {
    // create an empty generator
    aGenerator tmp = rGen.NewGen();

    // create an exact copy of rGen that will be modified
    aGenerator copyGen = aGenerator(rGen);

    // temporary entry state map
    std::map<Idx,StateSet> tmp_entrystatemap;

    // project tmp with respect to palphabet
    Project(copyGen, rProjectAlphabet); 
    // put deterministic result into tmp
    Deterministic(copyGen, tmp_entrystatemap, tmp);
    // write entry state map for minimized generator
    std::vector<StateSet> subsets;
    std::vector<Idx> newindices;
    // minimize states and rewrite result to rResGen
    StateMin(tmp, rResGen, subsets, newindices);
    // build entry state map
    std::vector<StateSet>::size_type i;
    std::map<Idx,StateSet>::iterator esmit;
    StateSet::iterator sit;
    for (i = 0; i < subsets.size(); ++i) {  
        StateSet tmpstates = rGen.NewStateSet();
        for (sit = subsets[i].Begin(); sit != subsets[i].End(); ++sit) {
            esmit = tmp_entrystatemap.find(*sit);
#ifdef FD_CHECKED
            if (esmit == tmp_entrystatemap.end()) {
                std::cerr << "algorithm error" << std::endl;
                abort();
            }
#endif
            // insert entry states in temporary StateSet
            tmpstates.Insert(esmit->second);
        }
        rEntryStatesMap.insert(std::make_pair(newindices[i], tmpstates));
    }
}


} // namespace faudes
