/* aparallel.h -- parallel composition */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef APARALLEL_H

#include "agenerator.h"
#include "parallel.h"
#include "project.h"

namespace faudes {

/**
 * Parallel composition of accessible states 
 * of aGenerators rGen1 and rGen2.
 *
 * Computes symbolic state names in rResGen, if symbolic state names are
 * enabled in both rGen1 and rGen2. The flagtables for the event properties 
 * and the state properties of the resulting generator are reset. If desired,
 * a function for assigning the event and state properties can be implemented here. 
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting parallel composition generator
 */
void Parallel(const aGenerator& rGen1, const aGenerator& rGen2, 
    aGenerator& rResGen);

/** 
 * inverse Projection of an aGenerator wrt to an event alphabet.
 * The event and state properties of the original generator are kept. There 
 * are no event properties for the additional events. 
 *
 * @param rGen1
 *   First generator
 * @param rEvents
 *   Event alphabet
 */

void InvProject(aGenerator& rGen1, const EventSet& rEvents);

/** 
 * inverse Projection of an aGenerator wrt the alphabet of another aGenerator.
 * The event properties and the state properties of the original generator 
 * are kept. The event properties of the additional events are adapted from the
 * second generator. 
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 */
 
void InvProject(aGenerator& rGen1, const aGenerator& rGen2);

/**
 * State Minimization
 * This function implements the (n*log n) set partitioning algorithm by
 * John E. Hopcroft. Given generator will be made accessible before computing
 * minimized generator. The event properties of the original generator are 
 * kept. There are no state properties set in the resulting generator. 
 *
 * @param rGen
 *   Generator
 * @param rResGen
 *   Minimized generator (result)
 *
 * @exception Exception
 *   Input automaton nondeterministic (id 505)
 */
void StateMin(aGenerator& rGen, aGenerator& rResGen);

/**
 * State Minimization
 * This function implements the (n*log n) set partitioning algorithm by
 * John E. Hopcroft. Given generator will be made accessible before computing
 * minimized generator.
 *
 * @param rGen
 *   aGenerator
 * @param rResGen
 *   Minimized aGenerator (result)
 * @param rSubsets
 *   Vector of subsets that will be constructed during running the algorithm
 *   (optional parameter)
 * @param rNewIndices
 *   Vector of new state indices corresponding to the subsets
 *   (optional parameter)
 *
 * @exception Exception
 *   Input automaton nondeterministic (id 505)
 */
void StateMin(aGenerator& rGen, aGenerator& rResGen, 
		std::vector<StateSet>& rSubsets, std::vector<Idx>& rNewIndices);
		
/**
 * Project aGenerator to alphabet rProjectAlphabet. The event properties
 * of the original generator are kept. There are no state properties set 
 * in the resulting generator. 
 *
 * @param rGen
 *   Reference to aGenerator
 * @param rProjectAlphabet
 *   Projection alphabet
 */
void Project(aGenerator& rGen, const EventSet& rProjectAlphabet);

/**
 * Make aGenerator deterministic. (function wrapper)  The event properties
 * of the original generator are kept. There are no state properties set 
 * in the resulting generator. 
 *
 * @param rGen
 *   Reference to aGenerator
 * @param rResGen
 *   Reference to resulting deterministic gaGnerator
 */
void Deterministic(const aGenerator& rGen, aGenerator& rResGen);

/**
 * Make aGenerator deterministic. (function wrapper)
 *
 * The second parameter is an empty std::map<Idx,StateSet> which holds all
 * the pairs of new states and their respective power states set. This
 * is used as a so called "entry state map" for deterministic projected
 * generators.
 *
 * @param rGen
 *   Reference to aGenerator
 * @param rEntryStatesMap
 *   Entry state map
 * @param rResGen
 *   Reference to resulting deterministic aGenerator
 */
void Deterministic(const aGenerator& rGen, std::map<Idx,StateSet>& rEntryStatesMap,
    aGenerator& rResGen);
    
/**
 * Minimized Deterministic projection. Does not modify aGenerator.
 * Calls project, determine and statemin. 
 *
 * @param rGen
 *   Reference to aGenerator
 * @param rProjectAlphabet
 *   Projection alphabet
 * @param rResGen
 *   Reference to resulting deterministic aGenerator
 */
void DetProject(const aGenerator& rGen, const EventSet& rProjectAlphabet, aGenerator& rResGen);

/**
 * Minimized Deterministic projection. Does not modify generator.
 * Calls project, determine and statemin. See functions for details.
 *
 * @param rGen
 *   Reference to cGenerator
 * @param rProjectAlphabet
 *   Projection alphabet
 * @param rEntryStatesMap
 *   Reference to entry states map, see Deterministic(..) (result)
 * @param rResGen
 *   Reference to resulting deterministic cGenerator (result)
 */
void DetProject(const aGenerator& rGen, const EventSet& rProjectAlphabet,
		std::map<Idx,StateSet>& rEntryStatesMap, aGenerator& rResGen);
}// namespace faudes

#define APARALLEL_H
#endif 

