/* baseset.h -- base class for EventSet and StateSet */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef BASESET_H

#include "definitions.h"
#include "debug.h"
#include <iostream>
#include <set>
#include <algorithm>

namespace faudes {

/**
 * Base Container for indices
 */
class BaseSet {
    public:
        /**
         * Virtual destructor
         */
        virtual ~BaseSet() {}
        /** 
         * Return name of BaseSet 
         * 
         * @return
         *   Name of BaseSet
         */
        std::string Name(void) const;
        
        /**
         * Set name of BaseSet
         *
         * @param rName
         *   BaseSet name
         */
        void Name(const std::string& rName);

        /**
         * Get Size of BaseSet
         *
         * @return 
         *   Number of indices in BaseSet
         */
        Idx Size(void) const;

        /** 
         * Check if the BaseSet ist Empty 
         *
         * @return 
         *   true if Empty
         */
        bool Empty(void) const;

        /** typedef for iterator */
        typedef std::set<Idx>::iterator iterator;

        /** typedef for const_iterator */
        typedef std::set<Idx>::const_iterator const_iterator;

        /** 
         * iterator to begin of mIndexSet
         *
         * @return 
         *   iterator
         */
        iterator Begin(void);

        /** 
         * const_iterator to begin of mIndexSet
         *
         * @return 
         *   const_iterator
         */
        const_iterator Begin(void) const;

        /** 
         * iterator to end of mIndexSet
         *
         * @return 
         *   iterator
         */
        iterator End(void);

        
        /** 
         * const_iterator to end of mIndexSet
         *
         * @return 
         *   const_iterator
         */
        const_iterator End(void) const;

        /**
         * Set Union. Result is current set.
         *
         * @param otherset
         *   Other BaseSet
         */
        void SetUnion(const BaseSet& otherset);

        /**
         * Set Intersection. Result is current set.
         *
         * @param otherset
         *   Other BaseSet
         */
        void SetIntersection(const BaseSet& otherset);

        /** Test for equality */
        bool operator == (const BaseSet& otherset) const;

        /** Test for unequality */
        bool operator != (const BaseSet& otherset) const;

        /** Test for subset */
        bool operator <= (const BaseSet& otherset) const;

        /** Test for superset */
        bool operator >= (const BaseSet& otherset) const;

        /** Order for sorting containers of BaseSet */
        bool operator < (const BaseSet& otherset) const;

        /**
        * Compute a Idx type signature for a StateSet. The value is computed by 
        * summing up the product of every index with an integer starting at
        * 1 (for the first index) and ending at Size() (for the last index).
        *
        * @return
        *   Idx type set signature
        */
        Idx Signature(void) const;

    protected:
        /** Name of BaseSet */
        std::string mMyName;

        /** Stored indices */
        std::set<Idx> mIndexSet;
};

} // namespace faudes

#define BASESET_H
#endif 
