/* cgenerator.cpp - controllable Generator */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


/* 
   + minor modifications due to tokenreader update
   + try to mimique old behaviour; not sure with exceptions
 
   tmoor, 20060413; GLPL 2.1 
*/



#include "cgenerator.h"

namespace faudes {

// cGenerator::cGenerator(void)
cGenerator::cGenerator(void) : aGenerator() {
	FD_DG("cGenerator(" << this << ")::cGenerator()");
}

// cGenerator::cGenerator(rOtherGen)
cGenerator::cGenerator(const Generator& rOtherGen) 
		: aGenerator(rOtherGen) {
	FD_DG("cGenerator(" << this << ")::cGenerator(" << &rOtherGen << ")");
}

// cGenerator::cGenerator(rOtherGen)
cGenerator::cGenerator(const cGenerator& rOtherGen) 
		: aGenerator(rOtherGen) {
	FD_DG("cGenerator(" << this << ")::cGenerator(" << &rOtherGen << ")");
}
	
// cGenerator::cGenerator(pFileName)
cGenerator::cGenerator(const char* pFileName) 
		: aGenerator() {
	FD_DG("cGenerator(" << this << ")::cGenerator(\"" << pFileName << "\")");
	Read(pFileName);
}

// virtual
// Write()
void cGenerator::Write(void) {
	TokenWriter tw;
	Write(tw);
}

// virtual  
// Write(file, openmode=default)
void cGenerator::Write(const char* pFileName, std::ios::openmode openmode) {
	try {
		TokenWriter tw(pFileName, openmode);
		Write(tw);
	}
	catch (faudes::Exception&) {
		std::stringstream errstr;
		errstr << "Exception opening/writing file \"" << pFileName << "\"";
		throw Exception("cGenerator::write", errstr.str(), 2);
	}
}

// virtual
// Write(tw)
void cGenerator::Write(TokenWriter& rTw) {
	std::map<Idx,Idx> minstateindexmap = MinStateIndexMap();
	// write generator
	rTw.WriteBegin("cGenerator");
	rTw.WriteBegin("Generator");
	rTw << mMyName;
	rTw << "\n";
	mAlphabet.Write(rTw);
	mStates.Write(rTw, minstateindexmap, true);
	WriteTransRel(rTw, minstateindexmap);
	mInitStates.Write(rTw, minstateindexmap, false);
	mMarkedStates.Write(rTw, minstateindexmap, false);
	rTw.WriteEnd("Generator");
	rTw.Write("\n");
	rTw.Write("\n");
	WriteControllableEvents(rTw);
	WriteObservableEvents(rTw);
	rTw.WriteEnd("cGenerator");
}

// WriteControllableEvents(rTw)
void cGenerator::WriteControllableEvents(TokenWriter& rTw) {
	EventSet calphabet = ControllableEvents();
	calphabet.Write(rTw);
}

// WriteObservableEvents(rTw)
void cGenerator::WriteObservableEvents(TokenWriter& rTw) {
	EventSet oalphabet = ObservableEvents();
	oalphabet.Write(rTw);
}

// virtual
// DWrite()
void cGenerator::DWrite(void) const {
	TokenWriter tw;
	tw.WriteBegin("cGenerator");
	tw.WriteBegin("Generator");
	tw << mMyName;
	tw << "\n";
	mAlphabet.Write(tw);
	mStates.DWrite(tw);
	DWriteTransRel(tw);
	mInitStates.DWrite(tw);
	mMarkedStates.DWrite(tw);
	tw.WriteEnd("Generator");
	tw << "\n";
	DWriteControllableEvents(tw);
	DWriteObservableEvents(tw);
	tw.WriteEnd("cGenerator");
}

// DWriteControllableEvents()
void cGenerator::DWriteControllableEvents(TokenWriter& tw) const {
	EventSet calphabet = ControllableEvents();
	calphabet.Write(tw);
}

// DWriteControllableEvents()
void cGenerator::DWriteObservableEvents(TokenWriter& tw) const {
	EventSet oalphabet = ObservableEvents();
	oalphabet.Write(tw);
}

// ToString()
std::string cGenerator::ToString() const {
	std::stringstream sstr;
	sstr << "<cGenerator>" 
		<< Generator::ToString() << std::endl
		<< std::endl
		<< "<Controllable>" << std::endl
		<< ControllableEvents().ToString() << std::endl
		<< "</Controllable>" << std::endl
		<< "<Observable>" << std::endl
		<< ObservableEvents().ToString() << std::endl
		<< "</Observable>" << std::endl
		<< "</cGenerator>";
	return sstr.str();
}

// virtual
// Read(pFileName)
void cGenerator::Read(const char* pFileName) {
  try {
    TokenReader tr(pFileName);
    Read(tr);
  }
  catch (faudes::Exception&) {  
    std::stringstream errstr;
    errstr << "Exception opening/reading file \"" << pFileName << "\"";
    throw Exception("cGenerator::Read", errstr.str(), 1);
  }
}

// virtual
// Read(rTr)
// this function is robust with respect to wrong formating. 
// Generators and aGenerators can be read
void cGenerator::Read(TokenReader& rTr) {
	FD_DG("cGenerator(" << this << ")::Read(rTr)");
	bool rewrite = false; // this value indicates if the generator has to be rewritten to get 
	// the correct format. It is just needed for the transition phase to the new format. 
	try{
        rTr.ReadBegin("cGenerator");  
	}
	catch(faudes::Exception&){// before change of format
	    std::cerr << "could not read cGenerator begin token" << std::endl;
        rTr.Rewind();
        rewrite = true;
	}

	Generator::Read(rTr);

    try{
        ReadControllableEvents(rTr);
        try{
            ReadObservableEvents(rTr);
        }
        catch(Exception&){
            std::cerr << "could not read observability properties token" << std::endl;
            rewrite = true;
        }
    }
    catch(faudes::Exception&){
        std::cerr << "could not read controllable events" << std::endl;
        rewrite = false;
    }
	   	       
	try{
	    rTr.ReadEnd("cGenerator");
	}
	catch(faudes::Exception&){// before change of format
	    std::cerr << "could not read cGenerator end token" << std::endl;
	}
	if(rewrite){
        this->Write(rTr.FileName().c_str() );
	    std::cerr << "generator rewritten in the correct format" << std::endl;
	}
}

// ReadControllableEvents(pFileName)
void cGenerator::ReadControllableEvents(const char* pFileName) {
    try{
        TokenReader tr(pFileName);
        ReadControllableEvents(tr);
    }
	catch (faudes::Exception&) {
		std::stringstream errstr;
		errstr << "Exception opening/reading file: no \"Controllable\" Section \""
        << pFileName << "\"";
		throw Exception("cGenerator::ReadControllableEvents", errstr.str(), 106);
	}
}

// ReadControllableEvents(rTr)
void cGenerator::ReadControllableEvents(TokenReader& rTr) {

	rTr.SeekBegin("Controllable");

	FD_DG("cGenerator(" << this << ")::ReadControllableEvents(" << rTr.FileName() << ")");
	EventSet calphabet = EventSet();
	EventSet ualphabet = EventSet();

	calphabet.Read(rTr,"Controllable");
	ualphabet = mAlphabet - calphabet;
	EventSet::iterator it;
	// check controllable events only
	for (it = calphabet.Begin(); it != calphabet.End(); ++it) {
	    FD_DG("controllable event: " << *it);
		SetControllable(*it);
	}
	// check uncontrollable events and set
	for (it = ualphabet.Begin(); it != ualphabet.End(); ++it) {
	    FD_DG("uncontrollable event: " << *it);
		SetUncontrollable(*it);
	}
}

// InsControllableEvent(index)
void cGenerator::InsControllableEvent(Idx index) {
	FD_DG("Generator(" << this << ")::InsControllableEvent(" << index << ")");
	InsAttributedEvent(index,mFlagEventControllable | mFlagEventObservableDefault);
}

// InsControllableEvent(rName)
Idx cGenerator::InsControllableEvent(const std::string& rName) {
	FD_DG("Generator(" << this << ")::InsControllableEvent(\"" << rName << "\")");
    Idx index = InsAttributedEvent(rName,mFlagEventControllable | mFlagEventObservableDefault);
	return index;
}

// InsUncontrollableEvent(index)
void cGenerator::InsUncontrollableEvent(Idx index) {
	FD_DG("Generator(" << this << ")::InsUncontrollableEvent(" << index << ")");
	InsAttributedEvent(index,mFlagEventUncontrollable & mFlagEventObservableDefault);
}

// InsUncontrollableEvent(rName)
Idx cGenerator::InsUncontrollableEvent(const std::string& rName) {
	FD_DG("Generator(" << this << ")::InsUncontrollableEvent(\"" << rName << "\")");
	Idx index = InsAttributedEvent(rName,mFlagEventUncontrollable & mFlagEventObservableDefault);
	return index;
}

// SetControllable(index)
void cGenerator::SetControllable(Idx index) {
    SetEventAttribute(index,mFlagEventControllable);
}

// SetControllable(rName)
void cGenerator::SetControllable(const std::string& rName) {
	SetControllable(EventIndex(rName));
}

// SetControllable(rEvents)
void cGenerator::SetControllable(const EventSet& rEvents) {
    EventSet::const_iterator eit;
    for(eit = rEvents.Begin(); eit!= rEvents.End(); ++eit){
	    SetControllable(*eit);
    }
}

// SetUncontrollable(index)
void cGenerator::SetUncontrollable(Idx index) {
    ClearEventAttribute(index,mFlagEventControllable);
}

// SetUncontrollable(rName)
void cGenerator::SetUncontrollable(const std::string& rName) {
	SetUncontrollable(EventIndex(rName));
}

// SetUncontrollable(rEvents)
void cGenerator::SetUncontrollable(const EventSet& rEvents) {
    EventSet::const_iterator eit;
    for(eit = rEvents.Begin(); eit!= rEvents.End(); ++eit){
	    SetUncontrollable(*eit);
    }
}

// Controllable(index)
bool cGenerator::Controllable(Idx index) const {
    if(GetEventAttribute(index, mFlagEventControllable, mFlagEventControllableDefault 
        == mFlagEventControllable) )
        return true;
    else
        return false;
}

// Controllable(rName)
bool cGenerator::Controllable(const std::string& rName) {
	return Controllable(EventIndex(rName));
}

// ControllableEvents()
EventSet cGenerator::ControllableEvents(void) const {
	EventSet calphabet = NewEventSet();
	calphabet.Name("Controllable");
	EventSet::const_iterator it;
	for (it = AlphabetBegin(); it != AlphabetEnd(); ++it) {
		if (Controllable(*it)) {
			calphabet.InsEvent(*it);
		}
	}
	return calphabet;
}

// UncontrollableEvents()
EventSet cGenerator::UncontrollableEvents(void) const {
	EventSet ualphabet = NewEventSet();
	EventSet::const_iterator it;
	for (it = AlphabetBegin(); it != AlphabetEnd(); ++it) {
		if (! Controllable(*it)) {
			ualphabet.InsEvent(*it);
		}
	}
	return ualphabet;
}

// ReadObservableEvents(pFileName)
void cGenerator::ReadObservableEvents(const char* pFileName) {
	try {
		TokenReader tr(pFileName);
		ReadObservableEvents(tr);
	}
	catch (std::ios::failure&) {
		std::stringstream errstr;
		errstr << "Exception opening/reading file: no \"Observable\" Section \""
        << pFileName << "\"";
		throw Exception("cGenerator::ReadObservableEvents", errstr.str(), 106);
	}
}

// ReadObservableEvents(rTr)
void cGenerator::ReadObservableEvents(TokenReader& rTr) {

	rTr.SeekBegin("Observable");

	FD_DG("cGenerator(" << this << ")::ReadObservableEvents(" << rTr.FileName() << ")");
	EventSet oalphabet = EventSet();
	EventSet ualphabet = EventSet();

	oalphabet.Read(rTr,"Observable");
	ualphabet = mAlphabet - oalphabet;
	EventSet::iterator it;
	// check observable events only
	for (it = oalphabet.Begin(); it != oalphabet.End(); ++it) {
	    FD_DG(std::cerr << "observable event: " << *it << std::endl);
		SetObservable(*it);
	}
	// check unobservable events and set
	for (it = ualphabet.Begin(); it != ualphabet.End(); ++it) {
	    FD_DG(std::cerr << "unobservable event: " << *it << std::endl);
		SetUnobservable(*it);
	}
}

// InsObservableEvent(index)
void cGenerator::InsObservableEvent(Idx index) {
	FD_DG("Generator(" << this << ")::InsObservableEvent(" << index << ")");
	InsAttributedEvent(index,mFlagEventObservable | mFlagEventControllableDefault);
}

// InsObservableEvent(rName)
Idx cGenerator::InsObservableEvent(const std::string& rName) {
	FD_DG("Generator(" << this << ")::InsObservableEvent(\"" << rName << "\")");
    Idx index = InsAttributedEvent(rName,mFlagEventObservable | mFlagEventControllableDefault);
	return index;
}

// InsUnobservableEvent(index)
void cGenerator::InsUnobservableEvent(Idx index) {
	FD_DG("Generator(" << this << ")::InsUnobservableEvent(" << index << ")");
	InsAttributedEvent(index,mFlagEventUnobservable  | mFlagEventControllableDefault);
}

// InsUnobservableEvent(rName)
Idx cGenerator::InsUnobservableEvent(const std::string& rName) {
	FD_DG("Generator(" << this << ")::InsUnobservableEvent(\"" << rName << "\")");
	Idx index = InsAttributedEvent(rName,mFlagEventUnobservable & mFlagEventControllableDefault);
	return index;
}

// SetObservable(index)
void cGenerator::SetObservable(Idx index) {
    SetEventAttribute(index,mFlagEventObservable);
}

// SetObservable(rName)
void cGenerator::SetObservable(const std::string& rName) {
	SetObservable(EventIndex(rName));
}

// SetObservable(rEvents)
void cGenerator::SetObservable(const EventSet& rEvents) {
    EventSet::const_iterator eit;
    for(eit = rEvents.Begin(); eit!= rEvents.End();++eit){
        SetObservable(*eit);
    }
}

// SetUnobservable(index)
void cGenerator::SetUnobservable(Idx index) {
    ClearEventAttribute(index,mFlagEventObservable);
}

// SetUnobservable(rName)
void cGenerator::SetUnobservable(const std::string& rName) {
	SetUnobservable(EventIndex(rName));
}

// SetUnobservable(rEvents)
void cGenerator::SetUnobservable(const EventSet& rEvents) {
    EventSet::const_iterator eit;
    for(eit = rEvents.Begin(); eit!= rEvents.End();++eit){
        SetUnobservable(*eit);
    }
}

// Observable(index)
bool cGenerator::Observable(Idx index) const {
    if(GetEventAttribute(index, mFlagEventObservable, mFlagEventObservableDefault 
        == mFlagEventObservable) )
        return true;
    else
        return false;
}

// Observable(rName)
bool cGenerator::Observable(const std::string& rName) {
	return Observable(EventIndex(rName));
}

// ObservableEvents()
EventSet cGenerator::ObservableEvents(void) const {
	EventSet oalphabet = NewEventSet();
	oalphabet.Name("Observable");
	EventSet::const_iterator it;
	for (it = AlphabetBegin(); it != AlphabetEnd(); ++it) {
		if (Observable(*it)) {
			oalphabet.InsEvent(*it);
		}
	}
	return oalphabet;
}

// UnobservableEvents()
EventSet cGenerator::UnobservableEvents(void) const {
	EventSet ualphabet = NewEventSet();
	EventSet::const_iterator it;
	for (it = AlphabetBegin(); it != AlphabetEnd(); ++it) {
		if (! Observable(*it)) {
			ualphabet.InsEvent(*it);
		}
	}
	return ualphabet;
}

// NewGen()
cGenerator cGenerator::NewGen(void) const {
	return NewcGenerator();
}

// NewGenp()
cGenerator* cGenerator::NewGenp(void) const {
	return NewcGeneratorp();
}

// NewcGenerator()
cGenerator cGenerator::NewcGenerator(void) const {
	return cGenerator();
}

// NewcGeneratorp()
cGenerator* cGenerator::NewcGeneratorp(void) const {
	return new cGenerator();
}
} // namespace faudes
