/* cgenerator.h - controllable Generator */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef CGENERATOR_H

#include "definitions.h"
#include "debug.h"
#include "agenerator.h"

namespace faudes {


/**
 * Modell of the common five-tupel consisting of alphabet, stateset,
 * transition relation, initial states, marked states with events marked
 * as controllable / uncontrollable
 *
 * File format of cGenerator is compatible to Generator but differs in the 
 * controllable events section. The controllable events are labeled 
 * "Controllable" and should have their place within a 
 * "EventProperties" section below the "Generator" entry.
 *
 * InsEvent and InsEvents call the Generator class function and do not set
 * the control property of the event. This can be checked with the 
 * ExistsControllableFlag cGenerator method.
 */
class cGenerator : public aGenerator {    
    public:
        /**
         * Creates an emtpy cGenerator object with static eventtable, statetable
         * and flagtable.
         */
        cGenerator(void);

        /** 
         * cGenerator from a Generator. Copy constructor 
         *
         * @param rOtherGen
         */
        cGenerator(const Generator& rOtherGen);
        
        /** 
         * Generator from a cGenerator. Copy constructor 
         *
         * @param rOtherGen
         */
        cGenerator(const cGenerator& rOtherGen);

        /**
         * Creates a cGenerator from file with static eventtable, statetable
         * and flagtable
         *
         * @param pFileName
         *    Filename
         *
         * @exception Exception
         *   If opening/reading fails an Exception object is thrown (id 2)
         *   If an uncontrollable event is already flagged controllable (id 100)
         */
        cGenerator(const char* pFileName);

        /**
         * Write cgenerator to console
         */
        virtual void Write(void);

        /**
         * Write cgenerator to file
         *
         * @param pFileName
         *   Filename to write
         * @param openmode
         *   ios::openmode (optional)
         *   Default is truncate file.
         *
         * @exception Exception
         *   If opening/writing fails an Exception object is thrown (id 2)
         */
        virtual void Write(const char* pFileName, 
                std::ios::openmode openmode = std::ios::out|std::ios::trunc);

        /**
         * Write cgenerator to TokenWriter
         *
         * @param rTw
         *   Reference to TokenWriter
         *
         * @exception ios::failure
         *   Thrown on i/o error.
         */
        virtual void Write(TokenWriter& rTw);

        /**
         * Write controllable alphabet to TokenWriter
         *
         * @param rTw
         *   Reference to TokenWriter
         *
         * @exception ios::failure
         *   Thrown on i/o error.
         */
        void WriteControllableEvents(TokenWriter& rTw);
        
        /**
         * Write observable alphabet to TokenWriter
         *
         * @param rTw
         *   Reference to TokenWriter
         *
         * @exception ios::failure
         *   Thrown on i/o error.
         */
        void WriteObservableEvents(TokenWriter& rTw);

        /**
         * (Debug-) write cgenerator to console.
         * This is just an alias function for Write()
         */
        virtual void DWrite(void) const;

        /**
         * (Debug-) write controllable alphabet to console
         * 
         * @param rTw
         *   Reference to TokenWriter
         *
         * @exception ios::failure
         *   Thrown on i/o error.
         */
        void DWriteControllableEvents(TokenWriter& rTw) const;

        /**
         * (Debug-) write observable alphabet to console
         * 
         * @param rTw
         *   Reference to TokenWriter
         *
         * @exception ios::failure
         *   Thrown on i/o error.
         */
        void DWriteObservableEvents(TokenWriter& rTw) const;
        
        /**
         * Create a string similar to the output of Write()
         *
         * @return
         *   std::string
         */
        std::string ToString() const;

        /**
         * Read cGenerator from file
         *
         * @param pFileName
         *   Filename containing generator definition
         *
         * @exception Exception
         *   If opening/reading fails an Exception object is thrown (id 1)
         *   Exception opening/reading controllable events in file (id 6)
         *   If an uncontrollable event is already flagged controllable (id 100)
         */
        virtual void Read(const char* pFileName);
        
        /**
         * Read cGenerator from TokenReader
         *
         * @param rTr
         *   Reference to TokenReader
         *
         * @exception ios::failure
         *   Thrown on i/o error.
         * @exception Exception
         *   Thrown on file format errors (id 50, 51)
         */
        virtual void Read(TokenReader& rTr);

        /**
         * Read controllable alphabet from file
         *
         * @param pFileName
         *   Filename containing controllale alphabet 
         *   "<Controllable>...</Controllable>"
         *
         * @exception Exception
         *   If opening/reading fails an Exception object is thrown (id 1)
         *   Exception opening/reading controllable events in file (id 6)
         *   If an uncontrollable event is already flagged controllable (id 100)
         */
        void ReadControllableEvents(const char* pFileName);

        /**
         * Read controllable alphabet from TokenReader
         *
         * @param rTr
         *   Reference to TokenReader
         *
         * @exception ios::failure
         *   Thrown on i/o error.
         * @exception Exception
         *   Thrown on file format errors (51)

         */
        void ReadControllableEvents(TokenReader& rTr);
    
        /**
         * Add an existing controllable event to generator. 
         * An entry in the global eventtable will be made.
         *
         * @param index
         *   Event index
         */
        void InsControllableEvent(Idx index);

        /**
         * Add new named controllable event to generator. 
         * An entry in the global eventtable will be made if event is new.
         *
         * @param rName
         *   Name of the event to add
         *
         * @return 
         *   New global unique index
         */
        Idx InsControllableEvent(const std::string& rName);
    
        /**
         * Add an existing uncontrollable event to generator. 
         * An entry in the global eventtable will be made.
         *
         * @param index
         *   Event index
         */
        void InsUncontrollableEvent(Idx index);

        /**
         * Add new named uncontrollable event to generator. 
         * An entry in the global eventtable will be made if event is new.
         *
         * @param rName
         *   Name of the event to add
         *
         * @return 
         *   New global unique index
         */
        Idx InsUncontrollableEvent(const std::string& rName);
        
        /**
         * Mark event controllable (by index)
         *
         * @param index
         *   Event index
         */
        void SetControllable(Idx index);

        /** 
         * Mark event controllable (by name)
         *
         * @param rName
         *   Event name    
         */
        void SetControllable(const std::string& rName);

        /** 
         * Mark set of events controllable (by index)
         *
         * @param rEvents
         *   EventSet
         */
        void SetControllable(const EventSet& rEvents);
        
        /**
         * Mark event uncontrollable (by index)
         * 
         * @param index
         *   Event index
         */
        void SetUncontrollable(Idx index);

        /**
         * Mark event uncontrollable (by name)
         * 
         * @param rName
         *   Event name
         */
        void SetUncontrollable(const std::string& rName);

        /** 
         * Mark set of events uncontrollable (by index)
         *
         * @param rEvents
         *   EventSet
         */
        void SetUncontrollable(const EventSet& rEvents);
          
        /**
         * Is event controllable (by index)
         *
         * @param index
         *   Event index
         *
         * @return
         *   True / false
         */
        bool Controllable(Idx index) const;

        /**
         * Is event controllable (by name)
         *
         * @param rName
         *   Event name
         *
         * @return
         *   True / false
         */
        bool Controllable(const std::string& rName);

        /** 
         * Get EventSet with controllable events
         *
         * @return
         *   EventSet of controllable events
         */
        EventSet ControllableEvents(void) const;

        /**
         * Get EventSet with uncontrollable events
         *
         * @return 
         *   EventSet of uncontrollable events
         */
        EventSet UncontrollableEvents(void) const;

        /**
         * Read observable alphabet from file
         *
         * @param pFileName
         *   Filename containing controllale alphabet 
         *   "<Observable>...</Observable>"
         *
         * @exception Exception
         *   If opening/reading fails an Exception object is thrown (id 1)
         *   Exception opening/reading controllable events in file (id 6)
         */
        void ReadObservableEvents(const char* pFileName);

        /**
         * Read observable alphabet from TokenReader
         *
         * @param rTr
         *   Reference to TokenReader
         *
         * @exception ios::failure
         *   Thrown on i/o error.
         * @exception Exception
         *   Thrown on file format errors (51)
         */
        void ReadObservableEvents(TokenReader& rTr);
    
        /**
         * Add an existing observable event to generator. 
         * An entry in the global eventtable will be made.
         *
         * @param index
         *   Event index
         */
        void InsObservableEvent(Idx index);

        /**
         * Add new named observable event to generator. 
         * An entry in the global eventtable will be made if event is new.
         *
         * @param rName
         *   Name of the event to add
         *
         * @return 
         *   New global unique index
         */
        Idx InsObservableEvent(const std::string& rName);
    
        /**
         * Add an existing unobservable event to generator. 
         * An entry in the global eventtable will be made.
         *
         * @param index
         *   Event index
         */
        void InsUnobservableEvent(Idx index);

        /**
         * Add new named unobservable event to generator. 
         * An entry in the global eventtable will be made if event is new.
         *
         * @param rName
         *   Name of the event to add
         *
         * @return 
         *   New global unique index
         */
        Idx InsUnobservableEvent(const std::string& rName);
        
        /**
         * Mark event observable (by index)
         *
         * @param index
         *   Event index
         */
        void SetObservable(Idx index);

        /** 
         * Mark event observable (by name)
         *
         * @param rName
         *   Event name    
         */
        void SetObservable(const std::string& rName);

        /**
         * Mark set of events observable (by index)
         *
         * @param index
         *   EventSet
         */
        void SetObservable(const EventSet& rEvents);
        
        /**
         * Mark event unobservable (by index)
         * 
         * @param index
         *   Event index
         */
        void SetUnobservable(Idx index);

        /**
         * Mark event unobservable (by name)
         * 
         * @param rName
         *   Event name
         */
        void SetUnobservable(const std::string& rName);

        /**
         * Mark set of events unobservable (by index)
         *
         * @param index
         *   EventSet
         */
        void SetUnobservable(const EventSet& rEvents);
          
        /**
         * Is event observable (by index)
         *
         * @param index
         *   Event index
         *
         * @return
         *   True / false
         */
        bool Observable(Idx index) const;

        /**
         * Is event observable (by name)
         *
         * @param rName
         *   Event name
         *
         * @return
         *   True / false
         */
        bool Observable(const std::string& rName);

        /** 
         * Get EventSet with observable events
         *
         * @return
         *   EventSet of controllable events
         */
        EventSet ObservableEvents(void) const;

        /**
         * Get EventSet with unobservable events
         *
         * @return 
         *   EventSet of uncontrollable events
         */
        EventSet UnobservableEvents(void) const;

        /**
         * Create Empty cGenerator with proper tables set on stack
         * Calls NewcGenerator()
         *
         * @return
         *   New cGenerator
         */
        cGenerator NewGen(void) const;

        /**
         * Create Empty cGenerator with proper tables set on heap
         * Calls NewcGeneratorp()
         *
         * @return
         *   New cGenerator*
         */
        cGenerator* NewGenp(void) const;

        /**
         * Create Empty cGenerator with proper tables set on stack
         *
         * @return
         *   New cGenerator
         */
        cGenerator NewcGenerator(void) const;

        /**
         * Create Empty cGenerator with proper tables set on heap
         *
         * @return
         *   New cGenerator*
         */
        cGenerator* NewcGeneratorp(void) const;

    private:

    protected:
        /** Controllability Bit */
        static const int mFlagEventControllable = 0x00000001;
        
        /** Uncontrollability Bit */
        static const int mFlagEventUncontrollable = 0x11111110;

        /** Default Bit for controllability */
        static const int mFlagEventControllableDefault = 0x00000000;
        
        /** Observability Bit */
        static const int mFlagEventObservable = 0x00000010;
        
        /** Unobservability Bit */
        static const int mFlagEventUnobservable = 0x11111101;
        
        /** Default Bit for observability */
        static const int mFlagEventObservableDefault = 0x00000010;

};

} // namespace faudes

#define CGENERATOR_H
#endif

