/* cmap.h - state name composition for parallel compositions  */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef COMPOSITIONMAP_H

#include "definitions.h"
#include "debug.h"
#include "generator.h"
#include "symboltable.h"

namespace faudes {

/** 
 * (Parallel) Composition Map
 * This class stores "<Idx,Idx> -> Idx" rcmap entries created in parallel
 * compositions in "Idx -> <Idx,Idx>" order to resolve the way a state was 
 * created with parallel compositions.
 */
class CompositionMap {
    public:
        /**
         * Constructor for StateSet with static StateSymbolTable
         */
        CompositionMap(void);

        /** typedef for iterator */
        typedef std::map< Idx, std::pair<Idx,Idx> >::iterator iterator;
    
        /** Begin iterator */
        iterator Begin(void);
        
        /** end iterator */
        iterator End(void);

        /**
         * Add the entries of a reverse composition map (rcmap) created in the
         * parallel composition function.
         *
         * @param rReverseCompositionMap
         *   Reverse composition map
         */
        void InsReverseMap(const std::map< std::pair<Idx,Idx>, Idx >& rReverseCompositionMap);

        /**
         * Recursively resolve parallel state composition and return corresponding 
         * string.
         *
         * @param index
         *   State index to resolve
         *
         * @return
         *   A "((1|state2)|(state3|3))|5)" like string showing the composition
         *   history of the state.
         */
        std::string StateName(Idx index) const;

        /**
         * Compute State names for all States of a Generator (calls StateName(index))
         *
         * @param rGen
         *   Reference to Generator
         */
        void CompStateNames(Generator& rGen) const;

        /**
         * Recursively delete index and the indices the stated was composed of.
         *
         * @param index
         *   State index
         *
         * @return
         *   True if state was found in composition map
         */
        bool DelState(Idx index);

        /**
         * Recursively delete all state composition indices of given generator
         *
         * @param rGen
         *   Generator
         *
         * @return
         *   True if all states where found in composition map
         */
        bool DelStates(const Generator& rGen);
        
    private:
        /** The real composition map */
        std::map< Idx, std::pair<Idx,Idx> > mCompositionMap;

        /** Pointer to statesymboltable */
        StateSymbolTable* mpStateSymbolTable;
};

} // namespace faudes

#define COMPOSITIONMAP_H
#endif 
