/* cparallel.h -- parallel composition */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef CPARALLEL_H

#include "cgenerator.h"
#include "parallel.h"
#include "project.h"

namespace faudes {

/**
 * Parallel composition of accessible states 
 * of cGenerators rGen1 and rGen2.
 *
 * Computes symbolic state names in rResGen, if symbolic state names are
 * enabled in both rGen1 and rGen2. It also takes care of the event
 * attributes (controllability) of the resulting generator
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting parallel composition generator
 */
void Parallel(const cGenerator& rGen1, const cGenerator& rGen2, 
    cGenerator& rResGen);

/**
 * Parallel composition of accessible states 
 * of cGenerators rGen1 and rGen2.
 *
 * Computes symbolic state names in rResGen, if symbolic state names are
 * enabled in both rGen1 and rGen2. It also takes care of the event
 * attributes (controllability) of the resulting generator
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rReverseCompositionMap
 *   reverse composition map 
 * @param rResGen
 *   Reference to resulting parallel composition generator
 */
void Parallel(const cGenerator& rGen1, const cGenerator& rGen2,
    std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
    cGenerator& rResGen);
      
/** 
 * Computation of the composed EventFlagTable of the resulting generator
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 *
 * @return 
 *   resulting EventFlagTable
 */
 
void cComposedEventFlagTable(const cGenerator& rGen1, const cGenerator& rGen2, aFlagTable& rEventFlagTable);

/** 
 * inverse Projection of a cGenerator wrt the alphabet of another cGenerator.
 * The event properties of the second cGenerator are adapted for the additional events.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 */
 
void InvProject(cGenerator& rGen1, const cGenerator& rGen2);

/** 
 * inverse Projection of a cGenerator wrt an event alphabet. Optionally,
 * the sets of controllable and/or observable events can be given. If
 * an event property is not given, it is set to "uncontrollable" or 
 * "unobservable" by default.
 *
 * @param rGen1
 *   First generator
 * @param rAlphabet
 *   event alphabet
 * @param rControllable
 *   controllable events
 * @param rObservable
 *   observable events
 */
 
void InvProject(cGenerator& rGen1, const EventSet& rAlphabet, 
    const EventSet rControllable = EventSet(), const EventSet rObservable = EventSet() );
}// namespace faudes

#define CPARALLEL_H
#endif 

