/* eventset.cpp -- container for storing events (alphabet) */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "eventset.h"

namespace faudes {

// empty constructor
EventSet::EventSet(void) {
    FD_DC("EventSet(" << this << ")::EventSet()");
    mpEventSymbolTable = &msEventSymbolTable;
    Name("Alphabet");
}

// read file constructor
EventSet::EventSet(const char* rFilename, const std::string& rLabel) {
    FD_DC("EventSet(" << this << ")::EventSet(" << rFilename << ")");
    mpEventSymbolTable = &msEventSymbolTable;
    Read(rFilename, rLabel);
}

// NewEventSet()
EventSet EventSet::NewEventSet(void) const {
    return EventSet();
}

// NewEventSetp()
EventSet* EventSet::NewEventSetp(void) const {
    return new EventSet();
}

// EventSet::mpEventSymbolTable
EventSymbolTable EventSet::msEventSymbolTable;

// EventSymbolTable()
EventSymbolTable* EventSet::EventSymbolTablep(void) {
    return mpEventSymbolTable;
}


// EventSet::StaticEventSymbolTable()
EventSymbolTable* EventSet::StaticEventSymbolTablep(void) {
    return &msEventSymbolTable;
}


// Write()
void EventSet::Write(void) const {
    TokenWriter tw;
    Write(tw);
}

// Write(rFilename, openmode)
void EventSet::Write(const std::string& rFilename, 
        std::ios::openmode openmode) const {
    try {
        TokenWriter tw(rFilename, openmode);
        Write(tw);
    }
    catch (std::ios::failure&) {
        std::stringstream errstr;
        errstr << "Exception opening/writing file \"" << rFilename << "\"";
        throw Exception("EventSet::Write", errstr.str(), 2);
    }
}

// Write(rFilename, rLabel, openmode)
void EventSet::Write(const std::string& rFilename, const std::string& rLabel,
        std::ios::openmode openmode) const {
    try {
        TokenWriter tw(rFilename, openmode);
        Write(tw, rLabel);
    }
    catch (std::ios::failure&) {
        std::stringstream errstr;
        errstr << "Exception opening/writing file \"" << rFilename << "\"";
        throw Exception("EventSet::Write", errstr.str(), 2);
    }
}

// Write(tw)
void EventSet::Write(TokenWriter& tw) const {
    Write(tw, Name());
}

// Write(tw, rLabel)
void EventSet::Write(TokenWriter& tw, const std::string& rLabel) const {
    const_iterator it;
    tw.WriteBegin(rLabel);
    for (it = Begin(); it != End(); ++it) {
        if (EventName(*it) != "") {
    	    Token token(Token::Name, EventName(*it));
    	    tw << token;
        }
        else {
    	    Token token(Token::Index, *it);
    	    tw << token;
        }
    }
    tw.WriteEnd(rLabel);
}

// DWrite()
void EventSet::DWrite(void) const {
    TokenWriter tw;
    const_iterator it;
    tw.WriteBegin(Name());
    for (it = Begin(); it != End(); ++it) {
        tw << Str(*it);
    }
    tw.WriteEnd(Name());
}

// ToString()
std::string EventSet::ToString(void) const {
    std::string ret;
    std::ostringstream ostr;
    const_iterator it;
    for (it = Begin(); it != End(); ++it) {
        if (EventName(*it) != "") {
    	    ostr << Str(*it) << " ";
        }
        else {
    	    FD_ERR("EventSet(" << this << ")::ToString(): index " << *it 
        		    << " don't got a name");
        }
    }
    ret = ostr.str();
    if (ret.size() > 0) ret.erase(ret.size()-1);
    return ret;
}

// Read(file)
void EventSet::Read(const std::string& rFilename) {
    Read(rFilename, "Alphabet");
}

// Read(rFilename, rLabel)
void EventSet::Read(const std::string& rFilename, const std::string& rLabel) {
    TokenReader tr(rFilename);
        tr.SeekBegin(rLabel); 
    Name(rLabel);
    Read(tr);
}

// Read(rTr)
void EventSet::Read(TokenReader& rTr) {
    Read(rTr, "Alphabet");
}

// read(rTr, rLabel)
void EventSet::Read(TokenReader& rTr, const std::string& rLabel) {
    Name(rLabel);
    Token token;
    mIndexSet.clear();
    while(rTr >> token) {
        if (token.Type() == Token::Name) {
    	    FD_DC("EventSet(" << this << ")::Read(..): inserting event \"" 
        		    << token.StringValue() << "\"");
    	    InsEvent(token.StringValue());
        }
        else if (token.Type() == Token::Index) {
    	    std::stringstream errstr;
    	    errstr << "Event token " << token.IndexValue() 
        		    << " is of type Token::Index. should be Token::Name";
    	    throw Exception("Eset::Read", errstr.str(), 50);
        }
        // return on </<endlabel>>
        else if ((token.Type() == Token::End) && 
    		    (token.StringValue() == rLabel)) {
    	    return;
        }
        else {
    	    std::stringstream errstr;
    	    errstr << "Token mismatch: (" << token.IndexValue() << ",\"" << token.StringValue()
        		    << "\") " << rTr.FileLine();
    	    throw Exception("Eset::Read", errstr.str(), 50);
        }
    }
}

// InsEvent(index)
void EventSet::InsEvent(Idx index) {
    FD_DC("EventSet(" << this << ")::InsEvent(" << index <<")");
    mIndexSet.insert(index);
}
        
// InsEvent(rName)
Idx EventSet::InsEvent(const std::string& rName) {
    FD_DC("EventSet(" << this << ")::InsEvent(" << rName <<")");
    // always insert in mpEventSymbolTable. 
    // doesn't matter if the event already exists
    Idx index = mpEventSymbolTable->InsEvent(rName);
    mIndexSet.insert(index);
    return index;
}

// DelEvent(index)
void EventSet::DelEvent(Idx index) {
    FD_DC("EventSet(" << this << ")::DelEvent(" << index <<")");
    if (mIndexSet.erase(index) == 0) {
        std::stringstream errstr;
        errstr << "index " << index << " not found in EventSet" << std::endl;
        throw Exception("EventSet::DelEvent", errstr.str(), 200);
    }
}

// DelEvent(rName)
void EventSet::DelEvent(const std::string& rName) {
    FD_DC("EventSet(" << this << ")::DelEvent(" << rName <<")");
    Idx index = mpEventSymbolTable->EventIndex(rName);
    if (index != 0) {
        mIndexSet.erase(index);
    }
    else {
        std::stringstream errstr;
        errstr << "name \"" << rName << "\" not found in EventSet" << std::endl;
        throw Exception("EventSet::DelEvent", errstr.str(), 202);
    }
}

// DelEvent(pos)
void EventSet::DelEvent(iterator pos) {
    mIndexSet.erase(pos);
}

// Clear()
void EventSet::Clear(void) {
    FD_DC("EventSet(" << this << ")::Clear()");
    mIndexSet.clear();
}

// EventName(index)
std::string EventSet::EventName(Idx index) const {
    return mpEventSymbolTable->EventName(index);
}

// EventName(index, name)
void EventSet::EventName(Idx index, const std::string& rName) {
    FD_DC("EventSet(" << this << ")::EventName(" << index << ", " << rName <<")");
#ifdef FAUDES_CHECKED
    if (! ExistsEvent(index)) {
        FD_ERR("EventSet::EventName(" << index << "," << rName << "): "
    		    << "index does not exist in this EventSet. aborting...");
        abort();
    }
#endif
    mpEventSymbolTable->EventName(index, rName);
}

// EventName(name, name)
void EventSet::EventName(const std::string& rName, 
        const std::string& rNewName) {
    FD_DC("EventSet(" << this << ")::EventName(" << rName << ", " 
    	    << rNewName <<")");
#ifdef FAUDES_CHECKED
    if (! ExistsEvent(rName)) {
        FD_ERR("EventSet::EventName(" << rName << "," << rNewName << "): "
    		    << "event " << "\"" << rName << "\" "
    		    << "does not exist in this EventSet. aborting...");
        abort();
    }
#endif
    mpEventSymbolTable->EventName(EventIndex(rName), rNewName);
}

// EventIndex(rName)
Idx EventSet::EventIndex(const std::string& rName) const {
    return mpEventSymbolTable->EventIndex(rName);
}

// ExistsEvent(index)
bool EventSet::ExistsEvent(Idx index) const {
    if (mIndexSet.find(index) != mIndexSet.end()) {
        return true;
    }
    return false;
}

// ExistsEvent(rName)
bool EventSet::ExistsEvent(const std::string& rName) const {
    if (mIndexSet.find(mpEventSymbolTable->EventIndex(rName)) != mIndexSet.end()) {
        return true;
    }
    return false;
}

// Find(index)
EventSet::iterator EventSet::Find(Idx index) {
    return mIndexSet.find(index);
}

// Find(index) const
EventSet::const_iterator EventSet::Find(Idx index) const {
    return mIndexSet.find(index);
}


// Find(rName)
EventSet::iterator EventSet::Find(const std::string& rName) {
    return mIndexSet.find(mpEventSymbolTable->EventIndex(rName));
}

// Find(rName) const
EventSet::const_iterator EventSet::Find(const std::string& rName) const {
    return mIndexSet.find(mpEventSymbolTable->EventIndex(rName));
}


// operator +
EventSet EventSet::operator + (const EventSet& rOtherSet) const {
    FD_DC("EventSet(" << this << ")::operator + (" << &rOtherSet << ")");
    EventSet res = NewEventSet();
    std::insert_iterator< std::set<Idx> > insit(res.mIndexSet, res.mIndexSet.begin());
    std::set_union(Begin(), End(), rOtherSet.Begin(), rOtherSet.End(), insit);
    res.Name(Name());
    return res;
}

// operator -
EventSet EventSet::operator - (const EventSet& rOtherSet) const {
    FD_DC("EventSet(" << this << ")::operator - (" << &rOtherSet << ")");
    EventSet res = NewEventSet();
    std::insert_iterator< std::set<Idx> > insit(res.mIndexSet, 
    	    res.mIndexSet.begin());
    std::set_difference(Begin(), End(), 
    	    rOtherSet.Begin(), rOtherSet.End(), insit);
    res.Name(Name());
    return res;
}

// operator *
EventSet EventSet::operator * (const EventSet& rOtherSet) const {
    FD_DC("EventSet(" << this << ")::operator * (" << &rOtherSet << ")");
    EventSet res = NewEventSet();
    std::insert_iterator< std::set<Idx> > insit(res.mIndexSet,
    	    res.mIndexSet.begin());
    std::set_intersection(Begin(), End(), 
    	    rOtherSet.Begin(), rOtherSet.End(), insit);
    res.Name(Name());
    return res;
}

// Str(index)
std::string EventSet::Str(Idx index) const {
    return mpEventSymbolTable->EventName(index)+"["+faudes::ToString(index)+"]";
}

} // namespace faudes
