/* eventset.h -- container for storing events (alphabet) */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef ESET_H

#include "definitions.h"
#include "debug.h"
#include "exception.h"
#include "baseset.h"
#include "symboltable.h"
#include "tokenreader.h"
#include "tokenwriter.h"
#include <fstream>
#include <sstream>
#include <iostream>
#include <set>
#include <assert.h>

namespace faudes {

/**
 * Container for events. Methods and data refers to global EventSymbolTable.
 */
class EventSet : public BaseSet {
    public:
        /**
         * Constructor for EventSet with static EventSymbolTable. 
         * The class behaviour is specified by usage of first constructor!
         */
      EventSet(void);
         
        /**
         * Constructor for EventSet with static EventSymbolTable. 
         * The class behaviour is specified by usage of first constructor!
         * This version reads a file with given label to find the alphabet
         * inside the file. If the label is omitted, "Alphabet" is assumed
         * to be the label to search for.
         *
         * @param rFilename
         *   Filename
         * @param rLabel
         *   Token label for the alphabet in the file; default value "Alphabet"
         */
        EventSet(const char* rFilename, 
    		    const std::string& rLabel = "Alphabet");
        /**
         * Return an EventSet with mpEventSymbolTable assigned on the stack
         *
         * @return 
         *   New empty EventSet on stack
         */
        EventSet NewEventSet() const;

        /**
         * Return a pointer to a EventSet on the heap with mpEventSymbolTable 
         * assigned
         *
         * @return
         *   Pointer to new empty EventSet on heap
         */
        EventSet* NewEventSetp() const;

        /**
         * Get Pointer to mpEventSymbolTable.
         *
         * @return
         *   Pointer to mpEventSymbolTable
         */
        EventSymbolTable* EventSymbolTablep(void);

        /**
         * Get pointer to static EventSymbolTable
         *
         * @return
         *   Pointer to static EventSymbolTable
         */
        static EventSymbolTable* StaticEventSymbolTablep(void);

        /**
         * Write EventSet to console
         */
        void Write(void) const;

        /**
         * Write EventSet to file opened with given openmode (default: truncate file)
         *
         * @param rFileName
         *   Filename
         * @param openmode
         *   std::ios::openmode
         *
         * @exception Exception
         *   If opening/writing fails an Exception object is thrown (id 2)
         */
        void Write(const std::string& rFileName, 
    		    std::ios::openmode openmode = std::ios::out|std::ios::trunc) const;

        /**
         * Write EventSet to file opened with label and 
         * given openmode (default: truncate file)
         *
         * @param rFileName
         *   Filename
         * @param rLabel
         *   Label for alphabet in file
         * @param openmode
         *   std::ios::openmode
         *
         * @exception Exception
         *   If opening/writing fails an Exception object is thrown (id 2)
         */
        void Write(const std::string& rFileName, const std::string& rLabel,
    		    std::ios::openmode openmode = std::ios::out|std::ios::trunc) const;


        /**
         * Write EventSet to TokenWriter. The name of the alphabet is used
         * as the label in the file.
         *
         * @param tw
         *   Reference to TokenWriter
         *
         * @exception std::ios::failure
         *   Thrown on i/o error.
         */
        void Write(TokenWriter& tw) const;
        
        /**
         * Write EventSet to TokenWriter with a given label
         *
         * @param tw
         *   Reference to TokenWriter
         * @param rLabel
         *   Label for alphabet in file
         *
         * @exception std::ios::failure
         *   Thrown on i/o error.
         */
        void Write(TokenWriter& tw, const std::string& rLabel) const;
        
        /**
         * Write EventSet to Console. Alias for Write()
         */
        void DWrite(void) const;

        /**
         * Write all events into a std::string
         *
         * @return
         *   String containing all events
         */
        std::string ToString(void) const;

        /**
         * Read EventSet from file. Reads label "Alphabet" by calling 
         * read(pFileName, label) with label == "Alphabet".

         * @param rFileName
         *   Filename
         *
         * @exception Exception
         *   If opening/reading fails an Exception object is thrown (id 1, 50, 51)
         */   
        void Read(const std::string& rFileName);

        /**
         * Read EventSet from file. Reads specified label by creating a tokenreader and
         * calling read(tr)
         *
         * @param rFileName
         *   Filename
         * @param rLabel
         *   token label to read from file
         *
         * @exception Exception
         *   If opening/reading fails an Exception object is thrown (id 1, 50, 51)
         */
        void Read(const std::string& rFileName, const std::string& rLabel);
    
        /**
         * Read EventSet from tokenreader. Clears eventset before. If file contains index 
         * false is returned. For label other than "Alphabet" use read(tr, label).
         *
         * @param tr
         *   Reference to tokenreader
         *
         * @exception std::ios::failure
         *   Thrown on i/o error.
         */
         void Read(TokenReader& tr);
    
        /**
         * Read EventSet from tokenreader. Clears eventset before. If file contains index 
         * false is returned.
         *
         * @param tr
         *   Reference to tokenreader
         * @param rLabel
         *   Label to read
         *
         * @exception std::ios::failure
         *   Thrown on i/o error.
         */
         void Read(TokenReader& tr, const std::string& rLabel);
        
        /**
         * Add an existing event to alphabet by index
         *
         * @param index
         *   Event index
         *
         * @return
         *   Success. Depending of global existence of event.
         */
        void InsEvent(Idx index);

        /** 
         * Add a named event to alphabet. If event isn't already in global
         * EventSymbolTable it will be added. 
         *
         * @param rName
         *   Event name
         *
         * @return 
         *   Index of (new) event
         */
        Idx InsEvent(const std::string& rName);

        /**
         * Delete event by index. Global EventSymbolTable stays untouched.
         *
         * @param index
         *   index
         * 
         * @exception
         *   Event index not found in EventSet (id 200)
         */
        void DelEvent(Idx index);

        /**
         * Delete event by name. Global EventSymbolTable stays untouched.
         *
         * @param rName
         *   named event
         *
         * @exception
         *   Event name not found in EventSet (id 202)
         */
        void DelEvent(const std::string& rName);

        /**
         * Delete event by iterator. Global EventSymbolTable stays untouched.
         *
         * @param pos
         *   EventSet::iterator
         */
        void DelEvent(iterator pos);

        /**
         * Clear all stored events
         */
        void Clear(void);

        /**
         * Set new name for existing event
         * FAUDES_CHECKED checks if index exists in EventSet. If not found the
         * program aborts.
         *
         * @param index
         *   Event index
         * @param rName
         *   New name
         *
         * @exception Exception
         *   Index not found in EventSymbolMap (id 102)
         *   Name already associated with another index (id 104)
         */
        void EventName(Idx index, const std::string& rName);

        
        /**
         * Set new name for existing event
         * FAUDES_CHECKED checks if EventIndex(rName) exists in EventSet. 
         * If not found the program aborts.
         *
         * @param rName
         *   Event name
         * @param rNewName
         *   New name
         *
         * @exception Exception
         *   Index not found in EventSymbolMap (id 102)
         *   Name already associated with another index (id 104)
         */
        void EventName(const std::string& rName, const std::string& rNewName);

        /**
         * Name lookup
         *
         * @param index
         *   Event index
         *
         * @return
         *   Corresponding name or empty std::string
         */
        std::string EventName(Idx index) const;

        /**
         * Index lookup
         *
         * @param rName
         *   Event name
         *
         * @return 
         *   Corresponding index or 0
         */
        Idx EventIndex(const std::string& rName) const;

        /**
         * Test existence of index
         *
         * @param index
         *   Event index
         *
         * @return 
         *   Success
         */
        bool ExistsEvent(Idx index) const;

        /** 
         * Test existence of name
         *
         * @param rName
         *   Event rName
         *
         * @return 
         *   Success
         */
        bool ExistsEvent(const std::string& rName) const;

        /**
         * Return EventSet::iterator to index
         * If not found returns End()
         *
         * @param index
         *   Event index
         *
         * @return
         *   EventSet::iterator
         */
        EventSet::iterator Find(Idx index);

        /**
         * Return EventSet::const_iterator to index
         * If not found returns End()
         *
         * @param index
         *   Event index
         *
         * @return
         *   EventSet::const_iterator
         */
        EventSet::const_iterator Find(Idx index) const;

        /**
         * Return EventSet::iterator to index of name
         * If not found returns End()
         *
         * @param rName
         *   Event name
         *
         * @return
         *   EventSet::iterator
         */
        EventSet::iterator Find(const std::string& rName);

        /**
         * Return EventSet::const_iterator to index of name
         * If not found returns End()
         *
         * @param rName
         *   Event name
         *
         * @return
         *   EventSet::const_iterator
         */
        EventSet::const_iterator Find(const std::string& rName) const;


        /**
         * Set union with another EventSet
         *
         * @return 
         *   Union EventSet
         */
        EventSet operator + (const EventSet& rOtherSet) const;

        /**
         * Set difference with another EventSet
         *
         * @return 
         *   Difference EventSet
         */
        EventSet operator - (const EventSet& rOtherSet) const;

        /**
         * Set intersection with another EventSet
         *
         * @return 
         *   Intersection EventSet
         */
        EventSet operator * (const EventSet& rOtherSet) const;

        /**
         * Return pretty printable event name for index.
         * Primary meant for debugging messages
         *
         * @param index
         *   Event index
         *
         * @return
         *   String
         */
        std::string Str(Idx index) const;

        /**
         * Create a copy with indices according to other EventSymbolTable
         *
         * @param pOtherEventSymbolTable
         *   Pointer to other EventSymbolTable
         * @param rReindexMap
         *   map<Idx,Idx> mapping original to new indices
         */
        //EventSet ReindexedCopy(EventSymbolTable* pOtherEventSymbolTable,
        //	    std::map<Idx,Idx>& rReindexMap) const; erasethis

    private:

        /** Pointer to local EventSymbolTable */
        EventSymbolTable* mpEventSymbolTable;

        /** Pointer to static global EventSymbolTable */
        static EventSymbolTable msEventSymbolTable;
};

} // namespace faudes

#define ESET_H
#endif 

