/* exception.h -- exception class */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef EXCEPTION_H

#include "definitions.h"
#include "debug.h"
#include <exception>
#include <string>

namespace faudes {

/**
 * Exception class for general use
 *
 * @param Function
 *   Function name (where did the exception happen)
 * @param Description
 *   Detailed description
 * @param ErrorId
 *   Error id
 *
 * ERROR IDs:
 *
 *    0: Error not defined or no error id given
 *    1: ios::failure thrown while opening/reading generator file
 *    2: ios::failure thrown while opening/writing generator file
 *    3: ios::failure thrown while opening/writing dotfile
 *    4: Error executing dot
 *    5: Dot output format unknown
 *    6: Exception opening/reading controllable events in file 
 *
 *   50: Token mismatch while reading generator
 *   51: Label not found (in generator file)
 *   52: Index not found in minindexmap
 *   54: read timing for empty clock set/transition relation
 *   55: invalid clock constraint found 
 *   56: invalid value for comperative operator found
 *   57: transition not found while reading timing
 *
 *  100: An uncontrollable event is already flagged controllable
 *       while reading cGenerator from file
 *  101: An controllable event is already flagged uncontrollable
 *       while reading cGenerator from file
 *  102: Event index not existant in EventSymbolTable
 *  104: Event Name already associated with another index
 *  105: Event name not existant in EventSymbolTable
 *  106: Clock index not existant in ClockSymbolTable
 *  107: Clock Name already associated with another index
 *  108: Clock name not existant in ClockSymbolTable
 *
 *  150: state/event attributes section not found in file
 *  151: error while reading event attributes in file
 *  152: error while reading state attributes in file
 *
 *  200: Event index not found in event domain of Generator/EventSet
 *  201: State index not found in Generator/StateSet
 *  202: Event name not found in event domain of Generator/EventSet
 *  203: State name not found in Generator/StateSet
 *  204: Clock name not found in clock set of Generator
 *  205: Clock index not found in clock set of Generator
 *
 *  500: Alphabets of generators don't match in supcon/controllable
 *  501: Plant generator nondeterminstic in SupConNB / IsControllable
 *  502: Index not found in entry states map
 *  503: Spec generator nondeterminstic in SupConNB / IsControllable
 *  504: Plant and spec generator are nondeterminstic in 
 *       SupConNB / IsControllable
 *  505: Input automaton nondeterministic in statemin
 *
 *  600: Idx overflow
 *
 *  700: Alphabet does not exist in SctScript class
 *  701: Genrator does not exist in SctScript class
 */
class Exception : public std::exception {
	public:
		/**
		 * Constructor
		 *
		 * Writes an error message on stderr.
		 * 
		 * @param rFunctionName
		 *   Function name (where did the exception happen)
		 * @param rDescription
		 *   Detailed error description
		 * @param errorId
		 *   Error id
		 */
		explicit Exception(const std::string& rFunctionName, 
				const std::string& rDescription, unsigned int errorId);

		/** Destructor */
		virtual ~Exception() throw();

		/** Returns Function */
		virtual const char* Where() const throw();

		/** Returns error description */
		virtual const char* What() const throw();

		/** Returns error id */
		virtual unsigned int Id() const throw();

	private:
		/** Function name */
		std::string FunctionName;

		/** Error description */
		std::string Description;

		/** Error id */
		unsigned int ErrorId;
};
		
} // namespace faudes

#define EXCEPTION_H
#endif 
