/* flagtable.h -- assign properties and attributs to states and events */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

#ifndef TFLAGTABLE_H

#include "definitions.h"
#include "debug.h"
#include <map>

namespace faudes {

/**  
 * FlagTable base class.
 *
 * Holds a flagfield (T) for indices implemented as a 
 * std::map<Idx,T>. 
 */
template <class TFlagType>
class TFlagTable {

    public:

        /** type definition for class iterator */
        typedef typename std::map<Idx,TFlagType>::iterator iterator;

        /** type definition for class const_iterator */
        typedef typename std::map<Idx,TFlagType>::const_iterator const_iterator;
        
        /** Creates a new FlagTable */
        TFlagTable(void) {
        }
        
        /**
         * Set flags for an index (set to 1).
         * The current flags of the index are combined with the new
         * flags by the OR bit operation. So every flag which should
         * be set must be 1 in the flag parameter. The other flags which
         * should stay untouched must be 0 in the flag parameter.
         * 
         * @param index
         *   Index
         * @param flags
         *   Flags to OR current flags
         */
        void SetFlags(Idx index, const TFlagType flags) {
            mFlagMap[index] |= flags;
        }

        /**
         * Unset Flags for an index (set to 0).
         * The current flags of the index are combined with the new 
         * flags by the AND bit operation. So every flag which is 0 in the
         * flag parameter will be set to 0 in the flag table. The other
         * flags which should stay untouched must be 1 in the parameter.
         * 
         * @param index
         *   Index
         * @param inverseflags
         *   Flags to AND current flags
         */
        void UnsetFlags(Idx index, const TFlagType inverseflags) {
            mFlagMap[index] &= inverseflags;
        }

        /**
         * Clear flags for an index (set to 0)
         * The current flags of the index are combined with new
         * flags by the AND bit operation after inverting the argument.
         * So every flag which should be cleared must be 1 in the flag
         * parameter. The other flags which should stay untouched
         * must be 0 in the flag parameter
         * 
         * @param index
         *   Index
         * @param flags
         *   Flags to INVERT and AND current flags
         */
        void ClrFlags(Idx index, const TFlagType flags) {
            mFlagMap[index] &= ~flags;
        }

        /**
         * Clear flags for all indices (set to 0)
         *
         * See ClrFlags(index, flags) for details.
         *
         * @param flags
         *   Flags to INVERT and AND current flags
         */
        void ClrFlags(const TFlagType flags) {
            iterator it;
            TFlagType inverseflags =~ flags;
            for (it = mFlagMap.begin(); it != mFlagMap.end(); ++it) {
              it->second &= inverseflags;
            }
        }

        /** 
         * Get the flags for an index. The second parameter specifies the
         * flags which should be returned (interesting flags must be 1, 
         * all other 0). A third parameter specifies the default value to 
         * return, if no entry is found for the index.
         * 
         * @param index
         *   Index 
         * @param flags
         *   Interesting flags
         * @param defaultflags
         *   Return this if no index was found
         *
         * @return
         *   Values of interesting flags
         */
        TFlagType GetFlags(Idx index, const TFlagType flags, 
                const TFlagType defaultflags) const {
            const_iterator it;
            it = mFlagMap.find(index);
            if (it == mFlagMap.end()) {
                return defaultflags;
            }
            else {
                return it->second & flags;
            }
        }

        /**
         * Test existence of an index
         *
         * @param index
         *   Index
         *
         * @return 
         *   True if index exists
         */
        bool ExistsIndex(Idx index) {
            iterator it = mFlagMap.find(index);
            if (it == mFlagMap.end()) {
                return false;
            }
            else {
                return true;
            }
        }

        /**
         * iterator to begin() of mFlagMap
         * 
         * @returns
         *   iterator to Begin of mFlagMap
         */
        iterator Begin(void){
            return mFlagMap.begin();
        }

        /**
         * const_iterator to begin() of mFlagMap
         * 
         * @returns
         *   const_iterator to Begin of mFlagMap
         */
        const_iterator Begin(void) const{
            return mFlagMap.begin();
        }

        /**
         * iterator to end() of mFlagMap
         * 
         * @returns
         *   iterator to end of mFlagMap
         */
        iterator End(void){
            return mFlagMap.end();
        }

        /**
         * const_iterator to end() of mFlagMap
         * 
         * @returns
         *   const_iterator to end of mFlagMap
         */
        const_iterator End(void) const{
            return mFlagMap.end();
        }
        
        /**
         * Clear complete map
         */
        void Clear() {
            mFlagMap.clear();
        }

        /**
         * Delete an entry in the flagmap
         * 
         * @param index
         */
        void DelFlag(Idx index) {
            mFlagMap.erase(index);
        }
            
         /**
         * Return a reference to the map
         * @return
         *     reference to mFlagMap
         */
         // Klaus Schmidt, 04.04.2006
//        std::map<Idx,TFlagType>& FlagMapr() {
//            return mFlagMap;
//        }    
  
            /**
         * Return  map
         * @return
         *     mFlagMap
         */
         // Klaus Schmidt, 04.04.2006
//        std::map<Idx,TFlagType> FlagMap() const{
//            return mFlagMap;
//        }        
        
        /** Klaus Schmidt 10.04.2006
         * Create a copy with new state indices. Maps from old to new indices 
         * are given as reference parameters.
         *
         * @param rFlagTable
         *   Empty flagtable which will become the copy
         * @param rStateReindexMap
         *   Empty map which will contain mapped state indices after copying
         */ 
        void Copy(TFlagTable& rFlagTable, std::map<Idx,Idx>& rReIndexMap) const{
           const_iterator const_it;
           for (const_it = mFlagMap.begin(); const_it != mFlagMap.end(); ++const_it) {
               rFlagTable.SetFlags(rReIndexMap[const_it->first],const_it->second );
           }
        } 
          
    protected:
        /** Flag lookup */
        typename std::map<Idx,TFlagType> mFlagMap;
};

} // namespace faudes

#define TFLAGTABLE_H
#endif

