/* generator.h -- generator class */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef GENERATOR_H

#include "definitions.h"
#include "debug.h"
#include "exception.h"
#include "symboltable.h"
#include "eventset.h"
#include "stateset.h"
#include "ttransset.h"
#include "token.h"
#include "tokenreader.h"
#include "tokenwriter.h"
#include <map>
#include <set>
#include <sstream>
#include <cstdlib>
#include <assert.h>

namespace faudes {

/**
 * Modell of the common five-tupel consisting of mAlphabet, mStates,
 * transition relation, initial states, marked states
 * 
 * @param mMyName
 *   Name of the generator
 * @param mAlphabet
 *   Alphabet of the generator
 * @param mStates
 *   Stateset of the generator
 * @param mTransRel
 *   Transition relation of the generator
 * @param mInitStates
 *   Initial states
 * @param mMarkedStates
 *   Marked states
 */

template <class TrnSet>
class TGenerator {
  public:

//     typedef TGenerator<Transition> Generator; 
    /**
     * Creates an emtpy Generator object with static event- and StateSymbolTable. 
     */
    TGenerator(void);

    /** 
     * Generator from a generator. Copy constructor. This creates a nearly 
     * 1:1 copy of the generator with identical indices. If one of the copy
     * gets out of scope or delete is called for an object in heap all state
     * names will be removed from mpStateSymbolTable.
     *
     * @param rOtherGen
     */
    TGenerator(const TGenerator& rOtherGen);

    /**
     * Creates a Generator from file with static event- and StateSymbolTable.
     *
     * @param pFileName
     *   Filename
     *
     * @exception Exception
     *   If opening/reading fails an Exception object is thrown (id 1, 50, 51)
     */
    TGenerator(const char* pFileName);

    /**
     * Destructor
     * Deletes state indices in mpStateSymbolTable
     */
    virtual ~TGenerator(void);

    /**
     * Register a same-state-index copy in mGeneratorCopySets
     *
     * @param pExistingGen
     *   Memory address of existing generator
     * @param pCopyGen
     *   Memory address of generator copy
     */
    static void RegisterGeneratorCopy(TGenerator<TrnSet>* pExistingGen, TGenerator<TrnSet>* pCopyGen);

    /**
     * Unregister the generator in mGeneratorCopySets. Returns true if there are no
     * further copys.
     *
     * @param pGen
     *   Memory address of Generator
     *
     * @return
     *   True if there are no further copys and statenames can savely be deleted
     *   in Generator's destruktor.
     */
    static bool UnRegisterGeneratorCopy(TGenerator<TrnSet>* pGen);

    /**
     * Get Pointer to mpEventSymbolTable.
     *
     * @return
     *   Pointer to mpEventSymbolTable
     */
    EventSymbolTable* EventSymbolTablep(void);

    /**
     * Get Pointer to mpEventSymbolTable. (const version)
     *
     * @return
     *   Pointer to mpEventSymbolTable
     */
    EventSymbolTable* EventSymbolTablep(void) const;

    /**
     * Get pointer to   EventSymbolTable
     *
     * @return
     *   Pointer to static EventSymbolTable
     */
    static EventSymbolTable* StaticEventSymbolTablep(void);

    /**
     * Get pointer to StateSymbolTable.
     *
     * @return
     *   Pointer to mpStateSymbolTable
     */
    StateSymbolTable* StateSymbolTablep(void);

    /**
     * Get pointer to StateSymbolTable. (const version)
     *
     * @return
     *   Pointer to mpStateSymbolTable
     */
    StateSymbolTable* StateSymbolTablep(void) const;

    /**
     * Get pointer to static StateSymbolTable.
     *
     * @return
     *   Pointer to static StateSymbolTable
     */
    static StateSymbolTable* StaticStateSymbolTablep(void);

    /**
     * Assignment operator. If symboltables do match the copy constructor 
     * that creates a 1:1 copy is used. In the other case the Copy 
     * method is called which creates new state indices and cares for correct
     * event indices.
     *
     * @param rOtherGen
     *   Other generator
     */
    virtual TGenerator<TrnSet>& operator = (const TGenerator<TrnSet>& rOtherGen);

    /**
     * Create a copy with new state indices. This is just a wrapper for the
     * other copy method with reindexmaps.
     *
     * @param rGen
     *   Empty generator which will become the copy
     */
    virtual void Copy(TGenerator<TrnSet>& rGen) const; 

    /**
     * Create a copy with new state indices. Maps from old to new indices 
     * are given as reference parameters.
     *
     * @param rGen
     *   Empty generator which will become the copy
     * @param rStateReindexMap
     *   Empty map which will contain mapped state indices after copying
     */
    virtual void Copy(TGenerator<TrnSet>& rGen, std::map<Idx,Idx>& rStateReindexMap) const; 
        
    /**
     * Return an EventSet with generator's EventSymbolTable assigned on stack
     *
     * @return
     *   New empty EventSet on stack
     */
    EventSet NewEventSet(void) const;

    /**
     * Return a pointer to a EventSet on heap with generator's
     * EventSymbolTable assigned
     *
     * @return
     *   Pointer to new empty EventSet on heap
     */
    EventSet* NewEventSetp(void) const;

    /**
     * Return a StateSet with generator's StateSymbolTable assigned on stack
     *
     * @return
     *   New empty StateSet on stack
     */
    StateSet NewStateSet(void) const;

    /**
     * Return a pointer to a StateSet on heap with generator's
     * EventSymbolTable assigned
     *
     * @return
     *   Pointer to new empty StateSet on heap
     */
    StateSet* NewStateSetp(void) const;

    /**
     * Write generator to console
     */
    virtual void Write(void);

    /**
     * Write generator to file
     *
     * @param pFileName
     *   Filename to write
     * @param openmode
     *   std::ios::openmode (optional)
     *   Default is truncate file.
     *
     * @exception Exception
     *   If opening/writing fails an Exception object is thrown (id 2)
     */
    virtual void Write(const char* pFileName, 
        std::ios::openmode openmode = std::ios::out|std::ios::trunc);

    /**
     * Write generator to TokenWriter
     *
     * @param rTw
     *   Reference to TokenWriter
     *
     * @exception std::ios::failure
     *   Thrown on i/o error.
     */
    virtual void Write(TokenWriter& rTw);

    /**
     * Writes the dot output format + png.
     *
     * @param rFileName
     *   Filename to write. A ".gen" at the end will be cutted off. 
     *
     * @exception Exception
     *   If opening/writing dotfile/png fails an Exception object is thrown 
     *   (id 3)
     *   Error executing dot (id 4)
     */
    virtual void DotWrite(const std::string& rFileName) const;

    /**
     * Writes the dot format and a output format of dot.
     * See http://www.graphviz.org/doc/info/output.html for details.
     *
     * @param rFileName
     *   Filename to write. A ".gen" at the end will be cutted off. 
     * @param rFileType
     *   String containing the file type. See url.
     *
     * @exception Exception
     *   If opening/writing dotfile/png fails an Exception object is thrown 
     *   (id 3)
     *   Error executing dot (id 4)
     *   Dot output format unknown (id 5)
     */
    virtual void DotWrite(const std::string& rFileName, 
        const std::string& rFileType) const;

    /**
     * Writes the dot output format + png. (Debugging! No minimal indices)
     *
     * @param rFileName
     *   Filename to write. A ".gen" at the end will be cutted off. 
     *
     * @exception Exception
     *   If opening/writing dotfile/png fails an Exception object is thrown 
     *   (id 3)
     *   Error executing dot (id 4)
     */
    virtual void DDotWrite(const std::string& rFileName) const;

    /**
     * Writes the dot format and a output format of dot.
     * (Debugging! No minimal indices)
     *
     * See http://www.graphviz.org/doc/info/output.html for details.
     *
     * @param rFileName
     *   Filename to write. A ".gen" at the end will be cutted off. 
     * @param rFileType
     *   String containing the file type. See url.
     *
     * @exception Exception
     *   If opening/writing dotfile/png fails an Exception object is thrown 
     *   (id 3)
     *   Error executing dot (id 4)
     *   Dot output format unknown (id 5)
     */
    virtual void DDotWrite(const std::string& rFileName, 
        const std::string& rFileType) const;

    /**
     * (Debug-) write generator to console.
     * This is just an alias function for Write()
     */
    virtual void DWrite(void) const;

    /**
     * Write alphabet into a std::string
     *
     * @return 
     *   String containing all events
     */
    std::string AlphabetToString(void) const;

    /**
     * Write mStates into a std::string
     *
     * @return
     *   String containing all states
     */
    std::string StatesToString(void) const;

    /**
     * Write mTransRel into a std::string
     * 
     * @return
     *  String containing all transitions 
     */
    std::string TransRelToString(void) const;

    /**
     * Write mInitStates into a std::string
     * 
     * @return
     *   String containing all initial states
     */
    std::string InitStatesToString(void) const;

    /**
     * Write mMarkedStates into a std::string
     * 
     * @return
     *   String containing all marked states
     */
    std::string MarkedStatesToString(void) const;

    /**
     * Read generator from file
     *
     * @param pFileName
     *   Filename containing generator definition
     *
     * @exception Exception
     *   If opening/reading fails an Exception object is thrown (id 2, 50, 51)
     */
    virtual void Read(const char* pFileName);

    /**
     * Read generator from TokenReader
     *
     * @param rTr
     *   Reference to TokenReader
     *
     * @exception ios::failure
     *   Thrown on i/o error.
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    virtual void Read(TokenReader& rTr);

    /** 
     * Set the generator's name 
     *
     * @param rName
     *   Generator name
     */
    void Name(const std::string& rName);

    /** 
     * Return generator's name 
     *
     * @return 
     *   Name of generator
     */
    std::string Name(void) const;

    /** 
     * Return generator's Id (mId)
     *
     * @return 
     *   (Idx type) id of Generator
     */
    Idx Id(void) const;

    /** 
     * Return whether the Generator is unique (has no copies)
     *
     * @return 
     *   true or false
     */
    bool IsUnique(void) const;

    /** 
     * Check if generator is valid 
     *
     * @return 
     *   Success
     */
    virtual bool Valid(void);

    /**
     * Clear alphabet, states and transitions 
     * (initial states and marked states also)
     */
    virtual void Clear(void);

    /**
     * Clear all states and transitions. Alphabet stays untouched.
     */
    void ClrStates(void);

    /**
     * Get number of events in mAlphabet
     *
     * @return
     *   Number of events
     */
    Idx AlphabetSize(void) const;

    /** 
     * Get generator size (# of states) 
     * 
     * @return 
     *   Number of states
     */
    Idx Size(void) const;

    /**
     * Get number of transitions
     *
     * @return
     *   Number of transitions
     */
    Idx TransRelSize(void) const;

    /**
     * Get number of initial states
     *
     * @return
     *   Number of initial states
     */
    Idx InitStatesSize(void) const;

    /**
     * Get number of marked states
     *
     * @return
     *   Number of marked states
     */
    Idx MarkedStatesSize(void) const;

    /** 
     * Check if generator is empty (no state)
     *
     * @return 
     *   True if mStates is empty
     */
    bool Empty(void) const;

    /**
     * Check if mAlphabet is Empty
     *
     * @return
     *   True if mAlphabet is empty
     */
    bool AlphabetEmpty(void) const;

    /**
     * Check if transition relation is empty
     *
     * @return
     *   True if transition relation is empty
     */
    bool TransRelEmpty(void) const;

    /**
     * Check if set of initial states is empty
     *
     * @return
     *   True if mInitStates is empty
     */
    bool InitStatesEmpty(void) const;

    /**
     * Check if set of marked states is empty
     *
     * @return
     *   True if mMarkedStates is empty
     */
    bool MarkedStatesEmpty(void) const;

    /**
     * If set to true, library functions may set names for new states.
     * If set to false, library functions will not set names for new states.
     *
     * @return
     *   False if ClrStateNames() was called
     */
    bool StateNamesEnabled(void) const;

    /**
     * See accessor function for details
     *
     * @param flag
     *   True enables statenames / false disables them
     */
    void StateNamesEnabled(bool flag);

    /**
     * iterator to Begin() of mAlphabet
     *
     * @return
     *   iterator to begin of mAlphabet
     */
    EventSet::iterator AlphabetBegin(void);

    /**
     * const_iterator to Begin() of mAlphabet
     *
     * @return
     *   const_iterator to begin of mAlphabet
     */
    EventSet::const_iterator AlphabetBegin(void) const;

    /**
     * iterator to End() of mAlphabet
     *
     * @return
     *   iterator to end of mAlphabet
     */
    EventSet::iterator AlphabetEnd(void);

    /**
     * const_iterator to End() of mAlphabet
     *
     * @return
     *   iterator to end of mAlphabet
     */
    EventSet::const_iterator AlphabetEnd(void) const;

    /**
     * iterator to Begin() of mStates
     *
     * @return
     *   iterator to begin of mStates
     */
    StateSet::iterator StatesBegin(void);

    /**
     * const_iterator to Begin() of mStates
     *
     * @return
     *   const_iterator to begin of mStates
     */
    StateSet::const_iterator StatesBegin(void) const;

    /**
     * iterator to End() of mStates
     *
     * @return
     *   iterator to end of mStates
     */
    StateSet::iterator StatesEnd(void);

    /**
     * const_iterator to End() of mStates
     *
     * @return
     *   const_iterator to end of mStates
     */
    StateSet::const_iterator StatesEnd(void) const;

    /**
     * iterator to Begin() of transition relation
     *
     * @return
     *   iterator to Begin of mTransRel
     */
    typename TrnSet::iterator TransRelBegin(void);

    /**
     * const_iterator to Begin() of transition relation
     *
     * @return
     *   const_iterator to Begin of mTransRel
     */
    typename TrnSet::const_iterator TransRelBegin(void) const;

    /**
     * iterator to End() of transition relation
     * 
     * @return
     *   iterator to End of mTransRel
     */
    typename TrnSet::iterator TransRelEnd(void);

    /**
     * const_iterator to End() of transition relation
     * 
     * @return
     *   const_iterator to End of mTransRel
     */
    typename TrnSet::const_iterator TransRelEnd(void) const;

    /**
     * iterator to begin of transitions with x1 as predecessor state.
     *
     * @param x1
     *   Predecessor state
     *
     * @return
     *   iterator to begin of transitions with x1
     */
    typename TrnSet::iterator TransRelBegin(Idx x1);

    /**
     * const_iterator to begin of transitions with x1 as predecessor state.
     *
     * @param x1
     *   Predecessor state
     *
     * @return
     *   const_iterator to begin of transitions with x1
     */
    typename TrnSet::const_iterator TransRelBegin(Idx x1) const;

    /**
     * iterator to end of transitions with x1 as predecessor state. 
     *
     * Notice:
     * Set the End(x1) iterator to a variable, so it won't be recalculated
     * every iteration.
     *
     * @param x1
     *   Predecessor state
     *
     * @return 
     *   iterator to end of transitions with x1 
     *   (one after last matching transition)
     */
    typename TrnSet::iterator TransRelEnd(Idx x1);

    /**
     * const_iterator to end of transitions with x1 as predecessor state. 
     *
     * Notice:
     * Set the End(x1) iterator to a variable, so it won't be recalculated
     * every iteration.
     *
     * @param x1
     *   Predecessor state
     *
     * @return 
     *   const_iterator to end of transitions with x1 
     *   (one after last matching transition)
     */
    typename TrnSet::const_iterator TransRelEnd(Idx x1) const;

    /**
     * iterator to begin of transitions with x1 as predecessor state and
     * event ev. 
     *
     * @param x1
     *   Predecessor state
     * @param ev
     *   Event
     *
     * @return
     *   iterator to begin of transitions with x1 and ev
     */
    typename TrnSet::iterator TransRelBegin(Idx x1, Idx ev);

    /**
     * const_iterator to begin of transitions with x1 as predecessor state and
     * event ev. 
     *
     * @param x1
     *   Predecessor state
     * @param ev
     *   Event
     *
     * @return
     *   const_iterator to begin of transitions with x1 and ev
     */
    typename TrnSet::const_iterator TransRelBegin(Idx x1, Idx ev) const;

    /**
     * iterator to end of transitions with x1 as predecessor state and
     * event ev. 
     *
     * Notice:
     * Set the End(x1,ev) iterator to a variable, so it won't be recalculated
     * every iteration.
     * 
     * @param x1
     *   Predecessor state
     * @param ev
     *   Event
     *
     * @return
     *   iterator to end of transitions with x1 and ev 
     *   (one after last matching transition)
     */
    typename TrnSet::iterator TransRelEnd(Idx x1, Idx ev);

    /**
     * const_iterator to end of transitions with x1 as predecessor state and
     * event ev.
     *
     * Notice:
     * Set the End(x1,ev) iterator to a variable, so it won't be recalculated
     * every iteration.
     * 
     * @param x1
     *   Predecessor state
     * @param ev
     *   Event
     *
     * @return
     *   const_iterator to end of transitions with x1 and ev 
     *   (one after last matching transition)
     */
    typename TrnSet::const_iterator TransRelEnd(Idx x1, Idx ev) const;

    /**
     * iterator to Begin() of mInitStates
     *
     * @returns
     *   iterator to begin of mInitStates
     */
    StateSet::iterator InitStatesBegin(void);

    /**
     * const_iterator to Begin() of mInitStates
     *
     * @returns
     *   const_iterator to begin of mInitStates
     */
    StateSet::const_iterator InitStatesBegin(void) const;

    /**
     * iterator to End() of mInitStates
     * 
     * @returns
     *   iterator to end of mInitStates
     */
    StateSet::iterator InitStatesEnd(void);

    /**
     * const_iterator to End() of mInitStates
     * 
     * @returns
     *   const_iterator to end of mInitStates
     */
    StateSet::const_iterator InitStatesEnd(void) const;

    /**
     * iterator to Begin() of mMarkedStates
     *
     * @returns
     *   iterator to Begin of mMarkedStates
     */
    StateSet::iterator MarkedStatesBegin(void);

    /**
     * const_iterator to Begin() of mMarkedStates
     *
     * @returns
     *   const_iterator to Begin of mMarkedStates
     */
    StateSet::const_iterator MarkedStatesBegin(void) const;

    /**
     * iterator to End() of mMarkedStates
     * 
     * @returns
     *   iterator to End of mMarkedStates
     */
    StateSet::iterator MarkedStatesEnd(void);

    /**
     * const_iterator to End() of mMarkedStates
     * 
     * @returns
     *   const_iterator to End of mMarkedStates
     */
    StateSet::const_iterator MarkedStatesEnd(void) const;

    /**
     * Set mAlphabet without consistency check.
     *
     * @param newalphabet
     *   EventSet with new alphabet
     */
    void InjectAlphabet(const EventSet& newalphabet);

    /**
     * Add an existing event to mAlphabet by index
     *
     * @param index
     *   Event index
     */
    void InsEvent(Idx index);

    /**
    * Add named event to generator. An entry in the mpEventSymbolTable will
    * be made if event is new.
    *
    * @param rName
    *   Name of the event to add
    *
    * @return 
    *   New unique index
    */
    Idx InsEvent(const std::string& rName);

    /**
     * Add new named events to generator.
     *
     * @param events
     *   EventSet
     */
    void InsEvents(const EventSet& events);

    /**
     * Delete event from generator by index. mpEventSymbolTable stays untouched.
     * Transitions containing event will be removed too.
     *
     * @param index
     *   Index of event
     *
     * @exception Exception
     *   Event index not found in generator (id 200)
     */
    void DelEvent(Idx index);

    /**
     * Delete event from generator by name. mpEventSymbolTable stays untouched.
     * Transitions containing event will be removed too.
     *
     * @param rName
     *    Name of event
     *
     * @exception Exception
     *   Event name not found in generator (id 202)
     */
    void DelEvent(const std::string& rName);

    /** 
     * Delete a set of events from generator. mpEventSymbolTable stays untouched.
     * Transitions containing events will be removed too.
     * 
     * @param rEvents
     *   EventSet containing events to remove
     */
    void DelEvents(const EventSet& rEvents);

    /** 
     * Delete event from Alphabet without consistency check. The event is only
     * deleted from mAlphabet but not from transition relation.
     *
     * @param index
     *   Index of event
     *
     * @exception Exception
     *   Event index not found in generator (id 200)
     */
    void DelEventFromAlphabet(Idx index);

    /** 
     * Add new anonymous state to generator 
     * 
     * @return 
     *   Index of new unique state
     */
    Idx InsState(void);

    /** 
     * Add new named state to generator. 
     * Remember: 
     * Names don't have to be unique and there is no name->index lookup.
     *
     * @param rName
     *   Name of the state to add
     *
     * @return 
     *   Index of new unique state
     */
    Idx InsState(const std::string& rName);

    /** 
     * Inject an existing state index into generator's mStates
     * Use with care! Should be used in speed optimized functions.
     *
     * @param index
     *   State index to inject
     */
    void InjectState(Idx index);

    /**
     * Inject a complete mStates without consistency checks
     *
     * @param rNewStates
     *    StateSet
     */
    void InjectStates(const StateSet& rNewStates);


    /**
     * Create new anonymous state and set as initial state
     *
     * @return 
     *   Index of new unique
     */
    Idx InsInitState(void);

    /**
     * Create a new named state and set as initial state
     *
     * @param rName
     *   Name of the state to add
     *
     * @return 
     *   Index of new unique state
     */
    Idx InsInitState(const std::string& rName);

    /**
     * Create new anonymous state and set as marked state
     *
     * @return 
     *   Index of new unique state
     */
    Idx InsMarkedState(void);

    /**
     * Create a new named state and set as marked state
     *
     * @param rName
     *   Name of the state to add
     *
     * @return 
     *   Index of new unique state
     */
    Idx InsMarkedState(const std::string& rName);

    /** 
     * Delete a state from generator by index. 
     * Cleans mStates, mInitStates, mMarkedStates, mTransRel and mpStateSymbolTable.
     *
     * @param index
     *   Index of state to delete. Will be erased in mpStateSymbolTable too.
     *
     * @exception Exception
     *   State index not found in generator (id 201)
     */
    void DelState(Idx index);

    /** 
     * Delete a state from generator by name.
     * Cleans mStates, mInitStates, mMarkedStates, mTransRel and mpStateSymbolTable.
     * (Expensive! Don't use!)
     *
     * @param rName
     *   Name of state to delete. Will be erased in mpStateSymbolTable too
     *
     * @exception Exception
     *   State name not found in generator (id 203)
     */
    void DelState(const std::string& rName);

    /**
     * Delete a set of states. Runs much faster than deleting every single state with
     * DelState(index). This method creates a reverse sorted transition relation and
          * calles the same method with two arguments.
     *
     * @param rDelStates
     *   StateSet containing states to delete
     */
    void DelStates(const StateSet& rDelStates);

    /**
     * Delete a set of states. Runs much faster than deleting every single state with
     * DelState(index).
     *
     * @param rDelStates
     *   StateSet containing states to delete
          * @param rTransRelX2EvX1
          *   Reverse sorted transition relation for identifying X2 states to delete.
     */
    void DelStates(const StateSet& rDelStates, TTransSet<TSortX2EvX1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>& rTransRelX2EvX1);

    /**
     * Delete a state from generator without consistency check. This removes the
     * state from mStates and mpStateSymbolTable but doesn't touch mTransRel,
     * mInitStates and mMarkedStates.
     *
     * @param index
     *   Index of state to delete. Will be erased in mpStateSymbolTable too
     *
     * @exception Exception
     *   State index not found in generator (id 201)
     */
    void DelStateFromStates(Idx index);

    /**
     * Delete a state from generator without consistency check. This removes the
     * state from mStates and mpStateSymbolTable but doesn't touch mTransRel,
     * mInitStates and mMarkedStates.
     * Index to delete is given by iterator. 
     *
     * @param pos
     *   StateSet::iterator
     */
    void DelStateFromStates(StateSet::iterator pos);

    /**
     * Set an existing state as initial state by index.
     *
     * Define FAUDES_CHECKED for consistency checks.
     *
     * @param index
     *   Index of state to set as initial state
     */
    void SetInitState(Idx index);

    /**
     * Set an existing state as initial state by name. 
     * (only for user interactions)
     *
     * @param rName
     *   Name of state to set as initial state
     *
     * @exception Exception
     *   State name not known in generator (id 203)
     */
    void SetInitState(const std::string& rName);

    /**
     * Replace mInitStates with StateSet given as parameter without consistency checks.
     *
     * @param rNewInitStates
     *   StateSet containing new mInitStates
     */
    void InjectInitStates(const StateSet& rNewInitStates);

    /**
     * Unset an existing state as initial state by index
     *
     * Define FAUDES_CHECKED for consistency checks.
     *
     * @param index
     *   State index
     *
     * @exception Exception
     *   State index not found in generator (id 201)
     */
    void ClrInitState(Idx index);

    /**
     * Unset an existing state as initial state by name (only for user
     * interactions)
     *
     * @param rName
     *   State name
     *
     * @exception Exception
     *   State name not known in Generator (id 503)
     */
    void ClrInitState(const std::string& rName);

    /**
     * Unset an existing state as initial state by iterator
     *
     * @param pos
     *   StateSet::iterator
     */
    void ClrInitState(StateSet::iterator pos);

    /**
     * Clear all mInitStates
     */
    void ClrInitStates(void); 

    /**
     * Set an existing state as marked state by index
     *
     * Define FAUDES_CHECKED for consistency checks.
     *
     * @param index
     *   Index of state to set as initial state
     */
    void SetMarkedState(Idx index);

    /**
     * Set an existing state as marked state by name. 
     * (only for user interactions)
     *
     * @param rName
     *   Name of state to set as marked state
     *
     * @exception Exception
     *   State name not known in generator (id 203)
     */
    void SetMarkedState(const std::string& rName);

    /**
     * Replace mMarkedStates with StateSet given as parameter without consistency checks.
     *
     * @param rNewMarkedStates
     *   StateSet containing new marked states
     */
    void InjectMarkedStates(const StateSet& rNewMarkedStates);

    /**
     * Unset an existing state as marked state by index
     *
     * Define FAUDES_CHECKED for consistency checks.
     *
     * @param index
     *   State index
     *
     * @exception Exception
     *   State index not found in generator (id 201)
     */
    void ClrMarkedState(Idx index);

    /**
     * Unset an existing state as marked state by name (only for user
     * interactions)
     *
     * @param rName
     *   State name
     *
     * @exception Exception
     *   State name not known in Generator (id 503)
     */
    void ClrMarkedState(const std::string& rName);

    /**
     * Unset an existing state as marked state by iterator
     *
     * @param pos
     *   StateSet::iterator
     */
    void ClrMarkedState(StateSet::iterator pos);

    /**
     * Clear all marked states
     */
    void ClrMarkedStates(void);

    /** 
     * Add a transition to generator by indices. States and event
     * must already exist!
     *
     * Define FAUDES_CHECKED for consistency checks.
     *
     * @param x1 
     *   Predecessor state index 
     * @param ev 
     *   Event index 
     * @param x2
     *   Successor state index
     */
    void SetTransition(Idx x1, Idx ev, Idx x2);

    /** 
     * Add a transition to generator by names. Statename and eventname
     * must already exist!
     *
     * @param rX1
     *   Predecessor state name
     * @param rEv
     *   Event name
     * @param rX2
     *   Successor state name
     *
     * @exception Exception
     *   Event name not found in event domain of Generator (id 202)
     *   State name not found in Generator (id 203)
     */
    void SetTransition(const std::string& rX1, const std::string& rEv, 
        const std::string& rX2);

    /** 
     * Add a transition to generator. States and event
     * must already exist!
     *
     * Define FAUDES_CHECKED for consistency checks.
     *
     * @param rTransition
     *   Transition
     */
    void SetTransition(const typename TrnSet::TransitionType& rTransition);

    /**
     * Remove a transition by indices
     *
     * @param x1
     *   Predecessor state index
     * @param ev
     *   Event index
     * @param x2
     *   Successor state index
     */
    void ClrTransition(Idx x1, Idx ev, Idx x2);

    /**
     * Remove a transition by transition object
     *
     * @param rTransition
     *   Transition object
     */
    virtual void ClrTransition(const typename TrnSet::TransitionType& rTransition);

    /**
     * Remove a transition by iterator
     *
     * @param it
     *   TransSet::iterator 
     */
    void ClrTransition(typename TrnSet::iterator it);

    /**
     * Clear all transitions
     */
    void ClrTransRel(void);

    /**
     * Set name for existing event
     *
     * @param index
     *   Event index
     * @param rName
     *   New name
     *
     * @exception Exception
     *   Index not found in EventSymbolMap (id 102)
     *   Name already associated with another index (id 104)
     */
    void EventName(Idx index, const std::string& rName);

    /** 
     * Set name for existing state
     *
     * @param index
     *   Index
     * @param rName
     *   Name
     *
     * @exception Exception
     *   Name already associated with another index (id 104)
     */
    void StateName(Idx index, const std::string& rName);

    /**
     * Clear name for existing state
     *
     * @param index
     *   State index
     */
    void ClrStateName(Idx index);

    /**
     * Clear name for existing state
     * (Expensive! Don't use!)
     *
     * @param rName
     *   State name
     */
    void ClrStateName(const std::string& rName);

    /** 
     * Test existence of event in mAlphabet
     *
     * @param index
     *   Event index
     *
     * @return
     *   true / false
     */
    bool ExistsEvent(Idx index) const;

    /** 
     * Test existence of event in mAlphabet
     *
     * @param rName
     *   Event name
     *
     * @return
     *   true / false
     */
    bool ExistsEvent(const std::string& rName) const;

    /**
     * Returns an iterator to event index in mAlphabet
     *
     * @param index
     *   Index to find
     *
     * @return
     *   EventSet::iterator to event index
     */
    EventSet::iterator FindEvent(Idx index);

    /**
     * Returns a const_iterator to event index in mAlphabet
     *
     * @param index
     *   Index to find
     *
     * @return
     *   EventSet::const_iterator to event index
     */
    EventSet::const_iterator FindEvent(Idx index) const;

    /**
     * Returns an iterator to event index in mAlphabet
     *
     * @param rName
     *   Event name of index to find
     *
     * @return
     *   EventSet::iterator to event index
     */
    EventSet::iterator FindEvent(const std::string& rName);

    /**
     * Returns a const_iterator to event index in mAlphabet
     *
     * @param rName
     *   Event name of index to find
     *
     * @return
     *   EventSet::const_iterator to event index
     */
    EventSet::const_iterator FindEvent(const std::string& rName) const;

    /**
     * Test existence of state in mStates
     * 
     * @param index
     *   State index
     *
     * @return
     *   true / false
     */
    bool ExistsState(Idx index) const;

    /**
     * Returns an iterator to state index in mStates 
     *
     * @param index
     *   Index to find
     *
     * @return
     *   StateSet::iterator to state index
     */
    StateSet::iterator FindState(Idx index);

    /**
     * Returns a const_iterator to state index in mStates 
     *
     * @param index
     *   Index to find
     *
     * @return
     *   StateSet::iterator to state index
     */
    StateSet::const_iterator FindState(Idx index) const;

    /**
     * Test existence of state in mInitStates
     * 
     * @param index
     *   State index
     *
     * @return
     *   true / false
     */
    bool ExistsInitState(Idx index) const;

    /**
     * Returns an iterator to state index in mInitStates 
     *
     * @param index
     *   Index to find
     *
     * @return
     *   StateSet::iterator to state index
     */
    StateSet::iterator FindInitState(Idx index);

    /**
     * Returns a const_iterator to state index in mInitStates 
     *
     * @param index
     *   Index to find
     *
     * @return
     *   StateSet::const_iterator to state index
     */
    StateSet::const_iterator FindInitState(Idx index) const;

    /**
     * Test existence of state in mMarkedStates
     * 
     * @param index
     *   State index
     *
     * @return
     *   true / false
     */
    bool ExistsMarkedState(Idx index) const;

    /**
     * Returns an iterator to state index in mMarkedStates
     *
     * @param index
     *   Index to find
     *
     * @return
     *   StateSet::iterator to state index
     */
    StateSet::iterator FindMarkedState(Idx index);

    /**
     * Returns a const_iterator to state index in mMarkedStates
     *
     * @param index
     *   Index to find
     *
     * @return
     *   StateSet::const_iterator to state index
     */
    StateSet::const_iterator FindMarkedState(Idx index) const;

    /** 
     * Event name lookup (in EventSymbolTable)
     *
     * @param index
     *
     * @return 
     *   Name or empty std::string
     */
    std::string EventName(Idx index) const;

    /**
     * Event index lookup (in EventSymbolTable)
     * 
     * @param rName
     *
     * @return 
     *   Valid index or 0
     */
    Idx EventIndex(const std::string& rName) const;

    /** 
     * State name lookup 
     *
     * @param index
     *
     * @return name or empty std::string
     */
    std::string StateName(Idx index) const;

    /**
     * State index lookup. This works only if state names are unique within
     * the stateset.
     *
     * @param rName
     *
     * @return
     *  Valid index or 0
     */
    Idx StateIndex(const std::string& rName) const;

    /**
     * Return copy of mAlphabet
     *
     * @return EventSet
     */
    EventSet Alphabet(void) const;

    /**
     * Return reference to mAlphabet
     *
     * @return EventSet
     */
    EventSet& Alphabetr(void);

    /**
     * Return copy of mStates
     *
     * @return StateSet
     */
    StateSet States(void) const;

    /**
     * Return reference to mStates
     *
     * @return StateSet
     */
    StateSet& Statesr(void);

    /**
     * Return copy of transition relation
     *
     * @return TransRel
     */
    TrnSet TransRel(void) const;

    /**
     * Return reference to transition relation
     *
     * @return TransRel
     */
    TrnSet& TransRelr(void);

    /** 
     * Return the transition relation sorted by ev, x1, x2
     *
     * @return
     *   Transition relation sorted by ev, x1, x2
     */
    TTransSet<TSortEvX1X2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> TransRelEvX1X2(void) const;

    /** 
     * Return the transition relation sorted by ev, x2, x1
     *
     * @return
     *   Transition relation sorted by ev, x2, x1
     */
    TTransSet<TSortEvX2X1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> TransRelEvX2X1(void) const;

    /** 
     * Return the transition relation sorted by x2, ev, x1
     *
     * @return
     *   Transition relation sorted by x2, ev, x1
     */
    TTransSet<TSortX2EvX1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> TransRelX2EvX1(void) const;

    /** 
     * Return the transition relation sorted by x2, x1, ev
     *
     * @return
     *   Transition relation sorted by x2, x1, ev
     */
    TTransSet<TSortX2X1Ev<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> TransRelX2X1Ev(void) const;

    /** 
     * Return the transition relation sorted by x1, x2, ev
     *
     * @return
     *   Transition relation sorted by x1, x2, ev
     */
    TTransSet<TSortX1X2Ev<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> TransRelX1X2Ev(void) const;

    /**
     * Return copy of initial states
     *
     * @return StateSet
     */
    StateSet InitStates(void) const;

    /**
     * Return reference to initial states
     *
     * @return StateSet
     */
    StateSet& InitStatesr(void);

    /**
     * Return copy of marked states
     *
     * @return StateSet
     */
    StateSet MarkedStates(void) const;

    /**
     * Return reference to marked states
     *
     * @return StateSet
     */
    StateSet& MarkedStatesr(void);

    /**
     * Set the alphabet to events executed in transitions
     */
    void MinimizeAlphabet(void);

    /**
     * Return used events (executed in transitions)
     *
     * @return EventSet
     */
    EventSet UsedEvents(void) const;

    /**
     * Return unused events
     *
     * @return EventSet
     */
    EventSet UnusedEvents(void) const;

    /**
     * Return active event set at state x1
     *
     * @param x1
     *   x1 Idx
     *
     * @return EventSet
     */
    virtual EventSet ActiveEventSet(Idx x1) const;

    /**
     * Return the states covered by transitions als indices
     *
     * @return StateSet
     */
    StateSet TransRelStateSpace(void) const;

    /**
     * Return the successor states of state x1
     *
     * @return StateSet
     */
    StateSet TransRelStateSpace(Idx x1) const;

    /**
     * Set default names for this generator's state indices
     * Idx=15 -> Name="15"
     * Only useful for debugging purposes.
     */
    void SetDefaultStateNames(void);

    /**
     * Remove all names from generator's state indices
     */
    void ClrStateNames(void);

    /**
     * Create another version of this generator with versioned event indices
     * and corresponding new unique event indices. State names stay the 
     * same but with different indices of course.
     *
     * @param version
     *   Integer value which corresponds to created version
     * @param rResGen
     *   Resulting versioned generator
     */
    void Version(Idx version, TGenerator<TrnSet>& rResGen);

    /**
     * Create another version of this generator with versioned event indices
     * and corresponding new unique event indices. State names stay the 
     * same but with different indices of course.
     *
     * @param rVersion
     *   String value which corresponds to created version
     * @param rResGen
     *   Resulting versioned generator
     */
    void Version(const std::string& rVersion, TGenerator<TrnSet>& rResGen);

    /**
     * Compute set of accessible states
     */
    StateSet AccessibleSet(void) const;

    /**
     * Make generator accessible
     *
     * @return
     *   True if generator contains at least one initial state
     */
    bool Accessible(void);

    /**
     * Check if generator is accessible
     *
     * @return
     *   True if generator contains at least one initial state
     */
    bool IsAccessible(void) const;

    /**
     * Compute set of Coaccessible states
     */
    StateSet CoaccessibleSet(void) const;

    /**
     * Make generator Coaccessible
     *
     * @return
     *   True if generator contains at least one marked state
     */
    bool Coaccessible(void);

    /**
     * Check if generator is Coaccessible
     *
     * @return
     *   True if generator contains at least one marked state
     */
    bool IsCoaccessible(void) const;

    /**
     * Computer set of trim states
     */
    StateSet TrimSet(void) const;

    /**
     * Make generator trim
     *
     * @return 
     *   True if generator contains at least one initial state and
     *   one marked state after internally calling Accessible() and
     *   Coaccessible().
     */
    bool Trim(void);

    /**
     * Check if generator is trim
     *
     * @return
     *   True if generator is trim
     */
    bool IsTrim(void) const;

    /**
     * Create empty Generator with proper tables set on stack
      * Calls NewGenerator()
     *
     * @return
     *   New Generator
     */
    TGenerator<TrnSet> NewGen(void) const;

    /**
     * Create empty Generator with proper tables set on heap
          * Calls NewGeneratorp()
     *
     * @return
     *   New Generator*
     */
    TGenerator<TrnSet>* NewGenp(void) const;

    /**
     * Create empty Generator with proper tables set on stack
     *
     * @return
     *   New Generator
     */
    TGenerator<TrnSet> NewGenerator(void) const;

    /**
     * Create empty Generator with proper tables set on heap
     *
     * @return
     *   New Generator*
     */
    TGenerator<TrnSet>* NewGeneratorp(void) const;

    /**
     * Get minimal index map for generator states
     *
     * @return
     *   map<Idx,Idx> from unique states to (local) minimal states
     */
    std::map<Idx,Idx> MinStateIndexMap(void) const;

    /** Order for sorting containers of Generators */
    bool operator < (const TGenerator<TrnSet>& rOtherGen) const {
      if (mId < rOtherGen.mId) {
        return true;
      }
      else {
        return false;
      }
    }

    /**
     * Create a string similar to the output of Write()
     *
     * @return
     *   std::string
     */
    std::string ToString() const;

    /**
     * Return pretty printable event name for index.
     * Primary meant for debugging messages
     *
     * @param index
     *   Event index
     *
     * @return
     *   std::string
     */
    std::string EStr(Idx index) const;

    /**
     * Return pretty printable state name for index.
     * Primary meant for debugging messages
     *
     * @param index
     *   State index
     *
     * @return
     *   std::string
     */
    std::string SStr(Idx index) const;

  
    // subclass may access all members
  protected:
    /** Name of generator */
    std::string mMyName;

    /** Number of generator */
    Idx mId;

    /** Number of generator objects */
    static Idx msObjectCount;

    /** Event symbol table */
    EventSymbolTable* mpEventSymbolTable;

    /** State symbol table */
    StateSymbolTable* mpStateSymbolTable;

    /** Generator alphabet */
    EventSet mAlphabet;

    /** Generator state set */
    StateSet mStates;

    /** Transition relation */
    TrnSet mTransRel;

    /** Initial states */
    StateSet mInitStates;

    /** Marked states */
    StateSet mMarkedStates;

    /** No state names */
    bool mStateNamesEnabled;

    /** Static repository for maintaining copys with same indices */
    static std::map<TGenerator<TrnSet>*, std::set<TGenerator<TrnSet>*>* > mGeneratorCopySets;

    /** Bool value for quick check, whether the Generator got copies */
        bool mIsUnique;

    /**
     * Write transition relation with respect to rMinIndexMap
     */
    void WriteTransRel(TokenWriter& rTw, 
        const std::map<Idx,Idx>& rMinIndexMap) const;

    /**
     * Debug write transition relation. Unique stateindices and eventnames
     * are used.
     */
    void DWriteTransRel(TokenWriter& rTw) const;

    /**
     * Read the generator's name from a file
     * 
     * @param pFileName
     *   File to read from
     *
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    void ReadGeneratorName(const char* pFileName);

    /**
     * Read the generator's name from a TokenReader
     * 
     * @param rTr
     *   Reference to TokenReader
     *
     * @exception std::ios::failure
     *   Thrown on i/o error.
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    void ReadGeneratorName(TokenReader& rTr);

    /**
     * Read the generator's alphabet from a file.
     *
     * @param pFileName
     *   File to read from
     *
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    void ReadAlphabet(const char* pFileName);

    /**
     * Read the generator's alphabet from a TokenReader
     *
     * @param rTr
     *   Reference to TokenReader
     *
     * @exception std::ios::failure
     *   Thrown on i/o error.
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    void ReadAlphabet(TokenReader& rTr);

    /**
     * Read the generator's mStates from a file. rMinIndexMap is created
     * while reading tokens.
     *
     * @param pFileName
     *   File to read from
     * @param rMinIndexMap
     *   Mapping from local indices in file to indices in program
     * @param rTmpStateNameMap
     *   Temporary map<std::string,Idx> for fast insertion of named states 
     *
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    void ReadStates(const char* pFileName, std::map<Idx,Idx>& rMinIndexMap, 
        std::map<std::string,Idx>& rTmpStateNameMap);

    /**
     * Read the generator's stateset from a TokenReader. 
     * rMinIndexMap is created  while reading tokens.
     *
     * @param rTr
     *   Reference to TokenReader
     * @param rMinIndexMap
     *   Mapping from local indices in file to indices in program
     * @param rTmpStateNameMap
     *   Temporary map<std::string,Idx> for fast insertion of named states 
     *
     * @exception ios::failure
     *   Thrown on i/o error.
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    void ReadStates(TokenReader& rTr, std::map<Idx,Idx>& rMinIndexMap, 
        std::map<std::string,Idx>& rTmpStateNameMap);

    /**
     * Read the generator's transition relation from a file. rMinIndexMap is
     * used for creating transrel.
     *
     * @param pFileName
     *   File to read from
     * @param rMinIndexMap
     *   Mapping from local indices in file to indices in program
     * @param rTmpStateNameMap
     *   Temporary map<std::string,Idx> for fast insertion of named states 
     *
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    virtual void ReadTransRel(const char* pFileName, 
        const std::map<Idx,Idx>& rMinIndexMap, 
        const std::map<std::string,Idx>& rTmpStateNameMap); // virtual added

    /**
     * Read the generator's transition relation from a TokenReader. 
     * rMinIndexMap is used for creating transrel.
     *
     * @param rTr
     *   Reference to TokenReader
     * @param rMinIndexMap
     *   Mapping from local indices in file to indices in program
     * @param rTmpStateNameMap
     *   Temporary map<std::string,Idx> for fast insertion of named states 
     *
     * @exception ios::failure
     *   Thrown on i/o error.
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    virtual void ReadTransRel(TokenReader& rTr, 
        const std::map<Idx,Idx>& rMinIndexMap, 
        const std::map<std::string,Idx>& rTmpStateNameMap); //virtual added


    /**
     * Read the generator's initial states from a file. rMinIndexMap is
     * used for creating minindex initial states.
     *
     * @param pFileName
     *   File to read from
     * @param rMinIndexMap
     *   Mapping from local indices in file to indices in program
     * @param rTmpStateNameMap
     *   Temporary map<std::string,Idx> for fast insertion of named states 
     *
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    void ReadInitStates(const char* pFileName,
        const std::map<Idx,Idx>& rMinIndexMap, 
        const std::map<std::string,Idx>& rTmpStateNameMap);

    /**
     * Read the generator's initial states from a TokenReader. 
     * rMinIndexMap is used for creating minindex initial states.
     *
     * @param rTr
     *   Reference to TokenReader
     * @param rMinIndexMap
     *   Mapping from local indices in file to indices in program
     * @param rTmpStateNameMap
     *   Temporary map<std::string,Idx> for fast insertion of named states 
     *
     * @exception ios::failure
     *   Thrown on i/o error.
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    void ReadInitStates(TokenReader& rTr, 
        const std::map<Idx,Idx>& rMinIndexMap, 
        const std::map<std::string,Idx>& rTmpStateNameMap);

    /**
     * Read the generator's marked states from a file. rMinIndexMap is
     * used for creating mMarkedStates.
     *
     * @param pFileName
     *   File to read from
     * @param rMinIndexMap
     *   Mapping from local indices in file to indices in program
     * @param rTmpStateNameMap
     *   Temporary map<std::string,Idx> for fast insertion of named states 
     *
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    void ReadMarkedStates(const char* pFileName,
        const std::map<Idx,Idx>& rMinIndexMap, 
        const std::map<std::string,Idx>& rTmpStateNameMap);

    /**
     * Read the generator's marked states from a TokenReader. 
     * rMinIndexMap is used for creating mMarkedStates.
     *
     * @param rTr
     *   Reference to TokenReader
     * @param rMinIndexMap
     *   Mapping from local indices in file to indices in program
     * @param rTmpStateNameMap
     *   Temporary map<std::string,Idx> for fast insertion of named states 
     *
     * @exception std::ios::failure
     *   Thrown on i/o error.
     * @exception Exception
     *   Thrown on file format errors (id 50, 51)
     */
    void ReadMarkedStates(TokenReader& rTr,
        const std::map<Idx,Idx>& rMinIndexMap, 
        const std::map<std::string,Idx>& rTmpStateNameMap);

    /**
     * Get accessible states by filling accessibleset recursive
     *
     * @param accessibleset
     *   Reference to StateSet with accessibel states
     * @param startState
     *   State index to start with
     */
    void CheckAccessible(StateSet& accessibleset, Idx startState) const;

    /**
     * Get Coaccessible states by filling Coaccessibleset recursive
     *
     * @param Coaccessibleset
     *   Reference to StateSet with coaccessibel states
     * @param rtrel
     *   Reverse transition relation where x1 <-> x2
     * @param startState
     *   State index to start with
     */
    void CheckCoaccessible(StateSet& Coaccessibleset, 
        const TTransSet<TSortX2EvX1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>& rtrel, Idx startState) const;


};


typedef TGenerator<TransSet> Generator; 




// Generator::Generator(void)
template <class TrnSet>
TGenerator<TrnSet>::TGenerator(void) 
  : mMyName("Generator"), 
    mpEventSymbolTable(EventSet::StaticEventSymbolTablep()),
    mpStateSymbolTable(StateSet::StaticStateSymbolTablep()), 
    mStateNamesEnabled(true),
    mIsUnique(true)
{  
  FD_DG("Generator(" << this << ")::Generator()");
  msObjectCount++; // FIXME: put in initialation list
  mId = msObjectCount;
  mStates.Name("States");
  mInitStates.Name("InitStates");
  mMarkedStates.Name("MarkedStates");
}

// Generator::Generator(rOtherGen)
template <class TrnSet>
TGenerator<TrnSet>::TGenerator(const TGenerator<TrnSet>& rOtherGen)
  : mMyName(rOtherGen.mMyName),
    mpEventSymbolTable(EventSet::StaticEventSymbolTablep()),
    mpStateSymbolTable(StateSet::StaticStateSymbolTablep()),
    mAlphabet(rOtherGen.mAlphabet),
    mStates(rOtherGen.mStates),
    mTransRel(rOtherGen.mTransRel),
    mInitStates(rOtherGen.mInitStates),
    mMarkedStates(rOtherGen.mMarkedStates),
    // copy mStateNamesEnabled status
    mStateNamesEnabled(rOtherGen.mStateNamesEnabled)
{
  FD_DG("Generator(" << this << ")::Generator(" << &rOtherGen << ")");
  msObjectCount++;
  mId = msObjectCount;
  // insert generator address in copy sets
  TGenerator<TrnSet>* pothergen = const_cast<TGenerator<TrnSet>*>(&rOtherGen);
  RegisterGeneratorCopy(pothergen, this);
}

// Generator::Generator(pFileName)
template <class TrnSet>
TGenerator<TrnSet>::TGenerator(const char* pFileName)
  : mpEventSymbolTable(EventSet::StaticEventSymbolTablep()),
    mpStateSymbolTable(StateSet::StaticStateSymbolTablep()),
    mStateNamesEnabled(true),
    mIsUnique(true)
{
  FD_DG("Generator(" << this << ")::Generator(" << pFileName << ")");
  Read(pFileName);
  ++msObjectCount;
  mId = msObjectCount;
}

// virtual
// Generator::~Generator
template <class TrnSet>
TGenerator<TrnSet>::~TGenerator(void) {
  FD_DG("Generator(" << this << ")::~Generator()");
  // UnRegisterGeneratorCopy returns true if there are no copys
  if (UnRegisterGeneratorCopy(this)) {
    StateSet::iterator it;
    for (it = StatesBegin(); it != StatesEnd(); ++it) {
      FD_DC("~Generator: mpStateSymbolTable->DelState(" << *it << ")");
      try {
          mpStateSymbolTable->DelState(static_cast<Idx>(*it));
      }
      catch (Exception&) {
        // just ignore exception
      }
    }
  }
}

// EventSymbolTable()
template <class TrnSet>
EventSymbolTable* TGenerator<TrnSet>::EventSymbolTablep(void) {
  return mpEventSymbolTable;
}

// EventSymbolTable() const
template <class TrnSet>
EventSymbolTable* TGenerator<TrnSet>::EventSymbolTablep(void) const {
  return mpEventSymbolTable;
}

// Generator::StaticEventSymbolTable()
template <class TrnSet>
EventSymbolTable* TGenerator<TrnSet>::StaticEventSymbolTablep(void) {
  return EventSet::StaticEventSymbolTablep();
}

// StateSymbolTablep()
template <class TrnSet>
StateSymbolTable* TGenerator<TrnSet>::StateSymbolTablep(void) {
  return mpStateSymbolTable;
}

// StateSymbolTablep() const
template <class TrnSet>
StateSymbolTable* TGenerator<TrnSet>::StateSymbolTablep(void) const {
  return mpStateSymbolTable;
}

// Generator::StaticStateSymbolTablep()
template <class TrnSet>
StateSymbolTable* TGenerator<TrnSet>::StaticStateSymbolTablep(void) {
  return StateSet::StaticStateSymbolTablep();
}

template <class TrnSet>
std::map<TGenerator<TrnSet>*, std::set<TGenerator<TrnSet>*>* > TGenerator<TrnSet>::mGeneratorCopySets;

// RegisterGeneratorCopy(existingId, copyId)
template <class TrnSet>
void TGenerator<TrnSet>::RegisterGeneratorCopy(TGenerator<TrnSet>* pExistingGen, TGenerator<TrnSet>* pCopyGen) {
  // unset both Generators as unique
  pExistingGen->mIsUnique = false;
  pCopyGen->mIsUnique = false;
  // find mId of existing generator in mGeneratorCopySets map
  typename std::map<TGenerator<TrnSet>*, std::set<TGenerator<TrnSet>*>* >::iterator mapit = 
      mGeneratorCopySets.find(pExistingGen);
  // if there is no entry
  if (mapit == mGeneratorCopySets.end()) {
    // create new copyset
    std::set<TGenerator<TrnSet>*>* copysetptr = new std::set<TGenerator<TrnSet>*>;
    // insert both mIds in copyset
    copysetptr->insert(pExistingGen);
    copysetptr->insert(pCopyGen);
    // register copyset for both mIds
    mGeneratorCopySets[pExistingGen] = copysetptr;
    mGeneratorCopySets[pCopyGen] = copysetptr;
  }
  // if there is an entry
  else {
    // insert mId of new copy in existing copyset
    mapit->second->insert(pCopyGen);
    // register copyset for mId of copy
    mGeneratorCopySets[pCopyGen] = mapit->second;
  }
}

// UnRegisterGeneratorCopy(id)
template <class TrnSet>
bool TGenerator<TrnSet>::UnRegisterGeneratorCopy(TGenerator<TrnSet>* pGen) {
  // find mId of generator in mGeneratorCopySets map
  typename std::map<TGenerator<TrnSet>*, std::set<TGenerator<TrnSet>*>* >::iterator mapit = 
      mGeneratorCopySets.find(pGen);
  // if there is no entry
  if (mapit == mGeneratorCopySets.end()) {
    // delete statenames = return true
    return true;
  }
  // if there is an entry
  else {
    // remove id from copyset
    mapit->second->erase(pGen);
    // if copyset empty now
    if (mapit->second->empty()) {
      // deallocate copyset
      delete(mapit->second);
      // remove from map
      mGeneratorCopySets.erase(pGen);
      // delete statenames = return true
      return true;
    }
    // there are more copys
    else {
      // remove from map
      mGeneratorCopySets.erase(pGen);
      // if only one copy left mark as unique
      if (mapit->second->size() == 1) {
        TGenerator<TrnSet>* tmpgen = *(mapit->second->begin());
        tmpgen->mIsUnique = true;
      }
      // don't delete statenames = return false
      return false;
    }
  }
}

// virtual
// operator =
template <class TrnSet>
TGenerator<TrnSet>& TGenerator<TrnSet>::operator = (const TGenerator<TrnSet>& rOtherGen) {
  FD_DG("Generator(" << this << ")::operator = " << &rOtherGen);

    mMyName = rOtherGen.mMyName;
    // copy generator five tuple
    mAlphabet = rOtherGen.mAlphabet;
    mStates = rOtherGen.mStates;
    mTransRel = rOtherGen.mTransRel;
    mInitStates = rOtherGen.mInitStates;
    mMarkedStates = rOtherGen.mMarkedStates;
    // copy mStateNamesEnabled status
    mStateNamesEnabled = rOtherGen.mStateNamesEnabled;
    // insert generator number in copy sets
    TGenerator<TrnSet>* pothergen = const_cast<TGenerator<TrnSet>*>(&rOtherGen);    
    RegisterGeneratorCopy(pothergen, this);
    //rOtherGen.Copy(*this);
 
  return *this;
}

// virtual
// Copy(gen) Klaus Schmidt 10.04.2006  
template <class TrnSet>
void TGenerator<TrnSet>::Copy(TGenerator<TrnSet>& rGen) const {
  // Helpers:
  //std::map<Idx,Idx> eventreindexmap;
  std::map<Idx,Idx> statereindexmap;
  Copy(rGen, statereindexmap);
} 
//virtual
// Copy(gen, rStateReindexMap&)  Klaus Schmidt 10.04.2006  
template <class TrnSet>
void TGenerator<TrnSet>::Copy(TGenerator<TrnSet>& rGen, std::map<Idx,Idx>& rStateReindexMap) const {
  FD_DG("Generator(" << this << ")::Copy(gen&)");
  // prepare result
  rGen.Clear();
  //rEventReindexMap.clear();
  rStateReindexMap.clear();
  // set members
  rGen.mMyName = mMyName;
  // copy mStates by creating new indices
  rGen.mStates = mStates.ReindexedCopy(rStateReindexMap);
  // copy mInitStates
  rGen.mInitStates = mInitStates.ReindexedCopy(rStateReindexMap);
  // copy mMarkedStates
  rGen.mMarkedStates = mMarkedStates.ReindexedCopy(rStateReindexMap);

  //if ((mpEventSymbolTable == rGen.mpEventSymbolTable) 
    //  and (mpStateSymbolTable == rGen.mpStateSymbolTable)) {
    // copy mAlphabet
  rGen.mAlphabet = mAlphabet;
  // copy transition relation
//  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::const_iterator tit;
  typename TrnSet::const_iterator tit;
  for (tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
      rGen.SetTransition(rStateReindexMap[tit->X1], tit->Ev, 
      rStateReindexMap[tit->X2]);
  }
//  else if ((mpEventSymbolTable != rGen.mpEventSymbolTable) 
//      and (mpStateSymbolTable != rGen.mpStateSymbolTable)) {
//    rGen.mAlphabet = mAlphabet.ReindexedCopy(rGen.EventSymbolTablep(),
//        rEventReindexMap);
//    // copy transition relation
//    TransSet::const_iterator tit;
//    for (tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
//      rGen.SetTransition(rStateReindexMap[tit->X1], rEventReindexMap[tit->Ev], 
//          rStateReindexMap[tit->X2]);
//    }
//  }
//  else {
//    FD_ERR("Generator::Copy(" << &rGen << "): symboltable mismatch");
//    abort();
//  }

  // mStateNamesEnabled status
  rGen.mStateNamesEnabled = mStateNamesEnabled;
} 

// NewEventSet()
template <class TrnSet>
EventSet TGenerator<TrnSet>::NewEventSet(void) const {
  return EventSet();
}

// NewEventSetp()
template <class TrnSet>
EventSet* TGenerator<TrnSet>::NewEventSetp(void) const {
  return new EventSet();
}

// NewStateSet()
template <class TrnSet>
StateSet TGenerator<TrnSet>::NewStateSet(void) const {
  return StateSet();
}

// NewStateSetp()
template <class TrnSet>
StateSet* TGenerator<TrnSet>::NewStateSetp(void) const {
  return new StateSet();
}

// Generator::msObjectCount
template <class TrnSet>
Idx TGenerator<TrnSet>::msObjectCount = 0;

// virtual
// Write()
template <class TrnSet>
void TGenerator<TrnSet>::Write(void) {
  TokenWriter tw;
  Write(tw);
}

// virtual
// Write(pFileName, openmode=default)
template <class TrnSet>
void TGenerator<TrnSet>::Write(const char* pFileName, std::ios::openmode openmode) {
  try {
    TokenWriter tw(pFileName, openmode);
    Write(tw);
  }
  catch (faudes::Exception&) {
    std::stringstream errstr;
    errstr << "Exception opening/writing file \"" << pFileName << "\"";
    throw Exception("Generator::write", errstr.str(), 2);
  }
}

// virtual
// Write(tw)
template <class TrnSet>
void TGenerator<TrnSet>::Write(TokenWriter& rTw) {
  // HELPERS:
  std::map<Idx,Idx> minstateindexmap = MinStateIndexMap();
  // write generator
  rTw.WriteBegin("Generator");
  rTw << mMyName;
  rTw << "\n";
  rTw << "\n";
  mAlphabet.Write(rTw);
  rTw << "\n";
  mStates.Write(rTw, minstateindexmap, true);
  rTw << "\n";
  WriteTransRel(rTw, minstateindexmap);
  rTw << "\n";
  mInitStates.Write(rTw, minstateindexmap, false);
  rTw << "\n";
  mMarkedStates.Write(rTw, minstateindexmap, false);
  rTw << "\n";
  rTw.WriteEnd("Generator");
}

//virtual
template <class TrnSet>
void TGenerator<TrnSet>::WriteTransRel(TokenWriter& rTw, 
    const std::map<Idx,Idx>& rMinIndexMap) const {
   mTransRel.Write(rTw, rMinIndexMap, Alphabet(), States());
}

// virtual
// DDotWrite(rFileName)
template <class TrnSet>
void TGenerator<TrnSet>::DDotWrite(const std::string& rFileName) const {
  DDotWrite(rFileName, "png");
}

// virtual
// DDotWrite(rFileName, rFileType)
template <class TrnSet>
void TGenerator<TrnSet>::DDotWrite(const std::string& rFileName, 
    const std::string& rFileType) const {
  StateSet::const_iterator lit;
//  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::const_iterator tit;
  typename TrnSet::const_iterator tit;
  // check file extension
  if (rFileType == "canon") {
  }
  else if (rFileType == "dot") { 
  }
  else if (rFileType == "xdot") {
  }
  else if (rFileType == "cmap") {
  }
  else if (rFileType == "dia") {
  }
  else if (rFileType == "fig") {
  }
  else if (rFileType == "gd") {
  }
  else if (rFileType == "gd2") {
  }
  else if (rFileType == "gif") {
  }
  else if (rFileType == "hpgl") {
  }
  else if (rFileType == "imap") {
  }
  else if (rFileType == "cmapx") {
  }
  else if (rFileType == "ismap") {
  }
  else if (rFileType == "jpg") {
  }
  else if (rFileType == "jpeg") {
  }
  else if (rFileType == "mif") {
  }
  else if (rFileType == "mp") {
  }
  else if (rFileType == "pcl") {
  }
  else if (rFileType == "pic") {
  }
  else if (rFileType == "plain") {
  }
  else if (rFileType == "plain-ext") {
  }
  else if (rFileType == "png") {
  }
  else if (rFileType == "ps") {
  }
  else if (rFileType == "ps2") {
  }
  else if (rFileType == "svg") {
  }
  else if (rFileType == "svgz") {
  }
  else if (rFileType == "vrml") {
  }
  else if (rFileType == "vtx") {
  }
  else if (rFileType == "wbmp") {
  }
  else {
    std::stringstream errstr;
    errstr << "Dot output format \"" << rFileType << "\" unknown";
    throw Exception("Generator::DotWrite", errstr.str(), 5);
  }
  std::string basename = rFileName;
  if (basename.rfind(".gen") < basename.size()) {
    basename.resize(basename.rfind(".gen"));
  }  
  std::string dotfilename = basename+".dot";
  std::string outfilename = basename+"."+rFileType;
  std::string dotcommand = "dot -T"+rFileType+" \""+dotfilename+"\" -o \""+outfilename+"\"";
  try {
    std::ofstream stream;
    stream.exceptions(std::ios::badbit|std::ios::failbit);
    stream.open(dotfilename.c_str());
    stream << "digraph \"" << Name() << "\" {" << std::endl;
    stream << "  rankdir=LR" << std::endl;
    stream << "  node [shape=circle];" << std::endl;
    stream << std::endl;
    stream << "  // istates" << std::endl;
    int i = 1;
    for (lit = InitStatesBegin(); lit != InitStatesEnd(); ++lit) {
      stream << "  dummyinit" << i << " [shape=point];" << std::endl;
      stream << "  dummyinit" << i << " -> \"" 
        << SStr(static_cast<Idx>(*lit)) << "\";" << std::endl; 
      i++;
    }
    stream << std::endl;
    stream << "  // mstates" << std::endl;
    for (lit = MarkedStatesBegin(); lit != MarkedStatesEnd(); ++lit) {
      stream << "  \"" << SStr(static_cast<Idx>(*lit)) 
        << "\" [shape=doublecircle];" << std::endl;
    }
    stream << std::endl;
    stream << "  // rest of stateset" << std::endl;
    for (lit = StatesBegin(); lit != StatesEnd(); ++lit) {
      // if not in mInitStates or mMarkedStates
      if (! (ExistsInitState(static_cast<Idx>(*lit)) || ExistsMarkedState(static_cast<Idx>(*lit)))) {
        stream << "  \"" << SStr(static_cast<Idx>(*lit)) 
          << "\";" << std::endl;
      }
    }
    stream << std::endl;
    stream << "  // transition relation" << std::endl;
    for (tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
      stream << "  \"" << SStr(tit->X1) 
        << "\" -> \"" << SStr(tit->X2)
        // << "\" [label=\"" << EStr(tit->Ev) << "\"];" << std::endl;
        << "\" [label=\"" << EventName(tit->Ev) << "\"];" << std::endl;
    }
    stream << "};" << std::endl;
    stream.close();
    if (system(dotcommand.c_str()) != 0) {
      throw Exception("Generator::DotWrite", 
          "Error creating png from dotfile", 4);
    }
  }
  catch (std::ios::failure&) {
    throw Exception("Generator::DotWrite", 
        "Exception opening/writing dotfile \""+dotfilename+"\"", 3);
  }
}

// virtual
// DotWrite(rFileName)
template <class TrnSet>
void TGenerator<TrnSet>::DotWrite(const std::string& rFileName) const {
  DotWrite(rFileName, "png");
}

// virtual
// DotWrite(rFileName, rFileType)
template <class TrnSet>
void TGenerator<TrnSet>::DotWrite(const std::string& rFileName, 
    const std::string& rFileType) const {
  std::map<Idx,Idx> minindexmap = MinStateIndexMap();
  StateSet::const_iterator lit;
//  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::const_iterator tit;
  typename TrnSet::const_iterator tit;
  // check file extension
  if (rFileType == "canon") {
  }
  else if (rFileType == "dot") { 
  }
  else if (rFileType == "xdot") {
  }
  else if (rFileType == "cmap") {
  }
  else if (rFileType == "dia") {
  }
  else if (rFileType == "fig") {
  }
  else if (rFileType == "gd") {
  }
  else if (rFileType == "gd2") {
  }
  else if (rFileType == "gif") {
  }
  else if (rFileType == "hpgl") {
  }
  else if (rFileType == "imap") {
  }
  else if (rFileType == "cmapx") {
  }
  else if (rFileType == "ismap") {
  }
  else if (rFileType == "jpg") {
  }
  else if (rFileType == "jpeg") {
  }
  else if (rFileType == "mif") {
  }
  else if (rFileType == "mp") {
  }
  else if (rFileType == "pcl") {
  }
  else if (rFileType == "pic") {
  }
  else if (rFileType == "plain") {
  }
  else if (rFileType == "plain-ext") {
  }
  else if (rFileType == "png") {
  }
  else if (rFileType == "ps") {
  }
  else if (rFileType == "ps2") {
  }
  else if (rFileType == "svg") {
  }
  else if (rFileType == "svgz") {
  }
  else if (rFileType == "vrml") {
  }
  else if (rFileType == "vtx") {
  }
  else if (rFileType == "wbmp") {
  }
  else {
    std::stringstream errstr;
    errstr << "Dot output format \"" << rFileType << "\" unknown";
    throw Exception("Generator::DotWrite", errstr.str(), 5);
  }
  std::string basename = rFileName;
  if (basename.rfind(".gen") < basename.size()) {
    basename.resize(basename.rfind(".gen"));
  }  
  std::string dotfilename = basename+".dot";
  std::string outfilename = basename+"."+rFileType;
  std::string dotcommand = "dot -T"+rFileType+" \""+dotfilename+"\" -o \""+outfilename+"\"";
  try {
    std::ofstream stream;
    stream.exceptions(std::ios::badbit|std::ios::failbit);
    stream.open(dotfilename.c_str());
    stream << "digraph \"" << Name() << "\" {" << std::endl;
    stream << "  rankdir=LR" << std::endl;
    stream << "  node [shape=circle];" << std::endl;
    stream << std::endl;
    stream << "  // istates" << std::endl;
    int i = 1;
    for (lit = InitStatesBegin(); lit != InitStatesEnd(); ++lit) {
      stream << "  dummyinit" << i << " [shape=point];" << std::endl;
      stream << "  dummyinit" << i << " -> \"" 
        << mInitStates.MinIdxStr(minindexmap, static_cast<Idx>(*lit)) << "\";" << std::endl; 
      i++;
    }
    stream << std::endl;
    stream << "  // mstates" << std::endl;
    for (lit = MarkedStatesBegin(); lit != MarkedStatesEnd(); ++lit) {
      stream << "  \"" << mMarkedStates.MinIdxStr(minindexmap, static_cast<Idx>(*lit)) 
        << "\" [shape=doublecircle];" << std::endl;
    }
    stream << std::endl;
    stream << "  // rest of stateset" << std::endl;
    for (lit = StatesBegin(); lit != StatesEnd(); ++lit) {
      // if not in mInitStates or mMarkedStates
      if (! (ExistsInitState(static_cast<Idx>(*lit)) || ExistsMarkedState(static_cast<Idx>(*lit)))) {
        stream << "  \"" << mStates.MinIdxStr(minindexmap, static_cast<Idx>(*lit)) 
          << "\";" << std::endl;
      }
    }
    stream << std::endl;
    stream << "  // transition relation" << std::endl;
    for (tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
      stream << "  \"" << mStates.MinIdxStr(minindexmap, tit->X1) 
        << "\" -> \"" << mStates.MinIdxStr(minindexmap, tit->X2)
        // << "\" [label=\"" << EStr(tit->Ev) << "\"];" << std::endl;
        << "\" [label=\"" << EventName(tit->Ev) << "\"];" << std::endl;
    }
    stream << "};" << std::endl;
    stream.close();
    if (system(dotcommand.c_str()) != 0) {
      throw Exception("Generator::DotWrite", 
          "Error creating png from dotfile", 4);
    }
  }
  catch (std::ios::failure&) {
    throw Exception("Generator::DotWrite", 
        "Exception opening/writing dotfile \""+dotfilename+"\"", 3);
  }
}

// virtual
// DWrite()
template <class TrnSet>
void TGenerator<TrnSet>::DWrite(void) const {
  TokenWriter tw;
  tw.WriteBegin("Generator");
  tw << mMyName;
  tw << "\n";
  mAlphabet.Write(tw);
  tw << "\n";
  mStates.DWrite(tw);
  tw << "\n";
  DWriteTransRel(tw);
  tw << "\n";
  mInitStates.DWrite(tw);
  tw << "\n";
  mMarkedStates.DWrite(tw);
  tw.WriteEnd("Generator");
  tw << "\n";
  std::cout << std::endl << "copys: ";
  typename std::map<TGenerator<TrnSet>*, std::set<TGenerator<TrnSet>*>* >::const_iterator mapit;
  for (mapit = mGeneratorCopySets.begin(); 
      mapit != mGeneratorCopySets.end(); ++mapit) {
    typename std::set<TGenerator<TrnSet>*>::const_iterator setit;
    std::cout << "{ ";
    for (setit = mapit->second->begin(); setit != mapit->second->end(); 
        ++setit) {
      std::cout << *setit << " ";
    }
    std::cout << "} ";
  }
  std::cout << std::endl;
}

template <class TrnSet>
void TGenerator<TrnSet>::DWriteTransRel(TokenWriter& rTw) const 
{
mTransRel.DWrite(rTw, Alphabet(), States());
}

template <class TrnSet>
std::string TGenerator<TrnSet>::ToString() const {
  std::stringstream sstr;
  sstr 
    << "<Generator>" << std::endl
    << Name() << std::endl
    << std::endl
    << "<Alphabet>" << std::endl
    << AlphabetToString() << std::endl
    << "</Alphabet>" << std::endl
    << std::endl
    << "<States>" << std::endl
    << StatesToString() << std::endl
    << "</States>" << std::endl
    << std::endl
    << "<TransRel>" << std::endl
    << TransRelToString() << std::endl
    << "</TransRel>" << std::endl
    << std::endl
    << "<InitStates" << std::endl
    << InitStatesToString() << std::endl
    << "</InitStates>" << std::endl
    << std::endl
    << "<MarkedStates>" << std::endl
    << MarkedStatesToString() << std::endl
    << "</MarkedStates>" << std::endl
    << std::endl
    << "</Generator>" << std::endl;
  return sstr.str();
}

// AlphabetToString()
template <class TrnSet>
std::string TGenerator<TrnSet>::AlphabetToString(void) const {
  return mAlphabet.ToString();
}

// StatesToString()
template <class TrnSet>
std::string TGenerator<TrnSet>::StatesToString(void) const {
  return mStates.ToString();
}

// TransRelToString()
template <class TrnSet>
std::string TGenerator<TrnSet>::TransRelToString(void) const {
  return mTransRel.ToString();
}

// InitStatesToString()
template <class TrnSet>
std::string TGenerator<TrnSet>::InitStatesToString(void) const {
  return mInitStates.ToString();
}

// MarkedStatesToString()
template <class TrnSet>
std::string TGenerator<TrnSet>::MarkedStatesToString(void) const {
  return mMarkedStates.ToString();
}

// virtual
// Read(pFileName)
template <class TrnSet>
void TGenerator<TrnSet>::Read(const char* pFileName) {
  try {
    TokenReader tr(pFileName);
    Read(tr);
  }
  catch (faudes::Exception&) {  
    std::stringstream errstr;
    errstr << "Exception opening/reading file \"" << pFileName << "\"";
    throw Exception("Generator::Read", errstr.str(), 1);
  }
}

// virtual
// Read(tr)
template <class TrnSet>
void TGenerator<TrnSet>::Read(TokenReader& rTr) {
  // HELPERS:
  std::map<Idx,Idx> minstateindexmap;
  std::map<std::string,Idx> tmpstringmap;
  // ALGORITHM:
  FD_DG("Generator(" << this << ")::Read(" << rTr.FileName() << ")");
  rTr.ReadBegin("Generator");
  // get the Generator "name" token
  ReadGeneratorName(rTr);
  // read mAlphabet
  ReadAlphabet(rTr);
  // read stateset
  ReadStates(rTr, minstateindexmap, tmpstringmap);
  // read transrel
  ReadTransRel(rTr, minstateindexmap, tmpstringmap);
  // read istates
  ReadInitStates(rTr, minstateindexmap, tmpstringmap);
  // read mstates
  FD_DV("geht?");
  ReadMarkedStates(rTr, minstateindexmap, tmpstringmap);
  // read end (syntax check only)
  rTr.ReadEnd("Generator");
  FD_DV("geht?");
}

// ReadGeneratorName(pFileName)
template <class TrnSet>
void TGenerator<TrnSet>::ReadGeneratorName(const char* pFileName) {
  TokenReader tr(pFileName);
  tr.SeekBegin("Generator");  // mimique old behaviour .. use "ReadBegin" instead   
  ReadGeneratorName(tr);
}

// ReadGeneratorName(tr)
template <class TrnSet>
void TGenerator<TrnSet>::ReadGeneratorName(TokenReader& rTr) {
  FD_DV("Generator(" << this << ")::ReadGeneratorName(\"" 
      << rTr.FileName() << "\")");
  try {
    mMyName=rTr.ReadName();
  } 
  catch (faudes::Exception&) {
    std::stringstream errstr;
    errstr << "Name of generator expected " << rTr.FileLine();
    throw Exception("Generator::ReadGeneratorName", errstr.str(), 50);
  }
}

// ReadAlphabet(pFileName)
template <class TrnSet>
void TGenerator<TrnSet>::ReadAlphabet(const char* pFileName) {
  FD_DV("Generator(" << this << ")::ReadAlphabet(\"" << pFileName << "\")");
  mAlphabet.Read(pFileName);
}

// ReadAlphabet(rTr) 
template <class TrnSet>
void TGenerator<TrnSet>::ReadAlphabet(TokenReader& rTr) {
  rTr.SeekBegin("Alphabet");  // mimique old behaviour .. use "ReadBegin" instead ?   
  FD_DV("Generator(" << this << ")::ReadAlphabet(\"" 
      << rTr.FileName() << "\")");
  mAlphabet.Read(rTr);
}

// ReadStates(pFileName) 
template <class TrnSet>
void TGenerator<TrnSet>::ReadStates(const char* pFileName, 
    std::map<Idx,Idx>& rMinIndexMap, 
        std::map<std::string,Idx>& rTmpStateNameMap) {
  TokenReader tr(pFileName);
  ReadStates(tr, rMinIndexMap, rTmpStateNameMap);
}

// ReadStates(tr) 
template <class TrnSet>
void TGenerator<TrnSet>::ReadStates(TokenReader& rTr, 
    std::map<Idx,Idx>& rMinIndexMap,
    std::map<std::string,Idx>& rTmpStateNameMap) {
  FD_DV("Generator(" << this << ")::ReadStates(\"" 
      << rTr.FileName() << "\")");
  rTr.SeekBegin("States");      // mimique old behaviour ... better use "ReadBegin" ?
  Token token;
  Idx fillindices;
  Idx minindex = 0;
  mStates.Clear();
  rMinIndexMap.clear();
  rTmpStateNameMap.clear();
  while((rTr >> token) && (token.Type() != Token::End)) {
    if (token.Type() == Token::Index) {
      fillindices = token.IndexValue();
      while (fillindices > 0) {
        minindex++;
        rMinIndexMap[minindex] = InsState();
        --fillindices;
      }
    }
    else if (token.Type() == Token::Name) {
      ++minindex;
      // first create new index...
      // ...then set name for index
      // this is much faster than calling InsState(string)
      Idx tmpindex = InsState();
      StateName(tmpindex, token.StringValue());

      rMinIndexMap[minindex] = tmpindex;
      rTmpStateNameMap[token.StringValue()] = tmpindex;
    }
  }
  mStates.Name("States");
}

// ReadTransRel(pFileName)
template <class TrnSet>
void TGenerator<TrnSet>::ReadTransRel(const char* pFileName, 
    const std::map<Idx,Idx>& min_index_map, 
    const std::map<std::string,Idx>& tmp_statename_map) {
  TokenReader tr(pFileName);
  ReadTransRel(tr, min_index_map, tmp_statename_map);
}

// ReadTransRel(tr)
template <class TrnSet>
void TGenerator<TrnSet>::ReadTransRel(TokenReader& rTr, 
    const std::map<Idx,Idx>& rMinIndexMap,
    const std::map<std::string,Idx>& rTmpStateNameMap) {
  FD_DV("Generator(" << this << ")::ReadTransRel(\"" << rTr.FileName() << "\")");
  try {
  mTransRel.Read(rTr, rMinIndexMap, rTmpStateNameMap, Alphabet());
  }
  catch (faudes::Exception&) {
    std::stringstream errstr;
    errstr << "Reading TransRel failed in " << rTr.FileLine();
    throw Exception("Generator::ReadTransRel", errstr.str(), 50); // FIXME: ID?
  }
}

// ReadInitStates(pFileName) 
template <class TrnSet>
void TGenerator<TrnSet>::ReadInitStates(const char* pFileName,
    const std::map<Idx,Idx>& rMinIndexMap, 
    const std::map<std::string,Idx>& rTmpStateNameMap) {
  TokenReader tr(pFileName);
  ReadInitStates(tr, rMinIndexMap, rTmpStateNameMap);
}

// ReadInitStates(tr) 
template <class TrnSet>
void TGenerator<TrnSet>::ReadInitStates(TokenReader& rTr, 
    const std::map<Idx,Idx>& rMinIndexMap, 
    const std::map<std::string,Idx>& rTmpStateNameMap) {
  FD_DV("Generator(" << this << ")::ReadInitStates(\"" << rTr.FileName() << "\")");
  rTr.SeekBegin("InitStates");  // better use "ReadBegin" ?
  // HELPERS:
  Token token;
  // ALGORITHM:
  mInitStates.Clear();
  while((rTr >> token) && (token.Type() != Token::End)) {
    if (token.Type() == Token::Index) {
      std::map<Idx,Idx>::const_iterator mit = rMinIndexMap.find(token.IndexValue());
      if (mit != rMinIndexMap.end()) {
        mInitStates.Insert(mit->second);
      }
      else {
        std::stringstream errstr;
        errstr << "Token " << token.IndexValue() << " not in rMinIndexMap"
          << rTr.FileLine();
        throw Exception("Generator::ReadInitStates", errstr.str(), 52);
      }
    }
    else if (token.Type() == Token::Name) {
      std::map<std::string,Idx>::const_iterator smit = 
          rTmpStateNameMap.find(token.StringValue());
      if (smit != rTmpStateNameMap.end()) {
        mInitStates.Insert(smit->second);
      }
      else {
        std::stringstream errstr;
        errstr << "Token " << token.StringValue() << " not known in stateset "
          << rTr.FileLine();
        throw Exception("Generator::ReadInitStates", errstr.str(), 52);
      }
    }
  }
  mInitStates.Name("InitStates");
}

// ReadMarkedStates(pFileName) 
template <class TrnSet>
void TGenerator<TrnSet>::ReadMarkedStates(const char* pFileName,
    const std::map<Idx,Idx>& rMinIndexMap, 
    const std::map<std::string,Idx>& rTmpStateNameMap) {
  TokenReader tr(pFileName);
  ReadMarkedStates(tr, rMinIndexMap, rTmpStateNameMap);
}

// ReadMarkedStates(tr) 
template <class TrnSet>
void TGenerator<TrnSet>::ReadMarkedStates(TokenReader& rTr, 
    const std::map<Idx,Idx>& rMinIndexMap, 
    const std::map<std::string,Idx>& rTmpStateNameMap) {
  FD_DV("Generator(" << this << ")::ReadMarkedStates(\"" 
      << rTr.FileName() << "\")");
  // HELPERS:
  Token token;
  // ALGORITHM:
  mMarkedStates.Clear();
  rTr.SeekBegin("MarkedStates"); // old behaviour
  while ((rTr >> token) && (token.Type() != Token::End)) {
    if (token.Type() == Token::Index) {
      std::map<Idx,Idx>::const_iterator mit = 
          rMinIndexMap.find(token.IndexValue());
      if (mit != rMinIndexMap.end()) {
        mMarkedStates.Insert(mit->second);
      }
      else {
        std::stringstream errstr;
        errstr << "Token " << token.IndexValue() << " not in rMinIndexMap"
          << rTr.FileLine();
        throw Exception("Generator::ReadMarkedStates", errstr.str(), 52);
      }
    }
    else if (token.Type() == Token::Name) {
      std::map<std::string,Idx>::const_iterator smit = 
          rTmpStateNameMap.find(token.StringValue());
      if (smit != rTmpStateNameMap.end()) {
        mMarkedStates.Insert(smit->second);
      }
      else {
        std::stringstream errstr;
        errstr << "Token " << token.StringValue() << " not known in stateset "
          << rTr.FileLine();
        throw Exception("TGenerator<TrnSet>::ReadMarkedStates", errstr.str(), 52);
      }
    }
  }
  mMarkedStates.Name("MarkedStates");
}

// Name(rName)
template <class TrnSet>
void TGenerator<TrnSet>::Name(const std::string& rName) {
  FD_DV("Generator(" << this << ")::Name(\"" << rName << "\")");
  mMyName = rName;
}

// Name()
template <class TrnSet>
std::string TGenerator<TrnSet>::Name(void) const {
  return mMyName;
}

// Id(
template <class TrnSet>
Idx TGenerator<TrnSet>::Id(void) const {
  return mId;
}

// IsUnique(
template <class TrnSet>
bool TGenerator<TrnSet>::IsUnique(void) const {
  return mIsUnique;
}

// virtual
// Valid()
template <class TrnSet>
bool TGenerator<TrnSet>::Valid(void) {
  FD_DV("Generator(" << this << ")::Valid()");
//  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::iterator tit;
  typename TrnSet::iterator tit;
  StateSet::iterator lit;
  for (tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
    if (! ExistsState(tit->X1)) return false;
    if (! ExistsEvent(tit->Ev)) return false;
    if (! ExistsState(tit->X2)) return false;
  }
  for (lit = InitStatesBegin(); lit != InitStatesEnd(); ++lit) {
    if (! ExistsState(static_cast<Idx>(*lit))) return false;
  }
  for (lit = MarkedStatesBegin(); lit != MarkedStatesEnd(); ++lit) {
    if (! ExistsState(static_cast<Idx>(*lit))) return false;
  }
  if (mAlphabet.Name() != "Alphabet") return false;
  if (mStates.Name() != "States") return false;
  if (mInitStates.Name() != "InitStates") return false;
  if (mMarkedStates.Name() != "MarkedStates") return false;
  return true;
}

// AlphabetSize()
template <class TrnSet>
Idx TGenerator<TrnSet>::AlphabetSize(void) const {
  return mAlphabet.Size();
}

// Size()
template <class TrnSet>
Idx TGenerator<TrnSet>::Size(void) const {
  return mStates.Size();
}

// virtual
// Clear()
template <class TrnSet>
void TGenerator<TrnSet>::Clear(void) {
  // UnRegisterGeneratorCopy returns true if there are no copys
  if (UnRegisterGeneratorCopy(this)) {
    StateSet::iterator it;
    for (it = StatesBegin(); it != StatesEnd(); ++it) {
      FD_DC("TGenerator<TrnSet>::Clear(): mpStateSymbolTable->DelState(" 
          << *it << ")");
      mpStateSymbolTable->DelState(static_cast<Idx>(*it));
    }
  }
  mAlphabet.Clear();
  mStates.Clear();
  mTransRel.Clear();
  mInitStates.Clear();
  mMarkedStates.Clear();
}

// ClrStates()
template <class TrnSet>
void TGenerator<TrnSet>::ClrStates(void) {
  // UnRegisterGeneratorCopy returns true if there are no copys
  if (UnRegisterGeneratorCopy(this)) {
    StateSet::iterator it;
    for (it = StatesBegin(); it != StatesEnd(); ++it) {
      FD_DC("TGenerator<TrnSet>::Clear(): mpStateSymbolTable->DelState(" 
          << *it << ")");
      mpStateSymbolTable->DelState(*it);
    }
  }
  mStates.Clear();
  mTransRel.Clear();
  mInitStates.Clear();
  mMarkedStates.Clear();
}

// TransRelSize()
template <class TrnSet>
Idx TGenerator<TrnSet>::TransRelSize(void) const {
  return mTransRel.Size();
}

// InitStatesSize()
template <class TrnSet>
Idx TGenerator<TrnSet>::InitStatesSize(void) const {
  return mInitStates.Size();
}

// MarkedStatesSize()
template <class TrnSet>
Idx TGenerator<TrnSet>::MarkedStatesSize(void) const {
  return mMarkedStates.Size();
}

// AlphabetEmpty()
template <class TrnSet>
bool TGenerator<TrnSet>::AlphabetEmpty(void) const {
  return mAlphabet.Empty();
}

// Empty()
template <class TrnSet>
bool TGenerator<TrnSet>::Empty(void) const {
  return mStates.Empty();
}

// TransRelEmpty()
template <class TrnSet>
bool TGenerator<TrnSet>::TransRelEmpty(void) const {
  return mTransRel.Empty();
}

// InitStatesEmpty()
template <class TrnSet>
bool TGenerator<TrnSet>::InitStatesEmpty(void) const {
  return mInitStates.Empty();
}

// MarkedStatesEmpty()
template <class TrnSet>
bool TGenerator<TrnSet>::MarkedStatesEmpty(void) const {
  return mMarkedStates.Empty();
}

// StateNamesEnabled()
template <class TrnSet>
bool TGenerator<TrnSet>::StateNamesEnabled(void) const {
  return mStateNamesEnabled;
}

// StateNamesEnabled(flag)
template <class TrnSet>
void TGenerator<TrnSet>::StateNamesEnabled(bool flag) {
  mStateNamesEnabled = flag;
}

// iterator AlphabetBegin()
template <class TrnSet>
EventSet::iterator TGenerator<TrnSet>::AlphabetBegin(void) {
  return mAlphabet.Begin();
}

// iterator AlphabetBegin() const
template <class TrnSet>
EventSet::const_iterator TGenerator<TrnSet>::AlphabetBegin(void) const {
  return mAlphabet.Begin();
}

// iterator AlphabetEnd()
template <class TrnSet>
EventSet::iterator TGenerator<TrnSet>::AlphabetEnd(void) {
  return mAlphabet.End();
}

// iterator AlphabetEnd() const
template <class TrnSet>
EventSet::const_iterator TGenerator<TrnSet>::AlphabetEnd(void) const {
  return mAlphabet.End();
}

// iterator StatesBegin()
template <class TrnSet>
StateSet::iterator TGenerator<TrnSet>::StatesBegin(void) {
  return mStates.Begin();
}

// iterator StatesBegin() const
template <class TrnSet>
StateSet::const_iterator TGenerator<TrnSet>::StatesBegin(void) const {
  return mStates.Begin();
}

// iterator StatesEnd()
template <class TrnSet>
StateSet::iterator TGenerator<TrnSet>::StatesEnd(void) {
  return mStates.End();
}

// iterator StatesEnd() const
template <class TrnSet>
StateSet::const_iterator TGenerator<TrnSet>::StatesEnd(void) const {
  return mStates.End();
}

// iterator TransRelBegin()
template <class TrnSet>
typename TrnSet::iterator TGenerator<TrnSet>::TransRelBegin(void) {
  return mTransRel.Begin();
}

// iterator TransRelBegin() const
template <class TrnSet>
typename TrnSet::const_iterator TGenerator<TrnSet>::TransRelBegin(void) const {
  return mTransRel.Begin();
}

// iterator TransRelEnd()
template <class TrnSet>
typename TrnSet::iterator TGenerator<TrnSet>::TransRelEnd(void) {
  return mTransRel.End();
}

// iterator TransRelEnd() const
template <class TrnSet>
typename TrnSet::const_iterator TGenerator<TrnSet>::TransRelEnd(void) const {
  return mTransRel.End();
}

// iterator TransRelBegin(x1)
template <class TrnSet>
typename TrnSet::iterator TGenerator<TrnSet>::TransRelBegin(Idx x1) {
  return mTransRel.Begin(x1);
}

// const_iterator TransRelBegin(x1) const
template <class TrnSet>
typename TrnSet::const_iterator TGenerator<TrnSet>::TransRelBegin(Idx x1) const {
  return mTransRel.Begin(x1);
}

// iterator TransRelEnd(x1)
template <class TrnSet>
typename TrnSet::iterator TGenerator<TrnSet>::TransRelEnd(Idx x1) {
  return mTransRel.End(x1);
}

// const_iterator TransRelEnd(x1) const
template <class TrnSet>
typename TrnSet::const_iterator TGenerator<TrnSet>::TransRelEnd(Idx x1) const {
  return mTransRel.End(x1);
}

// iterator TransRelBegin(x1, ev)
template <class TrnSet>
typename TrnSet::iterator TGenerator<TrnSet>::TransRelBegin(Idx x1, Idx ev) {
  return mTransRel.Begin(x1, ev);
}

// const_iterator TransRelBegin(x1, ev) const
template <class TrnSet>
typename TrnSet::const_iterator TGenerator<TrnSet>::TransRelBegin(Idx x1, Idx ev) const {
  return mTransRel.Begin(x1, ev);
}

// iterator TransRelEnd(x1, ev)
template <class TrnSet>
typename TrnSet::iterator TGenerator<TrnSet>::TransRelEnd(Idx x1, Idx ev) {
  return mTransRel.End(x1, ev);
}

// const_iterator TransRelEnd(x1, ev) const
template <class TrnSet>
typename TrnSet::const_iterator TGenerator<TrnSet>::TransRelEnd(Idx x1, Idx ev) const {
  return mTransRel.End(x1, ev);
}

// iterator InitStatesBegin()
template <class TrnSet>
StateSet::iterator TGenerator<TrnSet>::InitStatesBegin(void) {
  return mInitStates.Begin();
}

// const_iterator InitStatesBegin() const
template <class TrnSet>
StateSet::const_iterator TGenerator<TrnSet>::InitStatesBegin(void) const {
  return mInitStates.Begin();
}

// iterator InitStatesEnd()
template <class TrnSet>
StateSet::iterator TGenerator<TrnSet>::InitStatesEnd(void) {
  return mInitStates.End();
}

// const_iterator InitStatesEnd() const
template <class TrnSet>
StateSet::const_iterator TGenerator<TrnSet>::InitStatesEnd(void) const {
  return mInitStates.End();
}

// iterator MarkedStatesBegin()
template <class TrnSet>
StateSet::iterator TGenerator<TrnSet>::MarkedStatesBegin(void) {
  return mMarkedStates.Begin();
}

// const_iterator MarkedStatesBegin()
template <class TrnSet>
StateSet::const_iterator TGenerator<TrnSet>::MarkedStatesBegin(void) const {
  return mMarkedStates.Begin();
}

// iterator MarkedStatesEnd()
template <class TrnSet>
StateSet::iterator TGenerator<TrnSet>::MarkedStatesEnd(void) {
  return mMarkedStates.End(); 
}

// const_iterator MarkedStatesEnd() const
template <class TrnSet>
StateSet::const_iterator TGenerator<TrnSet>::MarkedStatesEnd(void) const {
  return mMarkedStates.End(); 
}

// InjectAlphabet( newalphabet)
template <class TrnSet>
void TGenerator<TrnSet>::InjectAlphabet(const EventSet& newalphabet) {
  mAlphabet = newalphabet;
  mAlphabet.Name("Alphabet");
}

// InsEvent(index)
template <class TrnSet>
void TGenerator<TrnSet>::InsEvent(Idx index) {
  FD_DG("Generator(" << this << ")::InsEvent(" << index << ")");
  mAlphabet.InsEvent(index);
}

// InsEvent(rName)
template <class TrnSet>
Idx TGenerator<TrnSet>::InsEvent(const std::string& rName) {
  FD_DG("Generator(" << this << ")::InsEvent(\"" << rName << "\")");
  return mAlphabet.InsEvent(rName);
}

// InsEvents(events)
template <class TrnSet>
void TGenerator<TrnSet>::InsEvents(const EventSet& events) {
  EventSet::const_iterator eit;
  for (eit = events.Begin(); eit != events.End(); ++eit) {
    InsEvent(*eit);
  }
}

// DelEvent(index)
template <class TrnSet>
void TGenerator<TrnSet>::DelEvent(Idx index) {
  FD_DG("Generator(" << this << ")::DelEvent(" << index << ")");
  mTransRel.EraseByEv(index);
  mAlphabet.DelEvent(index);
}

// DelEvent(rName)
template <class TrnSet>
void TGenerator<TrnSet>::DelEvent(const std::string& rName) {
  FD_DG("Generator(" << this << ")::DelEvent(\"" << rName << "\")");
  Idx index;
  try {
    index = mAlphabet.EventIndex(rName);
    mTransRel.EraseByEv(index);
    mAlphabet.DelEvent(index);
  }
  catch (Exception) {
    std::stringstream errstr;
    errstr << "event \"" << rName << "\" not found in generator \""
        << Name() << "\"";
    throw Exception("TGenerator<TrnSet>::DelEvent", errstr.str(), 202);
  }
}

// DelEvents(events)
template <class TrnSet>
void TGenerator<TrnSet>::DelEvents(const EventSet& rEvents) {
  FD_DG("Generator(" << this << ")::DelEvents(\"" 
      << rEvents.ToString() << "\")");
  EventSet::const_iterator it;
  for (it = rEvents.Begin(); it != rEvents.End(); ++it) {
    DelEvent(*it);
  }
}

// DelEventFromAlphabet(index)
template <class TrnSet>
void TGenerator<TrnSet>::DelEventFromAlphabet(Idx index) {
  FD_DG("Generator(" << this << ")::DelEventFromAlphabet(" 
      << index << ")");
  mAlphabet.DelEvent(index);
}

// InsState()
template <class TrnSet>
Idx TGenerator<TrnSet>::InsState(void) {
  FD_DG("Generator(" << this << ")::InsState()");
  return mStates.InsState();
}

// InsState(rName)
template <class TrnSet>
Idx TGenerator<TrnSet>::InsState(const std::string& rName) {
  FD_DG("Generator(" << this << ")::InsState(\"" << rName << "\")");
  return mStates.InsState(rName);
}

// InjectState(index)
template <class TrnSet>
void TGenerator<TrnSet>::InjectState(Idx index) {
  FD_DG("Generator(" << this << ")::InsState_direct(\"" 
      << SStr(index) << "\")");
  mStates.Insert(index);
}

// InjectStates(rNewStates)
template <class TrnSet>
void TGenerator<TrnSet>::InjectStates(const StateSet& rNewStates) {
  FD_DG("Generator(" << this << ")::InjectStates(" 
      << rNewStates.ToString() << ")");
  mStates = rNewStates;
  mStates.Name("States");
}

// InsInitState()
template <class TrnSet>
Idx TGenerator<TrnSet>::InsInitState(void) {
  FD_DG("Generator(" << this << ")::InsInitState()");
  Idx index;
  index = InsState();
  mInitStates.Insert(index);
  return index;
}

// InsInitState()
template <class TrnSet>
Idx TGenerator<TrnSet>::InsInitState(const std::string& rName) {
  FD_DG("Generator(" << this << ")::InsInitState(\"" << rName << "\")");
  Idx index;
  index = InsState(rName);
  mInitStates.Insert(index);
  return index;
}

// InsMarkedState()
template <class TrnSet>
Idx TGenerator<TrnSet>::InsMarkedState(void) {
  FD_DG("Generator(" << this << ")::InsMarkedState()");
  Idx index;
  index = InsState();
  mMarkedStates.Insert(index);
  return index;
}

// InsMarkedState()
template <class TrnSet>
Idx TGenerator<TrnSet>::InsMarkedState(const std::string& rName) {
  FD_DG("Generator(" << this << ")::InsMarkedState(\"" << rName << "\")");
  Idx index;
  index = InsState(rName);
  mMarkedStates.Insert(index);
  return index;
}

// DelState(index)
template <class TrnSet>
void TGenerator<TrnSet>::DelState(Idx index) {
  FD_DG("Generator(" << this << ")::DelState(" << index << ")");
  // mInitStates
  mInitStates.Erase(index);
  // mstates
  mMarkedStates.Erase(index);
  // transrel 
  mTransRel.EraseByX1OrX2(index);
  // mStates + global
  if (mIsUnique) {
    mStates.DelState(index);
  }
  else {
    mStates.Erase(index);
  }
}

// DelState(rName)
template <class TrnSet>
void TGenerator<TrnSet>::DelState(const std::string& rName) {
  FD_DG("Generator(" << this << ")::DelState(\"" << rName << "\")");
  Idx index;
  index = mStates.StateIndex(rName);
  if (index == 0) {
    std::stringstream errstr;
    errstr << "state name \"" << rName << "\" not found in generator \""
        << Name() << "\"";
    throw Exception("TGenerator<TrnSet>::DelState(name)", errstr.str(), 203);
  }
  else {
    return DelState(index);
  }
}

// DelStates(rDelStates)
template <class TrnSet>
void TGenerator<TrnSet>::DelStates(const StateSet& rDelStates) {
  TTransSet<TSortX2EvX1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> rtransrel = TransRelX2EvX1();
  // call method with two arguments
  DelStates(rDelStates, rtransrel);
}

// DelStates(rDelStates, rTransRelX2EvX1)
template <class TrnSet>
void TGenerator<TrnSet>::DelStates(const StateSet& rDelStates,  
    TTransSet<TSortX2EvX1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>& rTransRelX2EvX1) {
  FD_DG("Generator(" << this << ")::DelStates(" 
      << rDelStates.ToString() << ")");
  StateSet::const_iterator cit;
  StateSet::const_iterator cit_end;

  // symbolic state names
  if (IsUnique() && StateNamesEnabled()) {
    for (cit = rDelStates.Begin(); cit != rDelStates.End(); ++cit) {
      mpStateSymbolTable->DelState(*cit);
    }
  }

  // mStates:
  mStates = mStates - rDelStates;
  mStates.Name("States");

  // mInitStates:
  // a generator usually only has one initial state compared to the
  // overall number of states!
  StateSet::iterator it, it_end;
  it = InitStatesBegin();
  it_end = InitStatesEnd();
  while (it != it_end) {
    if (rDelStates.ExistsState(*it)) {
      StateSet::iterator tmpit = it;
      ++it;
      mInitStates.Erase(tmpit);
    }
    else {
      ++it;
    }
  }
  
  // mMarkedStates:
  // a generator usually only has few marked states compared to the
  // overall number of states!
  it = MarkedStatesBegin();
  it_end = MarkedStatesEnd();
  while (it != it_end) {
    if (rDelStates.ExistsState(*it)) {
      StateSet::iterator tmpit = it;
      ++it;
      mMarkedStates.Erase(tmpit);
    }
    else {
      ++it;
    }
  }

  // mTransRel:
  mTransRel.EraseByX1OrX2(rDelStates, mTransRel, rTransRelX2EvX1);
}

// DelStateFromStates(index)
template <class TrnSet>
void TGenerator<TrnSet>::DelStateFromStates(Idx index) {
  FD_DG("Generator(" << this << ")::DelStateFromStates(" << index << ")");
  // mStates + global
  if (mIsUnique) {
    mStates.DelState(index);
  }
  else {
    mStates.Erase(index);
  }
}

// DelStateFromStates(pos)
template <class TrnSet>
void TGenerator<TrnSet>::DelStateFromStates(StateSet::iterator pos) {
  FD_DG("Generator(" << this << ")::DelState(" << *pos << ")");
  if (mIsUnique) {
    mStates.DelState(pos);
  }
  else {
    mStates.Erase(pos);
  }
}

// SetInitState(index)
template <class TrnSet>
void TGenerator<TrnSet>::SetInitState(Idx index) {
  FD_DG("Generator(" << this << ")::SetInitState(" << index << ")");
#ifdef FAUDES_CHECKED
  if (! mStates.ExistsState(index)) {
    FD_ERR("TGenerator<TrnSet>::SetInitState: index " << index << " not in stateset");
    abort();
  }
#endif
  mInitStates.Insert(index);
}

// SetInitState(rName)
template <class TrnSet>
void TGenerator<TrnSet>::SetInitState(const std::string& rName) {
  FD_DG("Generator(" << this << ")::SetInitState(\"" << rName << "\")");
  Idx index = StateIndex(rName);
  if (index == 0) {
    std::stringstream errstr;
    errstr << "State name \"" << rName << "\" not known in Generator";
    throw Exception("TGenerator<TrnSet>::SetInitState(..)", errstr.str(), 203);
  }
  SetInitState(index);
}

// InjectInitStates(rNewInitStates)
template <class TrnSet>
void TGenerator<TrnSet>::InjectInitStates(const StateSet& rNewInitStates) {
  FD_DG("Generator(" << this << ")::InjectInitStates(" 
      << rNewInitStates.ToString() << ")");
  mInitStates = rNewInitStates;
  mInitStates.Name("InitStates");
}

// ClrInitState(index)
template <class TrnSet>
void TGenerator<TrnSet>::ClrInitState(Idx index) {
  FD_DG("Generator(" << this << ")::ClrInitState(" << index << ")");
#ifdef FAUDES_CHECKED
  if (! mStates.ExistsState(index)) {
    FD_ERR("TGenerator<TrnSet>::ClrInitState: index " << index 
        << " not in stateset");
    abort();
  }
#endif
  mInitStates.Erase(index);
}

// ClrInitState(rName)
template <class TrnSet>
void TGenerator<TrnSet>::ClrInitState(const std::string& rName) {
  FD_DG("Generator(" << this << ")::ClrInitState(\"" << rName << "\")");
  Idx index = StateIndex(rName);
  if (index == 0) {
    std::stringstream errstr;
    errstr << "State name \"" << rName << "\" not known in Generator";
    throw Exception("TGenerator<TrnSet>::ClrInitState(..)", errstr.str(), 203);
  }
  ClrInitState(index);
}

// ClrInitState(pos)
template <class TrnSet>
void TGenerator<TrnSet>::ClrInitState(StateSet::iterator pos) {
  FD_DG("Generator(" << this << ")::ClrInitState(" << *pos << ")");
  mInitStates.Erase(pos);
}

// ClrInitStates()
template <class TrnSet>
void TGenerator<TrnSet>::ClrInitStates(void) {
  mInitStates.Clear();
}

// SetMarkedState(index)
template <class TrnSet>
void TGenerator<TrnSet>::SetMarkedState(Idx index) {
  FD_DG("Generator(" << this << ")::SetMarkedState(" << index << ")");
#ifdef FAUDES_CHECKED
  if (! mStates.ExistsState(index)) {
    FD_ERR("TGenerator<TrnSet>::SetMarkedState: index " << index 
        << " not in stateset");
    abort();
  }
#endif
  mMarkedStates.Insert(index);
}

// SetMarkedState(rName)
template <class TrnSet>
void TGenerator<TrnSet>::SetMarkedState(const std::string& rName) {
  FD_DG("Generator(" << this << ")::SetMarkedState(\"" << rName << "\")");
  Idx index = StateIndex(rName);
  if (index == 0) {
    std::stringstream errstr;
    errstr << "State name \"" << rName << "\" not known in Generator";
    throw Exception("TGenerator<TrnSet>::SetMarkedState(..)", errstr.str(), 203);
  }
  SetMarkedState(index);
}

// InjectMarkedStates(rNewMarkedStates)
template <class TrnSet>
void TGenerator<TrnSet>::InjectMarkedStates(const StateSet& rNewMarkedStates) {
  FD_DG("Generator(" << this << ")::InjectMarkedStates(" 
      << rNewMarkedStates.ToString() << ")");
  mMarkedStates = rNewMarkedStates;
  mMarkedStates.Name("MarkedStates");
}

// ClrMarkedState(index)
template <class TrnSet>
void TGenerator<TrnSet>::ClrMarkedState(Idx index) {
  FD_DG("Generator(" << this << ")::ClrMarkedState(" << index << ")");
#ifdef FAUDES_CHECKED
  if (! mStates.ExistsState(index)) {
    FD_ERR("TGenerator<TrnSet>::ClrMarkedState: index " << index 
        << " not in stateset");
    abort();
  }
#endif
  mMarkedStates.Erase(index);
}

// ClrMarkedState(rName)
template <class TrnSet>
void TGenerator<TrnSet>::ClrMarkedState(const std::string& rName) {
  FD_DG("Generator(" << this << ")::ClrMarkedState(\"" << rName << "\")");
  Idx index = StateIndex(rName);
  if (index == 0) {
    std::stringstream errstr;
    errstr << "State name \"" << rName << "\" not known in Generator";
    throw Exception("TGenerator<TrnSet>::ClrMarkedState(..)", errstr.str(), 203);
  }
  ClrMarkedState(index);
}

// ClrMarkedState(pos)
template <class TrnSet>
void TGenerator<TrnSet>::ClrMarkedState(StateSet::iterator pos) {
  FD_DG("Generator(" << this << ")::ClrMarkedState(" << *pos << ")");
  mMarkedStates.Erase(pos);
}

// ClrMarkedStates()
template <class TrnSet>
void TGenerator<TrnSet>::ClrMarkedStates(void) {
  mMarkedStates.Clear();
}

// SetTransition(x1, ev, x2)
template <class TrnSet>
void TGenerator<TrnSet>::SetTransition(Idx x1, Idx ev, Idx x2) {
  FD_DG("Generator(" << this << ")::SetTransition(" 
      << x1 << "," << ev << "," << x2 << ")");
#ifdef FAUDES_CHECKED
  if (! mStates.ExistsState(x1)) {
    FD_ERR("TGenerator<TrnSet>::SetTransition: state " << x1 << " not in stateset");
    abort();
  }
  if (! mAlphabet.ExistsEvent(ev)) {
    FD_ERR("TGenerator<TrnSet>::SetTransition: event " << ev << " not in alphabet");
    abort();
  }
  if (! mStates.ExistsState(x2)) {
    FD_ERR("TGenerator<TrnSet>::SetTransition: state " << x2 << " not in stateset");
    abort();
  }
#endif
  mTransRel.Insert(x1, ev, x2);
}


// SetTransition(rX1, rEv, rX2)
template <class TrnSet>
void TGenerator<TrnSet>::SetTransition(const std::string& rX1, const std::string& rEv,
    const std::string& rX2) {
  FD_DG("Generator(" << this << ")::SetTransition(\""
      << rX1 << "\", \"" << rEv << "\", \"" << rX2 << "\")");
  Idx x1 = StateIndex(rX1);
  Idx x2 = StateIndex(rX2);
  if (x1 == 0) {
    FD_ERR("TGenerator<TrnSet>::SetTransition: state " << rX1 
        << " not in stateset");
    std::stringstream errstr;
    errstr << "State name " << rX1 << " not found in Generator";
    throw Exception("TGenerator<TrnSet>::SetTransition(..)", errstr.str(), 203);
  }
  if (! mAlphabet.ExistsEvent(rEv)) {
    FD_ERR("TGenerator<TrnSet>::SetTransition: event " << rEv << " not in alphabet");
    std::stringstream errstr;
    errstr << "Event name " << rEv << " not found in event domain of Generator";
    throw Exception("TGenerator<TrnSet>::SetTransition(..)", errstr.str(), 202);
  }
  if (x2 == 0) {
    FD_ERR("TGenerator<TrnSet>::SetTransition: state " << rX2 << " not in stateset");
    std::stringstream errstr;
    errstr << "State name " << rX2 << " not found in Generator";
    throw Exception("TGenerator<TrnSet>::SetTransition(..)", errstr.str(), 203);
  }
  mTransRel.Insert(StateIndex(rX1), EventIndex(rEv), StateIndex(rX2));
}


// SetTransition(rTransition)
template <class TrnSet>
void TGenerator<TrnSet>::SetTransition(const typename TrnSet::TransitionType& rTransition) {
  FD_DG("Generator(" << this << ")::SetTransition(" << rTransition.X1 << "," 
      << rTransition.Ev << "," << rTransition.X2 << ")");
#ifdef FAUDES_CHECKED
  if (! mStates.ExistsState(rTransition.X1)) {
    FD_ERR("TGenerator<TrnSet>::SetTransition: state " << rTransition.X1 
        << " not in stateset");
    abort();
  }
  if (! mAlphabet.ExistsEvent(rTransition.Ev)) {
    FD_ERR("TGenerator<TrnSet>::SetTransition: event " << rTransition.Ev 
        << " not in alphabet");
    abort();
  }
  if (! mStates.ExistsState(rTransition.X2)) {
    FD_ERR("TGenerator<TrnSet>::SetTransition: state " << rTransition.X2 
        << " not in stateset");
    abort();
  }
#endif
  mTransRel.Insert(rTransition);
}

// ClrTransition.X1, ev, x2)
template <class TrnSet>
void TGenerator<TrnSet>::ClrTransition(Idx x1, Idx ev, Idx x2) {
  FD_DG("Generator(" << this << ")::ClrTransition(" 
      << x1 << "," << ev << "," << x2 << ")");
  mTransRel.Erase(x1, ev, x2);
}

// ClrTransition(rTransition)
template <class TrnSet>
void TGenerator<TrnSet>::ClrTransition(const typename TrnSet::TransitionType& rTransition) {
  FD_DG("Generator(" << this << ")::ClrTransition(" << rTransition.X1 
      << "," << rTransition.Ev << "," << rTransition.X2 << ")");
  mTransRel.Erase(rTransition);
}

// ClrTransition(it)
template <class TrnSet>
void TGenerator<TrnSet>::ClrTransition(typename TrnSet::iterator it) {
  FD_DG("Generator(" << this << ")::ClrTransition(" << it->X1
      << "-" << it->Ev << "-" << it->X2 << ")");
  mTransRel.Erase(it);
}

// ClrTransRel()
template <class TrnSet>
void TGenerator<TrnSet>::ClrTransRel(void) {
  mTransRel.Clear();
}

// EventName(index, name)
template <class TrnSet>
void TGenerator<TrnSet>::EventName(Idx index, const std::string& rName) {
  FD_DG("Generator(" << this << ")::EventName(" 
      << index << ",\"" << rName << "\")");
  mAlphabet.EventName(index, rName);
}

// StateName(index, name)
template <class TrnSet>
void TGenerator<TrnSet>::StateName(Idx index, const std::string& rName) {
  FD_DG("Generator(" << this << ")::StateName(" 
      << index << ",\"" << rName << "\")");
  mStates.StateName(index, rName);
}

// ClrStateName(index)
template <class TrnSet>
void TGenerator<TrnSet>::ClrStateName(Idx index) {
  FD_DG("Generator(" << this << ")::ClrStateName(\"" << index << "\")");
  mStates.ClrStateName(index);
}

// ClrStateName(rName)
template <class TrnSet>
void TGenerator<TrnSet>::ClrStateName(const std::string& rName) {
  FD_DG("Generator(" << this << ")::ClrStateName(\"" << rName << "\")");
  mStates.ClrStateName(rName);
}

// ExistsEvent(index)
template <class TrnSet>
bool TGenerator<TrnSet>::ExistsEvent(Idx index) const {
  return mAlphabet.ExistsEvent(index);
}

// ExistsEvent(rName)
template <class TrnSet>
bool TGenerator<TrnSet>::ExistsEvent(const std::string& rName) const {
  return mAlphabet.ExistsEvent(rName);
}

// FindEvent(index)
template <class TrnSet>
EventSet::iterator TGenerator<TrnSet>::FindEvent(Idx index) {
  return mAlphabet.Find(index);
}

// FindEvent(index) const
template <class TrnSet>
EventSet::const_iterator TGenerator<TrnSet>::FindEvent(Idx index) const {
  return mAlphabet.Find(index);
}

// FindEvent(rName)
template <class TrnSet>
EventSet::iterator TGenerator<TrnSet>::FindEvent(const std::string& rName) {
  return mAlphabet.Find(rName);
}

// FindEvent(rName)
template <class TrnSet>
EventSet::const_iterator TGenerator<TrnSet>::FindEvent(const std::string& rName) const {
  return mAlphabet.Find(rName);
}

// ExistsState(index)
template <class TrnSet>
bool TGenerator<TrnSet>::ExistsState(Idx index) const {
  return mStates.ExistsState(index);
}

// FindState(index)
template <class TrnSet>
StateSet::iterator TGenerator<TrnSet>::FindState(Idx index) {
  return mStates.Find(index);
}

// FindState(index) const
template <class TrnSet>
StateSet::const_iterator TGenerator<TrnSet>::FindState(Idx index) const {
  return mStates.Find(index);
}

// ExistsInitState(index)
template <class TrnSet>
bool TGenerator<TrnSet>::ExistsInitState(Idx index) const {
  return mInitStates.ExistsState(index);
}

// FindInitState(index)
template <class TrnSet>
StateSet::iterator TGenerator<TrnSet>::FindInitState(Idx index) {
  return mInitStates.Find(index);
}

// FindInitState(index)
template <class TrnSet>
StateSet::const_iterator TGenerator<TrnSet>::FindInitState(Idx index) const {
  return mInitStates.Find(index);
}

// ExistsMarkedState(index)
template <class TrnSet>
bool TGenerator<TrnSet>::ExistsMarkedState(Idx index) const {
  return mMarkedStates.ExistsState(index);
}

// FindMarkedState(index)
template <class TrnSet>
StateSet::iterator TGenerator<TrnSet>::FindMarkedState(Idx index) {
  return mMarkedStates.Find(index);
}

// FindMarkedState(index)
template <class TrnSet>
StateSet::const_iterator TGenerator<TrnSet>::FindMarkedState(Idx index) const {
  return mMarkedStates.Find(index);
}

// EventName(index)
template <class TrnSet>
std::string TGenerator<TrnSet>::EventName(Idx index) const {
  return mAlphabet.EventName(index);
}

// EventIndex(rName)
template <class TrnSet>
Idx TGenerator<TrnSet>::EventIndex(const std::string& rName) const {
  return mAlphabet.EventIndex(rName);
}

// StateName(index)
template <class TrnSet>
std::string TGenerator<TrnSet>::StateName(Idx index) const {
  return mStates.StateName(index);
}

// StateIndex(rName)
template <class TrnSet>
Idx TGenerator<TrnSet>::StateIndex(const std::string& rName) const {
  return mStates.StateIndex(rName);
}

// Alphabet()
template <class TrnSet>
EventSet TGenerator<TrnSet>::Alphabet(void) const {
  return mAlphabet;
}

// Alphabetr()
template <class TrnSet>
EventSet& TGenerator<TrnSet>::Alphabetr(void) {
  return mAlphabet;
}

// States()
template <class TrnSet>
StateSet TGenerator<TrnSet>::States(void) const {
  return mStates;
}

// Statesr()
template <class TrnSet>
StateSet& TGenerator<TrnSet>::Statesr(void) {
  return mStates;
}
//FIXME: Retrurn TansRel
// TransRel()
template <class TrnSet>
TrnSet TGenerator<TrnSet>::TransRel(void) const {
  return mTransRel;
}

// TransRelr()
template <class TrnSet>
TrnSet& TGenerator<TrnSet>::TransRelr(void) {
  return mTransRel;
}

// TransRelEvX1X2()
template <class TrnSet>
TTransSet<TSortEvX1X2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> TGenerator<TrnSet>::TransRelEvX1X2(void) const {
  return mTransRel.TransSetEvX1X2();
}

// TransRelEvX2X1()
template <class TrnSet>
TTransSet<TSortEvX2X1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> TGenerator<TrnSet>::TransRelEvX2X1(void) const {
  return mTransRel.TransSetEvX2X1();
}

// TransRelX2EvX1()
template <class TrnSet>
TTransSet<TSortX2EvX1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> TGenerator<TrnSet>::TransRelX2EvX1(void) const {
  return mTransRel.TransSetX2EvX1();
}

// TransRelX2X1Ev()
template <class TrnSet>
TTransSet<TSortX2X1Ev<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> TGenerator<TrnSet>::TransRelX2X1Ev(void) const {
  return mTransRel.TransSetX2X1Ev();
}

// TransRelX1X2Ev()
template <class TrnSet>
TTransSet<TSortX1X2Ev<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> TGenerator<TrnSet>::TransRelX1X2Ev(void) const {
  return mTransRel.TransSetX1X2Ev();
}

// InitStates()
template <class TrnSet>
StateSet TGenerator<TrnSet>::InitStates(void) const {
  return mInitStates;
}

// InitStatesr()
template <class TrnSet>
StateSet& TGenerator<TrnSet>::InitStatesr(void) {
  return mInitStates;
}

// MarkedStates()
template <class TrnSet>
StateSet TGenerator<TrnSet>::MarkedStates(void) const {
  return mMarkedStates;
}

// MarkedStatesr()
template <class TrnSet>
StateSet& TGenerator<TrnSet>::MarkedStatesr(void) {
  return mMarkedStates;
}

// MinimizeAlphabet()
template <class TrnSet>
void TGenerator<TrnSet>::MinimizeAlphabet(void) {
  mAlphabet = UsedEvents();
}

// UsedEvents()
template <class TrnSet>
EventSet TGenerator<TrnSet>::UsedEvents(void) const {
  EventSet resultset = EventSet();
//  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::const_iterator it;
  typename TrnSet::const_iterator it;
  for (it = mTransRel.Begin(); it != mTransRel.End(); ++it) {
    resultset.InsEvent(it->Ev);
  }
  return resultset;
}

// UnusedEvents()
template <class TrnSet>
EventSet TGenerator<TrnSet>::UnusedEvents(void) const {
  return mAlphabet - UsedEvents();
}

// ActiveEventSet(x1)
template <class TrnSet>
EventSet TGenerator<TrnSet>::ActiveEventSet(Idx x1) const {
  EventSet result = NewEventSet();
//  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::const_iterator it;
  typename TrnSet::const_iterator it;
  for (it = TransRelBegin(x1); it != TransRelEnd(x1); ++it) {
    result.InsEvent(it->Ev);
  }
  return result;
}

// TransRelStateSpace()
template <class TrnSet>
StateSet TGenerator<TrnSet>::TransRelStateSpace(void) const {
  StateSet states = StateSet();
  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::const_iterator it;
  for (it=mTransRel.Begin(); it != mTransRel.End(); ++it) {
    states.Insert(it->X1);
    states.Insert(it->X2);
  }
  return states;
}

// TransRelStateSpace(x1)
template <class TrnSet>
StateSet TGenerator<TrnSet>::TransRelStateSpace(Idx x1) const {
  StateSet states = StateSet();
  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::const_iterator it = mTransRel.Begin(x1);
  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::const_iterator it_end = mTransRel.End(x1);
  for (; it != it_end; ++it) {
    states.Insert(it->X2);
  }
  return states;
}

// SetDefaultStateNames()
template <class TrnSet>
void TGenerator<TrnSet>::SetDefaultStateNames(void) {
  FD_DG("Generator(" << this << ")::SetDefaultStateNames()");
  StateSet::iterator it;
  for (it = mStates.Begin(); it != mStates.End(); ++it) {
    mpStateSymbolTable->SetDefaultStateName(*it);
  }
}

// ClrStateNames()
template <class TrnSet>
void TGenerator<TrnSet>::ClrStateNames(void) {
  FD_DG("Generator(" << this << ")::ClrStateNames()");
  StateSet::iterator it;
  for (it = mStates.Begin(); it != mStates.End(); ++it) {
    mpStateSymbolTable->ClrStateName(*it);
  }
  mStateNamesEnabled = false;
}

// Version()
template <class TrnSet>
void TGenerator<TrnSet>::Version(Idx version, TGenerator<TrnSet>& rResGen) {
  FD_DG("Generator(" << this << ")::Version(" << version << ")");
  // prepare Empty generator
  rResGen.Clear();
  EventSet::iterator eit;
  StateSet::iterator lit;
  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::iterator tit;
  std::map<Idx,Idx> eventoldnewmap;
  std::map<Idx,Idx> stateoldnewmap;
  std::ostringstream o;
  o << version;
  rResGen.Name(Name()+"_"+o.str());
  // create versioned mAlphabet
  for (eit = AlphabetBegin(); eit != AlphabetEnd(); ++eit) {
    eventoldnewmap[*eit] = rResGen.InsEvent(EventName(*eit)+"_"+o.str());
  }
  // create new stateset
  for (lit = StatesBegin(); lit != StatesEnd(); ++lit) {
    if (StateName(*lit) != "") {
      stateoldnewmap[*lit] = rResGen.InsState(StateName(*lit));
    }
    else {
      stateoldnewmap[*lit] = rResGen.InsState();
    }
  }
  // created versioned transrel
  for (tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
    rResGen.SetTransition(stateoldnewmap[tit->X1], eventoldnewmap[tit->Ev], 
        stateoldnewmap[tit->X2]);
  }
  // set istates
  for (lit = InitStatesBegin(); lit != InitStatesEnd(); ++lit) {
    rResGen.SetInitState(stateoldnewmap[*lit]);
  }
  // set mstates
  for (lit = MarkedStatesBegin(); lit != MarkedStatesEnd(); ++lit) {
    rResGen.SetMarkedState(stateoldnewmap[*lit]);
  }
  rResGen.mStateNamesEnabled = mStateNamesEnabled;
}

// Version()
template <class TrnSet>
void TGenerator<TrnSet>::Version(const std::string& rVersion, TGenerator<TrnSet>& rResGen) {
  FD_DG("Generator(" << this << ")::Version(" << rVersion << ")");
  // prepare Empty generator
  rResGen.Clear();
  EventSet::iterator eit;
  StateSet::iterator lit;
  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::iterator tit;
  std::map<Idx,Idx> eventoldnewmap;
  std::map<Idx,Idx> stateoldnewmap;
  rResGen.Name(Name()+"_"+rVersion);
  // create versioned mAlphabet
  for (eit = AlphabetBegin(); eit != AlphabetEnd(); ++eit) {
    eventoldnewmap[*eit] = rResGen.InsEvent(EventName(*eit)+"_"+rVersion);
  }
  // create new stateset
  for (lit = StatesBegin(); lit != StatesEnd(); ++lit) {
    if (StateName(*lit) != "") {
      stateoldnewmap[*lit] = rResGen.InsState(StateName(*lit));
    }
    else {
      stateoldnewmap[*lit] = rResGen.InsState();
    }
  }
  // created versioned transrel
  for (tit = TransRelBegin(); tit != TransRelEnd(); ++tit) {
    rResGen.SetTransition(stateoldnewmap[tit->X1], eventoldnewmap[tit->Ev], 
        stateoldnewmap[tit->X2]);
  }
  // set istates
  for (lit = InitStatesBegin(); lit != InitStatesEnd(); ++lit) {
    rResGen.SetInitState(stateoldnewmap[*lit]);
  }
  // set mstates
  for (lit = MarkedStatesBegin(); lit != MarkedStatesEnd(); ++lit) {
    rResGen.SetMarkedState(stateoldnewmap[*lit]);
  }
  rResGen.mStateNamesEnabled = mStateNamesEnabled;
}

// AccessibleSet()
template <class TrnSet>
StateSet TGenerator<TrnSet>::AccessibleSet(void) const {
  StateSet accessibleset = StateSet();
  StateSet::const_iterator it;
  for (it = InitStatesBegin(); it != InitStatesEnd(); ++it) {
    CheckAccessible(accessibleset, *it);
  }
  accessibleset.Name("accessibleset");
  return accessibleset;
}

// Accessible()
template <class TrnSet>
bool TGenerator<TrnSet>::Accessible(void) {
  StateSet accessibleset = AccessibleSet();
  StateSet not_accessible = mStates - accessibleset;
  StateSet::const_iterator it;
  // mTransRel
  //mTransRel.EraseByX1OrX2(not_accessible);
  for(it = not_accessible.Begin(); it != not_accessible.End(); ++it){
      mTransRel.EraseByX1OrX2(*it);
  }
  // mStates
  mStates = accessibleset;
  mStates.Name("States");
  // mInitStates
  mInitStates = mInitStates * accessibleset;
  mInitStates.Name("InitStates");
  // mMarkedStates
  mMarkedStates = mMarkedStates * accessibleset;
  mMarkedStates.Name("MarkedStates");
  // symbolic state name handling
  if (IsUnique() && StateNamesEnabled()) {
    for (it = not_accessible.Begin(); it != not_accessible.End(); ++it) {
      mpStateSymbolTable->DelState(*it);
    }
  }
  // return true if there is a initial state
  if (! mInitStates.Empty()) {
    FD_DF("TGenerator<TrnSet>::accessible: generator is accessible");
    return true;
  }
  else {
    FD_DF("TGenerator<TrnSet>::accessible: generator is accessible");
    return false;
  }

}

// IsAccessible()
template <class TrnSet>
bool TGenerator<TrnSet>::IsAccessible(void) const {
  if ((AccessibleSet() == mStates) && (! mInitStates.Empty())) {
    FD_DF("TGenerator<TrnSet>::accessible: generator is accessible");
    return true;
  }
  else {
    FD_DF("TGenerator<TrnSet>::accessible: generator is not accessible");
    return false;
  }
}

// CoaccessibleSet()
template <class TrnSet>
StateSet TGenerator<TrnSet>::CoaccessibleSet(void) const {
  StateSet coaccessibleset = StateSet();
  StateSet::const_iterator it;
  typename TTransSet<TSortX1EvX2<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::const_iterator tit;
  // build reverse transition relation
  TTransSet<TSortX2EvX1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType> rtrel = TransRelX2EvX1();
  for (it = MarkedStatesBegin(); it != MarkedStatesEnd(); ++it) {
    CheckCoaccessible(coaccessibleset, rtrel, *it);
  }
  coaccessibleset.Name("coaccessibleset");
  return coaccessibleset;
}

// Coaccessible()
template <class TrnSet>
bool TGenerator<TrnSet>::Coaccessible(void) {
  StateSet coaccessibleset = CoaccessibleSet();
  StateSet not_coaccessible = mStates - coaccessibleset;
  StateSet::const_iterator it;
  // mTransRel
  for(it = not_coaccessible.Begin(); it != not_coaccessible.End(); ++it){
      mTransRel.EraseByX1OrX2(*it);
  }
  // mStates
  mStates = coaccessibleset;
  mStates.Name("States");
  // mInitStates
  mInitStates = mInitStates * coaccessibleset;
  mInitStates.Name("InitStates");
  // mMarkedStates
  mMarkedStates = mMarkedStates * coaccessibleset;
  mMarkedStates.Name("MarkedStates");
  // symbolic state name handling
  if (IsUnique() && StateNamesEnabled()) {
    for (it = not_coaccessible.Begin(); it != not_coaccessible.End(); ++it) {
      mpStateSymbolTable->DelState(*it);
    }
  }
  // return true if there is a marked state
  if (! mMarkedStates.Empty()) {
    FD_DF("TGenerator<TrnSet>::coaccessible: generator is coaccessible");
    return true;
  }
  else {
    FD_DF("TGenerator<TrnSet>::coaccessible: generator is not coaccessible");
    return false;
  }
}

// IsCoaccessible()
template <class TrnSet>
bool TGenerator<TrnSet>::IsCoaccessible(void) const {
  if ((CoaccessibleSet() == mStates) && (! mMarkedStates.Empty())) {
    FD_DF("TGenerator<TrnSet>::coaccessible: generator is coaccessible");
    return true;
  }
  else {
    FD_DF("TGenerator<TrnSet>::coaccessible: generator is not coaccessible");
    return false;
  }
}

// TrimSet()
template <class TrnSet>
StateSet TGenerator<TrnSet>::TrimSet(void) const {
  FD_DF("TGenerator<TrnSet>::trimset: trim states: " 
      << (AccessibleSet() * CoaccessibleSet()).ToString());
  return AccessibleSet() * CoaccessibleSet();
}

// Trim()
template <class TrnSet>
bool TGenerator<TrnSet>::Trim(void) {
  FD_DF("TGenerator<TrnSet>::trim: generator states: " << mStates.ToString());
  bool accessiblebool = Accessible();
  bool coaccessiblebool = Coaccessible();
  FD_DF("TGenerator<TrnSet>::trim: trim states: " << mStates.ToString());
  if (accessiblebool && coaccessiblebool) {
    FD_DF("TGenerator<TrnSet>::trim: generator is trim");
    return true;
  }
  else {
    FD_DF("TGenerator<TrnSet>::trim: generator is not trim");
    return false;
  }
}

// IsTrim()
template <class TrnSet>
bool TGenerator<TrnSet>::IsTrim(void) const {
  if (IsAccessible() && IsCoaccessible()) {
    FD_DF("TGenerator<TrnSet>::trim: generator is trim");
    return true;
  }
  else {
    FD_DF("TGenerator<TrnSet>::trim: generator is not trim");
    return false;
  }
}

// NewGen()
template <class TrnSet>
TGenerator<TrnSet> TGenerator<TrnSet>::NewGen(void) const {
  return NewGenerator();
}

// NewGenp()
template <class TrnSet>
TGenerator<TrnSet>* TGenerator<TrnSet>::NewGenp(void) const {
  return NewGeneratorp();
}

// NewGenerator()
template <class TrnSet>
TGenerator<TrnSet> TGenerator<TrnSet>::NewGenerator(void) const {
  return TGenerator<TrnSet>();
}

// NewGeneratorp()
template <class TrnSet>
TGenerator<TrnSet>* TGenerator<TrnSet>::NewGeneratorp(void) const {
  return new TGenerator<TrnSet>();
}

// CheckAccessible(accessibleset, startState)
template <class TrnSet>
void TGenerator<TrnSet>::CheckAccessible(StateSet& accessibleset, Idx startState) const {
  if (! accessibleset.ExistsState(startState)) {
    accessibleset.Insert(startState);
    typename TrnSet::const_iterator it = TransRelBegin(startState);
    typename TrnSet::const_iterator it_end = TransRelEnd(startState);
    for (; it != it_end; ++it) {
      CheckAccessible(accessibleset, it->X2);
    }
  }
}

// CheckCoaccessible(coaccessibleset, rtransrel, startState)
template <class TrnSet>
void TGenerator<TrnSet>::CheckCoaccessible(StateSet& coaccessibleset, 
    const TTransSet<TSortX2EvX1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>& rtrel, Idx startState) const {
  if (! coaccessibleset.ExistsState(startState)) {
    coaccessibleset.Insert(startState);
    typename TTransSet<TSortX2EvX1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::const_iterator it = rtrel.BeginByX2(startState);
    typename TTransSet<TSortX2EvX1<typename TrnSet::TransitionType>, typename TrnSet::TransitionType>::const_iterator it_end = rtrel.EndByX2(startState);
    for (; it != it_end; ++it) {
      CheckCoaccessible(coaccessibleset, rtrel, it->X1);
    }
  }
}

// MinStateIndexMap()
template <class TrnSet>
std::map<Idx,Idx> TGenerator<TrnSet>::MinStateIndexMap(void) const {
  std::map<Idx,Idx> minindexmap;
  StateSet::const_iterator it;
  Idx minindex = 0;
  // if generator states got names
  if (StateNamesEnabled()) {
    // named initial states first
    for (it = InitStatesBegin(); it != InitStatesEnd(); ++it) {
      if (StateName(static_cast<Idx>(*it)) != "") {
        minindexmap[*it] = ++minindex;
      }
    }
    // then all other named states
    for (it = StatesBegin(); it != StatesEnd(); ++it) {
      if (minindexmap.count(*it) == 0) {
        if (StateName(static_cast<Idx>(*it)) != "") {
          minindexmap[*it] = ++minindex;
        }
      }
    }
    // at last all anonymous states
    for (it = StatesBegin(); it != StatesEnd(); ++it) {
      if (minindexmap.count(*it) == 0) {
        minindexmap[*it] = ++minindex;
      }
    }
  }
  // if generator states are all anonymous
  else {
    // all initial states first
    for (it = InitStatesBegin(); it != InitStatesEnd(); ++it) {
      minindexmap[*it] = ++minindex;
    }
    // then the rest
    for (it = StatesBegin(); it != StatesEnd(); ++it) {
      if (minindexmap.count(*it) == 0) {
        minindexmap[*it] = ++minindex;
      }
    }
  }
#ifdef FAUDES_DEBUG_CONTAINER
  std::map<Idx,Idx>::const_iterator _it;
  for (_it = minindexmap.begin(); _it != minindexmap.end(); ++_it) {
    FD_DC("TGenerator<TrnSet>::MinStateIndexMap: " << _it->first
        << " <-- " << SStr(_it->second));
  }
#endif
  return minindexmap;
}

// EStr(index)
template <class TrnSet>
std::string TGenerator<TrnSet>::EStr(Idx index) const {
  return mAlphabet.Str(index);
}

// SStr(index)
template <class TrnSet>
std::string TGenerator<TrnSet>::SStr(Idx index) const {
  return mStates.Str(index);
}


} // namespace faudes

#define GENERATOR_H
#endif

