/* memorybase.h -- memory management for generator smart pointers */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef MEMORYBASE_H

#include "definitions.h"
#include "debug.h"
#include <vector>
#include <set>

namespace faudes {

template <class GenType> 
class TGeneratorPool;

template <class GenType> 
class TGenPtr;

typedef std::vector<void*>::size_type ptr_idx;


///////////////////////////////////////////////////////////////////////////////
//
// class specification of TGenPtr<GenType>
//
///////////////////////////////////////////////////////////////////////////////


/** 
 * TGenPtr is a special smart pointer class for any kind of generators. 
 * To be precise it can be used for any type of object.
 * The generator type is specified by the template parameter "GenType". 
 *
 * The primary goals of the smart pointer's design are:
 *   * Deleting the generators's (heap) allocated memory when the last 
 *     corresponding TGenPtr which points to the memory gets out of scope.
 *   * Having the possibility to iterate over all generators in the smart
 *     pointers associated with a GeneratorPool
 * 
 * TGenPtrs are created by the TGeneratorPool class, which has to be 
 * subclassed and adapted for the special features of the generator 
 * type, which means e.g. providing an EventSymbolTable, StateSymbolTable, 
 * EventFlagTable and overriding the method which allocate generators.
 * For TGenPtrs itself just a type definition should be made.
 *
 * See cmemory.h for an example.
 * 
 * A TGenPtr internally got pointers to a generator and the corresponding 
 * TGeneratorPool. It also contains an integer type member "id" which 
 * describes the position in TGeneratorPool's generator*- and reference- 
 * management vectors.
 *
 * TGenPtrs have pointer operators "*" and "->" overloaded, so handling is 
 * transparent (compared to generator pointers).
 * 
 * TGenPtrs should always be passed by value, never use references. By always
 * using call-by-value, the TGenPtr's copy constructor ensures correct handling
 * of the internal generator's destructor.
 *
 * When using the assignment operator "=" or copy constructor, a new TGenPtr
 * pointing to the same generator is created and registered in the
 * TGeneratorPool. 
 *
 * 
 * Members:
 *
 * @param mPointerToGeneratorPool
 *   Pointer to TGeneratorPool. May be 0
 * @param rGen
 *   Pointer to internal generator. May be 0 if mPointerToGeneratorPool is also 0.
 * @param id
 *   Internal TGeneratorPool vector index "id". Unspecified if mPointerToGeneratorPool = 0.
 *   (It's set to some vector index maximum value for default.)
 */
template <class GenType>
class TGenPtr {
  friend class TGeneratorPool<GenType>;
  
  public:
    /**
     * Empty constructor
     */
    TGenPtr(void);

    /**
     * Copy constructor
     */
    TGenPtr(const TGenPtr& rOtherPtr);

    /**
     * TGenPtr destructor
     */
    ~TGenPtr();
    
    /** 
     * Assignment operator
     * 
     * @param rOtherPtr
     *   Other TGenPtr
     */
    TGenPtr& operator = (const TGenPtr& rOtherPtr);

    /**
     * "Pointer-to-Member" operator.
     * This dereferences the internal generator pointer.
     *
     * @return
     *   GenType&
     */
    GenType& operator *() const;

    /**
     * "Pointer-to-Member" operator.
     * This dereferences the internal GenType pointer for accessing
     * generators members.
     *
     * @return
     *   GenType*
     */
    GenType* operator ->() const;

    /**
     * Returns the pointer to the TGeneratorPool which created the smart pointer
     *
     * @return 
     *   TGeneratorPool*
     */
    TGeneratorPool<GenType>* GetPool(void) const;

    /**
     * Returns the pointer to the generator
     * 
     * Use with care:
     * * Don't free the memory the pointer is pointing to!
     * * Never assign the generator pointer to some variable which could live
     *   longer than the TGenPtrs pointing to generator!!!
     *
     * @return
     *   GenType*
     */
    GenType* Get(void) const;
  
    /**
     * Unregister internal generator in TGeneratorPool and return it. TGenPtr 
     * itself becomes also unregistered. It has to be registered again 
     * (with TGenPtr::reset) before reuse.
     * Take care: The gen Member variable of all TGenPtrs pointing to this 
     * generator will be set to 0!
     * 
     * This is the save way of "get"-ting (see TGenPtr::Get()) the generator pointer.
     */
    GenType* Release(void);

    /**
     * Unregister this TGenPtr in TGeneratorPool and reregister with an optional
     * new generator Pointer. 
     *
     * By supplying no generator* as argument this has just the same effect as 
     * calling the class destructor. TGenPtr becomes unregistered and the 
     * corresponding generator* freed if this TGenPtr was the last one in scope for
     * the id.
     *
     * @param p
     *   Pointer to generator (optional)
     */
    void Reset(GenType* p = 0);

    /**
     * Debugging output to console
     */
    void DWrite(void) const;

  private:

    /**
     * Create a TGenPtr containing a valid generator* directly. Only meant
     * for internal use in TGeneratorPool, as this doesn't register the
     * TGenPtr.
     *
     * This constructor is private so it can't be abused by a user.
     *
     * @param pPointerToGeneratorPool
     *   Pointer to TGeneratorPool
     * @param pPointerToGenerator
     *   Pointer to generator
     * @param id
     *   vector index id in TGeneratorPool
     */
    TGenPtr(TGeneratorPool<GenType>* pPointerToGeneratorPool, 
        GenType* pPointerToGenerator, ptr_idx id);

    /**
     * "Unset" a TGenPtr by setting internal pointers to 0 and id to the
     * default value.
     */
    void Unset(void);

    /** Pointer to the generator */  
    GenType* mPointerToGenerator;
    
    /** Internal index in TGeneratorPool */
    ptr_idx mId;

    /** Pointer to TGeneratorPool */
    TGeneratorPool<GenType>* mPointerToGeneratorPool;
};



///////////////////////////////////////////////////////////////////////////////
//
// class implementation of TGenPtr<GenType>
//
///////////////////////////////////////////////////////////////////////////////


// TGenPtr<GenType>::TGenPtr()
template <class GenType>
TGenPtr<GenType>::TGenPtr(void) 
  : mPointerToGenerator(0), 
    mId(TGeneratorPool<GenType>::mMaxId), 
    mPointerToGeneratorPool(0) {
  FD_DM("GenPtr<GenType>::GenPtr<" << this << ">(): "
      << "this.mPointerToGenerator=" << mPointerToGenerator << " "
      << "this.mId=" << mId << " "
      << "this.mPointerToGeneratorPool=" << mPointerToGeneratorPool << " "
      );
}

// TGenPtr<GenType>::TGenPtr(rOtherPtr)
template <class GenType>
TGenPtr<GenType>::TGenPtr(const TGenPtr& rOtherPtr) 
  : mPointerToGenerator(rOtherPtr.mPointerToGenerator),
    mId(rOtherPtr.mId), 
    mPointerToGeneratorPool(rOtherPtr.mPointerToGeneratorPool) {
  if (mPointerToGeneratorPool != 0) {
    mId = mPointerToGeneratorPool->RegisterGenPtr(this);  
  }
  else {
    mId = TGeneratorPool<GenType>::mMaxId;
  }
  FD_DM("GenPtr::GenPtr<" << this << ">(rOtherPtr): " 
      << "this.mPointerToGenerator=" << mPointerToGenerator << " "
      << "this.mId=" << mId << " "
      << "this.mPointerToGeneratorPool=" << mPointerToGeneratorPool 
      << " "
      << "rOtherPtr.mPointerToGenerator=" << rOtherPtr.mPointerToGenerator 
      << " "
      << "rOtherPtr.mId=" << rOtherPtr.mId << " "
      << "rOtherPtr.mPointerToGeneratorPool=" 
      << rOtherPtr.mPointerToGeneratorPool << " "
      );
}

// TGenPtr<GenType>::TGenPtr(pPointerToGeneratorPool, pPointerToGenerator, id)
template <class GenType>
TGenPtr<GenType>::TGenPtr(TGeneratorPool<GenType>* pPointerToGeneratorPool, 
    GenType* pPointerToGenerator, ptr_idx id) {
  mPointerToGenerator = pPointerToGenerator;
  mPointerToGeneratorPool = pPointerToGeneratorPool;
  mId = id;
  FD_DM("GenPtr::GenPtr<" << this << ">(pPointerToGeneratorPool*=" 
      << pPointerToGeneratorPool << " generator*=" 
      << pPointerToGenerator << " " 
      << " id=" << id << "): "
      << "this.mPointerToGenerator=" << mPointerToGenerator << " "
      << "this.mId=" << mId << " "
      << "this.mPointerToGeneratorPool=" << mPointerToGeneratorPool 
      << " ");
}

// TGenPtr<GenType>::~TGenPtr()
template <class GenType>
TGenPtr<GenType>::~TGenPtr(void) {
  FD_DM("GenPtr::~GenPtr<" << this << ">(): "
      << "this.mPointerToGenerator=" << mPointerToGenerator << " "
      << "this.mId=" << mId << " "
      << "this.mPointerToGeneratorPool=" << mPointerToGeneratorPool << " "
      );
  if (mPointerToGeneratorPool != 0) {
    mPointerToGeneratorPool->DelGenPtr(this);
  }
  else {
    // this should never happen
    if (mPointerToGenerator != 0) {
      FD_ERR("GenPtr::~GenPtr<" << this << ">(): "
          << "pool == 0 but gen != 0 !!! This should never happen");
      abort();
    }
  }
}

// operator =
template <class GenType>
TGenPtr<GenType>& TGenPtr<GenType>::operator = (
    //const TGenPtr<GenType>& rOtherPtr) {
    const TGenPtr& rOtherPtr) {
  if (this != &rOtherPtr) { 
    // free possibly allocated memory first
    if (mPointerToGeneratorPool != 0) {
      mPointerToGeneratorPool->DelGenPtr(this);
    }
    // set members
    mPointerToGenerator = rOtherPtr.mPointerToGenerator;
    mPointerToGeneratorPool = rOtherPtr.mPointerToGeneratorPool;
    mId = rOtherPtr.mId;
    if (mPointerToGeneratorPool != 0) {
      mId = mPointerToGeneratorPool->RegisterGenPtr(this);  
    }
    else {
      mId = TGeneratorPool<GenType>::mMaxId;
    }
    FD_DM("GenPtr::operator = <" << this << ">: " 
        << "this.mPointerToGenerator=" << mPointerToGenerator << " "
        << "this.mId=" << mId << " "
        << "this.mPointerToGeneratorPool=" << mPointerToGeneratorPool << " "
        << "rOtherPtr.mPointerToGenerator=" << rOtherPtr.mPointerToGenerator << " "
        << "rOtherPtr.mId=" << rOtherPtr.mId << " "
        << "rOtherPtr.mPointerToGeneratorPool=" << rOtherPtr.mPointerToGeneratorPool << " "
        );
  }  
  return *this;
}

// operator *()
template <class GenType>
GenType& TGenPtr<GenType>::operator *() const {
  if (mPointerToGenerator == 0) {
    FD_ERR("GenPtr::operator *: dereferencing an unallocated generator");
    abort();
  }
  return *mPointerToGenerator;
}

// operator ->()
template <class GenType>
GenType* TGenPtr<GenType>::operator ->() const {
  if (mPointerToGenerator == 0) {
    FD_ERR("GenPtr::operator ->(): dereferencing an unallocated generator");
    abort();
  }
  return mPointerToGenerator;
}

// GetPool()
template <class GenType>
TGeneratorPool<GenType>* TGenPtr<GenType>::GetPool(void) const {
  return mPointerToGeneratorPool;
}

// Get()
template <class GenType>
GenType* TGenPtr<GenType>::Get(void) const {
  return mPointerToGenerator;
}

// Release()
template <class GenType>
GenType* TGenPtr<GenType>::Release(void) {
  if (mPointerToGeneratorPool != 0) {
    mPointerToGeneratorPool->ReleaseGenPtr(mId);
    mPointerToGeneratorPool = 0;
  }
  else {
    FD_ERR("GenPtr::Release(): GenPtr was not registered in a GeneratorPool");
    exit(1);
  }
  GenType* result = mPointerToGenerator;
  mPointerToGenerator = 0;
  return result;
}

// Reset()
template <class GenType>
void TGenPtr<GenType>::Reset(GenType* p) {
  // test if pointer is valid
  if (p != 0) {
    // this should crash program is pointer is pointing "somewhere" in space
    p->Name();
  }
  mPointerToGenerator = p;
  if (mPointerToGeneratorPool != 0) {
    mPointerToGeneratorPool->DelGenPtr(this);
    mId = mPointerToGeneratorPool->RegisterGenPtr(this, true);
  }
  else {
    FD_ERR("GenPtr::Reset(generator*): "
        << "GenPtr was not registered in a GeneratorPool");
    exit(1);
  }
}

// Unset()
template <class GenType>
void TGenPtr<GenType>::Unset(void) {
  mPointerToGenerator = 0;
  mId = TGeneratorPool<GenType>::mMaxId;
  mPointerToGeneratorPool = 0;
}
    
// DWrite()
template <class GenType>
void TGenPtr<GenType>::DWrite(void) const {
  std::cerr << "========== GenPtr DWrite() ==========" << std::endl;
  std::cerr << "gen=" << mPointerToGenerator << std::endl;
  std::cerr << "id=" << mId << std::endl;
  std::cerr << "mPointerToGeneratorPool=" << mPointerToGeneratorPool << std::endl;
  std::cerr << "=====================================" << std::endl;
}



///////////////////////////////////////////////////////////////////////////////
//
// class specification of TGeneratorPool<GenType>
//
///////////////////////////////////////////////////////////////////////////////


/**
 * The TGeneratorPool "registers" and "unregisters" pointers to generators 
 * specified by template parameter "GenType". For user access the generators
 * are stored in smart pointers of the corresponding smart pointer class 
 * "TGenPtr". It holds information about references and handles the logic 
 * behind the smart pointers.
 *
 * TGeneratorPool<GenType> is an abstract base class and has to be subclassed 
 * by providing a GenType in the template argument and overriding 
 * the three variants of the method "newptr" and the method "newgen" in the
 * subclass which both care for generator memory allocation.
 *
 * See cmemory.h for an example.
 * 
 * Every generator* is assigned an integer type "id" which corresponds to the 
 * index in the two vectors mPointerToGeneratorsVector and mSmartPointerSets:
 * 
 * * mPointerToGeneratorsVector<GenType*>:
 *   Holds copys of the pointers to the generators in TGenPtrs 
 *   (member "mPointerToGenerator").
 * 
 * * mSmartPointerSets< std::set<TGenPtr*> >:
 *   Holds pointers to all TGenPtrs which have the 
 *   mPointerToGeneratorsVector[mId] (in this class) assigned to their 
 *   mPointerToGenerator member.
 *
 * Currently unused "id"s are stored in a set called "mFreeVectorIndices".
 *
 * End user methods are:
 * 
 * * TGenPtr TGeneratorPool::NewPtr()  (virtual)
 *   Returns a TGenPtr with an allocated generator
 * 
 * * TGenPtr TGeneratorPool::NewPtr(GenType* gen)  (virtual)
 *    Returns a TGenPtr created from an existing generator
 *
 * * TGenPtr TGeneratorPool::NewPtr(const char* filename)  (virtual)
 *   Returns a TGenPtr with an allocated generator from file.
 *
 * * GenType* TGeneratorPool::NewGen(void)  (virtual)
 *   Returns an allocated generator. Has to be deleted with "delete(..)" 
 *   again.
 *
 * * void dWrite()
 *   Writes TGeneratorPool status to console.
 *
 * Members:
 * mPointerToGeneratorsVector
 *   Vector of generator* pointers
 * mSmartPointerSets
 *   Vector of std::set<TGenPtr*>* pointers, holding pointers to the smartpointer
 *   objects
 * mFreeVectorIndices
 *   Set of free vector indices
 * maxid
 *   Maximum vector id. Used as default value in TGenPtr if TGenPtr currently
 *   not registered in TGeneratorPool
 */
 
template <class GenType>
class TGeneratorPool {
  friend class TGenPtr<GenType>;

  public:
    /**
     * Constructor
     */
    TGeneratorPool(void);

    /**
     * Destructor
     * Deallocates memory of generators
     */
    virtual ~TGeneratorPool(void);
    
    /**
     * Create new smart pointer with allocated internal generator
     * (Virtual function)
     */
    virtual TGenPtr<GenType> NewPtr(void);

    /**
     * Create new smart pointer from existing generator*
     * (Virtual function)
     */
    virtual TGenPtr<GenType> NewPtr(GenType* pPointerToGenerator);

    /**
     * Create new smart pointer by reading a generator file
     * (Virtual function)
     */
    virtual TGenPtr<GenType> NewPtr(const char* pFileName);

    /** iterator to access generator pointers in mPointerToGeneratorsVector
     *
     * Iterator position has to be checked for nullpointers before 
     * dereferencing!!!
     */
    typedef typename std::vector<GenType*>::iterator iterator;

    /** const iterator to access generator pointers in mPointerToGeneratorsVector
     *
     * Iterator position has to be checked for nullpointers before 
     * dereferencing!!!
     */
    typedef typename std::vector<GenType*>::const_iterator const_iterator;

    /**
     * Iterator to begin of generator* vector
     *
     * Iterator position has to be checked for nullpointers before 
     * dereferencing!!!
     *
     * @return
     *   TGeneratorPool<GenType>::iterator
     */
    iterator Begin(void);

    /**
     * Iterator to end of generator* vector
     *
     * Iterator position has to be checked for nullpointers before 
     * dereferencing!!!
     *
     * @return
     *   TGeneratorPool<GenType>::iterator
     */
    iterator End(void);

    /** 
     * Write TGeneratorPool's status to console
     */
    void DWrite(void) const;

  private:
      
    /**
     * Registers a smart pointer in TGeneratorPool. If "true" is given as second
     * argument, the generator will get new id. If called with one argument the
     * generator gets registered for the id stored in TGenPtr, if the id is
     * valid. In case of an unvalid id the generator is registered for a new id
     * anyway.
     *
     * @param pGenPtr
     *   Pointer to a TGenPtr. Usually the "this" variable of the TGenPtr asking 
     *   for registering in TGeneratorPool
     * @param usenewid
     *   Forces the creation of an unallocated id.
     */
    ptr_idx RegisterGenPtr(TGenPtr<GenType>* pGenPtr, bool usenewid = false);

    /**
     * Unregisters (releases) a TGenPtr from TGeneratorPool. In the case it's the 
     * last TGenPtr containing the generator* pointer, the id will be freed in 
     * all internal vectors and added to the mFreeVectorIndices set.
     * The TGenPtr itself stays untouched, the allocated memory of the generator
     * too.
     *
     * @param pGenPtr
     *   Pointer to TGenPtr
     */
    void ReleaseGenPtr(TGenPtr<GenType>* pGenPtr);

    /**
     * Unregisters (releases) a generator from TGeneratorPool. All TGenPtrs 
     * pointing to the corresponding generator will be unregistered. The 
     * generator* is returned after freeing entries in all internal vectors 
     * and adding id to the mFreeVectorIndices set.
     *
     * @param id
     *   Vector index "id" of a generator*
     *
     * @return
     *   Unregistered generator*
     */
    GenType* ReleaseGenPtr(ptr_idx id);

    /**
     * Unregisters a TGenPtr from TGeneratorPool. Calls releaseptr first and 
     * deletes allocated generator memory if given TGenPtr was the last one for
     * it's id in TGeneratorPool. TGenPtr stays untouched anyway.
     *
     * @param pGenPtr
     *   TGenPtr*
     */
    void DelGenPtr(TGenPtr<GenType>* pGenPtr);

    /**
     * "Gets" the next valid id in TGeneratorPool. This may be one if the ids
     * stored in the free set or the vectors' size if there is no mFreeVectorIndices id.
     * Returned id will be removed from mFreeVectorIndices set.
     *
     * @return
     *   Next valid vector index 
     */
    ptr_idx GetNextId(void);

    /**
     * Test if an id (vector index) contains valid entries in vectors
     *
     * @param id
     *   Vector index
     */
    bool ValidId(ptr_idx id);
    
    /** Vector storing generator* pointers */
    std::vector<GenType*> mPointerToGeneratorsVector;

    /** Vector storing sets of pointers to smart pointer objects */
    std::vector< std::set<TGenPtr<GenType>*>* > mSmartPointerSets;

    /** Set containing vector indices which are currently not in use */
    std::set<ptr_idx> mFreeVectorIndices;

    /** Maximum vector index */
    static ptr_idx mMaxId;
};



///////////////////////////////////////////////////////////////////////////////
//
// class implementation of TGeneratorPool<GenType>
//
///////////////////////////////////////////////////////////////////////////////


// TGeneratorPool::TGeneratorPool()
template <class GenType>
TGeneratorPool<GenType>::TGeneratorPool(void) {
}

// TGeneratorPool<GenType>::~GeneratorPool()
template <class GenType>
TGeneratorPool<GenType>::~TGeneratorPool(void) {
  // clean up
  typename std::vector<GenType*>::iterator gens_it;
  typename std::vector< std::set<TGenPtr<GenType>*>* >::iterator ptrs_it;
  typename std::set<TGenPtr<GenType>*>::iterator ptrset_it;
  for (gens_it = mPointerToGeneratorsVector.begin(); 
      gens_it != mPointerToGeneratorsVector.end(); ++gens_it) {
    delete(*gens_it);
    *gens_it = 0;
  }
  for (ptrs_it = mSmartPointerSets.begin(); 
      ptrs_it != mSmartPointerSets.end(); ++ptrs_it) {
    if (*ptrs_it != 0) {
      for (ptrset_it = (*ptrs_it)->begin(); 
          ptrset_it != (*ptrs_it)->end(); ++ptrset_it) {
        (*ptrset_it)->Unset();
      }
      delete(*ptrs_it);
      *ptrs_it = 0;
    }
  }
}

// Begin()
template <class GenType>
typename TGeneratorPool<GenType>::iterator 
    TGeneratorPool<GenType>::Begin(void) {
  return mPointerToGeneratorsVector.begin();
}

// End()
template <class GenType>
typename TGeneratorPool<GenType>::iterator 
    TGeneratorPool<GenType>::End(void) {
  return mPointerToGeneratorsVector.end();
}

// TGeneratorPool<GenType>::mMaxId
template <class GenType>
ptr_idx TGeneratorPool<GenType>::mMaxId = std::numeric_limits<ptr_idx>::max();

// NewPtr()
template <class GenType>
TGenPtr<GenType> TGeneratorPool<GenType>::NewPtr() {
  return NewPtr(new GenType);
}

// NewPtr(gen)
template <class GenType>
TGenPtr<GenType> TGeneratorPool<GenType>::NewPtr(GenType* pPointerToGenerator) {
  // get next free id
  const ptr_idx resid = GetNextId();
  // create GenPtr on stack
  TGenPtr<GenType> ptr(this, pPointerToGenerator, resid);
  // prepare data structures:
  // ptrset
  std::set< TGenPtr<GenType>* >* ptrset = new std::set< TGenPtr<GenType>* >;
  ptrset->insert(&ptr);
  // register ptr:
  // if no free index in mSmartPointerSets vector
  if (resid == mSmartPointerSets.size()) {
    mPointerToGeneratorsVector.push_back(pPointerToGenerator);
    mSmartPointerSets.push_back(ptrset);
  }
  // if there is a free index
  else {
    mPointerToGeneratorsVector[resid] = pPointerToGenerator;
    mSmartPointerSets[resid] = ptrset;
  }
  FD_DM("TGeneratorPool::NewPtr()");
  return ptr;
}

// NewPtr("filename")
template <class GenType>
TGenPtr<GenType> TGeneratorPool<GenType>::NewPtr(const char* pFileName) {
  return NewPtr(new GenType(pFileName));
}

// RegisterGenPtr(ptr, useNewId=false)
template <class GenType>
ptr_idx TGeneratorPool<GenType>::RegisterGenPtr(TGenPtr<GenType>* pGenPtr, 
    bool useNewId) {
  FD_DM("TGeneratorPool::RegisterGenPtr(" << pGenPtr << ", " << useNewId << "): "
      << "registering GenPtr in GeneratorPool");
  ptr_idx resid = pGenPtr->mId;
  // if resid is valid just add ptr to mSmartPointerSets[id]
  if (ValidId(resid) && (useNewId == false)) {
    FD_DM("TGeneratorPool::RegisterGenPtr(" << pGenPtr << "): "
        << "id " << resid << " already exists. only registering address");
    mSmartPointerSets[resid]->insert(pGenPtr);
  }
  // if resid is not valid create the next free id and set pGenPtr->mId to it
  else {
    FD_DM("TGeneratorPool::RegisterGenPtr(" << pGenPtr << "): "
        << "id " << resid << " not valid. creating new id.");
    resid = GetNextId();
    // prepare data structurs for container insertion:
    // mSmartPointerSets
    std::set< TGenPtr<GenType>* >* ptrset = new std::set< TGenPtr<GenType>* >;
    ptrset->insert(pGenPtr);
    // if no free index in mSmartPointerSets vector
    if (resid == mSmartPointerSets.size()) {
      mPointerToGeneratorsVector.push_back(pGenPtr->mPointerToGenerator);
      mSmartPointerSets.push_back(ptrset);
    }
    // if there is a free index
    else {
      mPointerToGeneratorsVector[resid] = pGenPtr->mPointerToGenerator;
      mSmartPointerSets[resid] = ptrset;
    }
  }
  return resid;
}
    
// ReleaseGenPtr(pGenPtr)
template <class GenType>
void TGeneratorPool<GenType>::ReleaseGenPtr(TGenPtr<GenType>* pGenPtr) {
  const ptr_idx id = pGenPtr->mId;
  if (ValidId(id)) {
    FD_DM("TGeneratorPool::ReleaseGenPtr(" << id << "): "
        << "mSmartPointerSets[" << id << "]=" << mSmartPointerSets[id] << " "
        );
    typename std::set< TGenPtr<GenType>* >::iterator ptrset_it;
    ptrset_it = mSmartPointerSets[id]->find(pGenPtr);
    if (ptrset_it != mSmartPointerSets[id]->end()) {
      // erase pGenPtr from mSmartPointerSets[id]
      mSmartPointerSets[id]->erase(ptrset_it);
      // if this was the last TGenPtr* for corresponding generator
      if (mSmartPointerSets[id]->empty()) {
        // delete ptrset
        delete(mSmartPointerSets[id]);
        mSmartPointerSets[id] = 0;
        // set generator = 0
        mPointerToGeneratorsVector[id] = 0;
        // insert in mFreeVectorIndices set
        mFreeVectorIndices.insert(id);
      }
      // if there are some more GenPtr*s for this generator 
      // just do nothing
    }
    else {
      FD_ERR("TGeneratorPool::ReleaseGenPtr(" << id << ", " << pGenPtr << "): "
          << "ptr not valid for id");
      exit(1);
    }
  }
  else {
    FD_ERR("TGeneratorPool::ReleaseGenPtr(" << id << ", " << pGenPtr << "): "
      << "id not valid");
    exit(1);
  } 
}

// ReleaseGenPtr(id)
template <class GenType>
GenType* TGeneratorPool<GenType>::ReleaseGenPtr(ptr_idx id) {
  if (ValidId(id)) {
    FD_DM("TGeneratorPool::ReleaseGenPtr(" << id << "): "
        << "mSmartPointerSets[" << id << "]=" << mSmartPointerSets[id] << " "
        );
    // prepare result:
    GenType* gen = mPointerToGeneratorsVector[id];
    // set id + gen in every GenPtr of id to default values
    typename std::set< TGenPtr<GenType>* >::iterator ptrset_it;
    for (ptrset_it = mSmartPointerSets[id]->begin(); ptrset_it != mSmartPointerSets[id]->end(); 
        ++ptrset_it) {
      (*ptrset_it)->mId = mMaxId;
      (*ptrset_it)->mPointerToGenerator = 0;
    }
    // delete set of GenPtr*
    delete(mSmartPointerSets[id]);
    mSmartPointerSets[id] = 0;
    // set mPointerToGeneratorsVector[id] = 0
    mPointerToGeneratorsVector[id] = 0;
    // insert id in mFreeVectorIndices 
    mFreeVectorIndices.insert(id);
    // return generator*
    return gen;
  }
  else {
    FD_ERR("TGeneratorPool::ReleaseGenPtr(" << id << "): "
      << "id not valid");
    exit(1);
  } 
}

// DelGenPtr(pGenPtr)
template <class GenType>
void TGeneratorPool<GenType>::DelGenPtr(TGenPtr<GenType>* pGenPtr) {
  // save generator pointer
  GenType* gen = pGenPtr->mPointerToGenerator;
  const ptr_idx id = pGenPtr->mId;
  // release smartpointer entries in GeneratorPool
  ReleaseGenPtr(pGenPtr);
  // delete allocated memory if no more GenPtrs point to this generator
  if (! ValidId(id)) {
    delete(gen);
    gen = 0;
  }
}

// GetNextId()
template <class GenType>
ptr_idx TGeneratorPool<GenType>::GetNextId(void) {
  std::set<ptr_idx>::iterator free_it;
  if (mFreeVectorIndices.empty()) {
    return mSmartPointerSets.size();
  }
  else {
    free_it = mFreeVectorIndices.begin();
    const ptr_idx id = *free_it;
    mFreeVectorIndices.erase(free_it);
    return id;
  }
}

// ValidId(id)
template <class GenType>
bool TGeneratorPool<GenType>::ValidId(ptr_idx id) {
  if (mFreeVectorIndices.find(id) != mFreeVectorIndices.end()) {
    FD_DM("TGeneratorPool::ValidId(" << id << "): "
        << "id marked as unassigned");
    return false;
  }
  if (id >= mSmartPointerSets.size()) {
    FD_DM("TGeneratorPool::ValidId(" << id << "): "
        << "id bigger than size of smartpointer vector");
    return false;
  }
  if (mPointerToGeneratorsVector[id] == 0) {
    FD_DM("TGeneratorPool::ValidId(" << id << "): "
        << "mPointerToGeneratorsVector[id] == 0");
    return false;
  }
  if (mSmartPointerSets[id] == 0) {
    FD_DM("TGeneratorPool::ValidId(" << id << "): "
        << "mSmartPointerSets[id] == 0");
    return false;
  }
  return true;
}

// DWrite()
template <class GenType>
void TGeneratorPool<GenType>::DWrite(void) const {
  typename std::vector<GenType*>::const_iterator gens_it;
  typename std::vector< std::set< TGenPtr<GenType>* >* >::const_iterator ptrs_it;
  typename std::set< TGenPtr<GenType>* >::const_iterator ptr_it;
  std::vector<unsigned int*>::const_iterator refs_it;
  std::set<ptr_idx>::const_iterator free_it;
  
  std::cerr << "============ GeneratorPool status ============" << std::endl;
  std::cerr << "free indices: ";
  for (free_it = mFreeVectorIndices.begin(); free_it != mFreeVectorIndices.end(); ++free_it) {
    std::cerr << *free_it << " ";
  }
  std::cerr << std::endl;
  std::cerr << "generator* vector: ";
  for (gens_it = mPointerToGeneratorsVector.begin(); gens_it != mPointerToGeneratorsVector.end(); ++gens_it) {
    std::cerr << *gens_it << " | ";
  }
  std::cerr << std::endl;
  std::cerr << "GenPtr* sets: ";
  for (ptrs_it = mSmartPointerSets.begin(); ptrs_it != mSmartPointerSets.end(); ++ptrs_it) {
    if (*ptrs_it != 0) {
      for (ptr_it = (*ptrs_it)->begin(); ptr_it != (*ptrs_it)->end();
          ++ptr_it) {
        std::cerr << *ptr_it << " ";
      }
    }
    else {
      std::cerr << "0* ";
    }
    std::cerr << "| ";
  }
  std::cerr << std::endl;
  std::cerr << "==============================================" << std::endl;
}
  
} // namespace faudes  

#define MEMORYBASE_H
#endif 
