/* parallel.h -- parallel composition */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef PARALLEL_H

#include "definitions.h"
#include "debug.h"
#include "generator.h"
#include "cmap.h"
#include <stack>
#include <map>
#include <set>

namespace faudes {

/**
 * Parallel composition of accessible states 
 * of generators rGen1 and rGen2.
 *
 * Computes symbolic state names in rResGen, if symbolic state names are
 * enabled in both rGen1 and rGen2
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting parallel composition generator
 */
void Parallel(const Generator& rGen1, const Generator& rGen2, 
    Generator& rResGen);


/**
 * Parallel composition of accessible states of 
 * generators rGen1 and rGen2.
 *
 * This version fills given reverse composition map with indices. For
 * storage of composition indices use the state composition container
 * provided.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rReverseCompositionMap
 *   Reverse composition map (map< pair<Idx,Idx>, Idx>)
 * @param rResGen
 *   Reference to resulting parallel composition generator
 */
void Parallel(const Generator& rGen1, const Generator& rGen2,
    std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
    Generator& rResGen);


/**
 * Fast shared event parallel composition of accessible states of 
 * deterministic generators rGen1 and rGen2. 
 * (sparallel = Shared event PARALLEL composition)
 * Only synchronized events are executed.
 *
 * Caution: use only with deterministic input generators
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting parallel composition generator
 */
void SParallel(const Generator& rGen1, const Generator& rGen2, 
    Generator& rResGen);


/**
 * Fast shared event parallel composition of accessible states of 
 * deterministic generators rGen1 and rGen2. 
 * (sparallel = Shared event PARALLEL composition)
 * Only synchronized events are be executed.
 *
 * Caution: use only with deterministic input generators
 *
 * This version fills given reverse composition map with indices. For
 * storage of composition indices use the state composition container
 * provided.
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rReverseCompositionMap
 *   Reverse composition map (map< pair<Idx,Idx>, Idx>)
 * @param rResGen
 *   Reference to resulting parallel composition generator
 */
void SParallel(const Generator& rGen1, const Generator& rGen2,
		std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
    Generator& rResGen);

} // namespace faudes

#define PARALLEL_H
#endif 

