/* project.h -- projection and subset construction */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef PROJECT_H

#include "definitions.h"
#include "debug.h"
#include "generator.h"
#include "localgen.h"
#include "statemin.h"
#include <stack>
#include <map>
#include <vector>
#include <list>
#include <utility>
#include <limits>

namespace faudes {

/**
 * Make initial states unique
 * To be called before determine
 *
 * @param rGen
 *   Reference to generator
 */
void UniqueInit(Generator& rGen);


/**
 * Check if generator is deterministic
 *
 * @param rGen
 *   Generator
 */
bool IsDeterministic(const Generator& rGen);

/**
 * Make generator deterministic. (function wrapper)
 *
 * @param rGen
 *   Reference to generator
 * @param rResGen
 *   Reference to resulting deterministic generator
 */
void Deterministic(const Generator& rGen, Generator& rResGen);

/**
 * Make generator deterministic. (function wrapper)
 *
 * The second parameter is an empty std::map<Idx,StateSet> which holds all
 * the pairs of new states and their respective power states set. This
 * is used as a so called "entry state map" for deterministic projected
 * generators.
 *
 * @param rGen
 *   Reference to generator
 * @param rEntryStatesMap
 *   Entry state map
 * @param rResGen
 *   Reference to resulting deterministic generator
 */
void Deterministic(const Generator& rGen, std::map<Idx,StateSet>& rEntryStatesMap,
    Generator& rResGen);

/**
 * Make generator deterministic. (real function)
 *
 * Second and third parameter hold the subsets + deterministic states
 * in vectors
 *
 * The multiway merge algorithm is based on a propsal in
 * "Ted Leslie, Efficient Approaches to Subset Construction, 
 *  Computer Science, University of Waterloo, 1995"
 *
 * @param rGen
 *   Reference to generator
 * @param rPowerStates
 *   Vector that holds the power states
 * @param rDetStates 
 *   Vector that holds the corresponding deterministic states
 * @param rResGen
 *   Reference to resulting deterministic generator
 */
void Deterministic(const Generator& rGen, std::vector<StateSet>& rPowerStates,
		std::vector<Idx>& rDetStates, Generator& rResGen);

/**
 * Project generator to alphabet palphabet
 *
 * @param rGen
 *   Reference to generator
 * @param rProjectAlphabet
 *   Projection alphabet
 */
void Project(Generator& rGen, const EventSet& rProjectAlphabet);

/**
 * Minimized Deterministic projection. Does not modify generator.
 * Calls project, determine and statemin. 
 *
 * @param rGen
 *   Reference to generator
 * @param rProjectAlphabet
 *   Projection alphabet
 * @param rResGen
 *   Reference to resulting deterministic generator
 */
void DetProject(const Generator& rGen, const EventSet& rProjectAlphabet, Generator& rResGen);

/**
 * Minimized Deterministic projection. Does not modify generator.
 * Calls project, determine and statemin. See functions for details.
 *
 * @param rGen
 *   Reference to generator
 * @param rProjectAlphabet
 *   Projection alphabet
 * @param rEntryStatesMap
 *   Reference to entry states map, see Deterministic(..) (result)
 * @param rResGen
 *   Reference to resulting deterministic generator (result)
 */
void DetProject(const Generator& rGen, const EventSet& rProjectAlphabet,
		std::map<Idx,StateSet>& rEntryStatesMap, Generator& rResGen);

/**
 * Inverse projection. This adds selfloop transition at every state for
 * all missing events.
 *
 * @param rGen
 *   Reference to generator
 * @param rProjectAlphabet
 *   Alphabet for inverse projection
 */
void InvProject(Generator& rGen, const EventSet& rProjectAlphabet);

} // namespace faudes

#define PROJECT_H
#endif 

