/* regular.cpp -- regular language operations */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

  
#include "regular.h"

namespace faudes {

// LanguageUnion(rGen1, rGen2, rResGen)
void LanguageUnion(const Generator& rGen1, const Generator& rGen2,
		Generator& rResGen) {
	FD_DF("lunion: forming union composition of \"" << rGen1.Name() 
			<< "\" and \"" << rGen2.Name() << "\"");
	// set up empty generator
	rResGen.Clear();
	// indexmap
	std::map<Idx,Idx> indexmap;
	// helpers
	StateSet::const_iterator lit;
	std::string name;

	// name
	rResGen.Name("("+rGen1.Name()+")v("+rGen2.Name()+")");
	// alphabet 
	rResGen.InjectAlphabet(rGen1.Alphabet()+rGen2.Alphabet());
	// states 
	for (lit = rGen1.StatesBegin(); lit != rGen1.StatesEnd(); ++lit) {
		name = rGen1.StateName(*lit);
		if (name != "") {
			indexmap[*lit] = rResGen.InsState(name+"_1");
		}
		else {
			indexmap[*lit] = rResGen.InsState();
		}
	}
	for (lit = rGen2.StatesBegin(); lit != rGen2.StatesEnd(); ++lit) {
		name = rGen2.StateName(*lit);
		if (name != "") {
			indexmap[*lit] = rResGen.InsState(name+"_2");
		}
		else {
			indexmap[*lit] = rResGen.InsState();
		}
	}
	// transrel
	TransSet::const_iterator tit;
	for (tit = rGen1.TransRelBegin(); tit != rGen1.TransRelEnd(); ++tit) {
		rResGen.SetTransition(indexmap[tit->X1], tit->Ev, indexmap[tit->X2]);
	}
	for (tit = rGen2.TransRelBegin(); tit != rGen2.TransRelEnd(); ++tit) {
		rResGen.SetTransition(indexmap[tit->X1], tit->Ev, indexmap[tit->X2]);
	}
	// istates
	for (lit = rGen1.InitStatesBegin(); lit != rGen1.InitStatesEnd(); ++lit) {
		rResGen.SetInitState(indexmap[*lit]);
	}
	for (lit = rGen2.InitStatesBegin(); lit != rGen2.InitStatesEnd(); ++lit) {
		rResGen.SetInitState(indexmap[*lit]);
	}
	// mstates
	for (lit = rGen1.MarkedStatesBegin(); lit != rGen1.MarkedStatesEnd(); 
      ++lit) {
		rResGen.SetMarkedState(indexmap[*lit]);
	}
	for (lit = rGen2.MarkedStatesBegin(); lit != rGen2.MarkedStatesEnd(); ++lit) {
		rResGen.SetMarkedState(indexmap[*lit]);
	}
}

// DetLanguageUnion(rGen1, rGen2, rResGen)
void DetLanguageUnion(const Generator& rGen1, const Generator& rGen2, 
    Generator& rResGen) {
	FD_DF("ldunion: forming union composition of \"" << rGen1.Name() 
			<< "\" and \"" << rGen2.Name() << "\"");
	// set up empty generator
	rResGen.Clear();
	Generator tmp = rGen1.NewGen();
	LanguageUnion(rGen1, rGen2, tmp);
	Deterministic(tmp, rResGen);
}

// LanguageIntersection(rGen1, rGen2, rResGen)
void LanguageIntersection(const Generator& rGen1, const Generator& rGen2, 
    Generator& rResGen) {
	FD_DF("lintersection: forming language intersection composition of \"" 
			<< rGen1.Name() << "\" and \"" << rGen2.Name() << "\"");
	// set up empty generator
	rResGen.Clear();
  // fast deterministic version
  if (IsDeterministic(rGen1) && IsDeterministic(rGen2)) {
    // do deterministic shared event parallel composition
    SParallel(rGen1, rGen2, rResGen);
  }
  // much slower nondeterministic version
  else {
    // create temporary copys
    Generator g1(rGen1);
    Generator g2(rGen2);
    EventSet unshared_g1 = g1.NewEventSet();
    EventSet unshared_g2 = g2.NewEventSet();
    // events only in g1
    unshared_g1 = g1.Alphabetr() - g2.Alphabetr();
    // events only in g2
    unshared_g2 = g2.Alphabetr() - g1.Alphabetr();
    EventSet::const_iterator eit;
    // remove transitions with unshared events of g1
    for (eit = unshared_g1.Begin(); eit != unshared_g1.End(); ++eit) {
      g1.DelEvent(*eit);
    }
    // remove transitions with unshared events of g1
    for (eit = unshared_g2.Begin(); eit != unshared_g2.End(); ++eit) {
      g2.DelEvent(*eit);
    }
    // do normal parallel composition
    Parallel(g1, g2, rResGen);
  }
 	rResGen.Name("("+rGen1.Name()+")^("+rGen2.Name()+")");
}

// Automata(rGen)
void Automata(Generator& rGen) {
	FD_DF("automata: comptrans: complete transition relation of \""
			<< rGen.Name() << "\"");
	// assertions
	UniqueInit(rGen);
	assert(IsDeterministic(rGen));
	// prepare result
	rGen.Name(rGen.Name()+" [automata]");
	// introduce dump state
	Idx dump;
	if (rGen.StateNamesEnabled()) {
		dump = rGen.InsState("Dump");
	}
	else {
		dump = rGen.InsState();
	}
	// introduce transtions to dumpstate
	StateSet::iterator lit;
	EventSet::iterator eit;
	for (lit = rGen.StatesBegin(); lit != rGen.StatesEnd(); ++lit) {
		for (eit = rGen.AlphabetBegin(); eit != rGen.AlphabetEnd(); ++eit) {
			if (rGen.TransRelBegin(*lit, *eit) == rGen.TransRelEnd(*lit, *eit)) {
				rGen.SetTransition(*lit, *eit, dump);
			}
		}
	}
}

// LanguageComplement(rGen)
void LanguageComplement(Generator& rGen) {
	FD_DF("automata: lcomplement: language complement of \""
			<< rGen.Name() << "\"");
	// prepare result
	rGen.Name(rGen.Name()+" [complement]");
	// complete transrel
	Automata(rGen);
	// invert marking
	rGen.InjectMarkedStates(rGen.Statesr() - rGen.MarkedStates());
}

// LanguageConcatenate(rGen1, rGen2, rResGen)
void LanguageConcatenate(const Generator& rGen1, const Generator& rGen2, 
		Generator& rResGen) {
	FD_DF("LanguageConcatenate: forming language conposition of \"" 
			<< rGen1.Name() << "\" and \"" << rGen2.Name() << "\"");
	// set up empty generator
	rResGen.Clear();
	// take union of alphabets
	rResGen.InjectAlphabet(rGen1.Alphabet() + rGen2.Alphabet());
	// newindexmap
	std::map<Idx,Idx> newindexmap;
	// helpers
	StateSet::const_iterator lit;
	StateSet::const_iterator lit1m, lit2i; // rGen1 marked states, rGen2 initial states
	TransSet::const_iterator tit;
	std::string name;

	// take union of states
	for (lit = rGen1.StatesBegin(); lit != rGen1.StatesEnd(); ++lit) {
		name = rGen1.StateName(*lit);
		if (name == "") {
			newindexmap[*lit] = rResGen.InsState();
		}
		else {
			newindexmap[*lit] = rResGen.InsState(name+"_1");
		}
	}
	for (lit = rGen2.StatesBegin(); lit != rGen2.StatesEnd(); ++lit) {
		name = rGen2.StateName(*lit);
		if (name == "") {
			newindexmap[*lit] = rResGen.InsState();
		}
		else {
			newindexmap[*lit] = rResGen.InsState(name+"_2");
		}
	}
	// remove rGen2 istates from rResGen.stateset and newindexmap as 
  // rGen2 istates but don't touch the marked ones
	for (lit = rGen2.InitStatesBegin(); lit != rGen2.InitStatesEnd(); ++lit) {
		if (! rGen2.ExistsMarkedState(*lit)) {
			rResGen.DelState(newindexmap[*lit]);
			newindexmap.erase(*lit);
		}
	}
	// set res istates to rGen1 istates
	for (lit = rGen1.InitStatesBegin(); lit != rGen1.InitStatesEnd(); ++lit) {
		rResGen.SetInitState(newindexmap[*lit]);
	}
	// set res mstates to rGen2 mstates
	for (lit = rGen2.MarkedStatesBegin(); lit != rGen2.MarkedStatesEnd(); 
      ++lit) {
		rResGen.SetMarkedState(newindexmap[*lit]);
	}
	// add rGen1 transitions
	for (tit = rGen1.TransRelBegin(); tit != rGen1.TransRelEnd(); ++tit) {
		rResGen.SetTransition(newindexmap[tit->X1], tit->Ev, newindexmap[tit->X2]);
	}
	// add rGen2 transitions starting at initial states.
	// iterate over rGen2 istates
	for (lit2i = rGen2.InitStatesBegin(); 
      lit2i != rGen2.InitStatesEnd(); ++lit2i) {
		// iterate over rGen2 transitions starting at istate
		for (tit = rGen2.TransRelBegin(*lit2i); 
        tit != rGen2.TransRelEnd(*lit2i); ++tit) {
			// iterate over rGen1 mstates
			for (lit1m = rGen1.MarkedStatesBegin(); 
          lit1m != rGen1.MarkedStatesEnd(); ++lit1m) {
				rResGen.SetTransition(newindexmap[*lit1m], tit->Ev, newindexmap[tit->X2]);
			}
		}
	}
	// add rGen2 transitions not starting at initial states
	for (tit = rGen2.TransRelBegin(); tit != rGen2.TransRelEnd(); ++tit) {
		// skip transitions from ummarked istates
		if (rGen2.ExistsInitState(tit->X1) 
        && (! rGen2.ExistsMarkedState(tit->X1))) {
			continue;
		}
		rResGen.SetTransition(newindexmap[tit->X1], tit->Ev, newindexmap[tit->X2]);
	}
}

// FullLanguage(rAlphabet, rResGen)
void FullLanguage(const EventSet& rAlphabet, Generator& rResGen) {
	FD_DF("automata: lfull: construct full language for alphabet \""
			<< rAlphabet.ToString() << "\"");
	// set up empty generator
	rResGen.Clear();
	// helpers
	Idx state;
	EventSet::const_iterator it;
	// alphabet
	rResGen.InjectAlphabet(rAlphabet);
	// stateset + istates
	state = rResGen.InsInitState("AllStates");
	// mstates
	rResGen.SetMarkedState(state);
	// transrel (selfloop)
	for (it = rAlphabet.Begin(); it != rAlphabet.End(); ++it) {
		rResGen.SetTransition(state, *it, state);
	}
}

// EmptyStringLanguage(rAlphabet, rResGen)
void EmptyStringLanguage(const EventSet& rAlphabet, Generator& rResGen) {
	FD_DF("automata: lemptystr: construct empty");
	// set up empty generator
	rResGen.Clear();
	// helpers
	Idx state;
	// rAlphabet
	rResGen.InjectAlphabet(rAlphabet);
	// stateset + istates
	state = rResGen.InsInitState("AllStates");
	// mstates
	rResGen.SetMarkedState(state);
}

// EmptyLanguage(rAlphabet, rResGen)
void EmptyLanguage(const EventSet& rAlphabet, Generator& rResGen) {
	FD_DF("lempty: construct empty");
	// set up empty generator
	rResGen.Clear();
	// rAlphabet
	rResGen.InjectAlphabet(rAlphabet);
}

// EmptyLanguage(rGen)
bool EmptyLanguage(const Generator& rGen) {
	StateSet accessibleset = rGen.AccessibleSet();
	StateSet coaccessibleset = rGen.CoaccessibleSet();
	return (accessibleset*coaccessibleset).Empty();
}

// LanguageInclusion(rGen1, rGen2)
bool LanguageInclusion(const Generator& rGen1, const Generator& rGen2) {
	FD_DF("linclusion: check weather \"" << rGen1.Name() << "\" includes \"" 
			<< rGen2.Name() << "\"");
	Generator rGen1_complement = rGen1.NewGen();
	rGen1_complement = rGen2;
	LanguageComplement(rGen1_complement);
	Generator tmp = rGen1.NewGen();
	Parallel(rGen1, rGen1_complement, tmp);  // L_a - L_b
	return EmptyLanguage(tmp);
}

// LanguageEquality(rGen1, rGen2)
bool LanguageEquality(const Generator& rGen1, const Generator& rGen2) {
	FD_DF("lequality: check for equality of languages of \"" 
			<< rGen1.Name() << "\" and \"" << rGen2.Name() << "\"");
	return (LanguageInclusion(rGen1,rGen2) && LanguageInclusion(rGen2,rGen1));
}

} // namespace faudes


