/* regular.h -- regular language operations */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef REGULAR_H

#include "definitions.h"
#include "debug.h"
#include "parallel.h"
#include "project.h"

namespace faudes {

/**
 * Union composition (result non-deterministic)
 * (textbook version: take union of all entities)
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting Generator of language union
 */
void LanguageUnion(const Generator& rGen1, const Generator& rGen2, 
    Generator& rResGen);

/**
 * Convenience function: Deterministic version of union composition
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to deterministic generator of language union
 */
void DetLanguageUnion(const Generator& rGen1, const Generator& rGen2,
		Generator& rResGen);

/**
 * Intersection composition (based on shared event parallel composition).
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting generator of language intersection
 */
void LanguageIntersection(const Generator& rGen1, const Generator& rGen2, 
		Generator& rResGen);

/**
 * Complete transition relation (i.e. turn generator into an automata).
 * Note: Assume input generator is deterministic
 *
 * @param rGen
 *   Generator
 */
void Automata(Generator& rGen);

/**
 * Language complement (assume generator is deterministic).
 *
 * @param rGen
 *   Generator
 */
void LanguageComplement(Generator& rGen);

/**
 * Concatenate
 * 
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 * @param rResGen
 *   Reference to resulting generator of language concatenation
 */
void LanguageConcatenate(const Generator& rGen1, const Generator& rGen2,
		Generator& rResGen);

/**
 * Compose full language from alphabet
 *
 * @param rAlphabet
 *   Alphabet
 * @param rResGen
 *   Reference to resulting generator with full language
 */
void FullLanguage(const EventSet& rAlphabet, Generator& rResGen);

/**
 * Compose empty string language (holds the empty string)
 *
 * @param rAlphabet
 *   Alphabet
 * @param rResGen
 *   Reference to resulting generator with empty string language
 */
void EmptyStringLanguage(const EventSet& rAlphabet, Generator& rResGen);

/**
 * Compose empty language
 *
 * @param rAlphabet
 *   Alphabet
 * @param rResGen
 *   Reference to resulting generator with empty language
 */
void EmptyLanguage(const EventSet& rAlphabet, Generator& rResGen);

/**
 * Test for empty language
 *
 * @param rGen
 *   Generator
 *
 * @return
 *   true = language empty, false = language not empty
 */
bool EmptyLanguage(const Generator& rGen);

/**
 * Test for language inclusion
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 *
 * @return
 *   True if language of rGen1 includes language of rGen2
 */
bool LanguageInclusion(const Generator& rGen1, const Generator& rGen2);

/**
 * Test for language equality
 * (inefficient. implicitly needs two parallel compositions).
 *
 * @param rGen1
 *   First generator
 * @param rGen2
 *   Second generator
 *
 * @return
 *   True if language of rGen1 is equal to language of rGen2
 */
bool LanguageEquality(const Generator& rGen1, const Generator& rGen2);

} // namespace faudes

#define REGULAR_H
#endif 

