/* stateset.cpp -- container for storing states */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "stateset.h"

namespace faudes {

// StateSet()
StateSet::StateSet(void) {
	FD_DC("StateSet(" << this << ")::StateSet()");
	mpStateSymbolTable = &msStateSymbolTable;
	Name("StateSet");
}

// NewStateSet()
StateSet StateSet::NewStateSet(void) const {
	return StateSet();
}

// NewStateSetp()
StateSet* StateSet::NewStateSetp(void) const {
	return new StateSet();
}

// StateSet::myStateSymbolTable
StateSymbolTable StateSet::msStateSymbolTable;

// StateSymbolTablep()
StateSymbolTable* StateSet::StateSymbolTablep(void) {
	return mpStateSymbolTable;
}

// StateSet::StaticStateSymbolTablep()
StateSymbolTable* StateSet::StaticStateSymbolTablep(void) {
	return &msStateSymbolTable;
}

// Write()
void StateSet::Write(void) const {
	DWrite();
}

// Write(rFileName)
void StateSet::Write(const char* rFileName) const {
	try {
		TokenWriter tw(rFileName, std::ios::out|std::ios::trunc);
		DWrite(tw); 
	}
	catch (std::ios::failure&) {
		std::stringstream errstr;
		errstr << "Exception opening/writing file \"" << rFileName << "\"";
		throw Exception("StateSet::Write", errstr.str(), 2);
	}
}

// Write(rFileName, openmode)
void StateSet::Write(const char* rFileName, std::ios::openmode openmode) const {
	try {
		TokenWriter tw(rFileName, openmode);
		DWrite(tw);
	}
	catch (std::ios::failure&) {
		std::stringstream errstr;
		errstr << "Exception opening/writing file \"" << rFileName << "\"";
		throw Exception("StateSet::Write", errstr.str(), 2);
	}
}

// Write(rFileName, openmode, rMinIndexMap&, sumAnonymousStates)
void StateSet::Write(const char* rFileName, std::ios::openmode openmode, 
		std::map<Idx,Idx>& rMinIndexMap, bool sumAnonymousStates) const {
	try {
		TokenWriter tw(rFileName, openmode);
		Write(tw, rMinIndexMap, sumAnonymousStates);
	}
	catch (std::ios::failure&) {
		std::stringstream errstr;
		errstr << "Exception opening/writing file \"" << rFileName << "\"";
		throw Exception("StateSet::Write", errstr.str(), 2);
	}
}

// Write(rTw&, rMinIndexMap&, sumAnonymousStates)
void StateSet::Write(TokenWriter& rTw, std::map<Idx,Idx>& rMinIndexMap,
		bool sumAnonymousStates) const {
	std::map<Idx,Idx>::const_iterator minit;
	rTw.WriteBegin(mMyName);
	// algorithm:
	// sum anonymous states (stateset)
	if (sumAnonymousStates) {
    // create a minindexmap<minindex,index> to print init states at first
    std::map<Idx,Idx> reversemap;
    for (minit = rMinIndexMap.begin(); 
        minit != rMinIndexMap.end(); ++minit) {
      reversemap[minit->second] = minit->first;
    }
    minit = reversemap.begin();
		while (minit != reversemap.end()) {
			while ((minit != reversemap.end()) && (StateName(minit->second) != "")) {
				rTw << StateName(minit->second);
				++minit;
			}
			Idx anoncount = 0;
			while ((minit != reversemap.end()) && (StateName(minit->second) == "")) {
				++anoncount;
				++minit;
			}
			if (anoncount > 0) {
				rTw << anoncount;
			}
		}
	}
	// Write every state index (istates, mstates)
	else {
    StateSet::const_iterator lit;
    for (lit = Begin(); lit != End(); ++lit) {
			std::string statename = StateName(*lit);
			if (statename != "") {
				rTw << statename;
			}
			else {
      	minit = rMinIndexMap.find(*lit);
	      if (minit == rMinIndexMap.end()) {
  	      std::cerr << "programming error: index " << *lit 
    	        << " could not be found in rMinIndexMap" << std::endl;
      	  abort();
      	}
				rTw << minit->second;
			}
		}
	}
	rTw.WriteEnd(mMyName);
}

// DWrite()
void StateSet::DWrite(void) const {
	TokenWriter tw;
	DWrite(tw);
}

// DWrite(rFileName, openmode)
void StateSet::DWrite(const char* rFileName, std::ios::openmode openmode) const {
	try {
		TokenWriter tw(rFileName, openmode);
		DWrite(tw);
	}
	catch (std::ios::failure&) {
		std::stringstream errstr;
		errstr << "Exception opening/writing file \"" << rFileName << "\"";
		throw Exception("StateSet::DWrite", errstr.str(), 2);
	}
}

// DWrite(rTw&)
void StateSet::DWrite(TokenWriter& rTw) const {
	const_iterator it;
	rTw.WriteBegin(mMyName);
	for (it = Begin(); it != End(); ++it) {
		if (StateName(*it) == "") {
			rTw << *it;
		}
		else {
			rTw << StateName(*it) + "(" + faudes::ToString(*it) + ")";
		}
	}
	rTw.WriteEnd(mMyName);
}

// ToString()
std::string StateSet::ToString(void) const {
	std::string ret;
	std::ostringstream ostr;
	const_iterator it;
	for (it = Begin(); it != End(); ++it) {
		ostr << Str(*it) << " ";
	}
	ret = ostr.str();
	if (ret.size() > 0) ret.erase(ret.size()-1);
	return ret;
}

// InsState()
Idx StateSet::InsState(void) {
	FD_DC("StateSet(" << this << ")::InsState()");
	Idx index;
	index = mpStateSymbolTable->InsState();
	mIndexSet.insert(index);
	return index;
}

// InsState(rName)
Idx StateSet::InsState(const std::string& rName) {
	FD_DC("StateSet(" << this << ")::InsState(\"" << rName << "\")");
	Idx index = StateIndex(rName);
	if (index == 0) {
		index = mpStateSymbolTable->InsState(rName);
		mIndexSet.insert(index);
	}
	return index;
}

// DelState(index)
void StateSet::DelState(Idx index) {
	FD_DC("StateSet(" << this << ")::DelState(" << index << ")");
	mpStateSymbolTable->DelState(index);
	if (mIndexSet.erase(index) == 0) {
		std::stringstream errstr;
		errstr << "index " << index << " not found in StateSet" << std::endl;
		Exception("StateSet::DelState", errstr.str(), 201);
	}
}

// DelState(rName)
void StateSet::DelState(const std::string& rName) {
	FD_DC("StateSet(" << this << ")::DelState(\"" << rName << "\")");
	iterator it;
	for (it = Begin(); it != End(); ++it) {
		if (mpStateSymbolTable->StateName(*it) == rName) {
			return DelState(*it);
		}
	}
	// if not found throw exception
	std::stringstream errstr;
	errstr << "statename \"" << rName << "\" not found in StateSet" << std::endl;
	throw Exception("StateSet::DelState", errstr.str(), 203);
}

// DelState(pos)
void StateSet::DelState(iterator pos) {
	FD_DC("StateSet(" << this << ")::DelState(" << *pos << ")");
	mpStateSymbolTable->DelState(*pos);
	mIndexSet.erase(pos);
}

// StateName(index, rName)
void StateSet::StateName(Idx index, const std::string& rName) {
	FD_DC("StateSet(" << this << ")::StateName(" 
			<< index << ",\"" << rName << "\")");
#ifdef FAUDES_CHECKED
	if (! ExistsState(index)) {
		FD_ERR("StateSet::StateName(" << index << "," << rName << "): "
				<< "index does not exist in this EventSet. aborting...");
		abort();
	}
#endif
	mpStateSymbolTable->StateName(index, rName);
}

// ClrStateName(index)
void StateSet::ClrStateName(Idx index) {
	FD_DC("StateSet(" << this << ")::ClrStateName(" << index << ")");
#ifdef FAUDES_CHECKED
	if (! ExistsState(index)) {
		FD_ERR("StateSet::StateName(" << index << "): "
				<< "index does not exist in this EventSet. aborting...");
		abort();
	}
#endif
	mpStateSymbolTable->ClrStateName(index);
}

// ClrStateName(rName)
void StateSet::ClrStateName(const std::string& rName) {
	FD_DC("StateSet(" << this << ")::ClrStateName(\"" << rName << "\")");
	iterator it;
	for (it = Begin(); it != End(); ++it) {
		if (mpStateSymbolTable->StateName(*it) == rName) {
			ClrStateName(*it);
			return;
		}
	}
#ifdef FAUDES_CHECKED
	// if state not found
	FD_ERR("StateSet::ClrStateName(" << rName << "): "
			<< "state name does not exist for an index in this EventSet. aborting...");
	abort();
#endif
}

// Insert(index)
void StateSet::Insert(Idx index) {
	FD_DC("StateSet(" << this << ")::Insert(" << index << ")");
	mIndexSet.insert(index);
}

// Insert(set)
void StateSet::Insert(const StateSet& rOtherSet) {
	FD_DC("StateSet(" << this << ")::Insert(" << rOtherSet.ToString() << ")");
	iterator it1 = Begin();
	const_iterator it2 = rOtherSet.Begin();
	while ((it1 != End()) && (it2 != rOtherSet.End())) {
		if (*it1 < *it2) {
			++it1;
		}
		else if (*it1 == *it2) {
			++it1;
			++it2;
		}
		else { // (*it1 > *it2)
			mIndexSet.insert(*it2);
			++it2;
		}
	}
	while (it2 != rOtherSet.End()) {
		mIndexSet.insert(End(), *it2);
		++it2;
	}
}
		
// Erase(index)
void StateSet::Erase(Idx index) {
	FD_DC("StateSet(" << this << ")::Erase(" << index << ")");
	mIndexSet.erase(index);
}

// Erase(pos)
void StateSet::Erase(iterator pos) {
	FD_DC("StateSet(" << this << ")::Erase(" << *pos << ")");
	mIndexSet.erase(pos);
}

// Clear()
void StateSet::Clear(void) {
	FD_DC("StateSet(" << this << ")::Clear()");
	mIndexSet.clear();
}

// StateName(index)
std::string StateSet::StateName(Idx index) const {
	return mpStateSymbolTable->StateName(index);
}

//StateIndex(rName)
Idx StateSet::StateIndex(const std::string& rName) const {
	const_iterator it;
	for (it = Begin(); it != End(); ++it) {
		if (mpStateSymbolTable->StateName(*it) == rName) {
			return *it;
		}
	}
	return 0;
}

// ExistsState(index)
bool StateSet::ExistsState(Idx index) const {
	if (mIndexSet.find(index) != mIndexSet.end()) {
		return true;
	}
	return false;
}

// Find(index)
StateSet::iterator StateSet::Find(Idx index) {
	return mIndexSet.find(index);
}

// Find(index) const
StateSet::const_iterator StateSet::Find(Idx index) const {
	return mIndexSet.find(index);
}


// ExistsState(rName)
bool StateSet::ExistsState(const std::string& rName) const {
	StateSet::const_iterator it;
	for (it = mIndexSet.begin(); it != mIndexSet.end(); ++it) {
		if (mpStateSymbolTable->StateName(*it) == rName) {
			return true;
		}
	}
	return false;
}

// includes(rOtherSet)
bool StateSet::Includes(const StateSet& rOtherSet) const {
	// use stl algorithm here
	return std::includes(Begin(), End(), 
			rOtherSet.Begin(), rOtherSet.End());
}

// Find(rName)
StateSet::iterator StateSet::Find(const std::string& rName) {
	StateSet::iterator it;
	for (it = mIndexSet.begin(); it != mIndexSet.end(); ++it) {
		if (mpStateSymbolTable->StateName(*it) == rName) {
			return it;
		}
	}
	return mIndexSet.end();
}

// Find(rName) const
StateSet::const_iterator StateSet::Find(const std::string& rName) const {
	StateSet::const_iterator it;
	for (it = mIndexSet.begin(); it != mIndexSet.end(); ++it) {
		if (mpStateSymbolTable->StateName(*it) == rName) {
			return it;
		}
	}
	return mIndexSet.end();
}


// operator +
StateSet StateSet::operator + (const StateSet& rOtherSet) const {
	FD_DC("StateSet(" << this << ")::operator + (" << &rOtherSet << ")");
	StateSet res = StateSet();
	std::insert_iterator< std::set<Idx> > insit(res.mIndexSet,
			res.mIndexSet.begin());
	std::set_union(Begin(), End(), 
			rOtherSet.Begin(), rOtherSet.End(), insit);
	res.Name(Name());
	return res;
}

// operator -
StateSet StateSet::operator - (const StateSet& rOtherSet) const {
	FD_DC("StateSet(" << this << ")::operator - (" << &rOtherSet << ")");
	StateSet res = StateSet();
	std::insert_iterator< std::set<Idx> > insit(res.mIndexSet,
			res.mIndexSet.begin());
	std::set_difference(Begin(), End(), 
			rOtherSet.Begin(), rOtherSet.End(), insit);
	res.Name(Name());
	return res;
}

// operator *
StateSet StateSet::operator * (const StateSet& rOtherSet) const {
	FD_DC("StateSet(" << this << ")::operator * (" << &rOtherSet << ")");
	StateSet res = StateSet();
	std::insert_iterator< std::set<Idx> > insit(res.mIndexSet,
			res.mIndexSet.begin());
	std::set_intersection(Begin(), End(),
			rOtherSet.Begin(), rOtherSet.End(), insit);
	res.Name(Name());
	return res;
}

// ReindexedCopy(pOtherStateSymbolTable*, rReindexMap) 
StateSet StateSet::ReindexedCopy(std::map<Idx,Idx>& rReindexMap) const {
	FD_DC("StateSet::reindex_copy( rReindexMap");
	// PREPARE RESULT:
	StateSet resultset;
	// HELPERS:
	StateSet::const_iterator it;
	// ALGORITHM:
	// if rReindexMap empty new indices have to be created first
	if (rReindexMap.empty()) {
		std::string name;
		Idx newindex;
		for (it = Begin(); it != End(); ++it) {
			name = StateName(*it);
			if (name == "") {
				newindex = resultset.InsState();
			}
			else {
				newindex = resultset.InsState(name);
			}
			rReindexMap[*it] = newindex;
		}
	}
	// else assign already created new indices
	else {
		std::map<Idx,Idx>::const_iterator mit;
		for (it = Begin(); it != End(); ++it) {
			mit = rReindexMap.find(*it);
			if (mit == rReindexMap.end()) {
				FD_ERR("key " << *it << " not in rReindexMap");
				abort();
			}
			resultset.Insert(mit->second);
		}
	}
	resultset.Name(Name());
	return resultset;
} 

// Str(index)
std::string StateSet::Str(Idx index) const {
	std::string name = mpStateSymbolTable->StateName(index);
	if (name == "") {
		return faudes::ToString(index)+"["+faudes::ToString(index)+"]";
	}
	else {
		return name+"["+faudes::ToString(index)+"]";
	}
}

// MinIdxStr(rMinIndexMap, index)
std::string StateSet::MinIdxStr(const std::map<Idx,Idx>& rMinIndexMap, 
		Idx index) const {
	std::string name = mpStateSymbolTable->StateName(index);
	std::map<Idx,Idx>::const_iterator it = rMinIndexMap.find(index);
	// if index got no name
	if (name == "") {
		if (it == rMinIndexMap.end()) {
			FD_ERR("StateSet::str(<rMinIndexMap>, " << index 
					<< "): index not found in rMinIndexMap");
			abort();
		}
		name = faudes::ToString(it->second);
	}
	return name;
}

} // namespace faudes
