/* stateset.h -- container for storing states */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef LSET_H

#include "definitions.h"
#include "debug.h"
#include "exception.h"
#include "baseset.h"
#include "symboltable.h"
#include "tokenreader.h"
#include "tokenwriter.h"
#include <fstream>
#include <sstream>
#include <iostream>
#include <set>
#include <map>
#include <assert.h>
#include <algorithm>

namespace faudes {

/**
 * Container for states. Can be used both for stateset and istates/mstates.
 * Methods for global unique states are:
 *   - InsState
 *   - DelState
 * STL-like Methods for local unique states are: 
 *   - Insert
 *   - Delete 
 */
class StateSet : public BaseSet {
	public:
		/**
		 * Constructor for StateSet with static StateSymbolTable. 
		 * The class behaviour is specified by usage of first constructor!
		 */
		StateSet(void);

		/**
		 * Return a StateSet with StateSymbolTable assigned on stack.
		 *
		 * @return
		 *   New Lset on stack
		 */
		StateSet NewStateSet() const;

		/**
		 * Return a pointer to a StateSet with StateSymbolTable assigned on heap.
		 * 
		 * @return
		 *   Pointer to new empty StateSet on heap
		 */
		StateSet* NewStateSetp() const;

		/**
		 * Get pointer to StateSymbolTable.
		 *
		 * @return
		 *   Pointer to mpStateSymbolTable
		 */
		StateSymbolTable* StateSymbolTablep(void);

		/**
		 * Get pointer to static StateSymbolTable.
		 *
		 * @return
		 *   Pointer to static StateSymbolTable
		 */
		static StateSymbolTable* StaticStateSymbolTablep(void);

		/**
		 * Write StateSet to console as stateset without creating minindexmap.
		 * (Just calls DWrite().)
		 */
		void Write(void) const;

		/**
		 * Write StateSet to a newly created file as stateset without creating 
		 * minindexmap.
		 * (Just calls DWrite(pFileName).)
		 *
		 * @param pFileName
		 *   Filename
		 *
		 * @exception Exception
		 *   If opening/writing fails an Exception object is thrown (id 2)
		 */
		void Write(const char* pFileName) const;

		/**
		 * Write StateSet to a file as stateset with openmode specified. No minindexmap 
		 * is created. Useful for writing logfiles.
		 * (Just calls DWrite(pFileName, openmode).)
		 *
		 * @param pFileName
		 *   Filename
		 * @param openmode
		 *   ios::openmode
		 *
		 * @exception Exception
		 *   If opening/writing fails an Exception object is thrown (id 2)
		 */
		void Write(const char* pFileName, std::ios::openmode openmode) const;

		/**
		 * Write StateSet to a file with openmode specified. See Write(tw, minindexmap) 
		 * for details about minindexmap behaviour and what
		 * sumAnonymousStates means
		 *
		 * @param pFileName
		 *   Filename
		 * @param openmode
		 *   ios::openmode
		 * @param rMinIndexMap
		 *   Reference to minimal index map<Idx,Idx>
		 * @param sumAnonymousStates
		 *   For States this should be true, for InitStates and MarkedStates
		 *   it must be false (prints only number of following anonymous states).
		 *
		 * @exception Exception
		 *   If opening/writing fails an Exception object is thrown (id 2)
		 */
		void Write(const char* pFileName, std::ios::openmode openmode, 
				std::map<Idx,Idx>& rMinIndexMap, bool sumAnonymousStates) const;

		/** 
		 * Write StateSet to TokenWriter. If sumAnonymousStates is true empty the StateSet 
		 * is written as a stateset. Else the StateSet is written as set of 
		 * istates/mstates. The name of the set becomes the token label.
		 *
		 * @param tw
		 *   Reference to TokenWriter
		 * @param rMinIndexMap
		 *   Reference to map of indices. The map is used for transforming 
		 *   the global unique indices into minimal indices (1, 2, 3, ...).
		 * @param sumAnonymousStates
		 *   For States this should be true, for InitStates and MarkedStates
		 *   it must be false (prints only number of following anonymous states).
		 *
		 * @exception ios::failure
		 *   Thrown on i/o error.
		 */
		void Write(TokenWriter& tw, std::map<Idx,Idx>& rMinIndexMap,
				bool sumAnonymousStates) const;

		/** 
		 * Write StateSet to console as it's stored in memory. 
		 * Usefull for debugging purposes.
		 */
		void DWrite(void) const;

		/**
		 * Write StateSet to pFileName as it's stored in memory. Specified openmode
		 * will be used for opening the file. Usefull for debugging / logging
		 * purposes.
		 *
		 * @param pFileName
		 *   Filename to write
		 * @param openmode
		 *   ios::openmode
		 *
		 * @exception Exception
		 *   If opening/writing fails an Exception object is thrown (id 2)
		 */
		void DWrite(const char* pFileName, std::ios::openmode openmode) const;

		/**
		 * Write StateSet to TokenWriter as it's stored in memory.
		 *
		 * @param tw
		 *   Reference to TokenWriter
		 *
		 * @exception ios::failure
		 *   Thrown on i/o error.
		 */
		void DWrite(TokenWriter& tw) const;

		/**
		 * Write all states into a string
		 *
		 * @return
		 *   String containing all states
		 */
		std::string ToString(void) const;

		/**
		 * Add new global state to StateSet
		 *
		 * @return
		 *   Index of new global unique state
		 */
		Idx InsState(void);

		/**
		 * Add new global state to StateSet
		 *
		 * @param rName
		 *   Name of the state to add
		 * 
		 * @return
		 *   Index of new global unique state
		 */
		Idx InsState(const std::string& rName);

		/**
		 * Delete a state from StateSet and global StateSymbolTable by index.
		 *
		 * @param index
		 *   Index to delete
		 */
		void DelState(Idx index);


		/**
		 * Delete a state from StateSet and global StateSymbolTable by name.
		 * (Expensive! Don't use!)
		 *
		 * @param rName
		 *   Named state to delete
		 */
		void DelState(const std::string& rName);

		/**
		 * Delete a state From StateSet and global statespace by iterator.
		 *
		 * @param pos
		 *   StateSet::iterator 
		 */
		void DelState(iterator pos);

		/** 
		 * Set name for existing state
		 * FAUDES_CHECKED turns on local existence check and aborts program if
		 * state not in StateSet
		 *
		 * @param index
		 *   Index
		 * @param rName
		 *   Name
		 *
		 * @exception Exception
		 *   Name already associated with another index (id 104)
		 */
		void StateName(Idx index, const std::string& rName);

		/**
		 * Clear name for existing state
		 * FAUDES_CHECKED turns on local existence check and aborts program if
		 * state not in StateSet
		 *
		 * @param index
		 *   State index
		 */
		void ClrStateName(Idx index);

		/**
		 * Clear name for existing state
		 * (Expensive! Don't use for performance critical functions!)
		 *
		 * @param rName
		 *   State name
		 */
		void ClrStateName(const std::string& rName);
		
		/** 
		 * Add a new index to mIndexSet
		 * FAUDES_CHECKED turns on StateSymbolTable existence check.
		 *
		 * @param index
		 *   Index to add
		 */
		 void Insert(Idx index);

		/**
		 * Union with another StateSet. Inserts Indices of rOtherSet,  not contained
		 * in this set.
		 * This is faster than inserting single indices when indices are 
		 * already stored in a StateSet.
		 *
		 * @param rOtherSet
		 *   Other StateSet
		 */
		void Insert(const StateSet& rOtherSet);

		/**
		 * Delete an index from mIndexSet. Global indextable stays untouched.
		 *
		 * @param index
		 *    Index to delete
		 */
		void Erase(Idx index);

		/**
		 * Delete an index from mIndexSet. Global indextable stays untouched.
		 * Index is given by iterator.
		 *
		 * @param pos
		 *   StateSet::iterator
		 */
		void Erase(iterator pos);

		/** 
		 * Clear all stored indices
		 */
		void Clear(void);

		/** 
		 * Name lookup
		 *
		 * @param index
		 *   State index
		 * 
		 * @return
		 *   Returns corresponding name or empty string
		 */
		std::string StateName(Idx index) const;

		/**
		 * Index lookup. Returns first index matching name. Only if names are unique 
		 * within this lset the result is valid. This method should only be used in 
		 * special cases, where performance doesn't matter!
		 *
		 * @param rName
		 *   State name
		 * 
		 * @returns
		 *   Corresponding index or 0
		 */
		Idx StateIndex(const std::string& rName) const;

		/**
		 * Test existence of index
		 *
		 * @param index
		 *   State index
		 *
		 * @return 
		 *   Success
		 */
		bool ExistsState(Idx index) const;

		/**
		 * Test existence of name (expensive! use only interactive!)
		 *
		 * @param rName
		 *   State name
		 *
		 * @return
		 *   Success
		 */
		bool ExistsState(const std::string& rName) const;

		/**
		 * Subset test
		 * The includes method returns true if every index in rOtherSet is also 
		 * contained in this set.
		 *
		 * @param rOtherSet
		 *    Other StateSet
		 *
		 * @return
		 *   True if rOtherSet is a subset
		 */
		bool Includes(const StateSet& rOtherSet) const;

		/**
		 * Find index. Returns an iterator.
		 *
		 * @param index
		 *   State index
		 *
		 * @return
		 *   iterator to element or mIndexSet.end()
		 */
		StateSet::iterator Find(Idx index);

		/**
		 * Find index. Returns a const_iterator
		 *
		 * @param index
		 *   State index
		 *
		 * @return
		 *   const_iterator to element or mIndexSet.end()
		 */
		StateSet::const_iterator Find(Idx index) const;

		/**
		 * Return iterator to index of name. If not found returns end().
		 * Expensive! Use only for user interaction!
		 *
		 * @param rName
		 *   State name
		 * 
		 * @return 
		 *   StateSet::iterator
		 */
		StateSet::iterator Find(const std::string& rName);

		/**
		 * Return const_iterator to index of name. If not found returns end().
		 * Expensive! Use only for user interaction!
		 *
		 * @param rName
		 *   State name
		 * 
		 * @return 
		 *   StateSet::const_iterator
		 */
		StateSet::const_iterator Find(const std::string& rName) const;

		/**
		 * Set union with another StateSet
		 *
		 * @return 
		 *   Union StateSet
		 */
		StateSet operator + (const StateSet& rOtherSet) const;

		/**
		 * Set difference with another StateSet
		 *
		 * @return 
		 *   Difference StateSet
		 */
		StateSet operator - (const StateSet& rOtherSet) const;

		/**
		 * Set intersection with another StateSet
		 *
		 * @return
		 *   Intersection StateSet
		 */
		StateSet operator * (const StateSet& rOtherSet) const;
		
				/**
		 * Copy an StateSet by creating new indices for the resulting set,
		 * corresponding to a given StateSymbolTable. 
		 *
		 * @param rReindexMap
		 *   Result map<Idx,Idx> of corresponding states in original and copy.
		 *   If the map is empty new indices are created during copying. In the 
		 *   other case the mapped indices are inserted in the new set.
		 *
		 * @return
		 *   New StateSet with unique global indices
		 */
		StateSet ReindexedCopy(std::map<Idx,Idx>& rReindexMap) const; 
		
		/**
		 * Return pretty printable state name for index of form 
		 *   "name[index]"
		 * where name is index when state is anonymous. Primary meant for debugging 
		 * messages
		 *
		 * @param index
		 *   State index
		 *
		 * @return
		 *   String
		 */
		std::string Str(Idx index) const;

		/**
		 * Return pretty printable state name for index of form
		 *   "name"
		 * where name is state name or minimal index of state (starting with 1).
		 *
		 * @param rMinIndexMap
		 *   map<Idx,Idx> mapping global states to minimal indices
		 * @param index
		 *   Global state index
		 * 
		 * @return
		 *   String
		 */
		std::string MinIdxStr(const std::map<Idx,Idx>& rMinIndexMap, 
				Idx index) const;

	private:

		/** Pointer to local StateSymbolTable */
		StateSymbolTable* mpStateSymbolTable;
		
		/** Pointer to static global StateSymbolTable */
		static StateSymbolTable msStateSymbolTable;
};

}

#define LSET_H
#endif 

