/* supcon.cpp -- supremal controllable sublanguage and controllablity */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */
   

#include "supcon.h"

namespace faudes {

/*
******************************
* SUPCON: USER FUNCTIONS     *
******************************
*/


// SupConNB(rPlantGen, rSpecGen, rResGen)
void SupConNB(const cGenerator& rPlantGen, const Generator& rSpecGen, 
    cGenerator& rResGen) {

	// HELPERS:
	std::map< std::pair<Idx,Idx>, Idx> rcmap;

	// ALGORITHM:
	SupConNB(rPlantGen, rSpecGen, rcmap, rResGen);
	if (rPlantGen.StateNamesEnabled() && rSpecGen.StateNamesEnabled()) {
		CompositionMap cmap = CompositionMap();
		cmap.InsReverseMap(rcmap);
		cmap.CompStateNames(rResGen);
	}
}


// SupConNB(rPlantGen, rSpecGen, rReverseCompositionMap, rResGen)
void SupConNB(const cGenerator& rPlantGen, const Generator& rSpecGen,
		std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, cGenerator& rResGen) {
	FD_DF("SupConNB(" << &rPlantGen << "," << &rSpecGen << ")");

	// PREPARE RESULT:
	
	rResGen.Clear();
	rResGen.Name("SupConNB(("+rPlantGen.Name()+"),("+rSpecGen.Name()+"))");
	rResGen.InjectAlphabet(rPlantGen.Alphabet());
	rResGen.SetEventFlagTable(rPlantGen.EventFlagTable() ); // notnice

	// HELPERS:

	// controllable events
	const EventSet ualph = rPlantGen.UncontrollableEvents();
	FD_DF("SupConNB: controllable events: " << rPlantGen.ControllableEvents().ToString());
	FD_DF("SupConNB: uncontrollable events: " << ualph.ToString());

	// CONSISTENCY CHECK:

	// alphabets must match
	if (rPlantGen.Alphabet() != rSpecGen.Alphabet()) {
		EventSet only_in_plant = rPlantGen.Alphabet() - rSpecGen.Alphabet();
		EventSet only_in_spec = rSpecGen.Alphabet() - rPlantGen.Alphabet();
		std::stringstream errstr;
		errstr << "Alphabets of generators do not match. Only in plant: " 
				<< only_in_plant.ToString() << ". Only in spec: "
				<< only_in_spec.ToString() << ".";
		throw Exception("SupConNB", errstr.str(), 500);
	}
	
	// plant and spec must be deterministic
	bool plant_det = IsDeterministic(rPlantGen);
	bool spec_det = IsDeterministic(rSpecGen);

	if ((plant_det == false) && (spec_det == true)) {
		std::stringstream errstr;
		errstr << "Plant generator must be deterministic, "
				<< "but is nondeterministic";
		throw Exception("SupConNB", errstr.str(), 501);
	}
	else if ((plant_det == true) && (spec_det == false)) {
		std::stringstream errstr;
		errstr << "Spec generator must be deterministic, "
				<< "but is nondeterministic";
		throw Exception("SupConNB", errstr.str(), 503);
	}
	else if ((plant_det == false) && (spec_det == false)) {
		std::stringstream errstr;
		errstr << "Plant and spec generator must be deterministic, "
				<< "but both are nondeterministic";
		throw Exception("SupConNB", errstr.str(), 504);
	}

	// ALGORITHM:
	SupconParallel(rPlantGen, rSpecGen, ualph, rReverseCompositionMap, rResGen);
	// make resulting generator trim until it's fully controllable
	while (1) {
		if (rResGen.Empty()) {
			return;
		}
		Idx state_num = rResGen.Size();
		bool supcon = SupConUnchecked(rPlantGen, rResGen);
		bool trim = rResGen.Trim();
		if (supcon && trim && (rResGen.Size() == state_num)) {
			return;
		}
	}
}


// SupConNBNonDet(rPlantGen, rSpecGen, rReverseCompositionMap, 
//     pDetPlantGen, pDetSpecGen, rResGen)
void SupConNBNonDet(const cGenerator& rPlantGen, const Generator& rSpecGen,
		std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap,
		cGenerator rDetPlantGen, cGenerator rDetSpecGen, 
    std::map<Idx,StateSet>& rPowerStatesPlant, 
    std::map<Idx,StateSet>& rPowerStatesSpec,
		bool& rDetPlantBool, bool& rDetSpecBool, cGenerator& rResGen) {
	FD_DF("SupConNBNonDet(" << &rPlantGen << "," << &rSpecGen << ")");

	// PREPARE RESULT:

	rResGen.Clear();
	rResGen.Name("SupConNBNonDet(("+rPlantGen.Name()+"),("+rSpecGen.Name()+"))");
	rResGen.InjectAlphabet(rPlantGen.Alphabet());
	rResGen.SetEventFlagTable(rPlantGen.EventFlagTable() ); // notnice
	
	// HELPERS:
  rPowerStatesPlant.clear();
  rPowerStatesSpec.clear();

	// controllable events
	const EventSet ualph = rPlantGen.UncontrollableEvents();
	FD_DF("SupConNB: controllable events: " << rPlantGen.ControllableEvents().ToString());
	FD_DF("SupConNB: uncontrollable events: " << ualph.ToString());

	// pointers to actual generators
	cGenerator* plantp;
	Generator* specp;


	// CONSISTENCY CHECK:

	// alphabets must match
	if (rPlantGen.Alphabet() != rSpecGen.Alphabet()) {
		throw Exception("SupConNBNonDet", "Alphabets of generators don't match", 500);
	}

	// plant and spec determinism
	rDetPlantBool = IsDeterministic(rPlantGen);
	rDetSpecBool = IsDeterministic(rSpecGen);
	
	// create deterministic generators and set pointers to the actual ones
	if ((rDetPlantBool == false) && (rDetSpecBool == true)) {
		Deterministic(rPlantGen, rPowerStatesPlant, rDetPlantGen);		
		plantp = &rDetPlantGen;
		specp = const_cast<Generator*>(&rSpecGen);
	}
	else if ((rDetPlantBool == true) && (rDetSpecBool == false)) {
		Deterministic(rSpecGen, rPowerStatesSpec, rDetSpecGen);
		plantp = const_cast<cGenerator*>(&rPlantGen);
		specp = &rDetSpecGen;
	}
	else if ((rDetPlantBool == false) && (rDetSpecBool == false)) {
		Deterministic(rPlantGen, rPowerStatesPlant, rDetPlantGen);		
		Deterministic(rSpecGen, rPowerStatesSpec, rDetSpecGen);
		plantp = &rDetPlantGen;
		specp = &rDetSpecGen;
	}
	else {
		plantp = const_cast<cGenerator*>(&rPlantGen);
		specp = const_cast<Generator*>(&rSpecGen);
	}

	// ALGORITHM:

	// compute supervisor
	SupconParallel(*plantp, *specp, ualph, rReverseCompositionMap, rResGen);

	// make resulting generator trim until it's fully controllable
	while (! (SupConUnchecked(*plantp, rResGen) && rResGen.Trim()));
}

// SupCon(rPlantGen, rSpecGen, rResGen)
void SupCon(const cGenerator& rPlantGen, const Generator& rSpecGen, 
    cGenerator& rResGen) {

	// HELPERS:
	std::map< std::pair<Idx,Idx>, Idx> rcmap;

	// ALGORITHM:
	SupCon(rPlantGen, rSpecGen, rcmap, rResGen);
}


// SupCon(rPlantGen, rSpecGen, rReverseCompositionMap, rResGen)
void SupCon(const cGenerator& rPlantGen, const Generator& rSpecGen,
		std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, cGenerator& rResGen) {
	FD_DF("SupCon(" << &rPlantGen << "," << &rSpecGen << ")");

	// PREPARE RESULT:
	
	rResGen.Clear();
	rResGen.Name("SupCon(("+rPlantGen.Name()+"),("+rSpecGen.Name()+"))");
	rResGen.InjectAlphabet(rPlantGen.Alphabet());
	rResGen.SetEventFlagTable(rPlantGen.EventFlagTable() ); // notnice
	
	// HELPERS:

	// controllable events
	const EventSet ualph = rPlantGen.UncontrollableEvents();
	FD_DF("SupCon: controllable events: " << rPlantGen.ControllableEvents().ToString());
	FD_DF("SupCon: uncontrollable events: " << ualph.ToString());

	// CONSISTENCY CHECK:

	// alphabets must match
	if (rPlantGen.Alphabet() != rSpecGen.Alphabet()) {
		EventSet only_in_plant = rPlantGen.Alphabet() - rSpecGen.Alphabet();
		EventSet only_in_spec = rSpecGen.Alphabet() - rPlantGen.Alphabet();
		std::stringstream errstr;
		errstr << "Alphabets of generators do not match. Only in plant: " 
				<< only_in_plant.ToString() << ". Only in spec: "
				<< only_in_spec.ToString() << ".";
		throw Exception("SupCon", errstr.str(), 500);
	}

	// plant and spec must be deterministic
	bool plant_det = IsDeterministic(rPlantGen);
	bool spec_det = IsDeterministic(rSpecGen);

	if ((plant_det == false) && (spec_det == true)) {
		std::stringstream errstr;
		errstr << "Plant generator must be deterministic, "
				<< "but is nondeterministic";
		throw Exception("SupCon", errstr.str(), 501);
	}
	else if ((plant_det == true) && (spec_det == false)) {
		std::stringstream errstr;
		errstr << "Spec generator must be deterministic, "
				<< "but is nondeterministic";
		throw Exception("SupCon", errstr.str(), 503);
	}
	else if ((plant_det == false) && (spec_det == false)) {
		std::stringstream errstr;
		errstr << "Plant and spec generator must be deterministic, "
				<< "but both are nondeterministic";
		throw Exception("SupConNB", errstr.str(), 504);
	}

	// ALGORITHM:

	// parallel composition
	SupconParallel(rPlantGen, rSpecGen, ualph, rReverseCompositionMap, rResGen);
	// make resulting generator controllable
	SupConUnchecked(rPlantGen, rResGen);
}



/*
********************************************
* Fast parallel composition for SupCon     *
********************************************
*/


// SupconParallel(rPlantGen, rSpecGen, rUAlph, rReverseCompositionMap, rResGen)
void SupconParallel(const cGenerator& rPlantGen, const Generator& rSpecGen,
		const EventSet& rUAlph, 
		std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
		cGenerator& rResGen) {
	FD_DF("SupconParallel(" << &rPlantGen << "," << &rSpecGen << ")");
	
	// HELPERS:

	// todo stack
	std::stack< std::pair<Idx,Idx> > todo;
	// set of forbidden states
	StateSet forbidden = rPlantGen.NewStateSet();
	// actual pair, new pair
	std::pair<Idx,Idx> currentstates, newstates;
	// other stuff
	Idx tmpstate;
	std::map< std::pair<Idx,Idx>, Idx>::iterator rcmapit;
	StateSet::const_iterator lit1, lit2;
	TransSet::const_iterator titg, titg_end, tith, tith_end;

	// ALGORITHM:
  
  if (rPlantGen.InitStatesEmpty()) {
    FD_DF("SupconParallel: plant got no initial states. "
        << "parallel composition contains empty language.");
    return;
  }

  if (rSpecGen.InitStatesEmpty()) {
    FD_DF("SupconParallel: spec got no initial states. "
        << "parallel composition contains empty language.");
    return;
  }

	// create initial state
	currentstates = std::make_pair(*rPlantGen.InitStatesBegin(), *rSpecGen.InitStatesBegin());
	todo.push(currentstates);
	if (rPlantGen.ExistsMarkedState(*rPlantGen.InitStatesBegin()) 
				&& rSpecGen.ExistsMarkedState(*rSpecGen.InitStatesBegin())) {
		rReverseCompositionMap[currentstates] = rResGen.InsInitState();
		rResGen.SetMarkedState(rReverseCompositionMap[currentstates]);
		FD_DF("SupconParallel: NEW IMSTATE: (" << rPlantGen.SStr(currentstates.first)
				<< "|" << rSpecGen.SStr(currentstates.second) << ") -> "	<< rReverseCompositionMap[currentstates]);
	}
	else {
		rReverseCompositionMap[currentstates] = rResGen.InsInitState();
		FD_DF("SupconParallel: NEW ISTATE: (" << rPlantGen.SStr(currentstates.first) 
				<< "|" << rSpecGen.SStr(currentstates.second) << ") -> " << rReverseCompositionMap[currentstates]);
	}

	// first do parallel composition of allowed states
	// by deleting forbidden states on the fly.
	// this creates an accessible new generator
	FD_DF("SupconParallel: *** processing reachable states ***");
	while (! todo.empty()) {
		// get next reachable pair of states from todo stack
		currentstates = todo.top();
		todo.pop();
		FD_DF("SupconParallel: todo pop: (" << rPlantGen.SStr(currentstates.first) 
				<< "|" << rSpecGen.SStr(currentstates.second) << ") -> " << rReverseCompositionMap[currentstates]);
		
		titg = rPlantGen.TransRelBegin(currentstates.first);
		titg_end = rPlantGen.TransRelEnd(currentstates.first);
		tith = rSpecGen.TransRelBegin(currentstates.second);
		tith_end = rSpecGen.TransRelEnd(currentstates.second);

#ifdef FAUDES_DEBUG_FUNCTION
		// print all transitions of current states
		FD_DF("SupConParallel: transitions from current states:");
		for (;titg != titg_end; ++titg) { 
			FD_DF("SupConParallel: g: " << rPlantGen.SStr(titg->X1) << "-" << rPlantGen.EStr(titg->Ev) 
					<< "-" << rPlantGen.SStr(titg->X2));
		}
		for (;tith != tith_end; ++tith) {
			FD_DF("SupConParallel: h: " << rSpecGen.SStr(tith->X1) << "-" << rSpecGen.EStr(tith->Ev) 
					<< "-" << rSpecGen.SStr(tith->X2));
		}
		titg = rPlantGen.TransRelBegin(currentstates.first);
		tith = rSpecGen.TransRelBegin(currentstates.second);
#endif                                                                      
		// process all h transitions while there could be matching g transitions
		while ((tith != tith_end) && (titg != titg_end)) {
			FD_DF("SupconParallel: actual g-transition: " << rPlantGen.SStr(titg->X1) 
					<< "-" << rPlantGen.EStr(titg->Ev) << "-" << rPlantGen.SStr(titg->X2));
			FD_DF("SupconParallel: actual h-transition: " << rSpecGen.SStr(tith->X1) 
					<< "-" << rSpecGen.EStr(tith->Ev) << "-" << rSpecGen.SStr(tith->X2));
			// execute common events
			if (titg->Ev == tith->Ev) {
				FD_DF("SupconParallel: executing common event " 
						<< rPlantGen.EStr(titg->Ev));
				newstates = std::make_pair(titg->X2, tith->X2);
				rcmapit = rReverseCompositionMap.find(newstates);
				// add to todo list if state is new
				if (rcmapit == rReverseCompositionMap.end()) {
					todo.push(newstates);
					// if marked state
					if (rPlantGen.ExistsMarkedState(newstates.first) && 
							rSpecGen.ExistsMarkedState(newstates.second)) {
						tmpstate = rResGen.InsMarkedState();
						FD_DF("SupconParallel: NEW MSTATE: (" 
								<< rPlantGen.SStr(newstates.first) << "|"
								<< rSpecGen.SStr(newstates.second) << ") -> " << tmpstate);
					}
					// if "normal" state
					else {
						tmpstate = rResGen.InsState();
						FD_DF("SupconParallel: NEW STATE: (" 
								<< rPlantGen.SStr(newstates.first) << "|" 
								<< rSpecGen.SStr(newstates.second) << ") -> " << tmpstate);
					}
					rReverseCompositionMap[newstates] = tmpstate;
					FD_DF("SupconParallel: todo push: (" << rPlantGen.SStr(newstates.first)
							<< "|" << rSpecGen.SStr(newstates.second) << ") -> " << tmpstate);
				}
				// if state already exists
				else {
					tmpstate = rcmapit->second;
				}
				// if successor state not in forbidden set add transition
				if (! forbidden.ExistsState(tmpstate)) {
					FD_DF("SupconParallel: ADDING TRANSITION " 
							<< rPlantGen.SStr(rReverseCompositionMap[currentstates]) << "-" << rPlantGen.EStr(titg->Ev) 
							<< "-" << rPlantGen.SStr(tmpstate));
					rResGen.SetTransition(rReverseCompositionMap[currentstates], titg->Ev, tmpstate);
					FD_DF("SupconParallel: incrementing g transrel");
					++titg;
					FD_DF("SupconParallel: incrementing h transrel");
					++tith;
				}
				// if successor state in forbidden and event uncontrollable 
				// delete state
				else if (rUAlph.ExistsEvent(titg->Ev)) {
					FD_DF("SupconParallel: successor " << rSpecGen.SStr(tmpstate) 
							<< "in forbidden and common event " << rSpecGen.EStr(titg->Ev) 
							<< " uncontrollable:");
					FD_DF("SupconParallel: forbidden insert" << rPlantGen.SStr(tmpstate));
					forbidden.Insert(tmpstate);
#ifdef FAUDES_CHECKED
					// do not end while loops here for detecting all h transitions 
					// not in g
					FD_DF("SupconParallel: incrementing g transrel (FAUDES_CHECKED)");
					++titg;
					FD_DF("SupconParallel: incrementing h transrel (FAUDES_CHECKED)");
					++tith;
#else
					// exit all loops
					titg = titg_end;
					tith = tith_end;
#endif
				}
        // else if successor state in forbidden and event controllable
        else {
					FD_DF("SupconParallel: incrementing g transrel");
          ++titg;
					FD_DF("SupconParallel: incrementing h transrel");
          ++tith;
        }
			}
			// if g got some more transitions try to resync events
			else if (titg->Ev < tith->Ev) {
				FD_DF("SupconParallel: asynchronous execution of event " 
						<< rPlantGen.EStr(titg->Ev) << " in g while " << rSpecGen.EStr(tith->Ev)
						<< " in h");
				// if uncontrollable transition leaves specification
				// delete state from res and put into forbiddenset
				if (rUAlph.ExistsEvent(titg->Ev)) {
					FD_DF("SupconParallel: asynchronous event " << rPlantGen.EStr(titg->Ev) 
							<< " in g is uncontrollable");
					tmpstate = rReverseCompositionMap[currentstates];	
					FD_DF("SupconParallel: forbidden insert" << rPlantGen.SStr(tmpstate));
					forbidden.Insert(tmpstate);
					// exit all loops
					titg = titg_end;
					tith = tith_end;
					break;
				}
				FD_DF("SupconParallel: incrementing g transrel");
				++titg;
			} // if specification leaves plant model emit warning
			else {
#ifdef FAUDES_CHECKED
				FD_WARN("SupconParallel:  transition " << rSpecGen.SStr(tith->X1) 
						<< "-" << rSpecGen.EStr(tith->Ev) << "-" << rSpecGen.SStr(tith->X2) 
						<< "in specification h not found in g");
#endif
				FD_DF("SupconParallel: incrementing h transrel");
				++tith;
			}
		}
		// if g got some more transitions not in h
		while (titg != titg_end) {
			FD_DF("SupconParallel: asynchronous execution of event " 
					<< rPlantGen.EStr(titg->Ev) << " in g at end of h");
			FD_DF("SupconParallel: actual g-transition: " << rPlantGen.SStr(titg->X1) 
					<< "-" << rPlantGen.EStr(titg->Ev) << "-" << rPlantGen.SStr(titg->X2));
			FD_DF("SupconParallel: actual h-transition: end");
			// if uncontrollable transition leaves specification
			if (rUAlph.ExistsEvent(titg->Ev)) {
				tmpstate = rReverseCompositionMap[currentstates];
				FD_DF("SupconParallel: asynchron executed uncontrollable end "
						<< "event " << rPlantGen.EStr(titg->Ev) << " leaves specification:");
				FD_DF("SupconParallel: forbidden insert" << rPlantGen.SStr(tmpstate));
				forbidden.Insert(tmpstate);
				// exit this loop
				// if FAUDES_CHECKED not defined this means exiting all loops
				break;
			}
			FD_DF("SupconParallel: incrementing g transrel");
			++titg;
		}
#ifdef FAUDES_CHECKED
		// if h got some more transitions not in g
		while (tith != tith_end) {
			FD_WARN("SupconParallel:  transition " << rSpecGen.SStr(tith->X1) << "-" 
					<< rSpecGen.EStr(tith->Ev) << "-" << rSpecGen.SStr(tith->X2) 
					<< "in specification h not found in g");
			FD_DF("SupconParallel: incrementing h transrel");
			++tith;
		}
#endif
	}
	FD_DF("SupconParallel: deleting forbidden states...");
	StateSet::const_iterator sit;
	for(sit = forbidden.Begin(); sit!= forbidden.End(); sit++)
	   rResGen.DelState(*sit); 
	//rResGen.DelStates(forbidden);
}


/*
*****************************************
* SupConUnchecked: REAL SUPCON FUNCTION *
*****************************************
*/


// SupConUnchecked(rPlantGen, rSupGen)
bool SupConUnchecked(const cGenerator& rPlantGen, Generator& rSupGen) {
  FD_DF("SupConUnchecked...");

	// HELPERS:

	// set of forbidden states
	StateSet forbidden = rPlantGen.NewStateSet();

	// ALGORITHM:
  
  // return true if plant or parallelcomp contain no initial states
  if (rPlantGen.InitStatesEmpty() || rSupGen.InitStatesEmpty()) {
    return true;
  }
   
	// test controllability
	IsControllable(rPlantGen, rSupGen, forbidden);
	FD_DF("SupCon: delete forbidden states: " << rSupGen.Statesr().ToString());
	StateSet::const_iterator sit;
	for(sit = forbidden.Begin(); sit != forbidden.End(); sit ++){
	    rSupGen.DelState(*sit);
	}
	//system("PAUSE");
	// return true if no states had to be deleted
	if (forbidden.Empty()) {
		FD_DF("SupCon: controllable");
		return true;
	}
	// return false if states had to be deleted
	else {
		FD_DF("SupCon: not controllable");
		return false;
	}
}


/*
**************************************
* IsControllable (controllablity test) *
**************************************
*/

// IsControllable(rPlantGen, rSpecGen)
bool IsControllable(const cGenerator& rPlantGen, const Generator& rSpecGen) {
	FD_DF("IsControllable(" << &rSpecGen << "," << &rPlantGen << ")");

	// HELPERS:
	// controllable events
	const EventSet rUAlph = rPlantGen.UncontrollableEvents();
	FD_DF("IsControllable: controllable events: " 
			<< rPlantGen.ControllableEvents().ToString());
	FD_DF("IsControllable: uncontrollable events: " << rUAlph.ToString());
	// forbidden set
	StateSet forbidden = rPlantGen.NewStateSet();

	// CONSISTENCY CHECKS:
	ControllableConsistencyCheck(rPlantGen, rSpecGen);

	// ALGORITHM:
	return IsControllableUnchecked(rPlantGen, rSpecGen, rUAlph, forbidden);
}


// IsControllable(rPlantGen, rSpecGen, rUAlph)
bool IsControllable(const cGenerator& rPlantGen, const Generator& rSpecGen, 
		const EventSet& rUAlph) {
	FD_DF("IsControllable(" << &rSpecGen << "," << &rPlantGen << ")");

	// HELPERS:
	// forbidden set
	StateSet forbidden = rPlantGen.NewStateSet();

	// CONSISTENCY CHECKS:
	ControllableConsistencyCheck(rPlantGen, rSpecGen);

	// ALGORITHM:
	return IsControllableUnchecked(rPlantGen, rSpecGen, rUAlph, forbidden);
}

// IsControllable(rPlantGen, rSpecGen, forbidden)
bool IsControllable(const cGenerator& rPlantGen, const Generator& rSpecGen, 
		StateSet& forbidden) {
	FD_DF("IsControllable(" << &rSpecGen << "," << &rPlantGen << ")");

	// HELPERS:
	// controllable events
	const EventSet rUAlph = rPlantGen.UncontrollableEvents();
	FD_DF("IsControllable: controllable events: " 
			<< rPlantGen.ControllableEvents().ToString());
	FD_DF("IsControllable: uncontrollable events: " << rUAlph.ToString());

	// CONSISTENCY CHECKS:
	ControllableConsistencyCheck(rPlantGen, rSpecGen);

	// ALGORITHM:
	return IsControllableUnchecked(rPlantGen, rSpecGen, rUAlph, forbidden);
}


// IsControllable(rPlantGen, rSpecGen, rUAlph, forbidden)
bool IsControllable(const cGenerator& rPlantGen, const Generator& rSpecGen, 
		const EventSet& rUAlph, StateSet& rForbidden) {
	FD_DF("IsControllable(" << &rSpecGen << "," << &rPlantGen << ")");

	// CONSISTENCY CHECKS:
	ControllableConsistencyCheck(rPlantGen, rSpecGen);

	// ALGORITHM:
	return IsControllableUnchecked(rPlantGen, rSpecGen, rUAlph, rForbidden);
}


// ControllableConsistencyCheck(rPlantGen, rSpecGen, rUAlph)
void ControllableConsistencyCheck(const cGenerator& rPlantGen,
		const Generator& rSpecGen) {

	// alphabets must match
	if (rPlantGen.Alphabet() != rSpecGen.Alphabet()) {
		throw Exception("ControllableConsistencyCheck", 
				"Alphabets of generators don't match", 500);
	}

	// plant and spec must be deterministic
	bool plant_det = IsDeterministic(rPlantGen);
	bool spec_det = IsDeterministic(rSpecGen);

	if ((plant_det == false) && (spec_det == true)) {
		std::stringstream errstr;
		errstr << "Plant generator must be deterministic, "
				<< "but is nondeterministic";
		throw Exception("ControllableConsistencyCheck", errstr.str(), 501);
	}
	else if ((plant_det == true) && (spec_det == false)) {
		std::stringstream errstr;
		errstr << "Spec generator must be deterministic, "
				<< "but is nondeterministic";
		throw Exception("ControllableConsistencyCheck", errstr.str(), 503);
	}
	else if ((plant_det == false) && (spec_det == false)) {
		std::stringstream errstr;
		errstr << "Plant and spec generator must be deterministic, "
				<< "but both are nondeterministic";
		throw Exception("ControllableConsistencyCheck", errstr.str(), 504);
	}
}

// IsControllableUnchecked(rPlantGen, rSpecGen, rUAlph, forbidden)
bool IsControllableUnchecked(const cGenerator& rPlantGen, 
		const Generator& rSpecGen, const EventSet& rUAlph, StateSet& forbidden) {
	FD_DF("IsControllableUnchecked(" << &rSpecGen << "," << &rPlantGen << ")");

	// HELPERS:

	// todo stack
	std::stack<Idx> todog, todoh;
	// set of already discovered states
	StateSet discovered = rPlantGen.NewStateSet();
	// reverse sorted transition relation build on the fly
	TransSetX2EvX1 rtransrel;

	// PREPARE RESULT:
	forbidden.Clear();

	// ALGORITHM:

	// return false (uncontrollable) if there is no initial state
	if (rPlantGen.InitStatesEmpty() || rSpecGen.InitStatesEmpty()) {
		return false;
	}

	// push combined initial state on todo stack
	todog.push(*rPlantGen.InitStatesBegin());
	todoh.push(*rSpecGen.InitStatesBegin());
	FD_DF("IsControllable: todo push: (" << rPlantGen.SStr(*rPlantGen.InitStatesBegin()) << "|"
			<< rSpecGen.SStr(*rSpecGen.InitStatesBegin()) << ")");

	// process todo stack
	while (! todog.empty()) {
		// get top elements from todo stack
		Idx currentg = todog.top();
		Idx currenth = todoh.top();
		todog.pop();
		todoh.pop();
		FD_DF("IsControllable: todo pop: (" << rPlantGen.SStr(currentg) << "|" 
				<< rSpecGen.SStr(currenth) << ")");

#ifdef FAUDES_DEBUG_FUNCTION
		TransSet::const_iterator _titg, _tith;
		// print all transitions of current states
		FD_DF("IsControllable: transitions from current states:");
		for (_titg = rPlantGen.TransRelBegin(currentg); _titg != rPlantGen.TransRelEnd(currentg); ++_titg) 
			FD_DF("IsControllable: g: " << rPlantGen.SStr(_titg->X1) << "-" 
					<< rPlantGen.EStr(_titg->Ev) << "-" << rPlantGen.SStr(_titg->X2));
		for (_tith = rSpecGen.TransRelBegin(currenth); _tith != rSpecGen.TransRelEnd(currenth); ++_tith) 
			FD_DF("IsControllable: h: " << rSpecGen.SStr(_tith->X1) << "-" 
					<< rSpecGen.EStr(_tith->Ev) << "-" << rSpecGen.SStr(_tith->X2));
#endif 

		// process all h transitions while there could be matching g transitions
		TransSet::const_iterator titg = rPlantGen.TransRelBegin(currentg);
		TransSet::const_iterator titg_end = rPlantGen.TransRelEnd(currentg);
		TransSet::const_iterator tith = rSpecGen.TransRelBegin(currenth);
		TransSet::const_iterator tith_end = rSpecGen.TransRelEnd(currenth);
		while ((tith != tith_end) && (titg != titg_end)) {
			FD_DF("IsControllable: actual g-transition: " << rPlantGen.SStr(titg->X1) 
					<< "-" << rPlantGen.EStr(titg->Ev) << "-" << rPlantGen.SStr(titg->X2));
			FD_DF("IsControllable: actual h-transition: " << rSpecGen.SStr(tith->X1) 
					<< "-" << rSpecGen.EStr(tith->Ev) << "-" << rSpecGen.SStr(tith->X2));
			// execute common events
			if (titg->Ev == tith->Ev) {
				FD_DF("IsControllable: executing common event " << rPlantGen.EStr(titg->Ev));
				// add to todo list if state is undiscovered
				if (! discovered.ExistsState(currenth)) {
					todog.push(titg->X2);
					todoh.push(tith->X2);
					FD_DF("IsControllable: todo push: (" << rPlantGen.SStr(titg->X2) << "|"
							<< rSpecGen.SStr(tith->X2) << ")");
				}
				// if h successor state not in forbidden set add transition to rtransrel
				if (! forbidden.ExistsState(tith->X2)) {
					rtransrel.Insert(*tith);
					FD_DF("IsControllable: incrementing g transrel");
					++titg;
					FD_DF("IsControllable: incrementing h transrel");
					++tith;
				}
				// if successor state is forbidden and event uncontrollable 
				else if (rUAlph.ExistsEvent(titg->Ev)) {
					FD_DF("IsControllable: successor state " << rSpecGen.SStr(tith->X2) << 
							" forbidden and event " << rPlantGen.EStr(titg->Ev) << " uncontrollable:");
					FD_DF("IsControllable: TraverseUncontrollableBackwards(" << rSpecGen.SStr(currenth) << ")");
					TraverseUncontrollableBackwards(rUAlph, rtransrel, forbidden, currenth);
#ifdef FAUDES_CHECKED
					// just increment transrel iterators to find all h transitions not in g
					FD_DF("IsControllable: incrementing g transrel (FAUDES_CHECKED)");
					++titg;
					FD_DF("IsControllable: incrementing h transrel (FAUDES_CHECKED)");
					++tith;
#else
					// exit all loops
					titg = titg_end;
					tith = tith_end;
#endif
					break;
				}
                // else if successor state in forbidden and event controllable
                else {
        					FD_DF("IsControllable: incrementing g transrel");
                  ++titg;
        					FD_DF("IsControllable: incrementing h transrel");
                  ++tith;
                }
			}
			// if g got some more transitions try to resync events
			else if (titg->Ev < tith->Ev) {
				FD_DF("IsControllable: asynchronous execution of event " 
						<< rPlantGen.EStr(titg->Ev) << " in g while " << rSpecGen.EStr(tith->Ev)
						<< " in h");
				// if uncontrollable transition leaves specification
				// delete state from rResGen and put into forbiddenset
				if (rUAlph.ExistsEvent(titg->Ev)) {
					FD_DF("IsControllable: asynchronous event " << rPlantGen.EStr(titg->Ev) 
							<< " in g is uncontrollable");
					FD_DF("IsControllable: TraverseUncontrollableBackwards(" << rSpecGen.SStr(currenth) << ")");
					TraverseUncontrollableBackwards(rUAlph, rtransrel, forbidden, currenth);
					// exit all loops over g transrel
					titg = titg_end;
					break;
				}
				FD_DF("IsControllable: incrementing g transrel");
				++titg;
			}
			// if specification leaves plant model emit warning
			else {
#ifdef FAUDES_CHECKED
				FD_WARN("IsControllable: transition " << rSpecGen.SStr(tith->X1) << "-" 
						<< rSpecGen.EStr(tith->Ev) << "-" << rSpecGen.SStr(tith->X2)
						<< "in specification h not found in g");
#endif
				FD_DF("IsControllable: incrementing h transrel");
				++tith;
			}
		}
		// process other transitions not in h
		while (titg != titg_end) {
			FD_DF("IsControllable: asynchronous execution of event " 
					<< rPlantGen.EStr(titg->Ev) << " in g at end of h");
			FD_DF("IsControllable: actual g-transition: " << rPlantGen.SStr(titg->X1) 
					<< "-" << rPlantGen.EStr(titg->Ev) << "-" << rPlantGen.SStr(titg->X2));
			FD_DF("IsControllable: actual h-transition: end");
			// if uncontrollable transition leaves specification
			if (rUAlph.ExistsEvent(titg->Ev)) {
				FD_DF("IsControllable: asynchronous execution of uncontrollable event "
						<< rPlantGen.EStr(titg->Ev) << " in g");
				FD_DF("IsControllable: TraverseUncontrollableBackwards(" << rPlantGen.SStr(currenth) << ")");
				TraverseUncontrollableBackwards(rUAlph, rtransrel, forbidden, currenth);
				// exit this loop
				break;
			}
			FD_DF("IsControllable: incrementing g transrel");
			++titg;
		}
#ifdef FAUDES_CHECKED
		// process other transitions not in g
		while (tith != tith_end) {
			FD_WARN("IsControllable: transition " << rSpecGen.SStr(tith->X1) << "-" 
					<< rSpecGen.EStr(tith->Ev) << "-" << rSpecGen.SStr(tith->X2)
					<< "in specification h not found in g");
			FD_DF("IsControllable: incrementing h transrel");
			++tith;
		}
#endif
		discovered.Insert(currenth);
	}

	// compute complete set of forbidden states
	forbidden = rSpecGen.States() - ( discovered - forbidden );
	
	if (forbidden.Empty()) {
		FD_DF("IsControllable: controllable");
		return true;
	}
	else {
		FD_DF("IsControllable: not controllable");
		return false;
	}
}


/*
***************************
* HELPER FOR CONTROLLABLE *
***************************
*/

// TraverseUncontrollableBackwards(rUAlph, rtransrel, forbidden, current)
void TraverseUncontrollableBackwards(const EventSet& rUAlph, 
		TransSetX2EvX1& rtransrel, StateSet& forbidden, Idx current) {
	FD_DF("TraverseUncontrollableBackwards: " << forbidden.Str(current));

	// HELPERS:
	std::stack<Idx> todo;
	TransSetX2EvX1::iterator rtit = rtransrel.BeginByX2(current);
	TransSetX2EvX1::iterator rtit_end = rtransrel.EndByX2(current);

	// ALGORITHM:
	forbidden.Insert(current);
	todo.push(current);
	FD_DF("TraverseUncontrollableBackwards: current forbidden set: " 
			<< forbidden.ToString());
	// process todo stack
	while (! todo.empty()) {
		// get top state from todo stack
		current = todo.top(); todo.pop();
		rtit_end = rtransrel.EndByX2(current);
		// iteration over all transitions with x2 == current
		for (rtit = rtransrel.BeginByX2(current); rtit != rtit_end; ++rtit) {
			// if uncontrollable event and predecessor state not already forbidden
			if ((rUAlph.ExistsEvent(rtit->Ev)) && (! forbidden.ExistsState(rtit->X1))) {
				FD_DF("TraverseUncontrollableBackwards: todo push " << forbidden.Str(rtit->X1));
				todo.push(rtit->X1);
				FD_DF("TraverseUncontrollableBackwards: forbidden insert: " 
						<< forbidden.Str(rtit->X1));
				forbidden.Insert(rtit->X1);
			}
		}
	}
}

} // namespace faudes
