/* supcon.h -- supremal controllable sublanguage and controllablity */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */

 
#ifndef SUBCON_H

#include "definitions.h"
#include "debug.h"
#include "cgenerator.h"
#include "project.h"
#include "parallel.h"
#include "cmap.h"
#include <stack>

namespace faudes {
    
/**
 * Nonblocking Supremal Controllable Sublanguage (wrapper function)
 *
 * Computes symbolic state names in resulting supervisor automaton
 * if symbolic state names are enabled in rPlantGen and rSpecGen
 *
 * @param rPlantGen
 *   Plant cGenerator
 * @param rSpecGen
 *   Specification Generator
 * @param rResGen
 *   Reference to resulting cGenerator, the
 *   minimal restrictive nonblocking supervisor
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   plant nondeterministic (id 501)
 *   spec nondeterministic (id 503)
 *   plant and spec nondeterministic (id 504)
 */
void SupConNB(const cGenerator& rPlantGen, const Generator& rSpecGen, 
		cGenerator& rResGen);


/** 
 * Nonblocking Supremal Controllable Sublanguage.
 * 
 * Only for deterministic plant + spec. Throws exception if plant or
 * spec is nondeterministic.
 *
 * Real SupConNB function
 *
 * Finds the "largest" sublanguage of h for that language of g is
 * controllable with respect to h. Differing from theory the marked
 * language of h may not be a sublanguage of g but both must share the
 * same alphabet.
 *
 * See "C.G CASSANDRAS AND S. LAFORTUNE. Introduction to Discrete Event 
 * Systems. Kluwer, 1999." for base algorithm.
 *
 * This version creates a "reverse composition map" given as reference
 * parameter.
 * 
 * @param rPlantGen
 *   Plant cGenerator 
 * @param rSpecGen
 *   Specification Generator 
 * @param rReverseCompositionMap
 *   std::map< std::pair<Idx,Idx>, Idx> as in the parallel composition function
 * @param rResGen
 *   Reference to resulting cGenerator, the
 *   minimal restrictive nonblocking supervisor
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   plant nondeterministic (id 501)
 *   spec nondeterministic (id 503)
 *   plant and spec nondeterministic (id 504)
 */
void SupConNB(const cGenerator& rPlantGen, const Generator& rSpecGen, 
		std::map< std::pair<Idx,Idx>,Idx >& rReverseCompositionMap, cGenerator& rResGen);


/** 
 * Nonblocking Supremal Controllable Sublanguage.
 * 
 * For deterministic and nondeterministic plant or spec. 
 *
 * - If rPlantGen is deterministic, rDetPlantGen will be unchanged.
 *   rDetPlantBool will be set to false.
 *
 * - If rSpecGen is deterministic, rDetSpecGen will be unchanged
 *   rDetSpecBool will be set to false
 *
 * - If rPlantGen is nondeterministic, rDetPlantGen will contain deterministic
 *   plant generator. rDetPlantBool will be set to true. rPowerStatesPlant will
 *   contain the mapping of deterministic states to subsets of nondeterministic
 *   states
 *
 * - If rSpecGen is nondeterministic, rDetPlantGen will contain deterministic
 *   spec generator. rDetSpecBool will be set to true. rPowerStatesSpec will
 *   contain the mapping of deterministic states to subsets of nondeterministic
 *   states
 *
 * Finds the "largest" sublanguage of h for that language of g is
 * controllable with respect to h. Differing from theory the marked
 * language of h may not be a sublanguage of g but both must share the
 * same alphabet.
 *
 * See "C.G CASSANDRAS AND S. LAFORTUNE. Introduction to Discrete Event 
 * Systems. Kluwer, 1999." for base algorithm.
 *
 * This version creates a "reverse composition map" given as reference
 * parameter.
 * 
 * @param rPlantGen
 *   Plant cGenerator 
 * @param rSpecGen
 *   Specification Generator 
 * @param rReverseCompositionMap
 *   std::map< std::pair<Idx,Idx>, Idx> as in the parallel composition function
 * @param rDetPlantGen
 *   Deterministic plant generator (if rPlantGen is nondeterministic)
 * @param rDetSpecGen
 *   Deterministic spec generator (if rSpecGen is nondeterministic)
 * @param rPowerStatesPlant
 *   Mapping of deterministic states to set of nondeterministic states
 * @param rPowerStatesSpec
 *   Mapping of deterministic states to set of nondeterministic states
 * @param rDetPlantBool
 *   true if rPlantGen is deterministic, false if not
 * @param rDetSpecBool
 *   true if rSpecGen is deterministic, false if not
 * @param rResGen
 *   Reference to resulting cGenerator, the
 *   minimal restrictive nonblocking supervisor
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 */
void SupConNBNonDet(const cGenerator& rPlantGen, const Generator& rSpecGen,
		std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap,
		cGenerator rDetPlantGen, cGenerator rDetSpecGen, 
    std::map<Idx,StateSet>& rPowerStatesPlant, 
    std::map<Idx,StateSet>& rPowerStatesSpec,
		bool& rDetPlantBool, bool& rDetSpecBool, cGenerator& rResGen);

/**
 * Supremal Controllable Sublanguage (wrapper function)
 *
 * @param rPlantGen
 *   Plant cGenerator
 * @param rSpecGen
 *   Specification Generator
 * @param rResGen
 *   Reference to resulting cGenerator, the
 *   minimal restrictive supervisor
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   plant nondeterministic (id 501)
 *   spec nondeterministic (id 503)
 *   plant and spec nondeterministic (id 504)
 */
void SupCon(const cGenerator& rPlantGen, const Generator& rSpecGen, 
		cGenerator& rResGen);


/** 
 * Supremal Controllable Sublanguage.
 * 
 * Only for deterministic plant + spec. Throws exception if plant or
 * spec is nondeterministic.
 *
 * Real SupCon function
 *
 * Finds the "largest" sublanguage of h for that language of g is
 * controllable with respect to h. Differing from theory the marked
 * language of h may not be a sublanguage of g but both must share the
 * same alphabet.
 *
 * See "C.G CASSANDRAS AND S. LAFORTUNE. Introduction to Discrete Event 
 * Systems. Kluwer, 1999." for base algorithm.
 *
 * This version creates a "reverse composition map" given as reference
 * parameter.
 * 
 * @param rPlantGen
 *   Plant cGenerator 
 * @param rSpecGen
 *   Specification Generator 
 * @param rReverseCompositionMap
 *   std::map< std::pair<Idx,Idx>, Idx> as in the parallel composition function
 * @param rResGen
 *   Reference to resulting cGenerator, the
 *   minimal restrictive supervisor
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   plant nondeterministic (id 501)
 *   spec nondeterministic (id 503)
 *   plant and spec nondeterministic (id 504)
 */
void SupCon(const cGenerator& rPlantGen, const Generator& rSpecGen, 
		std::map< std::pair<Idx,Idx>,Idx >& rReverseCompositionMap, cGenerator& rResGen);

/** 
 * Fast parallel composition for computation for the SupConNB. 
 * Composition stops at transition paths that leave the specification
 * by uncontrollable events in plant.
 * 
 * @param rPlantGen
 *   Plant cGenerator 
 * @param rSpecGen
 *   Specification Generator 
 * @param rUAlph
 *   Uncontrollable Events
 * @param rReverseCompositionMap
 *   std::map< std::pair<Idx,Idx>, Idx> as in the parallel composition function
 * @param rResGen
 *   Reference to resulting cGenerator, the
 *   less restrictive supervisor
 */
void SupconParallel(const cGenerator& rPlantGen, const Generator& rSpecGen,
		const EventSet& rUAlph, std::map< std::pair<Idx,Idx>, Idx>& rReverseCompositionMap, 
		cGenerator& rResGen);

/**
 * Supremal Controllable Sublangauge (Real SupCon function; unchecked)
 *
 * Both, plant and spec MUST be deterministic and share the same alphabet!!!
 *
 * Most likely will result in blocking states.
 *
 * @param rPlantGen
 *   Plant generator
 * @param rSupGen
 *   Specification generator
 *
 * @return
 *   True if no state has been delete. Else false.
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   Plant generator nondeterministic (id 501)
 *   Specification generator nondeterministic (id 503)
 *   Plant & Spec generator nondeterministic (id 504)
 */
bool SupConUnchecked(const cGenerator& rPlantGen, Generator& rSupGen);

/**
 * Controllability (wrapper function)
 *
 * Only for deterministic plant + spec!!! Throws exception if plant or
 * spec is nondeterministic.
 *
 * Checks if language of specification h is controllable with respect to
 * language of generator g.
 *
 * @param rPlantGen
 *   Plant generator
 * @param rSpecGen
 *   Specification generator
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   Plant generator nondeterministic (id 501)
 *   Specification generator nondeterministic (id 503)
 *   Plant & Spec generator nondeterministic (id 504)
 *
 * @return 
 *   true / false
 */
bool IsControllable(const cGenerator& rPlantGen, const Generator& rSpecGen);

/**
 * Controllability (wrapper function)
 *
 * Only for deterministic plant + spec!!! Throws exception if plant or
 * spec is nondeterministic.
 *
 * Checks if language of specification h is controllable with respect to
 * language of generator g.
 *
 * Uncontrollable event set has to be given as parameter.
 *
 * @param rPlantGen
 *   Plant generator
 * @param rSpecGen
 *   Specification generator
 * @param rUAlph
 *   Uncontrollable events
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   Plant generator nondeterministic (id 501)
 *   Specification generator nondeterministic (id 503)
 *   Plant & Spec generator nondeterministic (id 504)
 *
 * @return 
 *   true / false
 */
bool IsControllable(const cGenerator& rPlantGen, const Generator& rSpecGen, 
		const EventSet& rUAlph);

/**
 * Controllability (wrapper function)
 *
 * Only for deterministic plant + spec!!! Throws exception if plant or
 * spec is nondeterministic.
 *
 * Checks if language of specification h is controllable with respect to
 * language of generator g.
 *
 * Set of forbidden states is given as a result parameter and will be
 * filled during computation.
 *
 * @param rPlantGen
 *   Plant generator
 * @param rSpecGen
 *   Specification generator
 * @param forbidden
 *   Set of forbidden states
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   Plant generator nondeterministic (id 501)
 *   Specification generator nondeterministic (id 503)
 *   Plant & Spec generator nondeterministic (id 504)
 *
 * @return 
 *   true / false
 */
bool IsControllable(const cGenerator& rPlantGen, const Generator& rSpecGen, 
		StateSet& forbidden);

/**
 * Controllability (wrapper function)
 *
 * Only for deterministic plant + spec!!! Throws exception if plant or
 * spec is nondeterministic.
 *
 * Checks if language of specification h is controllable with respect to
 * language of generator g. 
 *
 * Uncontrollable event set has to be given as parameter.
 * Last parameter forbidden contains the set of forbidden states after
 * computation.
 *
 * @param rPlantGen
 *   Plant generator
 * @param rSpecGen
 *   Specification generator
 * @param rUAlph
 *   Uncontrollable events
 * @param forbidden
 *   Set of forbidden states
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   Plant generator nondeterministic (id 501)
 *   Specification generator nondeterministic (id 503)
 *   Plant & Spec generator nondeterministic (id 504)
 *
 * @return 
 *   true / false
 */
bool IsControllable(const cGenerator& rPlantGen, const Generator& rSpecGen, 
		const EventSet& rUAlph, StateSet& forbidden);

/**
 * Consistency checks for controllablity tests.
 * Checks if alphabets match and if generators are deterministic.
 *
 * @param rPlantGen
 *   Plant generator
 * @param rSpecGen
 *   Specification generator
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   Plant generator nondeterministic (id 501)
 *   Specification generator nondeterministic (id 503)
 *   Plant & Spec generator nondeterministic (id 504)
 */
void ControllableConsistencyCheck(const cGenerator& rPlantGen, 
		const Generator& rSpecGen);

/**
 * Controllability (real function; no consistency checks)
 *
 * Only for deterministic plant + spec!!! Throws exception if plant or
 * spec is nondeterministic.
 *
 * Checks if language of specification h is controllable with respect to
 * language of generator g. 
 *
 * Uncontrollable event set has to be given as parameter.
 * Last parameter forbidden contains the set of forbidden states after
 * computation.
 *
 * @param rPlantGen
 *   Plant generator
 * @param rSpecGen
 *   Specification generator
 * @param rUAlph
 *   Uncontrollable events
 * @param forbidden
 *   Set of forbidden states
 *
 * @exception Exception
 *   Alphabets of generators don't match (id 500)
 *   Plant generator nondeterministic (id 501)
 *   Specification generator nondeterministic (id 503)
 *   Plant & Spec generator nondeterministic (id 504)
 *
 * @return 
 *   true / false
 */
bool IsControllableUnchecked(const cGenerator& rPlantGen, const Generator& rSpecGen, 
		const EventSet& rUAlph, StateSet& forbidden);


/**
 * Helper function for controllable. Put state given as parameter
 * current and all uncontrollable predecessor states into forbidden set.
 *
 * @param rUAlph
 *   Set of uncontrollable events
 * @param rtransrel
 *   Reverse sorted transition relation
 * @param forbidden
 *   Set of forbidden states in composition generator
 * @param current
 *   Current state
 */
void TraverseUncontrollableBackwards(const EventSet& rUAlph, 
		TransSetX2EvX1& rtransrel, StateSet& forbidden, Idx current);

} // namespace faudes

#define SUBCON_H
#endif 

