/* symboltable.cpp -- symboltables for event- and statenames */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "symboltable.h"
#include <iostream>

namespace faudes {

// SymbolTable::OpcNames()  // untested
void SymbolTable::OpcNames(void) {
	std::map<char,char>::iterator cmit;
	// name map 
	std::map<Idx,std::string>::iterator nmit;
	// index map
	std::map<std::string,Idx>::iterator imit;
	std::map<std::string,Idx> newindexmap;
	// other
	std::string::iterator i;
	
	// iteration over mNameMap
	for (nmit = mNameMap.begin(); nmit != mNameMap.end(); ++nmit) {
		// iterate over every string in mNameMap
		for (i = nmit->second.begin(); i != nmit->second.end(); ++i) {
			// if character in mForbiddenCharReplaceMap replace it
			cmit = mForbiddenCharReplaceMap.find(*i);
			if (cmit != mForbiddenCharReplaceMap.end()) {
				nmit->second.replace(i, i, 1, cmit->second);
			}
		}
	}

	// iteration over mIndexMap
	// as map order may not be destroyed we create new mIndexMap and swap maps
	for (imit = mIndexMap.begin(); imit != mIndexMap.end(); ++imit) {
		// check string
		std::string name = imit->first;
		for (i = name.begin(); i != name.end(); ++i) {
			cmit = mForbiddenCharReplaceMap.find(*i);
			if (cmit != mForbiddenCharReplaceMap.end()) {
				name.replace(i, i, 1, cmit->second);
			}
		}
		// create new pair in newindexmap
		newindexmap[name] = imit->second;
	}
	mIndexMap.swap(newindexmap);
}


// EventSymbolTable::EventSymbolTable()
EventSymbolTable::EventSymbolTable(void) : SymbolTable() {
}

// EventSymbolTable::InsEvent()
Idx EventSymbolTable::InsEvent(void) {
	return SymbolTable::InsIndex();
}

// EventSymbolTable::InsEvent(rName)
Idx EventSymbolTable::InsEvent(const std::string& rName) {
	if (mNextIndex == mMaxIndex) {
		FD_ERR("EventSymbolTable::InsEvent: index overflow in " << mMyName);
		exit(1);
	}
  Idx tmpindex = EventIndex(rName);
  if (tmpindex != 0) {
    return tmpindex;
  }
  mIndexMap[rName] = mNextIndex;
	 mNameMap[mNextIndex] = rName;
  return mNextIndex++;
}

// EventSymbolTable::DelEvent(index)
void EventSymbolTable::DelEvent(Idx index) {
	std::map<Idx,std::string>::iterator it = mNameMap.find(index);
	if (it == mNameMap.end()) {
		std::stringstream errstr;
		errstr << "Event index " << index << " not found in EventSymbolTable";
		throw Exception("EventSymbolTable::DelEvent(index)", errstr.str(), 102);
	}
	else {
		mIndexMap.erase(it->second);
		mNameMap.erase(it);
	}
}

// EventSymbolTable::DelEvent(rName)
void EventSymbolTable::DelEvent(const std::string& rName) {
	std::map<std::string,Idx>::iterator it = mIndexMap.find(rName);
	if (it == mIndexMap.end()) {
		std::stringstream errstr;
		errstr << "Event name \"" << rName << "\" not found in EventSymbolTable";
		throw Exception("EventSymbolTable::DelEvent(index)", errstr.str(), 105);
	}
	else {
		mNameMap.erase(it->second);
		mIndexMap.erase(it);
	}
}

// EventSymbolTable::EventIndex(rName)
Idx EventSymbolTable::EventIndex(const std::string& rName) const {
	std::map<std::string,Idx>::const_iterator it = mIndexMap.find(rName);
	if (it == mIndexMap.end()) {
    return 0;
	}
	else return it->second;
}

// EventSymbolTable::EventName
void EventSymbolTable::EventName(Idx index, const std::string& rName) {
	std::map<std::string,Idx>::iterator it1 = mIndexMap.find(rName);
	std::map<Idx,std::string>::iterator it2 = mNameMap.find(index);
	if ((it1 != mIndexMap.end()) && (it1->second != index)) {
		std::stringstream errstr;
		errstr << "Event name \"" << rName 
				<< "\" already associated with index " << it1->second;
		throw Exception("EventSymbolTable::EventName(index,name)", 
				errstr.str(), 104);
	}
	else if (it2 == mNameMap.end()) {
		std::stringstream errstr;
		errstr << "Index " << index << " does not exist in EventSymbolTable";
		throw Exception("EventSymbolTable::EventName(index,name)", 
				errstr.str(), 102);
	}
	it2->second = rName;
	mIndexMap[rName] = index;
}

// EventSymbolTable::ExistsEvent(index)
bool EventSymbolTable::ExistsEvent(Idx index) {
	std::map<Idx,std::string>::iterator it = mNameMap.find(index);
	if (it != mNameMap.end()) {
		return true;
	}
	else {
		return false;
	}
}

// EventSymbolTable::ExistsEvent(rName)
bool EventSymbolTable::ExistsEvent(const std::string& rName) const {
	std::map<std::string,Idx>::const_iterator it = mIndexMap.find(rName);
	if (it == mIndexMap.end()) return false;
	return true;
}

// StateSymbolTable::SetDefaultStateName(index)
void StateSymbolTable::SetDefaultStateName(Idx index) {
	ClrStateName(index); // throws exception if index not found
	std::ostringstream o;
	std::string dname;
	o << index;
	dname = o.str();
	mNameMap[index] = dname;
}

// StateSymbolTable::ClrStateNames()
void StateSymbolTable::ClrStateNames(void) {
	mNameMap.clear();
}

} // namespace faudes
