/* symboltable.h -- symboltables for event- and statenames */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef SYMBOLTABLE_H

#include "definitions.h"
#include "debug.h"
#include "exception.h"
#include <algorithm>
#include <map>
#include <set>
#include <limits>
#include <iostream>
#include <sstream>

namespace faudes {

/**  
 * SymbolTable base class
 */
class SymbolTable {
	public:
		/** Creates a new SymbolTable */
		SymbolTable(void) :
				mMaxIndex(std::numeric_limits<Idx>::max()), mNextIndex(1) { }

		/** typedef for iterator */
		typedef std::set<Idx>::iterator iterator;

		/** Return name of SymbolTable */
		std::string Name(void) const {
			return mMyName;
		}
		
		/** Set name of statetable */
		void Name(const std::string& rName) {
			mMyName = rName;
		}
		
		/** Get maximum index */
		Idx MaxIndex(void) const {
			return mMaxIndex;
		}
		
		/** 
		 * Set maximum index 
		 *
		 * @param index
		 *   New maximum index
		 *
		 * @return 
		 *   Success
		 */
		bool MaxIndex(Idx index) {
			if (index <= std::numeric_limits<Idx>::max()) {
				mMaxIndex = index;
				return true;
			}
			else {
				return false;
			}
		}
		
		/** Get the highest index in use */
		Idx LastIndex(void) const {
			return mNextIndex - 1;
		}
		
		/** Return next free index value and increment mNextIndex */
		Idx InsIndex(void) {
			if (mNextIndex == mMaxIndex) {
				FD_ERR("index overflow in symboltable " << mMyName);
				FD_ERR("terminating program");
				exit(1);
			}
			return mNextIndex++;
		}
	
		/**
		 * Remove non conform characters from name strings for OPC code output, etc.
		 * Uses the forbiddencharmap defined below.
     *
     * Untested!
		 */
		void OpcNames(void);
	
	protected:
		/** Name of the SymbolTable */
		std::string mMyName;
		
		/** Index lookup */
		std::map<std::string,Idx> mIndexMap;
		
		/** Name lookup */
		std::map<Idx,std::string> mNameMap;

		/** Upper limit */
		Idx mMaxIndex;
		
		/** Last used index + 1 */
		Idx mNextIndex;

		/** 
		 * Forbidden char map for OPC output and other formats with a restricted
		 * character set.
		 */
		std::map<char,char> mForbiddenCharReplaceMap;
};



/**  
 * EventSymbolTable class. Event names have to be unique.
 * Both name lookup and reverse lookup is supplied.
 */
class EventSymbolTable : public SymbolTable {
	public:
		/** Creates a new EventSymbolTable */
		EventSymbolTable(void);

		/** 
		 * Allocates next free index value and increment mNextIndex.
		 * Program exit on index overflow.
		 *
		 * @return
		 *   Next free index
		 */
		Idx InsEvent(void);
	
		/** 
		 * Add name with next free index value and increment mNextIndex.
		 * Both name lookup and reverse lookup will be added. If name already
		 * exists, existing index will be returned.
		 * Program exit on index overflow.
		 *
		 * @param rName
		 *   Event name to associate new index with
		 *
		 * @return 
		 *   Index of event
		 */
		Idx InsEvent(const std::string& rName);
	
		/** 
		 * Delete an event by index 
		 *
		 * @param index
		 *   Index to delete
		 *
		 * @exception Exception
		 *   Event index not found in EventSymbolTable (id 102)
		 */
		void DelEvent(Idx index);
		
		/** 
		 * Delete an event by name
		 *
		 * @param rName
		 * 
		 * @exception Exception
		 *   Event name not found in EventSymbolTable (id 105)
		 */
		void DelEvent(const std::string& rName);

		/** 
		 * Name lookup
		 *
		 * @param index
		 *   Event index
		 *
		 * @return
		 *   Name or empty string ""
		 */
		std::string EventName(Idx index) const {
			std::map<Idx,std::string>::const_iterator it = mNameMap.find(index);
			if (it == mNameMap.end()) return "";
			return it->second;
		}

		/** 
		 * Index lookup
		 *
		 * @param rName
		 *   Event name
		 * 
		 * @return 
		 *   Index or 0 if rName not found
		 */
		Idx EventIndex(const std::string& rName) const;

		/** 
		 * Set event name for existing index 
		 *
		 * @param index
		 *   Event index
		 * @param rName
		 *   New event name
		 *
		 * @exception Exception
		 *   Index not found in EventSymbolMap (id 102)
		 *   Name already associated with another index (id 104)
		 */
		void EventName(Idx index, const std::string& rName);
		
		/** 
		 * Existance of index 
		 *
		 * @param index
		 *   Event index
		 *
		 * @return 
		 *   Success
		 */
		bool ExistsEvent(Idx index);
	
		/** 
		 * Existance of name
		 *
		 * @param rName
		 *   Event name
		 *
		 * @return
		 *  Success
		 */
		bool ExistsEvent(const std::string& rName) const;
};



/**  
 * StateSymbolTable class. State names may not be unique.
 * Only name lookup is supplied.
 */
class StateSymbolTable : public SymbolTable {
	public:
		/** Creates a new StateSymbolTable */
		StateSymbolTable(void) : SymbolTable() {
		}

		/** 
		 * Return next free index value and increment mNextIndex.
		 * Program exit on index overflow.
		 */
		Idx InsState(void) {
			return SymbolTable::InsIndex();
		}
		
		/** 
		 * Add name with next free index value and increment mNextIndex.
		 * Only name lookup will be added. Program exit on index overflow.
		 * If empty string is given, no mNameMap entry will be created.
		 *
		 * @param rName
		 *   State name
		 *
		 * @return 
		 *   Success
		 */
		Idx InsState(const std::string& rName) {
			if (mNextIndex == mMaxIndex) {
				FD_ERR("StateSymbolTable::InsState: index overflow in " << mMyName);
				exit(1);
			}
			if (rName == "") {
				return SymbolTable::InsIndex();
			}
			else {
				mNameMap[mNextIndex] = rName;
				return mNextIndex++;
			}
		}

		/** 
		 * Delete a state by index 
		 *
		 * @param index
		 *   State index to delete
		 */
		void DelState(Idx index) {
			mNameMap.erase(index);
		}

		/** 
		 * Name lookup
		 *
		 * @param index
		 *   Event index
		 *
		 * @return
		 *   Name or empty string ""
		 */
		std::string StateName(Idx index) const {
			std::map<Idx,std::string>::const_iterator it = mNameMap.find(index);
			if (it == mNameMap.end()) return "";
			return it->second;
		}
		
		/** 
		 * Set name for index 
		 *
		 * @param index
		 *   State index
		 * @param rName
		 *   New state name
		 */
		void StateName(Idx index, const std::string& rName) {
      mNameMap[index] = rName;
    }

		/** 
		 * Existance of index 
		 *
		 * @param index
		 *   Event index
		 *
		 * @return 
		 *   Success
		 */
		bool ExistsState(Idx index);
		// FIXME: definition missing
	
		/**
		 * Set default name ("1", "2", "3", ...) for index. The name is the 
		 * string equivalent to the index then.
		 *
		 * @param index
		 *   Index for which to set a default integer name
		 */
		void SetDefaultStateName(Idx index);

		/**
		 * Clear name for state index 
		 *
		 * @param index
		 *   Index which name should be cleared
		 */
		void ClrStateName(Idx index) {
			mNameMap.erase(index);
		}
		
		/**
		 * Remove all names from global state indices
		 */
		void ClrStateNames(void);
};

} // namespace faudes

#define SYMBOLTABLE_H
#endif
