/* token.cpp -- token class for file io */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "token.h"

namespace faudes {

// Token::Token()
Token::Token(void) 
		: mType(None), mStringValue(""), mIndexValue(0) { }

// Token::Token(type, name)
Token::Token(TokenType type, const std::string& rName)
		: mType(type), mStringValue(rName), mIndexValue(0) {
	// check fi the type is correct
  assert((mType != Index) && (mType != None));
}

// Token::Token(type, index)
Token::Token(TokenType type, Idx index)
		: mType(type), mStringValue(""), mIndexValue(index) {
	// check if the type is correct
  assert(mType == Index);
}

// destructor
Token::~Token(void){
}

// ReadIndex(rStream)
bool Token::ReadIndex(std::ifstream& rStream){
	char c;
	Idx v=0;
	bool ok=false;
	// check the whole rStream
	while(rStream){
	        // check eof
   	        if (rStream.eof()) {mType=None; return(false);}; 
		// check next charakter
		c = rStream.peek(); 
		// break if space
		if (isspace(c)) break;
		// break if it is not a digit
		if (!isdigit(c)) break;
		// get the next charakter
		rStream.get();
		// compute the corresponding Idx
		v = 10 * v + ( c - '0');
		ok = true; 
	} // termination with the end of the rStream or a non-digit or a space
	// assign the Idx value in Token
	mIndexValue = v;
	return(ok);
}

// Write(pStream)
void Token::Write(std::ofstream* pStream){
	// distuingish the token type
	switch (mType) {
		// <mStringValue>	
		case Begin:
			*pStream << '<' << mStringValue << '>';
			break;
		// </mStringValue>
		case End:
			*pStream << "</" << mStringValue << '>';
			break;
		// "mStringValue"
		case Name:
			*pStream <<  ExpandString("\""+mStringValue+"\"", FD_NAMELEN);
			break;
		// mIndexValue
		case Index:
			*pStream << ExpandString(ToString(mIndexValue), FD_NAMELEN);
			break;
		// all other unexpected cases
		default:
			assert(0);
	}
}

// Write()
void Token::Write(void){
	// distuingish the token type
	switch(mType) {
		// <mStringValue>	
		case Begin:
			std::cout << '<' << mStringValue << '>';
			break;
		// </mStringValue>
		case End:
			std::cout << "</" << mStringValue << '>';
			break;
		// "mStringValue"
		case Name:
			std::cout <<  ExpandString("\""+mStringValue+"\"",FD_NAMELEN);
			break;
		// mIndexValue
		case Index:
			std::cout << ExpandString(ToString(mIndexValue), FD_NAMELEN);
			break;
		// all other unexpected cases
		default:
			assert(0);
	}
}


// ReadString(rStream, char)
bool Token::ReadString(std::ifstream& rStream, char stop){
	char c;
	std::string v = "";
	bool ok = false;
	// check the whole rStream
	while (rStream) {
	        // check eof
   	        if (rStream.eof()) {mType=None; return(false);}; 
		// look one charakter ahead
		c = rStream.peek();
		// unsuccessful termination if c is a control character
		if (iscntrl(c)) return(0);
		// else get the next character
		rStream.get(); 
		// break if the stop character is reached
		if(c == stop) break; 
		// else append the character to the current string
		v.append(1,c);
		ok = true;
		// termination if the end of the rStream or the stop character is reached
	} 
	// assign the string value in Token
	mStringValue = v;
	return(ok);
}


// ReadSpace(rStream)
int Token::ReadSpace(std::ifstream& rStream){
	char c = '\0';
	int lc = 0;
	// check the whole rStream
	while (rStream) {
		while (rStream) {
 	                // check eof
   	                if (rStream.eof()) {return(lc);}; 
			// look one character ahead
			c = rStream.peek();
			// break if the next character is 
			if((! isspace(c)) && (! iscntrl(c))) break;
			// not a space or a control character
			// get the next space or control character
			rStream.get();
			// count the next lines
			if(c == '\n') ++lc;
		}
		// if the next charakter starts a comment
		if (c != '%') break;
		while (rStream) {
 	                // check eof
   	                if (rStream.eof()) {return(lc);}; 
			// get the next character
			c = rStream.get();
			// count the next lines
			if (c == '\n') ++lc;
			// terminate with the next control character
			if (iscntrl(c)) break;
		}
	}
	// termination if the next character is not a space, a control character 
	//or a '%'.
	return (lc); 
}

// Read(rStream)
int Token::Read(std::ifstream& rStream){
	char c1, c2;
	int lc = 0;
	// the token is initialized with the type "None"
	mType = None;
 	// check eof
   	if (rStream.eof()) return(lc); 
	// read all spaces in the rStream
	lc += ReadSpace(rStream);
 	// check eof
   	if (rStream.eof()) return(lc); 
	// get the first useful character
	c1 = rStream.peek();
	if (isdigit(c1)) {
		// token is an index if the character is a digit
		if (ReadIndex(rStream)) mType = Index;
		// character is a digit
	} 
	// token is a string if it starts with '"'
	else if (c1 == '"') {
		rStream.get();
		// read the string until '"'
		if (ReadString(rStream,'"')) mType = Name;
	} 
	else if (c1 == '<') {
		rStream.get();
    	        // check eof
   	        if (rStream.eof()) return(lc); 
		c2 = rStream.peek(); 
		// token is of type "begin" if it starts with <
		if (c2 != '/') {
			if (ReadString(rStream,'>')) mType = Begin;
		} 
		else {  
			rStream.get();
			// token is of type "end" if it starts with </
			c2=rStream.peek();
			// read the string until '>'
			if (ReadString(rStream,'>')) mType = End;
		}
	}
	FD_DV("Token::Read: char1='" << c1 << "' mType=" << (int) mType 
			<< "\" mStringValue=\"" << mStringValue 
			<< "\" mIndexValue=" <<mIndexValue <<"\n");
	return(lc);
}

} // namespace faudes
