/* token.h -- token class for file io */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef TOKEN_H

#include "definitions.h"
#include "debug.h"
#include "stringhelper.h"
#include <string>
#include <iostream>
#include <fstream>
#include <assert.h>

namespace faudes {

/**
 * Tokens are used to read / write events and states from / to files. A Token
 * consists of a type and a corresponding index or name.
 *
 * @param ttype
 *   TokenType of the Token
 * @param tsvalue
 *   Token value of type std::string
 * @param tivalue
 *   Token value of type Idx
 */
class Token {
  public:

		/**
		 * Empty constructor
		 */
    Token(void);
		
		/**
		 * Copy constructor
		 */
    Token(const Token& rToken);

		/**
		 * Token types:
		 *   none:    
		 *   begin:   <label>
		 *   name:    "name"
		 *   index:   1234
		 *   end:     </label>
		 */
    enum TokenType {None, Begin, Name, Index, End};

		/**
		 * String constructor
		 * 
		 * @param type
		 *   Token type
		 * @param rName
		 *   String to fill the Token
		 */
    Token(TokenType type, const std::string& rName);

		/**
		 * Index constructor
		 *
		 * @param type
		 *  Token type
		 * @param idx
		 *   Index to fill the Token
		 */
    Token(TokenType type, Idx idx);

		/**
		 * Token destructor
		 */
    ~Token(void);
    
		/**
		 * Get index value of a indextoken
		 *
		 * @return
		 *   Token index value
		 */
    Idx IndexValue(void) const {
			return(mIndexValue);
		}

		/**
		 * Get string value of a stringtoken
		 *
		 * @return
		 *   Token std::string value
		 */
    const std::string& StringValue(void) const {
			return(mStringValue);
		}

		/**
		 * Get token Type
		 *
		 * @return
		 *   Token TokenType
		 */
    TokenType Type(void) const {
			return(mType);
		}

		/**
		 * Read Token from input file stream
		 * 
		 * @param rStream
		 *   Pointer to std::ifstream
		 */
    int Read(std::ifstream& rStream);

		/**
		 * Write Token to output file stream
		 *
		 * @param pStream // FIXME
		 *   Pointer to ofstream
		 */
    void Write(std::ofstream* pStream); // FIXME

		/**
		 * Write Token to stdout. Mainly for debugging purposes.
		 */
		void Write(void);
    
  private:
		/** Token type */
    TokenType mType; 
		
		/** Token std::string value (if Token is of type Name) */
    std::string mStringValue;

		/** Token index value (if Token is of type Index) */
    Idx mIndexValue;

		/**
		 * Read an index value from an input file stream
		 * 
		 * @param rStream
		 *   Reference to std::ifstream
		 * 
		 * @return 
		 *   Success
		 */
    bool ReadIndex(std::ifstream& rStream); 

		/** Read a std::string value from an input file stream
		 *
		 * @param rStream
		 *   Reference to std::ifstream
		 * @param stop
		 *   Stop character
		 * 
		 * @return 
		 *   Success
		 */
    bool ReadString(std::ifstream& rStream, char stop);

		/**
		 * Read (ignore) spaces and comments in an input file stream
		 *
		 * @param rStream
		 *   Reference to std::ifstream
		 *
		 * @return
		 *   Number of lines read
		 */   
    int  ReadSpace(std::ifstream& rStream);
};

} // namespace faudes

#define TOKEN_H
#endif


