/* tokenreader.cpp -- read files with tokens */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


/* 
   + new api based on sections (readbegin, readend, etc)
   + try only to throw faudes::Exception 
 
   tmoor, 20060413; GLPL 2.1 
*/


#include "tokenreader.h"

namespace faudes {

// TokenReader(rFilename)
TokenReader::TokenReader(const std::string& rFilename) 
		: mFileName(rFilename) {
  	mStream.exceptions(std::ios::badbit|std::ios::failbit);
        try{
	        mStream.open(rFilename.c_str());
	} 
        catch (std::ios::failure&) {
		std::stringstream errstr;
                errstr << "Exception opening/reading file in \""<< FileLine() << "\"";
		throw Exception("TokenReader::TokenReader", errstr.str(), 1);
        }
        Rewind();
}

// Rewind()
void TokenReader::Rewind(void) {
        try {
                mStream.seekg(0);
                mHasPeekToken=false;
                mLevel=0;
                mLevelPos.clear();
                mLevelPos.push_back(0);
                mLineCount=1;
        }
        catch (std::ios::failure&) {
		std::stringstream errstr;
                errstr << "Exception opening/reading file in \""<< FileLine() << "\"";
		throw Exception("TokenReader::TokenReader", errstr.str(), 1);
        }
}


// FileName()
std::string TokenReader::FileName(void) const {
	return mFileName;
}

// Peek(token)
bool TokenReader::Peek(Token& token) {
        if(mHasPeekToken) { 
     	        token=mPeekToken;
                return true;
	}
        try{
  	        mLineCount += token.Read(mStream);
	} 
        catch (std::ios::failure&) {
	  	std::stringstream errstr;
                errstr << "Exception opening/reading file in \""<< FileLine() << "\"";
	  	throw Exception("TokenReader::Peek", errstr.str(), 1);
        }
        mPeekToken=token;
  	if (mPeekToken.Type() == Token::None) {
	        return false;
	} 
        mHasPeekToken=true;
	return true;
}

// Get(token)
bool TokenReader::Get(Token& token) {
        bool res;
        // get token either from peek buffer or stream
        if(mHasPeekToken) {   
     	        res=true;
                token=mPeekToken;
                mHasPeekToken=false;
	} else {
                res=Peek(token);
                mHasPeekToken=false;

	}    
        // track level of nested sections
        if(res) {
               if (token.Type() == Token::Begin) {
                      mLevel++;
                      mLevelPos.push_back(mStream.tellg());
	       }
               if (token.Type() == Token::End) {
                      mLevel--;
                      mLevelPos.pop_back();
	       }
	}
        return res;
}

// SeekBegin(label)
bool TokenReader::SeekBegin(const std::string& rLabel) {
	// search for begin at any descending level
        int level=mLevel;
	Token token;
	for (;;) {
	        // exception: did not get a token at all (incl. eof)
	        if(!Peek(token)) {
			std::stringstream errstr;
			errstr << "Section \"" << rLabel << "\" expected at " << FileLine() << "no token at all";
			throw Exception("TokenReader::ReadBegin", errstr.str(), 51);
		}
	        // exception: current section ends
		if ((token.Type() == Token::End) && (mLevel == level)) {
			std::stringstream errstr;
			errstr << "Section \"" << rLabel << "\" expected at " << FileLine() << "current section ended unexpected. Found: " << token.StringValue() << " Type " << token.Type();
			throw Exception("TokenReader::ReadBegin", errstr.str(), 51);
		}
                Get(token);
	        // success: found begin section
		if ((token.Type() == Token::Begin) && (token.StringValue() == rLabel)) {
                        return true;
		}
	}
	return false;
}
 
// ReadBegin(label)
bool TokenReader::ReadBegin(const std::string& rLabel) {
        FD_DG("looking for Section \"" << rLabel << "\"");
        try {
               int level=mLevel;
               bool firstgo=true;
	       long int startpos=mStream.tellg();  
               FD_DG("section level " << level << " current pos " << startpos << " begin of section " << mLevelPos[level]);
    	       Token token;
    	       // search for begin at current level
	       for (;;) {
	               // exception: did not get a token at all (incl eof)
	               if(!Peek(token)) {
			       std::stringstream errstr;
			       errstr << "Section \"" << rLabel << "\" expected at " << FileLine() << " no token at all";
			       throw Exception("TokenReader::ReadBegin Peek", errstr.str(), 51);
		       }
	               // success: found begin section
		       if ((token.Type() == Token::Begin) && (token.StringValue() == rLabel) && (mLevel==level)) {
                               Get(token);
                               return true;
		       }
	               // rewind once when current section ends
		       if ((token.Type() == Token::End) && (mLevel == level) && firstgo) {
		               mStream.seekg(mLevelPos[level]);                       
                               firstgo=false;
                               mHasPeekToken=false;
                               continue;
		       }
                       // exception: did not find begin label
                       if((mStream.tellg()>startpos) && (!firstgo)) {
			       std::stringstream errstr;
			       errstr << "Section \"" << rLabel << "\" expected at " << FileLine() << " did not find begin lable";
			       throw Exception("TokenReader::ReadBegin Missed", errstr.str(), 51);
		       }
                       // skip this token
                       Get(token);
	       }
	}
        // catch my seek/tell errors
        catch (std::ios::failure&) {
	       std::stringstream errstr;
	       errstr << "Section \"" << rLabel << "\" expected at " << FileLine();
	       throw Exception("TokenReader::ReadBegin Rewind", errstr.str(), 51);
        }
	return false;

}
 

// Eos(label)
bool TokenReader::Eos(const std::string& rLabel) {
	// peek token and check for end of section
	Token token;
        Peek(token);
	if ((token.Type() == Token::End) && (token.StringValue() == rLabel))
                return true; 
        return false;
}
  

// ReadEnd(label)
bool TokenReader::ReadEnd(const std::string& rLabel) {
	// search for end at current level
        int level=mLevel;
	Token token;
	for (;;) {
	        // exception: did not get a token at all
	        if(!Get(token)) {
			std::stringstream errstr;
			errstr << "end of Section \"" << rLabel << "\" expected at " << FileLine();
			throw Exception("TokenReader::ReadEnd", errstr.str(), 51);
		}
	        // success: found end of current section
		if ((token.Type() == Token::End) && (token.StringValue() == rLabel) && (mLevel==level-1)) {
                        return true;
       		}
	        // exception: current section ends with unexpected label
                if(mLevel<level) {
			std::stringstream errstr;
			errstr << "end of Section \"" << rLabel << "\" expected at " << FileLine();
			throw Exception("TokenReader::ReadEnd", errstr.str(), 51);
		        return false;
		}
	}
	return false;
}

// ReadIndex()
Idx TokenReader::ReadIndex(void) {
        Token token;			
        Get(token);
        if(token.Type()!=Token::Index) {
	        std::stringstream errstr;
		errstr << "Index expected at " << FileLine();
		throw Exception("TokenReader::TokenReader", errstr.str(), 50);
	}
        return token.IndexValue();
}

// ReadName()
const std::string& TokenReader::ReadName(void) {
        Token token;			
        Get(token);
        if(token.Type()!=Token::Name) {
	        std::stringstream errstr;
		errstr << "Name expected at " << FileLine();
		throw Exception("TokenReader::TokenReader", errstr.str(), 50);
	}
        mLastString=token.StringValue();
        return(mLastString);
}


 

// line()
int TokenReader::Line(void) const {
	return mLineCount;
}

// FileLine()
std::string TokenReader::FileLine(void) const {
 	return " ("+ mFileName + ":" + ToString(mLineCount) +")";
}

} // namespace faudes
