/* tokenreader.h -- read files with tokens */

/* FAU Discrete Event Systems Library (libfaudes)

	 Copyright (C) 2006  Bernd Opitz
	 Exclusive copyright is granted to Klaus Schmidt

	 This library is free software; you can redistribute it and/or
	 modify it under the terms of the GNU Lesser General Public
	 License as published by the Free Software Foundation; either
	 version 2.1 of the License, or (at your option) any later version.

	 This library is distributed in the hope that it will be useful,
	 but WITHOUT ANY WARRANTY; without even the implied warranty of
	 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	 Lesser General Public License for more details.

	 You should have received a copy of the GNU Lesser General Public
	 License along with this library; if not, write to the Free Software
	 Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef TOKENREADER_H

#include <vector>

#include "definitions.h"
#include "debug.h"
#include "exception.h"
#include "token.h"

namespace faudes {

	/**
	 * Convenience class for reading tokens
	 *
	 * Throws a faudes::Exception when an error occurs.
	 * Catch the exception this way:
	 * 
	 * try {
	 *   <some tokenreader operation>
	 * }
	 * catch (faudes::Exception& ex) {
	 *   cerr << "Error reading file (details: " << ex.What() << ")" << endl;
	 * }
	 */

	class TokenReader {
		public:
			/**
			 * Creates a TokenReader object for reading complete file
			 *
			 * @param rFilename
			 *   File to read
			 *
			 * @exception Exception
			 *   ios exception (id 1)
			 */
			TokenReader(const std::string& rFilename);

			/**
			 * Get state of TokenReader
			 *
			 * @return
			 *   stream.good()
			 */
			bool good(void) const;

			/**
			 * Get the filename
			 * 
			 * @return 
			 *   Filename
			 */
			std::string FileName(void) const;

			/**
			 * Peek next token. False is returned at eof.
			 *
			 * @param token
			 *   Reference to token
			 * 
			 * @exception Exception
			 *   faudes exception ios (id 1)
			 *
			 * @return
			 *   true - if token got a value
			 *   none - if end of tokens is reached
			 */
			bool Peek(Token& token);

			/**
			 * Get next token. False is returned at eof.
			 *
			 * @param token
			 *   Reference to token
			 * 
			 * @exception Exception
			 *   faudes exception ios (id 1)
			 *
			 * @return
			 *   true - if token got a value
			 *   none - if end of tokens is reached
			 */
			bool Get(Token& token);

			/**
			 * Rewind stream (must be a seekable stream)
			 *
			 * @exception Exception
			 *   faudes exception ios (id 1)
			 */
			void Rewind(void);

			/**
			 * Open a section on any descending level. Stops right after end token.
			 *
			 * @param rLabel
			 *   Token label to specify section
			 *
			 * @exception Exception
			 *   Token mismatch (id 50)
			 *   Token label not found (id 51)
			 *
			 * @return
			 *   True when found
			 */
			bool SeekBegin(const std::string& rLabel);

			/**
			 * Open a section on same level. Stops right after end token.
			 *
			 * @param rLabel
			 *   Token label to specify section
			 *
			 * @exception Exception
			 *   Token mismatch (id 50)
			 *   Token label not found (id 51)
			 *
			 * @return
			 *   True when found
			 */
			bool ReadBegin(const std::string& rLabel);

			/**
			 * Close current section. Stops right after end token.
			 *
			 * @param rLabel
			 *   Token label to specify section
			 *
			 * @exception Exception
			 *   Token mismatch (id 50)
			 *   Token label not found (id 51)
			 *
			 * @return
			 *   True when found
			 */
			bool ReadEnd(const std::string& rLabel);

			/**
			 * Peek a token and check wether it ends specified section;
			 * will not get the token
			 *
			 * @param rLabel
			 *   Token label to specify section
			 *
			 * @exception Exception
			 *   faudes exception ios (id 1)
			 *
			 * @return
			 *   True at end of section
			 */
			bool Eos(const std::string& rLabel);

			/**
			 * Read index  token
			 *		  
			 * @exception Exception
			 *   Token mismatch (id 50)
			 *
			 * @return
			 *   value of index token
			 */
			Idx ReadIndex(void);

			/**
			 * Read name token
			 *		  
			 * @exception Exception
			 *   Token mismatch (id 50)
			 *
			 * @return
			 *   value of name token
			 */
			const std::string& ReadName(void);

			/** Operator for get */
			bool operator >> (Token& token) {
				return Get(token);
			}

			/**
			 * Return number of lines read
			 * 
			 * @return
			 *   Number of lines read
			 */ 
			int Line(void) const;

			/**
			 * Return <filename>:<line>
			 */
			std::string FileLine(void) const;

		private:
			/** Filename */
			std::string mFileName;

			/** Input file stream */
			std::ifstream mStream;

			/** Line counter */
			int mLineCount;

			/** Level (of nested sections)*/
			int mLevel;

			/** file positions of sections */
			std::vector<long int> mLevelPos;

			/** recent string buffer */
			std::string mLastString;

			/** validity of peek buffer */
			bool mHasPeekToken;

			/** peek buffer */
			Token mPeekToken;
	};

} // namespace faudes

#define TOKENREADER_H
#endif 
