/* tokenwriter.cpp -- write tokens to file */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#include "tokenwriter.h"

namespace faudes {

// TokenWriter()
TokenWriter::TokenWriter(void) 
		: mFileName("stdout"), mColumns(80/FD_NAMELEN), mColCount(0) {
	mStream.copyfmt(std::cout);
	mStream.clear(std::cout.rdstate());
	typedef std::basic_ios<char> ___basic_ios_char_; // trick for vc++
	mStream.___basic_ios_char_::rdbuf(std::cout.rdbuf());
	
}

// TokenWriter(rFilename, openmode)
TokenWriter::TokenWriter(const std::string& rFilename, 
		std::ios::openmode openmode) 
		: mFileName(rFilename), mColumns(80/FD_NAMELEN), mColCount(0) {
	mStream.exceptions(std::ios::badbit|std::ios::failbit);
	mStream.open(rFilename.c_str(), openmode); 
}

// Columns()
int TokenWriter::Columns(void) const {
	return mColumns;
}

// Columns()
void TokenWriter::Columns(int columns) {
	mColumns = columns;
}

// Write(rToken)
void TokenWriter::Write(Token& rToken) {
    try{
	    if ((rToken.Type() == Token::Begin) || (rToken.Type() == Token::End)) {
    		mStream << std::endl;
    		rToken.Write(&mStream);
    		mStream << std::endl;
    		mColCount = 0;
    	}
    	else {
    		if (mColCount == mColumns) {
    			mColCount = 0;
    			mStream << std::endl;
    		}
    		mColCount++;
    		rToken.Write(&mStream);
    		mStream << " ";
    	}
   	}
    catch (std::ios::failure&) {
        std::stringstream errstr;
        errstr << "Exception opening/writing file \"" << mFileName << "\"";
        throw Exception("Generator::write", errstr.str(), 2);
    }
}

// Write(rName)
void TokenWriter::Write(const std::string& rName) {
    try{
    	if (mColCount == mColumns) {
    		mColCount = 0;
    		mStream << std::endl;
    	}
    	mColCount++;
    	if ((rName == "\n") || (rName == "\r\n")) {
    		mStream << std::endl;
    	}
    	else {
    		mStream << ExpandString("\""+rName+"\"",FD_NAMELEN);
    		mStream << " ";
    	}
   	}
    catch (std::ios::failure&) {
        std::stringstream errstr;
        errstr << "Exception opening/writing file \"" << mFileName << "\"";
        throw Exception("Generator::write", errstr.str(), 2);
    }
}

// Write(index)
void TokenWriter::Write(Idx index) {
    try{
    	if (mColCount == mColumns) {
    		mColCount = 0;
    		mStream << std::endl;
    	}
    	mColCount++;
    	mStream << ExpandString(ToString(index), FD_NAMELEN);
    	mStream << " ";
   	}
    catch (std::ios::failure&) {
        std::stringstream errstr;
        errstr << "Exception opening/writing file \"" << mFileName << "\"";
        throw Exception("Generator::write", errstr.str(), 2);
    }
}

// writebegin(rLabel)
void TokenWriter::WriteBegin(const std::string& rLabel) {
    try{
    	mStream << "<";
    	mStream << rLabel;
    	mStream << ">";
    	mStream << std::endl;
    	mColCount = 0;
   	}
    catch (std::ios::failure&) {
        std::stringstream errstr;
        errstr << "Exception opening/writing file \"" << mFileName << "\"";
        throw Exception("Generator::write", errstr.str(), 2);
    }
}

// writeend(rLabel)
void TokenWriter::WriteEnd(const std::string& rLabel) {
    try{
    	mStream << std::endl;
    	mStream << "</";
    	mStream << rLabel;
    	mStream << ">";
    	mStream << std::endl;
    	mColCount = 0;
   	}
    catch (std::ios::failure&) {
        std::stringstream errstr;
        errstr << "Exception opening/writing file \"" << mFileName << "\"";
        throw Exception("Generator::write", errstr.str(), 2);
    }
}

// destructor
TokenWriter::~TokenWriter(void) {
	mStream.close();
}

} // namespace faudes
