/* tokenwriter.h -- write tokens to file */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef TOKENWRITER_H

#include "definitions.h"
#include "debug.h"
#include "token.h"
#include <fstream>
#include <iostream>

namespace faudes {

/**
 * Convenience class for writing tokens
 *
 * Throws an std::ios::failure exception when an error happens.
 * Catch the exception this way:
 * 
 * try {
 *   <some tokenwriter operation>
 * }
 * catch (std::ios::failure& ex) {
 *   cerr << "Exception opening/writing file" << endl;
 * }
 */
class TokenWriter {
	public:
	/**
	 * Console TokenWriter constructor
	 */
	TokenWriter(void);

	/** 
	 * File TokenWriter constructor
	 *
	 * @param rFilename
	 *   File to write
	 * @param openmode
	 *   std::ios::openmode
	 */
	TokenWriter(const std::string& rFilename, 
			std::ios::openmode openmode = std::ios::out|std::ios::trunc);

	/** Destructor. Calls close */
	~TokenWriter(void);
	
	/**
	 * Get number of columns in a line
	 *
	 * @return
	 *   # of columns in a line
	 */
	int Columns(void) const;

	/**
	 * Set number of columns in a line 
	 *
	 * @param columns
	 *   # of columns in a line
	 */
	void Columns(int columns);

	/**
	 * Write next token
	 *
	 * @param rToken
	 *   Token to write
	 */
	void Write(Token& rToken);

	/**
	 * Write name 
	 *
	 * @param rString
	 *   String to write
	 */
	void Write(const std::string& rString);

	/**
	 * Write index
	 *
	 * @param index
	 *   Index to write
	 */
	void Write(Idx index);

	/**
	 * Write begin label
	 * 
	 * @param rLabel
	 *   End label, e.g. "Alphabet"
	 */
	void WriteBegin(const std::string& rLabel);

	/**
	 * Write end label
	 * 
	 * @param rLabel
	 *   End label, e.g. "Alphabet"
	 */
	void WriteEnd(const std::string& rLabel);
	
	/** 
	 * Operator for writing tokens
	 *
	 * @param rToken
	 *   Token to write
	 */
	void operator << (Token& rToken) {
		Write(rToken);
	}

	/** 
	 * Operator for writing std::strings to a stream 
	 *
	 * @param rString
	 *   String to write
	 */
	void operator << (const std::string& rString) {
		Write(rString);
	}

	/** 
	 * Operator for writing Idxs to a stream
	 *
	 * @param index
	 *   Index to write
	 */
	void operator << (const Idx index) {
		Write(index);
	}

	private:
		/** ofstream object */
		std::ofstream mStream;

		/** Filename */
		std::string mFileName;

		/** Number of columns */
		int mColumns;

		/** Column counter */
		int mColCount;
};

} // namespace faudes

#define TOKENWRITER_H
#endif 

