/* ttransset.h -- store a set of transitions (transition relation) */

/* FAU Discrete Event Systems Library (libfaudes)

   Copyright (C) 2006  Bernd Opitz
   Exclusive copyright is granted to Klaus Schmidt

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA */


#ifndef TTRANSSET_H

#include "definitions.h"
// #include "debug.h"
#include "eventset.h"
#include "stateset.h"
#include "sstream"
#include <set>
#include <iostream>
#include "tokenreader.h"


namespace faudes {

/**
 * Triple for a current state, event and next state; 
 * provides lexographic order
 */
class Transition {

    public:
    /** Creates an empty Transition object */
    Transition(void) {
    }
    
    /*
     * Creates a Transition object with assigned values 
     * @param x1
     *   Index of current state
     * @param ev
     *   event index
     * @param x2
     *   Index of next state
     */
    Transition(Idx x1, Idx ev,Idx x2) :
        X1(x1), Ev(ev), X2(x2) {  }

    /** Lexographic order for sorting container of Transition */
    inline bool operator < (const Transition& othertrans) const {
      if (X1 < othertrans.X1) return(true);
      if (X1 > othertrans.X1) return(false);
      if (Ev < othertrans.Ev) return(true);
      if (Ev > othertrans.Ev) return(false);
      if (X2 < othertrans.X2) return(true);
      return(false);
    } 

    /** Current state */
    Idx X1;
    
    /** Event */
    Idx Ev; 
    
    /** Next state */
    Idx X2;
};




/* Binary predicate for sorting transitions in order Ev, X1, X2 */
template <class Trn=Transition>
struct TSortEvX1X2 {
  inline bool operator () (const Trn& left, const Trn& right) const {
    if (left.Ev < right.Ev) return true;
    if (left.Ev > right.Ev) return false;
    if (left.X1 < right.X1) return true;
    if (left.X1 > right.X1) return false;
    if (left.X2 < right.X2) return true;
    else return false;
  }
};

/* Binary predicate for sorting transitions in order Ev, X2, X1 */
template <class Trn=Transition>
struct TSortEvX2X1 {
  inline bool operator () (const Trn& left, const Trn& right) const {
    if (left.Ev < right.Ev) return true;
    if (left.Ev > right.Ev) return false;
    if (left.X2 < right.X2) return true;
    if (left.X2 > right.X2) return false;
    if (left.X1 < right.X1) return true;
    else return false;
  }
};

/* Binary predicate for sorting transitions in order X2, Ev, X1 */
template <class Trn=Transition>
struct TSortX2EvX1 {
  inline bool operator () (const Trn& left, const Trn& right) const {
    if (left.X2 < right.X2) return true;
    if (left.X2 > right.X2) return false;
    if (left.Ev < right.Ev) return true;
    if (left.Ev > right.Ev) return false;
    if (left.X1 < right.X1) return true;
    else return false;
  }
};

/* Binary predicate for sorting transitions in order X2, X1, Ev */
template <class Trn=Transition>
struct TSortX2X1Ev {
  inline bool operator () (const Trn& left, const Trn& right) const {
    if (left.X2 < right.X2) return true;
    if (left.X2 > right.X2) return false;
    if (left.X1 < right.X1) return true;
    if (left.X1 > right.X1) return false;
    if (left.Ev < right.Ev) return true;
    else return false;
  }
};

/* Binary predicate for sorting transitions in order X1, X2, Ev */
template <class Trn=Transition>
struct TSortX1X2Ev {
  inline bool operator () (const Trn& left, const Trn& right) const {
    if (left.X1 < right.X1) return true;
    if (left.X1 > right.X1) return false;
    if (left.X2 < right.X2) return true;
    if (left.X2 > right.X2) return false;
    if (left.Ev < right.Ev) return true;
    else return false;
  }
};

/* Binary predicate for sorting transitions in order X1, Ev X2.
   Standard sorting for completeness */
template <class Trn=Transition>
struct TSortX1EvX2 {
  inline bool operator () (const Trn& left, const Trn& right) const {
    if (left.X1 < right.X1) return true;
    if (left.X1 > right.X1) return false;
    if (left.Ev < right.Ev) return true;
    if (left.Ev > right.Ev) return false;
    if (left.X2 < right.X2) return true;
    else return false;
  }
};


/** 
 * Container of Transitions
 * Default predicate for sorting is less<Transition>
 */
//template <class Cmp=std::less<Transition>, class Trn=Transition>
template <class Cmp=TSortX1EvX2<Transition>, class Trn=Transition>
class TTransSet {
  public:
    /** Creates an empty TTransSet object */
    TTransSet(void); 

    /** Virtual destructor */
    virtual ~TTransSet() {}
    
    /** 
     * Copy constructor 
     *
     * @param otherTransrel
     */
    TTransSet(const TTransSet& otherTransrel); //

    /** 
     * Number of transitions 
     *
     * @return
     *   Number of transitions in object
     */
    Idx Size(void) const;

    /**
     * Check if transition relation is empty
     *
     * @return
     *   True if there is no transition
     */
    bool Empty(void) const;
    
    /** 
     * Get Transition set 
     *
     * @return
     *   Set of Transition
     */
    std::set<Trn, Cmp> GetTransSet(void) const;



    /** iterator to access transitions */
    typedef typename std::set<Trn,Cmp>::iterator iterator;

    /** reverse_iterator to access transitions */
    typedef typename std::set<Trn,Cmp>::reverse_iterator reverse_iterator;
    
    /** const_iterator to access transitions */
    typedef typename std::set<Trn,Cmp>::const_iterator const_iterator;

    /** const_reverse_iterator to access transitions */
    typedef typename std::set<Trn,Cmp>::const_reverse_iterator 
        const_reverse_iterator;

    
    /** 
     * iterator to begin of set 
     *
     * @return
     *   TTransSet<Cmp, Trn>::iterator
     */
    iterator Begin(void);

    /** 
     * const_iterator to begin of set 
     *
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator Begin(void) const;

    /** 
     * iterator to end of set 
     *
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator End(void);  

    /** 
     * const_iterator to end of set 
     *
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator End(void) const;  

    /**
     * reverse_iterator that yields the transitions in reverse order starting at
     * the last element and ending after the first. See set<...>::RBegin(void).
     * Returns the "begin". 
     *
     * @return
     *   TTransSet<Cmp>::reverse_iterator
     */
     reverse_iterator RBegin(void);

    /**
     * const_reverse_iterator that yields the transitions in reverse order starting at
     * the last element and ending after the first. See set<...>::RBegin(void).
     * Returns the "begin". 
     *
     * @return
     *   TTransSet<Cmp>::const_reverse_iterator
     */
     const_reverse_iterator RBegin(void) const;

    /**
     * reverse_iterator that yields the transitions in reverse order starting at
     * the last element and ending after the first. See set<...>::rend(void).
     * Returns the "end".
     *
     * @return
     *   TTransSet<Cmp>::reverse_iterator
     */
     reverse_iterator REnd(void);

    /**
     * const_reverse_iterator that yields the transitions in reverse order starting at
     * the last element and ending after the first. See set<...>::rend(void).
     * Returns the "end".
     *
     * @return
     *   TTransSet<Cmp>::const_reverse_iterator
     */
     const_reverse_iterator REnd(void) const;

    /** 
     * iterator to first Transition specified by current state X1.
     *
     * @param x1
     *   Predecessor state index
     *
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator Begin(Idx x1);

    /** 
     * const_iterator to first Transition specified by current state X1.
     *
     * @param x1
     *   Predecessor state index
     *
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator Begin(Idx x1) const;

    /** 
     * iterator to first Transition after current state x1.
     *
     * Notice:
     * Set the End(x1) iterator to a variable, so it won't be recalculated
     * every iteration.
     *
     * @param x1
     *   Predecessor state index
     *
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator End(Idx x1);

    /** 
     * const_iterator to first Transition after current state x1.
     *
     * Notice:
     * Set the End(x1) iterator to a variable, so it won't be recalculated
     * every iteration.
     *
     * @param x1
     *   Predecessor state index
     *
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator End(Idx x1) const;

    /** 
     * iterator to first Transitions specified by x1, ev 
     *
     * @param x1
     *   Predecessor state index
     * @param ev
     *   Event index
     *
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator Begin(Idx x1, Idx ev);

    /** 
     * const_iterator to first Transitions specified by x1, ev 
     *
     * @param x1
     *   Predecessor state index
     * @param ev
     *   Event index
     *
     * @return
     *   TTransSet<Cmp>::const_iterator 
     */
    const_iterator Begin(Idx x1, Idx ev) const;

    /** 
     * iterator to first Transition after current state x1, event ev 
     *
     * Notice:
     * Set the End(x1,ev) iterator to a variable, so it won't be 
     * recalculated every iteration.
     *
     * @param x1
     *   Predecessor state index
     * @param ev
     *   Event index
     *
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator End(Idx x1, Idx ev);

    /** 
     * const_iterator to first Transition after current state x1, event ev 
     *
     * Notice:
     * Set the End(x1,ev) iterator to a variable, so it won't be 
     * recalculated every iteration.
     *
     * @param x1
     *   Predecessor state index
     * @param ev
     *   Event index
     *
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator End(Idx x1, Idx ev) const;

    /**
     * iterator to first Transition specified by event ev.
     * Works only on a TransSet sorted by ev first
     * 
     * @param ev
     *   Event index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator BeginByEv(Idx ev);

    /**
     * const_iterator to first Transition specified by event ev.
     * Works only on a TransSet sorted by ev first
     * 
     * @param ev
     *   Event index
     *   
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator BeginByEv(Idx ev) const;

    /**
     * iterator to first Transition specified by event > ev
     * Works only on a TransSet sorted by ev first
     *
     * Notice:
     * Set the EndByEv(ev) iterator to a variable, so it won't be 
     * recalculated in every iteration.
     * 
     * @param ev
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator EndByEv(Idx ev);

    /**
     * const_iterator to first Transition specified by event > ev
     * Works only on a TransSet sorted by ev first
     *
     * Notice:
     * Set the EndByEv(ev) const_iterator to a variable, so it won't be 
     * recalculated in every iteration.
     * 
     * @param ev
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator EndByEv(Idx ev) const;

    /**
     * iterator to first Transition specified by event + x1.
     * Works only on a TransSet sorted by ev, x1
     * 
     * @param ev
     *   Event index
     * @param x1
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator BeginByEvX1(Idx ev, Idx x1);

    /**
     * const_terator to first Transition specified by event + x1.
     * Works only on a TransSet sorted by ev, x1
     * 
     * @param ev
     *   Event index
     * @param x1
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator BeginByEvX1(Idx ev, Idx x1) const;

    /**
     * iterator to first Transition specified by event > ev and
     * successor state > x1
     * Works only on a TransSet sorted by ev, x1
     *
     * Notice:
     * Set the EndByEvX2(ev,x1) iterator to a variable, so it won't be 
     * recalculated in every iteration.
     * 
     * @param ev
     *   Event index
     * @param x1
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator EndByEvX1(Idx ev, Idx x1);

    /**
     * const_iterator to first Transition specified by event > ev and
     * successor state > x1
     * Works only on a TransSet sorted by ev, x1
     *
     * Notice:
     * Set the EndByEvX2(ev,x1) iterator to a variable, so it won't be 
     * recalculated in every iteration.
     * 
     * @param ev
     *   Event index
     * @param x1
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator EndByEvX1(Idx ev, Idx x1) const;

    /**
     * iterator to first Transition specified by event + x2.
     * Works only on a TransSet sorted by ev, x2
     * 
     * @param ev
     *   Event index
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator BeginByEvX2(Idx ev, Idx x2);

    /**
     * const_iterator to first Transition specified by event + x2.
     * Works only on a TransSet sorted by ev, x2
     * 
     * @param ev
     *   Event index
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator BeginByEvX2(Idx ev, Idx x2) const;

    /**
     * iterator to first Transition specified by event > ev and
     * successor state > x2
     * Works only on a TransSet sorted by ev, x2
     *
     * Notice:
     * Set the EndByEvX2(ev,x2) iterator to a variable, so it won't be 
     * recalculated in every iteration.
     * 
     * @param ev
     *   Event index
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator EndByEvX2(Idx ev, Idx x2);

    /**
     * const_iterator to first Transition specified by event > ev and
     * successor state > x2
     * Works only on a TransSet sorted by ev, x2
     *
     * Notice:
     * Set the EndByEvX2(ev,x2) const_iterator to a variable, so it won't be 
     * recalculated in every iteration.
     * 
     * @param ev
     *   Event index
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator EndByEvX2(Idx ev, Idx x2) const;

    /**
     * iterator to first Transition specified by predecessor state x2.
     * Works only on a TransSet sorted by x2 first
     * 
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator BeginByX2(Idx x2);

    /**
     * const_iterator to first Transition specified by predecessor state x2.
     * Works only on a TransSet sorted by x2 first
     * 
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator BeginByX2(Idx x2) const;

    /**
     * iterator to first Transition specified by predecessor state > x2.
     * Works only on a TransSet sorted by x2 first
     *
     * Notice:
     * Set the EndByX2(x2) iterator to a variable, so it won't be 
     * recalculated in every iteration.
     * 
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator EndByX2(Idx x2);

    /**
     * const_iterator to first Transition specified by predecessor state > x2.
     * Works only on a TransSet sorted by x2 first
     *
     * Notice:
     * Set the EndByX2(x2) const_iterator to a variable, so it won't be 
     * recalculated in every iteration.
     * 
     * @param x2
     *   Predecessor state index
     *   
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator EndByX2(Idx x2) const;

    /**
     * iterator to first Transition specified by predecessor state x2 + ev.
     * Works only on a TransSet sorted by x2, ev
     * 
     * @param x2
     *   Predecessor state index
     * @param ev
     *   Event index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator BeginByX2(Idx x2, Idx ev);

    /**
     * const_iterator to first Transition specified by predecessor state x2 + ev.
     * Works only on a TransSet sorted by x2, ev
     * 
     * @param x2
     *   Predecessor state index
     * @param ev
     *   Event index
     *   
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator BeginByX2(Idx x2, Idx ev) const;

    /**
     * iterator to first Transition specified by predecessor state > x2 and
     * event > ev.
     * Works only on a TransSet sorted by x2, ev
     *
     * Notice:
     * Set the EndByX2(x2,ev) iterator to a variable, so it won't be 
     * recalculated in every iteration.
     * 
     * @param x2
     *   Predecessor state index
     * @param ev
     *   Event index
     *   
     * @return
     *   TTransSet<Cmp>::iterator
     */
    iterator EndByX2(Idx x2, Idx ev);

    /**
     * const_iterator to first Transition specified by predecessor state > x2 and
     * event > ev.
     * Works only on a TransSet sorted by x2, ev
     *
     * Notice:
     * Set the EndByX2(x2,ev) const_iterator to a variable, so it won't be 
     * recalculated in every iteration.
     * 
     * @param x2
     *   Predecessor state index
     * @param ev
     *   Event index
     *   
     * @return
     *   TTransSet<Cmp>::const_iterator
     */
    const_iterator EndByX2(Idx x2, Idx ev) const;

	void DWrite(TokenWriter& rTwconst,
		    const EventSet& alphabet,
		    const StateSet& states) const;
	
	void Write(TokenWriter& rTw, 
		   const std::map<Idx,Idx>& rMinIndexMap,
		   const EventSet& alphabet,
		   const StateSet& states) const;
	
	void Read(TokenReader& rTr,
		  std::map<Idx,Idx> rMinIndexMap,
		  std::map<std::string,Idx> rTmpStateNameMap,
		  const EventSet& alphabet);

    /** Clear all transitions */
    void Clear(void);
    
    /** 
     * Add a Transition directly 
     * FAUDES_CHECKED turns on existence check for transition.
     * 
     * @param rTransition 
     *   Reference to transition object
     */
    void Insert(const Trn& rTransition);
    
    /** 
     * Add a Transition by indices
     * FAUDES_CHECKED turns on existence check for transition.
     *
     * @param x1
     *   Predecessor state
     * @param ev
     *   Event
     * @param x2
     *   Successor state
     */
    void Insert(Idx x1, Idx ev, Idx x2);
    
    /** 
     * Remove a Transition directly 
     *
     * @return
     *   Success
     */
    bool Erase(const Trn& t);
    
    /** 
     * Remove a Transition by x1, ev, x2
     *
     * @return
     *   Success
     */
    bool Erase(Idx x1, Idx ev, Idx x2);

    /**
     * Remove a Transition by iterator it
     */
    void Erase(iterator it);

    /**
     * Remove all transitions containing predecessor state x1
     *
     * @param x1
     *   State index
     */
    void EraseByX1(Idx x1);

    /**
     * Remove all transitions containing successor state x2
     *
     * @param x2
     *   State index
     */
    void EraseByX2(Idx x2);

    /**
     * Remove all transitions containing event ev
     *
     * @param ev
     *   Event index
     */
    void EraseByEv(Idx ev);
    
    /**
     * Remove all transitions containing state x
     * 
     * @param x
     *   State index
     */
    void EraseByX1OrX2(Idx x1);


    /**
     * Set existence test
     *
     * @param t
     *   Transition
     *
     * @return
     *   bool
     */
    bool ExistsTransition(const Trn& t) const;

    /**
     * Set existence test
     *
     * @param x1
     *   Predecessor state Idx
     * @param ev
     *   Event Idx
     * @param x2
     *   Successor state Idx
     *
     * @return
     *   bool
     */
    bool ExistsTransition(Idx x1, Idx ev, Idx x2) const;

    /**
     * Tests if a transition with predecessor or successor state
     * x exists.
     *
     * @param x
     *   State Idx
     *
     * @return
     *   bool
     */
    bool ExistsTransition(Idx x) const;


    /**
     * Get x1, ev, x2 sorted transition relation
     *
     * @return
     *   Transition relation sorted by x1, ev, x2
     */
    TTransSet<TSortX1EvX2<Trn>, Trn> TransSetX1EvX2(void) const;

    /**
     * Get ev, x1, x2 sorted transition relation
     *
     * @return
     *   Transition relation sorted by ev, x1, x2
     */
    TTransSet<TSortEvX1X2<Trn>, Trn> TransSetEvX1X2(void) const;

    /**
     * Get ev, x2, x1 sorted transition relation
     *
     * @return
     *   Transition relation sorted by ev, x2, x1
     */
    TTransSet<TSortEvX2X1<Trn>, Trn> TransSetEvX2X1(void) const;

    /**
     * Get x2, ev, x1 sorted transition relation
     *
     * @return
     *   Transition relation sorted by x2, ev, x1
     */
    TTransSet<TSortX2EvX1<Trn>, Trn> TransSetX2EvX1(void) const;

    /**
     * Get x2, x1, ev sorted transition relation
     *
     * @return
     *   Transition relation sorted by x2, x1, ev
     */
    TTransSet<TSortX2X1Ev<Trn>, Trn> TransSetX2X1Ev(void) const;

    /**
     * Get x2, x1, ev sorted transition relation
     *
     * @return
     *   Transition relation sorted by x1, x2, ev
     */
    TTransSet<TSortX1X2Ev<Trn>, Trn> TransSetX1X2Ev(void) const;

    /** 
     * Get state space covered by transitions 
     *
     * @return
     *   Set of state indices covered by transitions
     */
    std::set<Idx> StateSpace(void) const;

    /**
     * Get state space covered by x2 states for given x1
     * 
     * @param x1
     *   x1 state
     *
     * @return
     *   Set of state indices
     */
    std::set<Idx> StateSpaceX2(Idx x1) const;

    /**
     * Get state space covered by x2 states for given x1, ev
     * 
     * @param x1
     *   x1 state
     * @param ev
     *   event
     *
     * @return
     *   Set of state indices
     */
    std::set<Idx> StateSpaceX2(Idx x1, Idx ev) const;

    /**
     * Get active event set of state x1
     *
     * @param x1
     *   x1 state
     *
     * @return 
     *   set<Idx> containing active event set of state x1
     */
     EventSet ActiveEvents(Idx x1) const;

    /**
     * Write all transitions into a string
     *
     * @return
     *   String containing all transitions
     */
    virtual std::string ToString(void) const; // FIXME: remove virtual when moved to TransSet

    typedef Trn TransitionType;

  protected:
    /** Set of Transition objects */
    std::set<Trn,Cmp> mTransSet;
};


/** Type definition for default sorted TTransSet */
// typedef TTransSet<> TransSet;

// typedef TTransSet<std::less<Transition>, Transition> TransSet;

typedef TTransSet<TSortX1EvX2<Transition>, Transition> TransSet;

/** Type definition for default sorted TTransSet */
typedef TTransSet<TSortX1EvX2<Transition>, Transition> TransSetX1EvX2;

/** Type definition for ev, x1, x2 sorted TTransSet */
typedef TTransSet<TSortEvX1X2<Transition>, Transition> TransSetEvX1X2;

/** Type definition for ev, x2, x1 sorted TTransSet */
typedef TTransSet<TSortEvX2X1<Transition>, Transition> TransSetEvX2X1;

/** Type definition for x2, ev, x1 sorted TTransSet */
typedef TTransSet<TSortX2EvX1<Transition>, Transition> TransSetX2EvX1;

/** Type definition for x2, x1, ev sorted TTransSet */
typedef TTransSet<TSortX2X1Ev<Transition>, Transition> TransSetX2X1Ev;

/** Type definition for x1, x2, ev sorted TTransSet */
typedef TTransSet<TSortX1X2Ev<Transition>, Transition> TransSetX1X2Ev;




// TTransSet<Cmp>::Transrel(void)
template<class Cmp, class Trn> 			
TTransSet<Cmp, Trn>::TTransSet(void) {
}

// TTransSet<Cmp>::TTransSet(othertransrel)
template<class Cmp, class Trn> 
TTransSet<Cmp, Trn>::TTransSet(const TTransSet& otherTransrel) {
  iterator it;
  mTransSet = otherTransrel.GetTransSet();
}

// Size()
template<class Cmp, class Trn> 
Idx TTransSet<Cmp, Trn>::Size(void) const {
  return (Idx)mTransSet.size();
}

// Empty()
template<class Cmp, class Trn>
bool TTransSet<Cmp, Trn>::Empty(void) const {
  return mTransSet.empty();
}

// TransSet()	
template<class Cmp, class Trn> 
std::set<Trn, Cmp> TTransSet<Cmp, Trn>::GetTransSet(void) const {
  return mTransSet;
}

// iterator Begin()
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::Begin(void) {
  return mTransSet.begin();
}

// iterator Begin() const
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::Begin(void) const {
  return mTransSet.begin();
}

// iterator End()
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::End(void) {
  return mTransSet.end();
}

// iterator End() const
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::End(void) const {
  return mTransSet.end();
}

// reverse_iterator RBegin()
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::reverse_iterator TTransSet<Cmp, Trn>::RBegin(void) {
  return mTransSet.rbegin();
}

// reverse_iterator RBegin()
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::const_reverse_iterator TTransSet<Cmp, Trn>::RBegin(void) const {
  return mTransSet.rbegin();
}

// iterator REnd()
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::reverse_iterator TTransSet<Cmp, Trn>::REnd(void) {
  return mTransSet.rend();
}

// iterator REnd() const
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::const_reverse_iterator TTransSet<Cmp, Trn>::REnd(void) const {
  return mTransSet.rend();
}

// iterator Begin(x1)
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::Begin(Idx x1) {
  Trn tlx(x1,0,0);
  return mTransSet.lower_bound(tlx);
}

// iterator Begin(x1) const
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::Begin(Idx x1) const {
  Trn tlx(x1,0,0);
  return mTransSet.lower_bound(tlx);
}

// iterator End(x1)
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::End(Idx x1) {
  Trn tlx(x1+1,0,0);
  return mTransSet.lower_bound(tlx);
}

// iterator End(x1) const
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::End(Idx x1) const {
  Trn tlx(x1+1,0,0);
  return mTransSet.lower_bound(tlx);
}

// iterator Begin(x1,ev)
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::Begin(Idx x1, Idx ev) {
  Trn tlx(x1,ev,0);
  return mTransSet.lower_bound(tlx);
}

// iterator Begin(x1,ev) const
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::Begin(Idx x1, Idx ev) const {
  Trn tlx(x1,ev,0);
  return mTransSet.lower_bound(tlx);
}

// iterator End(x1,ev)
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::End(Idx x1, Idx ev) {
  Trn tlx(x1,ev+1, 0);
  return mTransSet.lower_bound(tlx);
}

// iterator End(x1,ev) const
template<class Cmp, class Trn> 
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::End(Idx x1, Idx ev) const {
  Trn tlx(x1,ev+1, 0);
  return mTransSet.lower_bound(tlx);
}

// iterator BeginByEv(ev)
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::BeginByEv(Idx ev) {
  Trn tlx(0,ev,0);
  return mTransSet.lower_bound(tlx);
}

// iterator BeginByEv(ev) const
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::BeginByEv(Idx ev) const {
  Trn tlx(0,ev,0);
  return mTransSet.lower_bound(tlx);
}

// iterator EndByEv(ev)
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::EndByEv(Idx ev) {
  Trn tlx(0,ev+1,0);
  return mTransSet.lower_bound(tlx);
}

// iterator EndByEv(ev) const
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::EndByEv(Idx ev) const {
  Trn tlx(0,ev+1,0);
  return mTransSet.lower_bound(tlx);
}

// iterator BeginByEvX1(ev,x1)
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::BeginByEvX1(Idx ev, Idx x1) {
  Trn tlx(x1,ev,0);
  return mTransSet.lower_bound(tlx);
}

// iterator BeginByEvX1(ev,x1) const
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::BeginByEvX1(Idx ev, Idx x1) const {
  Trn tlx(x1,ev,0);
  return mTransSet.lower_bound(tlx);
}

// iterator EndByEvX1(ev,x1)
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::EndByEvX1(Idx ev, Idx x1) {
  Trn tlx(x1+1,ev,0);
  return mTransSet.lower_bound(tlx);
}

// iterator EndByEvX1(ev,x1) const
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::EndByEvX1(Idx ev, Idx x1) const {
  Trn tlx(x1+1,ev,0);
  return mTransSet.lower_bound(tlx);
}

// iterator BeginByEvX2(ev,x2)
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::BeginByEvX2(Idx ev, Idx x2) {
  Trn tlx(0,ev,x2);
  return mTransSet.lower_bound(tlx);
}

// iterator BeginByEvX2(ev,x2) const
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::BeginByEvX2(Idx ev, Idx x2) const {
  Trn tlx(0,ev,x2);
  return mTransSet.lower_bound(tlx);
}

// iterator EndByEvX2(ev,x2)
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::EndByEvX2(Idx ev, Idx x2) {
  Trn tlx(0,ev,x2+1);
  return mTransSet.lower_bound(tlx);
}

// iterator EndByEvX2(ev,x2) const
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::EndByEvX2(Idx ev, Idx x2) const {
  Trn tlx(0,ev,x2+1);
  return mTransSet.lower_bound(tlx);
}

// iterator BeginByX2(x2)
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::BeginByX2(Idx x2) {
  Trn tlx(0,0,x2);
  return mTransSet.lower_bound(tlx);
}

// iterator BeginByX2(x2) const
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::BeginByX2(Idx x2) const {
  Trn tlx(0,0,x2);
  return mTransSet.lower_bound(tlx);
}

// iterator EndByX2(x2)
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::EndByX2(Idx x2) {
  Trn tlx(0,0,x2+1);
  return mTransSet.lower_bound(tlx);
}

// iterator EndByX2(x2) const
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::EndByX2(Idx x2) const {
  Trn tlx(0,0,x2+1);
  return mTransSet.lower_bound(tlx);
}

// iterator BeginByX2(x1,ev)
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::BeginByX2(Idx x2, Idx ev) {
  Trn tlx(0,ev,x2);
  return mTransSet.lower_bound(tlx);
}

// iterator BeginByX2(x1,ev) const
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::BeginByX2(Idx x2, Idx ev) const {
  Trn tlx(0,ev,x2);
  return mTransSet.lower_bound(tlx);
}

// iterator EndByX2(x1,ev)
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::iterator TTransSet<Cmp, Trn>::EndByX2(Idx x2, Idx ev) {
  Trn tlx(0,ev+1,x2);
  return mTransSet.lower_bound(tlx);
}

// iterator EndByX2(x1,ev) const
template<class Cmp, class Trn>
typename TTransSet<Cmp, Trn>::const_iterator TTransSet<Cmp, Trn>::EndByX2(Idx x2, Idx ev) const {
  Trn tlx(0,ev+1,x2);
  return mTransSet.lower_bound(tlx);
}

template<class Cmp, class Trn>
void TTransSet<Cmp, Trn>::DWrite(TokenWriter& rTw,
		      const EventSet& alphabet,
		      const StateSet& states) const

{
  const_iterator tit;
  int oldcolumns = rTw.Columns();
  rTw.Columns(3);
  rTw.WriteBegin("TransRel");
  const_iterator test = Begin();
  for (tit = Begin(); tit != End(); ++tit) {
    if (states.StateName(tit->X1) == "" ) {
      rTw << tit->X1;
    }
    else {
      rTw << states.StateName(tit->X1) + "(" +faudes::ToString(tit->X1) + ")";
    }
    rTw << alphabet.EventName(tit->Ev);
    if (states.StateName(tit->X2) == "" ) {
      rTw << tit->X2;
    }
    else {
      rTw << states.StateName(tit->X2) + "(" +faudes::ToString(tit->X2) + ")";
    }
  }
  rTw.WriteEnd("TransRel");
  rTw.Columns(oldcolumns);
}


template<class Cmp, class Trn>
void TTransSet<Cmp, Trn>::Write(TokenWriter& rTw, 
		     const std::map<Idx,Idx>& rMinIndexMap,
		     const EventSet& alphabet,
		     const StateSet& states) const
{
  const_iterator tit;
  std::map<Idx,Idx>::const_iterator mapit;
  int oldcolumns = rTw.Columns();
  rTw.Columns(3);
  rTw.WriteBegin("TransRel");
  bool smalltransrel = false;
  if (Size() < FD_SMALLTRANSREL) {
    smalltransrel = true;
  }
  for (tit = Begin(); tit != End(); ++tit) {

    // write x1
    mapit = rMinIndexMap.find(tit->X1);
    if (mapit == rMinIndexMap.end()) {
      std::cerr << "programming error: index " << tit->X1
          << " could not be found in minindexmap" << std::endl;
      abort();
    }
    if (smalltransrel) {
      std::string x1 = states.StateName(tit->X1);
      if (x1 != "") {
        rTw << x1;
      }
      else {
        rTw << mapit->second;
      }
    }
    else {
      rTw << mapit->second;
    }

    // write ev
    rTw << alphabet.EventName(tit->Ev);

    // write x2
    mapit = rMinIndexMap.find(tit->X2);
    if (mapit == rMinIndexMap.end()) {
      std::cerr << "programming error: index " << tit->X2
          << " could not be found in minindexmap" << std::endl;
      abort();
    }
    if (smalltransrel) {
      std::string x2 = states.StateName(tit->X2);
      if (x2 != "") {
        rTw << x2;
      }
      else {
        rTw << mapit->second;
      }
    }
    else {
      rTw << mapit->second;
    }

  }
  rTw.WriteEnd("TransRel");
  rTw.Columns(oldcolumns);
}



template<class Cmp, class Trn>
void TTransSet<Cmp, Trn>::Read(TokenReader& rTr,
		    std::map<Idx,Idx> rMinIndexMap,
		    std::map<std::string,Idx> rTmpStateNameMap,
		    const EventSet& alphabet)

{

  FD_DV("TransRel(" << this << ")::Read(\"" << rTr.FileName() << "\")");
  rTr.SeekBegin("TransRel");    // better use "ReadBegin" ?
  // HELPERS:
  Token token;
  Idx x1 = 0, ev = 0, x2 = 0;
  int readstate = 0; // 1, 2 or 3 (x1, ev or x2)
  // ALGORITHM: 
  Clear();
  do {
    switch (readstate) {
      // 0: just read a token first
      case 0:
        break;
        // 1: the x1 token
      case 1: 
        if (token.Type() == Token::Index) {
          std::map<Idx,Idx>::const_iterator it = rMinIndexMap.find(token.IndexValue());
          if (it != rMinIndexMap.end()) {
            x1 = it->second;
          }
          else {
            std::stringstream errstr;
            errstr << "token " << token.IndexValue() << " not found in rMinIndexMap "
              << rTr.FileLine();
            throw Exception("TransRel::Read", errstr.str(), 52);
          }
        }
        else if (token.Type() == Token::Name) {
          std::map<std::string,Idx>::const_iterator smit = rTmpStateNameMap.find(token.StringValue());
          if (smit != rTmpStateNameMap.end()) {
            x1 = smit->second;
          }
          else {
            std::stringstream errstr;
            errstr << "token " << token.StringValue() << " not known in stateset " 
              << rTr.FileLine();
            throw Exception("TransRel::Read", errstr.str(), 52);
          }
        }
        break;
        // 2: the ev token
      case 2: 
        if (token.Type() == Token::Index) {
          std::stringstream errstr;
          errstr << "No event indices in transrel allowed "	<< rTr.FileLine();
          throw Exception("TransRel::Read", errstr.str(), 50);
        }
        else if (token.Type() == Token::Name) {
          ev = alphabet.EventIndex(token.StringValue());
          if (ev == 0) {
            std::stringstream errstr;
            errstr << "token " << token.StringValue() << " not known in mAlphabet" 
              << rTr.FileLine();
            throw Exception("TransRel::Read", errstr.str(), 50);
          }
        }
        break;
        // 3: the x2 token
      case 3:
        if (token.Type() == Token::Index) {
          std::map<Idx,Idx>::const_iterator it = rMinIndexMap.find(token.IndexValue());
          if (it != rMinIndexMap.end()) {
            x2 = it->second;
          }
          else {
            std::stringstream errstr;
            errstr << "token " << token.IndexValue() << " not in rMinIndexMap" 
              << rTr.FileLine();
            throw Exception("TransRel::Read", errstr.str(), 52);
          }
        }
        else if (token.Type() == Token::Name) {
          std::map<std::string,Idx>::const_iterator smit = rTmpStateNameMap.find(token.StringValue());
          if (smit != rTmpStateNameMap.end()) {
            x2 = smit->second;
          }
          else {
            std::stringstream errstr;
            errstr << "token " << token.StringValue() << " not known in stateset " 
              << rTr.FileLine();
            throw Exception("TransRel::Read", errstr.str(), 52);
          }
        }
        // add transition to transrel
        Insert(x1, ev, x2);
        // reset readstate after x2 token
        readstate = 0;
        break;
        // default: should not happen
      default:
        std::cerr << "should not happen" << std::endl;
        assert(false);
        break;
    }
    readstate++;
  } while ((rTr >> token) && (token.Type() != Token::End));
  // check if next expected token would have been a a x1 token
  if (readstate != 1) {
    std::stringstream errstr;
    errstr << "Wrong number of tokens in TransRel " << rTr.FileLine();
    throw Exception("TransRel::Read", errstr.str(), 50);
  }

}



// Clear()
template<class Cmp, class Trn>
void TTransSet<Cmp, Trn>::Clear(void) {
  FD_DC("TTransSet(" << this << ")::Clear()");
  mTransSet.clear();
}

// insert(transition)
template<class Cmp, class Trn> 
void TTransSet<Cmp, Trn>::Insert(const Trn& t) {
  mTransSet.insert(t);
}

// insert(x1,ev,x2)
template<class Cmp, class Trn> 
void TTransSet<Cmp, Trn>::Insert(Idx x1, Idx ev, Idx x2) {
  FD_DC("TTransSet(" << this << ")::Insert(" << x1 << "-" << ev << "-" << x2 << ")");
  mTransSet.insert(Trn(x1, ev, x2));
}

// Erase(transition)
template<class Cmp, class Trn> 
bool TTransSet<Cmp, Trn>::Erase(const Trn& t) {
  typename TTransSet<Cmp, Trn>::iterator it = mTransSet.find(t);
  if (it == End()) {
    return false;
  }
  else {
    mTransSet.erase(it);
    return true;
  }
}
    
// Erase(x1,ev,x2)
template<class Cmp, class Trn> 
bool TTransSet<Cmp, Trn>::Erase(Idx x1, Idx ev, Idx x2) {
  FD_DC("TTransSet(" << this << ")::Erase(" << x1 << "-" << ev << "-" << x2 << ")");
  Trn t(x1, ev, x2);
  typename TTransSet<Cmp, Trn>::iterator it = mTransSet.find(t);
  if (it == End()) {
    return false;
  }
  mTransSet.erase(it);
  return true;
}

// Erase(it)
template<class Cmp, class Trn> 
void TTransSet<Cmp, Trn>::Erase(iterator it) {
  FD_DC("TTransSet(" << this << ")::Erase(" << it->X1 << "-" << it->Ev 
      << "-" << it->X2 << ")");
  mTransSet.erase(it);
}

// EraseByX1(x)
template<class Cmp, class Trn> 
void TTransSet<Cmp, Trn>::EraseByX1(Idx x1) {
  FD_DC("TTransSet(" << this << ")::EraseByX1(" << x1 << ")");
  iterator lower, upper;
  Trn tl(x1,0,0);
  Trn tu(x1+1,0,0);
  lower = mTransSet.lower_bound(tl);
  upper = mTransSet.upper_bound(tu);
  mTransSet.erase(lower, upper);
}

// EraseByX2(x)
template<class Cmp, class Trn> 
void TTransSet<Cmp, Trn>::EraseByX2(Idx x2) {
  FD_DC("TTransSet(" << this << ")::EraseByX2(" << x2 << ")");
  iterator it, tmpit;
  it = Begin();
  while (it != End()) {
    if (it->X2 == x2) {
      tmpit = it;
      it++;
      mTransSet.erase(tmpit);
    }
    else {
      it++;
    }
  }
}

// EraseByEv(ev)
template<class Cmp, class Trn> 
void TTransSet<Cmp, Trn>::EraseByEv(Idx ev) {
  FD_DC("TTransSet(" << this << ")::EraseByEv(" << ev << ")");
  iterator it, tmpit;
  it = Begin();
  while (it != End()) {
    if (it->Ev == ev) {
      tmpit = it;
      it++;
      mTransSet.erase(tmpit);
    }
    else {
      it++;
    }
  }
}



// EraseByX1OrX2(x)
template<class Cmp, class Trn> 
void TTransSet<Cmp, Trn>::EraseByX1OrX2(Idx x) {
  FD_DC("TTransSet(" << this << ")::EraseByX1OrX2(" << x << ")");
  iterator it, tmpit;
  it = Begin();
  while (it != End()) {
    if ((it->X1 == x) || (it->X2 == x)) {
      tmpit = it;
      it++;
      mTransSet.erase(tmpit);
    }
    else {
      it++;
    }
  }
}

// ExistsTransition(t)()
template<class Cmp, class Trn>
bool TTransSet<Cmp, Trn>::ExistsTransition(const Trn& t) const {
  if (mTransSet.find(t) != mTransSet.end()) {
    return true;
  }
  else {
    return false;
  }
}

// ExistsTransition(x1, ev, x2)()
template<class Cmp, class Trn>
bool TTransSet<Cmp, Trn>::ExistsTransition(Idx x1, Idx ev, Idx x2) const {
  if (mTransSet.find(Trn(x1,ev,x2)) != mTransSet.end()) {
    return true;
  }
  else {
    return false;
  }
}

// ExistsTransition(x)()
template<class Cmp, class Trn>
bool TTransSet<Cmp, Trn>::ExistsTransition(Idx x) const {
  const_iterator it;
  for (it = Begin(); it != End(); ++it) {
    if ((it->X1 == x) || (it->X2 == x)) {
      return true;
    }
  }
  return false;
}


// TransSetX1EvX2()
template<class Cmp, class Trn>
TTransSet<TSortX1EvX2<Trn>, Trn> TTransSet<Cmp, Trn>::TransSetX1EvX2(void) const {
  TTransSet<TSortX1EvX2<Trn>, Trn> ret;
  const_iterator it;
  for (it = Begin(); it != End(); ++it) {
    ret.Insert(it->X1, it->Ev, it->X2);
  }
  return ret;
}

// TransSetEvX1X2()
template<class Cmp, class Trn>
TTransSet<TSortEvX1X2<Trn>, Trn> TTransSet<Cmp, Trn>::TransSetEvX1X2(void) const {
  TTransSet<TSortEvX1X2<Trn>, Trn> ret;
  const_iterator it;
  for (it = Begin(); it != End(); ++it) {
    ret.Insert(it->X1, it->Ev, it->X2);
  }
  return ret;
}

// TransSetEvX2X1()
template<class Cmp, class Trn>
TTransSet<TSortEvX2X1<Trn>, Trn> TTransSet<Cmp, Trn>::TransSetEvX2X1(void) const {
  TTransSet<TSortEvX2X1<Trn>, Trn> ret;
  const_iterator it;
  for (it = Begin(); it != End(); ++it) {
    ret.Insert(it->X1, it->Ev, it->X2);
  }
  return ret;
}

// TransSetX2EvX1()
template<class Cmp, class Trn>
TTransSet<TSortX2EvX1<Trn>, Trn> TTransSet<Cmp, Trn>::TransSetX2EvX1(void) const {
  TTransSet<TSortX2EvX1<Trn>, Trn> ret;
  const_iterator it;
  for (it = Begin(); it != End(); ++it) {
    ret.Insert(it->X1, it->Ev, it->X2);
  }
  return ret;
}

// TransSetX2X1Ev
template<class Cmp, class Trn>
TTransSet<TSortX2X1Ev<Trn>, Trn> TTransSet<Cmp, Trn>::TransSetX2X1Ev(void) const {
  TTransSet<TSortX2X1Ev<Trn>, Trn> ret;
  const_iterator it;
  for (it = Begin(); it != End(); ++it) {
    ret.Insert(it->X1, it->Ev, it->X2);
  }
  return ret;
}

// TransSetX1X2Ev
template<class Cmp, class Trn>
TTransSet<TSortX1X2Ev<Trn>, Trn> TTransSet<Cmp, Trn>::TransSetX1X2Ev(void) const {
  TTransSet<TSortX1X2Ev<Trn>, Trn> ret;
  const_iterator it;
  for (it = Begin(); it != End(); ++it) {
    ret.Insert(it->X1, it->Ev, it->X2);
  }
  return ret;
}

// StateSpace()
template<class Cmp, class Trn> 
std::set<Idx> TTransSet<Cmp, Trn>::StateSpace(void) const {
  std::set<Idx> states;
  const_iterator it;
  for (it=Begin(); it!=End(); ++it) {
    states.insert(it->X1);
    states.insert(it->X2);
  }
  return states;
}

// StateSpaceX2(x1)
template<class Cmp, class Trn> 
std::set<Idx> TTransSet<Cmp, Trn>::StateSpaceX2(Idx x1) const {
  std::set<Idx> states;
  iterator it = Begin(x1);
  iterator it_end = End(x1);
  while (it != it_end) {
    states.insert(it->X2);
    ++it;
  }
  return states;
}

// StateSpaceX2(x1, ev)
template<class Cmp, class Trn> 
std::set<Idx> TTransSet<Cmp, Trn>::StateSpaceX2(Idx x1, Idx ev) const {
  std::set<Idx> states;
  iterator it = Begin(x1, ev);
  iterator it_end = End(x1, ev);
  while (it != it_end) {
    states.insert(it->X2);
    ++it;
  }
  return states;
}

template<class Cmp, class Trn> 
EventSet TTransSet<Cmp, Trn>::ActiveEvents(Idx x1) const {
  const_iterator it = Begin(x1);
  const_iterator it_end = End(x1);
  EventSet result;
  for (; it != it_end; ++it) {
    result.InsEvent(it->Ev);
  }
  return result;
}


// ToString() // FIXME: move to TransSet && find out how to make TimedTransSet to String
template<class Cmp, class Trn>
std::string TTransSet<Cmp, Trn>::ToString(void) const {
  std::ostringstream ostr;
  std::string ret;
  const_iterator it;
  for (it = Begin(); it != End(); ++it) {
    ostr << it->X1 << "-" << it->Ev << "-" << it->X2 << " ";
  }
  ret = ostr.str();
  if (ret.size() > 0) ret.erase(ret.size()-1);
  return ret;
}

}

#define TTRANSSET_H
#endif

