//////////////////////////////////////////////////////////////////////////////
// 1_generator_example.cpp --
//   Walkthrough of the Generator methods
//////////////////////////////////////////////////////////////////////////////


#include "libfaudes.hpp"


// for simplicity we make the faudes namespace available to our program

using namespace faudes;

/*

General Description of the Generator Class:
===========================================

The Generator class implements the 5-tuple automaton
  G := (Set of States X, Alphabet Sigma, Transition Relation Delta,
        Set of Initial States X_0, Set of Marked States X_m)
that generates a regular language L. X_0 and X_m are subsets of X.

Basically the class consists of five members:
  * StateSet mStates         (set of states)
  * EventSet mAlphabet       (the alphabet of the generator)
  * TransSet mTransRel       (the transition relation)
  * StateSet mInitStates     (the set of initial states)
  * StateSet mMarkedStates   (the set of marked states)

Events are hold by objects of the class EventSet, states by the class
StateSet and transitions by the class TransSet. The TransSet class actually
is only a type definition for the template class TTransSet<TSort> with
template parameter TSort = std::less. TSort is a binary predicate, that
specifies the sorting order of Transition objects in the sorted set of
transitions.

All events and states are represented by integer values (indices). Event
indices are (have to be) consistent within an event domain. State indices
are globally unique. This is a design decision, to enable the
implementation of mappings of state indices in algorithm output automata
to state indices in algorithm input automata. By definition (for this
library) event indices must have symbolic names assigned, while state
indices may but are not forced to have symbolic names assigned.

For a portable handling of event and state indices, the actually used
integer type is defined by the faudes type "Idx".

The consistent mapping of event indices to symbolic event names is
achieved by a static EventSymbolTable object. Every EventSet and Generator
object got a pointer to this EventSymbolTable object and calls the
EventSymbolTable's methods for consistent event handling. The
EventSymbolTable object internally holds a mapping of indices to symbolic
names and the corresponding reverse mapping of symbolic names to indices.

Analogous to the mapping of event indices to symbolic names, the mapping
of state indices to symbolic state names is held by a static
StateSymbolTable. Every StateSet and Generator got a pointer to the static
StateSymbolTable. To keep state indices unique, a StateSymbolTable
contains an internal integer counter, that holds the last assigned state
index and becomes incremented each time a new state is created in a
Generator (actually in the Generator's mStates member).

Differing from an EventSymbolTable a StateSymbolTable holds only a one way
mapping of state indices to state names (because a symbolic state name
corresponds to different indices in different automata). For this reason
only fast symbolic name resolution for state indices is provided. The
reverse resolution of symbolic names to indices is emulated by a search
through the indices of a StateSet until a matching symbolic name can be
resolved (or not). Therefore this name to index resolve functionality may
and is only required to be used in user interactions.

Anyway, the assignment of symbolic names to event and state indices is
a user feature and not required for the functionality of the algorithms in the
faudes library.

Reason for implementing the Generator as sets of indices and not as sets
of strings are performance considerations. Integers can be compared much
faster than strings and require less memory to store.

There are Generator methods to configure the behaviour of state name
computation in algorithms, that construct new automata. Also state names
can be cleared (one or all) and reassigned.

Usually all Generators hold unique state indices. Sometimes there may be
the requirement for creating temporary 1:1 copies of a Generator in
algorithms. To enable this feature, the copy constructor of the Generator
class (and the derived classes aGenerator and cGenerator) creates 1:1 copies of
the Generator's members and register the copies in a static data structure
within the Generator. Every time a Generator object is destroyed and it's
destructor is called, the actual "copy" is unregistered. If there was only
a single "copy" (meaning no duplicate) the symbolic state names of the
state indices of the Generator are removed from the StateSymbolTable. For
this reason temporary copies of Generators may only be used for removing
states, not inserting new states.

*/



/////////////////
// main program
/////////////////

int main() {

  ////////////////////////////////////////////
  // Constructors (part 1) and filling example
  ////////////////////////////////////////////

  // At first we create an empty Generator object
  Generator g1;

  // We can also create a Generator by reading a Generator file
  Generator g2("data/simplemachine.gen");

	// do some random "user interaction" stuff with the Generator g1
	g1.InsState("s1");
	g1.InsState("s2");
	g1.InsState("s3");
	g1.InsEvent("a");
	g1.InsEvent("b");
	g1.SetTransition("s1", "a", "s2");
	g1.SetTransition("s2", "a", "s3");
	g1.SetTransition("s3", "b", "s1");
	g1.SetInitState("s1");
	g1.SetMarkedState("s2");
	g1.SetMarkedState("s3");



  ////////////////////////////////////////////
  // Access to the symbol tables
  ////////////////////////////////////////////

  // Get a pointer to the static EventSymbolTable of the Generator

  EventSymbolTable* etable = g1.EventSymbolTablep();

  // Get a pointer to the static StateSymbolTable of the Generator

  StateSymbolTable* stable = g1.StateSymbolTablep();



  ///////////////////////////////////////////////////
  // Constructors (part 2) & Copying and versioning
  ///////////////////////////////////////////////////

  // Create a a 1:1 copy of the Generator with the copy constructor. The
  // state indices in the copy are exactly the same as in the original.
  // Such copies may only be used temporary in algorithms. For proper
  // deletion of symbolic state names in the Generator's destructor only
  // the last remaining copy of a Generator may be used for inserting new
  // states. However deleting states in temporary copies is flawless. For
  // non-temporary copies the "Copy" method must be used.

  Generator g_copy(g1);
  g_copy = g1; // same result

  // Copy method: Create a copy of a Generator with new unique state indices

  Generator g3;
  g1.Copy(g3);

  // Create a Generator copy with versioned events (for testing algorithms):

  // versioning by an integer. E.g. for integer 3 events {"a", "b", "c"}
  // become {"a_3", "b_3", "c_3"}.

  Generator version1;
  g1.Version(3, version1);

  // versioning by a string. "a" -> "a_versionstring"

  Generator version2;
  g1.Version("str", version2);



  ///////////////////////////////////////////////
  // Methods for console and file input/output
  ///////////////////////////////////////////////

  // Write a Generator to console

  g1.Write();

  // Write a Generator to file

  g1.Write("tmp_generator.gen");

  // Write a Generator to file by appending the output

  g1.Write("tmp_generator.gen", std::ios::out|std::ios::app);

  // Create dotfile for graphviz and call graphviz for writing .png.
  // State indices begin with 1

  g1.DotWrite("tmp_generator"); // becomes tmp_generator.dot / tmp_generator.png

  // Create dotfile for graphviz and call graphviz for other output format
  // State indices begin with 1

  g1.DotWrite("tmp_generator2", "svg"); // svg for example

  // same as DotWrite but also plot the real state indices (Debug-DotWrite)

  g1.DDotWrite("tmp_generator_debug"); // also other file format possible here

  // debug output of Generator to console

  g1.DWrite();

  // Create a debug string for an event with symbolic name + index

  std::string str_singleevent = g1.EStr(2);

  // Create a debug string for a state with symbolic name + index.
  // If there is no symblic name, a symbolic name of the index is constructed.

  std::string str_singlestate = g1.SStr(3);

  // build (debugging) string of events in the Generator's alphabet

  std::string str_alph =  g1.AlphabetToString();

  // build (debugging) string of states in the Generator's set of states
  std::string str_states = g1.StatesToString();

  // ... TransRelToString(), InitStatesToString(), MarkedStatesToString()

  // read file into existing Generator; Clears old content

  Generator g4;
  g4.Read("data/simplemachine.gen");



  //////////////////////////////////////
  // Accessing the Generator's members
  //////////////////////////////////////

  // get the Generator's name

  std::string str_name = g1.Name();

  // set new name for Generator

  g1.Name("NewName");

  // Retrieve a copy of the Generator's alphabet "mAlphabet"

  EventSet eset_copy_alph = g1.Alphabet();

  // Retrieve a reference to the Generator's alphabet "mAlphabet"

  EventSet& eset_ref_alph = g1.Alphabetr();

  // Retrieve a copy of the Generator's set of states "mStates"

  StateSet sset_copy_states = g1.States();

  // Retrieve a reference to the Generator's set of states "mStates"

  StateSet& sset_ref_states = g1.Statesr();

  // Retrieve a copy of the Generator's transition relation "mTransRel"

  TransSet tset_copy_trel = g1.TransRel();

  // Retrieve a reference to the Generator's transition relation "mTransRel"

  TransSet& tset_ref_trel = g1.TransRelr();

  // Retrieve a copy of the Generator's set of initial states "mInitStates"

  StateSet sset_copy_istates = g1.InitStates();

  // Retrieve a reference to the Generator's set of initial states "mInitStates"

  StateSet& sset_ref_istates = g1.InitStatesr();

  // Retrieve a copy of the Generator's set of marked states "mMarkedStates"

  StateSet sset_copy_mstates = g1.MarkedStates();

  // Retrieve a reference to the Generator's set of marked states "mMarkedStates"

  StateSet& sset_ref_mstates = g1.MarkedStatesr();

  // Retrieve the unique id of the Generator (statically counted)

  Idx genid = g1.Id();



  //////////////////////////////////////////////////////////////////////////////
  // Accessing and modifying the 5-tuple Generator (X, Sigma, Delta, X0 and Xm)
  //////////////////////////////////////////////////////////////////////////////

  // Insert an existing event into the Generator's alphabet (mAlphabet)

  g1.InsEvent(1);   // of course index 1 is already in the alphabet here...

  // Insert a new or existing event by it's symbolic name in the alphabet

  g1.InsEvent("newevent");

  // Insert a bunch of events (EventSet) and get the integer index if requested

  EventSet eset1;
  eset1.InsEvent("newevent1");
  Idx idx_tmp = eset1.InsEvent("newevent2");
  g1.InsEvents(eset1);

  // Delete an event from Generator (mAlphabet & mTransRel)

  g1.DelEvent(idx_tmp); // by index
  g1.DelEvent("newevent2"); // by symbolic name

  // Delete a bunch of events

  g1.DelEvents(eset1); // .. of course we have already deleted them before...

  // Insert a new state. The state gets a integer index that is unique within
  // the Generator's event domain (EventSymbolTable/StateSymbolTable)

  idx_tmp = g1.InsState(); // anonymous state
  idx_tmp = g1.InsState("newstate2"); // named state

  // Insert a new state as initial state

  idx_tmp = g1.InsInitState();
  idx_tmp = g1.InsInitState("newinitstate");

  // ... same for marked states

  idx_tmp = g1.InsMarkedState();
  idx_tmp = g1.InsMarkedState("newmarkedstate");

  // delete single states from Generator

  g1.DelState(idx_tmp); // by index (relatively fast, for algorithms)
  g1.DelState("newinitstate"); // by symbolic name, if name assigned

  // (deletion by symbolic name is slow; only for user interactions; see
  // stateset example for details)

  // delete a bunch of states. This is much faster than deleting single states
  // (when deleting single states, the whole transition relation has to be
  // completely searched every time)

  StateSet stateset1;
  StateSet stateset2;
  StateSet stateset3;
  stateset1.Insert(1);
  stateset2.Insert(2);
  stateset3.Insert(3);
  g1.DelStates(stateset1);

  // for further proceeding we insert some new states and events...

  Idx idx_s10 = g1.InsState("s10");
  Idx idx_s11 = g1.InsState("s11");
  Idx idx_s12 = g1.InsState("s12");
  Idx idx_e10 = g1.InsEvent("e10");
  Idx idx_e11 = g1.InsEvent("e11");

  // set a state that already exists in Generator as initial state

  g1.SetInitState(idx_s10);

  // set a state that already exists in Generator as marked state

  g1.SetMarkedState(idx_s11);

  // unset an existing state as initial state (does not remove from mStates)

  g1.ClrInitState(idx_s10); // also possibly by symbolic name (slow)

  // unset an existing state as marked state (does not remove from mStates)

  g1.ClrMarkedState(idx_s10); // also possible by symbolic name (slow)

  // clear all initial states (mStates stays untouched)

  //   g1.ClrInitStates();  // we do not really do it here, so it's commented

  // clear all marked states (mStates stays untouched)

  //   g1.ClrMarkedStates();  // we do not really do it here, so it's commented

  // set a transition for existing states and events

  g1.SetTransition(idx_s10, idx_e10, idx_s11); // by indices
  g1.SetTransition("s10", "e11", "s10"); // by symbolic names (slow)

  // clear a transition (does not touch mStates, mInitStates and mMarkedStates)

  g1.ClrTransition(idx_s10, idx_e10, idx_s11); // by index

  // transitions can also be cleared by names (slower) or by an assigned
  // TransSet::iterator (very fast!)

  // clear all transitions

  //   g1.ClrTransRel(); // we do not really want to do that now...

  // clear the symbolic name for a state in the StateSymbolTable

  g1.ClrStateName(idx_s10); // also symbolic name possible (slow)

  // exists event index/name in mAlphabet?

  bool bool_eventexists1 = g1.ExistsEvent("e11"); // "fast" for user interactions
  bool bool_eventexists2 = g1.ExistsEvent(2); // fast! for algorithms

  // ... FindEvent(..) returns an iterator. This is for optimized algorithms only...

  // exists state in mStates?

  bool bool_stateexists1 = g1.ExistsState(4); // no string version here

  // .. also FindState(..) method, that returns an iterator available...

  // check if a state is an initial state

  bool bool_initstateexists = g1.ExistsInitState(4); // also symbolic name possible (slow)

  // check if a state is a marked state

  bool bool_ismarkedstate = g1.ExistsMarkedState(4); // also symbolic name possible (slow)

  // ... FindInitState(..) and FindMarkedState(..) return iterators; for
  // optimized algorithms

  // look up event name for index in the EventSymbolTable of the event domain

  std::string str_eventname1 = g1.EventName(1);

  // look up event index for name in the EventSymbolTable of the event domain

  Idx idx_eventindex = g1.EventIndex("e11");

  // get symbolic name assigned to state (returns "" if no name assigned).
  // with FAUDES_CHECKED defined, an the program terminates if the index is not
  // contained in mStates

  std::string str_tmp = g1.StateName(idx_s10);

  // get index for symbolic state name. only possible for state names of states in
  // the Generator (slow, only for user interactions)

  idx_tmp = g1.StateIndex("s12");

  // clear Generator (including alphabet)

  g4.Clear();

  // get the number of events in the Generator's alphabet

  Idx idx_eventnum = g1.AlphabetSize();

  // get the number of states

  Idx idx_statenum = g1.Size();

  // get the number of transitions

  Idx idx_transnum = g1.TransRelSize();

  // ... InitStatesSize(), MarkedStatesSize()

  // is the alphabet of the Generator empty?

  bool bool_alphempty = g1.AlphabetEmpty();

  // is the Generator empty (number of states == 0) ?

  bool bool_isempty = g1.Empty();

  // ... TransRelEmpty, InitStatesEmpty, MarkedStatesEmpty



  ///////////////////////
  // Iterators
  ///////////////////////

  // remember:
  // never change the current value of an iterator in a sorted set!!!
  // instead remove value and insert new!

  // there is always a iterator and a const_iterator

  // iteration over alphabet indices (Member "mAlphabet")

  EventSet::iterator eit;
  for (eit = g1.AlphabetBegin(); eit != g1.AlphabetEnd(); ++eit) {
    // do something here... but never change the iterators value
    std::cout << *eit << std::endl;
  }

  // iteration over state indices (Member "mStates")

  StateSet::iterator sit;
  for (sit = g1.StatesBegin(); sit != g1.StatesEnd(); ++sit) {
    // do something here... but never change the iterators value
    std::cout << *sit << std::endl;
  }

  // iteration over complete transition relation (member "mTransRel")

  TransSet::iterator tit;
  for (tit = g1.TransRelBegin(); tit != g1.TransRelEnd(); ++tit) {
    // do something here... but never change the iterators value
    std::cout << tit->X1 << " " << tit->Ev << " " << tit->X2 << std::endl;
  }

  // iteration over transitions from a given state. for not computing
  // the end of the iteration in every step we compute this before

  idx_tmp = g1.StateIndex("s1");
  TransSet::iterator tit_end;
  tit = g1.TransRelBegin(idx_tmp);
  tit_end = g1.TransRelEnd(idx_tmp);
  for (; tit != tit_end; ++tit) {
    // do something here... but never change the iterators value
    std::cout << tit->X1 << " " << tit->Ev << " " << tit->X2 << std::endl;
  }

  // ... also possible: transitions of given state index + given event index

  //   TransRelBegin(x1, ev) - TransRelEnd(x1, ev)

  // ... InitStatesBegin() - InitStatesEnd()  (member "mInitStates")

  // ... MarkedStatesBegin() - MarkedStatesEnd() (member "mMarkedStates")



  ////////////////////////////////////////////////////////////
  // Retrieve copies of the Generator's transition releation
  // in different sorting orders than X1 -> Ev -> X2
  ////////////////////////////////////////////////////////////

  // Important: Take care which iterators you use with different sorted
  // transition relations. While the Begin() and the End() iterators are
  // always valid, there exist different iterators to access only parts
  // of different sorted transition relations.

  // Retrieve a copy that is sorted by Ev -> X1 -> X2 by the binary
  // predicate TSortEvX1X2. "TransSetEvX1X2" is a convenience type definition
  // for "TTransSet<TSortEvX1X2>"

  TransSetEvX1X2 tset_evx1x2 = g1.TransRelEvX1X2();

  // Retrieve a copy that is sorted by Ev -> X2 -> X1 by the binary
  // predicate TSortEvX2X1. "TransSetEvX2X1" is a convenience type definition
  // for "TTransSet<TSortEvX2X1>"

  TransSetEvX2X1 tset_evx2x1 = g1.TransRelEvX2X1();

  // Retrieve a copy that is sorted by X2 -> Ev -> X1 by the binary
  // predicate TSortX2EvX1. "TransSetX2EvX1" is a convenience type definition
  // for "TTransSet<TSortX2EvX1>"
  
  TransSetX2EvX1 tset_x2evx1 = g1.TransRelX2EvX1();
  
  // Retrieve a copy that is sorted by X2 -> X1 -> Ev by the binary
  // predicate TSortX2X1Ev. "TransSetX2X1Ev" is a convenience type definition
  // for "TTransSet<TSortX2X1Ev>"

  TransSetX2X1Ev tset_x2x1ev = g1.TransRelX2X1Ev();

  // Retrieve a copy that is sorted by X1 -> X2 -> Ev by the binary
  // predicate TSortX1X2Ev. "TransSetX1X2Ev" is a convenience type definition
  // for "TTransSet<TSortX1X2Ev>"

  TransSetX1X2Ev tset_x1x2ev = g1.TransRelX1X2Ev();



  ////////////////////////
  // Convenience Methods
  ////////////////////////

  // Remove all events from mAlphabet, that do not have a transition in
  // mTransRel g1.MinimizeAlphabet(); we actually do not want this here

  // Get an EventSet containing all the events that drive a transition in the
  // automaton

  EventSet eset_usedevents = g1.UsedEvents();

  // Get an EventSet containing all the events that do not drive a transition
  // in the automaton

  EventSet eset_unusedevents = g1.UnusedEvents();

  // Return the active event set at a given state as an EventSet

  EventSet eset_activeeventset = g1.ActiveEventSet(idx_s12);

  // Return a StateSet containing all the states that are connected by
  // transitions. This has no relation to the language, the automaton
  // actually generates.

  StateSet sset_trel_sspace = g1.TransRelStateSpace();

  // Return a StateSet containing all the successor states of a given predecessor
  // state.

  StateSet sset_successors = g1.TransRelStateSpace(idx_s12);



  /////////////////////////////////
  // Symbolic state name handling
  /////////////////////////////////

  // Are symbolic state names enabled? depending on this boolean value
  // library functions like Determine or StateMin may create symbolic
  // state names or not.

  bool bool_statenamesenabled = g1.StateNamesEnabled();

  // Disable state name creation in resulting generators for functions in
  // the faudes library, that support this feature (nearly all) with
  // "false". Enable state name creation with "true".

  g1.StateNamesEnabled(true); // anyway .. true is the default value

  // Clear existing symbolic statenames for states in the Generator. Actually
  // the state names are cleared in the StateSymbolTable of the Generator's
  // event domain.

  //  g1.ClrStateNames(); // we do not actually want to do this

  // Set symbolic names for all states in the generator. The symbolic name becomes
  // the equivalent string representation of the state's integer index. This is
  // only usefull for debugging purposes.

  g1.SetDefaultStateNames();



  ///////////////////////////////////
  // Accessible, Coaccessible, Trim
  ///////////////////////////////////

  // StateSet containing all accessible states

  StateSet sset_accessibleset = g1.AccessibleSet();

  // Make the Generator accessible by removing transitions and states. If
  // the Generator containes at least one initial state the Generator
  // can be made accessible and the return value is "true", else "false".

  bool bool_isaccessiblenow = g1.Accessible();

  // Is the Generator accessible. (No modification of the Generator)

  bool bool_isaccessible = g1.IsAccessible();

  // StateSet containing all coaccessible states

  StateSet sset_coaccessibleset = g1.CoaccessibleSet();

  // Make the Generator coaccessible by removing transitions and states. If
  // the Generator containes at least one initial state the Generator
  // can be made coaccessible and the return value is "true", else "false".

  bool bool_iscoaccessiblenow = g1.Coaccessible();

  // Is the Generator coaccessible. (No modification of the Generator)

  bool bool_iscoaccessible = g1.IsCoaccessible();

  // StateSet containing all trim states

  StateSet sset_trimset = g1.TrimSet();

  // Make the Generator trim by removing transitions and states. If
  // the Generator containes at least one initial state the Generator
  // can be made trim and the return value is "true", else "false".

  bool bool_istrimnow = g1.Trim();

  // Is the Generator trim. (No modification of the Generator)

  bool bool_istrim = g1.IsTrim();

  return 0;
}



