//////////////////////////////////////////////////////////////////////////////
// 2_eventset_example.cpp --
//   Simple examples of using the EventSet container to store alphabets.
//   It is assumed the Generator example was already studied by so far.
//////////////////////////////////////////////////////////////////////////////


#include "libfaudes.hpp"

// for simplicity we want to just write "EventSet", not "faudes::EventSet"

using namespace faudes;


/*

General Description of the EventSet class:

The EventSet class consists of an internal sorted set of unique elements
of the type std::set<Idx>, where Idx is the type definition for the
actually used event index integer type in the library. For symbolic state
name resolution and insertion of indices, the EventSet also holds a
static pointer to a (static) EventSymbolTable object.

Different from the StateSet class (with a StateSymbolTable), an
EventSymbolTable stores both, a mapping of integer indices to symbolic names
and a mapping of symbolic names to integer indices. Therefore the lookup in
both directions is fast. The EventSymbolTable keeps the mapping of indices
and symbolic names consistent within several EventSet objects within an
event domain. See doxygen documentation of the EventSet class for details.

*/



int main() {

	// Create EventSet objects
	
	EventSet alphabet1;
	EventSet alphabet2;
	EventSet alphabet3;

	// New events always have to be inserted by calling the InsEvent method
	// with a symbolic name. This assigns the next free index to the
	// symbolic name in the static EventSymbolTable. If the symbolic name
	// of the event is already known in the EventSymbolTable, the existing
	// index is returned.

	Idx ev1 = alphabet1.InsEvent("a");
	Idx ev2 = alphabet1.InsEvent("b");
	Idx ev3 = alphabet1.InsEvent("c");
	Idx ev4 = alphabet1.InsEvent("d");

	Idx ev5 = alphabet2.InsEvent("c"); // ev3 == ev5
	Idx ev6 = alphabet2.InsEvent("d"); // ev4 == ev6
	Idx ev7 = alphabet2.InsEvent("e");
	Idx ev8 = alphabet2.InsEvent("f");

	// Events that already have an index assigned can be inserted by their
	// index. This is usually the case in algorithms. This also does not
	// require looking up the index for a symbolic name. The InsEvent called
	// with an index returns void.

	alphabet2.InsEvent(ev1);
 
	std::cout << "event index ev2: " << ev2 << std::endl;
	
	// iterator usage. use const_iterator whenever possible.

	EventSet::const_iterator eit;
	for (eit = alphabet1.Begin(); eit != alphabet1.End(); ++eit) {
		std::cout << alphabet1.EventName(*eit) << ": " << *eit << std::endl;
	}

	// read a alphabet from file

	alphabet3.Read("data/simplemachine.gen");

	// read a alphabet from file at object construction

	EventSet alphabet4("data/simplemachine.gen");

	// print a string containing all events in an alphabet

	std::cout << "alphabet1: " << alphabet1.ToString() << std::endl;
	std::cout << "alphabet2: " << alphabet2.ToString() << std::endl;
	std::cout << "alphabet3: " << alphabet3.ToString() << std::endl;
	std::cout << "alphabet4: " << alphabet4.ToString() << std::endl;
	
	// write a alphabet to file

	alphabet2.Write("tmp_alphabet.alp");

	// debug output to console

	alphabet1.DWrite();

	// set inclusion by overloaded <= operator

	if (alphabet1 <= alphabet2) {
		std::cout << "alphabet1 includes alphabet2" << std::endl;
	}
	else {
		std::cout << "alphabet2 does not include alphabet2" << std::endl;
	}
	
	// We can do the same with the stl algorithm "includes". This is just an
	// example for using stl algorithms with libfaudes.

	if (std::includes(alphabet1.Begin(), alphabet1.End(), alphabet2.Begin(), alphabet2.End())) {
		std::cout << "alphabet1 includes alphabet2" << std::endl;
	}
	else {
		std::cout << "alphabet2 does not include alphabet2" << std::endl;
	}

	// delete an event by name

	alphabet2.DelEvent("e");

	// delete an event by index

	alphabet2.DelEvent(alphabet2.EventIndex("f"));

	// clear a eventset

	alphabet4.Clear();

	// test existence of event

	if (alphabet2.ExistsEvent("d")) {
		std::cout << "alphabet2: event d exists" << std::endl;
	}

	std::cout << "alphabet1: " << alphabet1.ToString() << std::endl;
	std::cout << "alphabet2: " << alphabet2.ToString() << std::endl;

	// The EventSet class is capable of set operations:

	// set union

	alphabet4 = alphabet1 + alphabet2;
	std::cout << "set union: " << alphabet4.ToString() << std::endl;

	// set difference

	alphabet4 = alphabet1 - alphabet2;
	std::cout << "set difference: " << alphabet4.ToString() << std::endl;

	// set intersection

	alphabet4 = alphabet1 * alphabet2;
	std::cout << "set intersection: " << alphabet4.ToString() << std::endl;


	return 0;
}
