//////////////////////////////////////////////////////////////////////////////
// 3_stateset_example.cpp --
//   Simple examples of using the StateSet container to store states.
//   It is assumed the Generator example was already studied by so far.
//////////////////////////////////////////////////////////////////////////////


#include "libfaudes.hpp"

// for simplicity we want to just write "StateSet", not "faudes::StateSet"

using namespace faudes;


/*

General Description of the StateSet class:

The StateSet class consists of an internal sorted set of unique elements
of the type std::set<Idx>, where Idx is the type definition for the
actually used integer type in the library. For symbolic state name
resolution and insertion globally unique state indices, the
StateSet also holds a pointer to the StateSymbolTable of the event domain.

Different from the EventSymbolTable class, the StateSymbolTable class only
got a mapping of state indices to symbolic names, but no reverse mapping.
Therefore a index lookup for a symbolic state name works only by search
through a StateSet until the state name is found or not. This results in
very slow lookups of indices for given symbolic names. Therefore this
functionality is only meant to be used in user interactions. See doxygen
documentation of the StateSet class for details.

*/


int main() {

	// Create a StateSet

	StateSet stateset1;

	// New globally unique states are put into a stateset by the InsState
	// method. It can either be called with no parameter or with a symbolic
	// name (string). In both cases a new globally unique state index is
	// generated an stored in the StateSet object. To implement globally
	// unique state indices, the static StateSymbolTable object got a
	// integer type counter, that is incremented each time, a new state is
	// created in a StateSet object.

	// By actually calling the method the newly created state index is put
	// into the StateSet object's internal set and can also be retrieved as
	// return value by the caller.

	Idx state1 = stateset1.InsState("start");
	Idx state2 = stateset1.InsState("stop");
	Idx state3 = stateset1.InsState("failure");

	// insert an anonymous state 

	Idx state4 = stateset1.InsState();

	// The InsState method is primary used in conjuction with a Generator's
	// stateset (Generator member: mStates). Here globally unique
	// indices have to be created when inserting a new state into a
	// generator. In contrast the implementation of a Generator's set of
	// initial states or set of marked states requires to insert indices,
	// that already exist in the Generator's stateset. Therefore a second
	// insertion method is available in the StateSet class, the "Insert"
	// method.

	// The Insert method is always called with an index as parameter and
	// simply puts the index in the internal set. This also enables the
	// StateSet to be used as a simple container to store existing or
	// nonexisting state indices. In contrast to the "InsState" method, no
	// new indices are created by the static StateSymbolTable.

	StateSet stateset2; 
	stateset2.Insert(100); 
	stateset2.Insert(2343435);

	// assign a (new) name to a state index

	stateset1.StateName(state4, "dumpstate");

	// iterator usage. use const_iterator whenever possible.

	StateSet::const_iterator sit;
	for (sit = stateset1.Begin(); sit != stateset1.End(); ++sit) {
		std::cout << stateset1.StateName(*sit) << ": " << *sit << std::endl;
	}

	// get the index of a state by a full search of the whole stateset for a
	// matching name. this method should only be used in user interactions

	Idx state5 = stateset1.StateIndex("stop");
	std::cout << "stop index = " << state5 << std::endl;

	// reading a StateSet from a file is not possible. This functionality
	// is handled by the Generator classes, which contain StateSet objects.

	// print a string containing all states in a stateset

	std::cout << "stateset1: " << stateset1.ToString() << std::endl;
	
	// write a stateset to file

	stateset1.Write("tmp_stateset.txt");

	// debug output to console

	stateset1.DWrite();

	// delete a state by name (slow, search through all states in StateSet!)

	stateset1.DelState("failure");

	// clear the symbolic name of a state (removes entry from StateSymbolTable)

	// stateset1.ClrStatename("start"); // we actually don't want to do this

	// delete a state by index (fast!)

	stateset1.DelState(state2);

	// copy a StateSet

	StateSet stateset3 = stateset1;

	// clear a StateSet

	stateset1.Clear();

	// test existence of state

	if (stateset3.ExistsState(state1)) {
		std::cout << "stateset3: state " << state1 << "exists" << std::endl;
	}

	return 0;
}
